/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/delay.h>
#include <linux/workqueue.h>
#include <linux/firmware.h>
#include "vastai_pci_boot.h"
#include "vastai_version.h"
#include "sg100_base.h"
#include "csram_partition.h"
#include "sg100_smcu.h"
#include "sg100_pcie_wrap.h"
#include "vastai_pci.h"
#include "hw_config.h"
#include "sg100_cmd.h"
#include "vastai_pci_test.h"
#include "vastai_state.h"
#include "vastai_sg100_bbox.h"
#include "config.h"
#include "vastai_dmi_table.h"
#include "vastai_ras.h"
#include "sdma.h"
#define CONFIG_VASTAI_DOWNLOAD_FW_FROM_HEX
#ifdef CONFIG_VASTAI_DOWNLOAD_FW_FROM_HEX
#include "vastai_fw_hex.h"
#endif
extern struct list_head vastai_card_info_list;
extern bool outbound_sriov_en;
static bool is_find_pkg_info(struct pci_dev *pdev, u8 pkg_id,
				struct vastai_card_info **card_info, struct vastai_pkg_info **pkg_info)
{
	bool isfind = false;
	struct vastai_card_info *card_loop = NULL;

	list_for_each_entry(card_loop, &vastai_card_info_list, card_node) {
		if((card_loop->bdf_info.dev_id == (pdev->devfn >> 3)) &&
			(card_loop->bdf_info.bus_id == (pdev->bus->number)) &&
			(card_loop->bdf_info.domain_id == pci_domain_nr(pdev->bus))) {
			struct vastai_pkg_info *pkg_loop = NULL;

			*card_info = card_loop;
			list_for_each_entry(pkg_loop, &card_loop->vastai_pkg_info_list, pkg_node) {
				if(pkg_loop->pkg_id == pkg_id) {
					*pkg_info = pkg_loop;
					isfind = true;
					break;
				}
			}
			break;
		}
	}

	return isfind;
}

static bool is_find_die_info(struct pci_dev *pdev, u8 pkg_id, u8 die_id, struct vastai_die_info **die_info)
{
	bool isfind = false;
	struct vastai_die_info *die_loop   = NULL;
	struct vastai_pkg_info *pkg_info   = NULL;
	struct vastai_card_info *card_info = NULL;

	if(is_find_pkg_info(pdev, pkg_id, &card_info, &pkg_info)) {
		list_for_each_entry(die_loop, &pkg_info->vastai_die_info_list, die_node) {
			if(die_loop->die_id == die_id) {
				*die_info = die_loop;
				isfind = true;
				break;
			}
		}
	}

	return isfind;
}

static bool is_find_fn_info(struct vastai_die_info *die_info, struct vastai_pci_info *priv)
{
	bool isfind = false;
	struct vastai_pci_info *fn_loop = NULL;

	list_for_each_entry(fn_loop, &die_info->vastai_fn_info_list, fn_node[die_info->die_id]) {
		if(fn_loop == priv) {
			isfind = true;
			break;
		}
	}

	return isfind;
}

static int vastai_create_pkg_info_tree(struct pci_dev *pdev, u8 pkg_id, struct vastai_pkg_info **pkg_info)
{
	struct vastai_card_info *card_info = NULL;

	if(is_find_pkg_info(pdev, pkg_id, &card_info, pkg_info))
		return 0;

	if(!card_info) {
		VASTAI_PCI_ERR(
			NULL, DUMMY_DIE_ID,
			"create_pkg_info_tree card_info is NULL.\n");
		return -ENOMEM;
	}

	*pkg_info = kzalloc(sizeof(struct vastai_pkg_info), GFP_KERNEL);
	if (!*pkg_info) {
		VASTAI_PCI_ERR(
			NULL, DUMMY_DIE_ID,
			"create_pkg_info_tree NOMEM.\n");
		return -ENOMEM;
	}

	(*pkg_info)->pkg_id = pkg_id;
	(*pkg_info)->card_info = card_info;
	INIT_LIST_HEAD(&(*pkg_info)->vastai_die_info_list);

	list_add_tail(&((*pkg_info)->pkg_node), &(card_info->vastai_pkg_info_list));

	return 0;
}

static int vastai_create_die_info_tree(struct pci_dev *pdev, u8 pkg_id, u8 die_id, struct vastai_die_info **die_info)
{
	int ret = 0;
	struct vastai_pkg_info  *pkg_info  = NULL;
	struct vastai_card_info *card_info = NULL;

	if(is_find_pkg_info(pdev, pkg_id, &card_info, &pkg_info)) {
		if(!pkg_info) {
			VASTAI_PCI_ERR(
				NULL, DUMMY_DIE_ID,
				"create_die_info_tree card_info is NULL.\n");
			return -ENOMEM;
		}

		if(is_find_die_info(pdev, pkg_id, die_id, die_info))
			return 0;
	} else {
		ret = vastai_create_pkg_info_tree(pdev, pkg_id, &pkg_info);
		if(ret)
			return ret;
	}

	*die_info = kzalloc(sizeof(struct vastai_die_info), GFP_KERNEL);
	if (!(*die_info)) {
		VASTAI_PCI_ERR(
			NULL, DUMMY_DIE_ID,
			"create_die_info_tree NOMEM.\n");
		return -ENOMEM;
	}

	(*die_info)->die_id = die_id;
	(*die_info)->pkg_info = pkg_info;
	INIT_LIST_HEAD(&(*die_info)->vastai_fn_info_list);

	list_add_tail(&((*die_info)->die_node), &(pkg_info->vastai_die_info_list));

	return 0;
}

extern struct mutex vastai_card_info_list_lock;
static int vastai_create_fn_info_tree(struct pci_dev *pdev, u8 pkg_id, u8 die_id,
					u8 fn_id, struct vastai_pci_info *priv)
{
	int ret = 0;
	struct vastai_die_info  *die_info  = NULL;

	mutex_lock(&vastai_card_info_list_lock);
	if(is_find_die_info(pdev, pkg_id, die_id, &die_info)) {
		if(!die_info) {
			VASTAI_PCI_ERR(
				NULL, DUMMY_DIE_ID,
				"create_fn_info_tree die_info is NULL.\n");

			mutex_unlock(&vastai_card_info_list_lock);
			return -ENOMEM;
		}

		if(is_find_fn_info(die_info, priv)) {
			mutex_unlock(&vastai_card_info_list_lock);
			return ret;
		}
	} else {
		ret = vastai_create_die_info_tree(pdev, pkg_id, die_id, &die_info);
		if(ret) {
			mutex_unlock(&vastai_card_info_list_lock);
			return ret;
		}
	}

	priv->die_info[die_info->die_id] = die_info;
	list_add_tail(&(priv->fn_node[die_info->die_id]), &(die_info->vastai_fn_info_list));
	mutex_unlock(&vastai_card_info_list_lock);

	return ret;
}


#if 0 // by keye
/*
 * smcu boot stage:
 * bit31(gpio3/pkg sel), bit30(gpio2/last die), bit29(dieid/slave die)
 * die_num die_id val      bit31 bit30 bit29
 * 1-die   die0   3'h2     0     1     0
 * 2-die   die0   3'h0     0     0     0
 * 2-die   die1   3'h3     0     1     1
 * 4-die   die0   3'h4     1     0     0
 * 4-die   die1   3'h1     0     0     1
 * 4-die   die2   3'h5     1     0     1
 * 4-die   die3   3'h7     1     1     1
 */
struct vastai_die_cfg die_cfg[VASTAI_DIE_CONFIG_TABLE_NUM] = {
	{2, VASTAI_SINGLE_DIE, VASTAI_DIE0}, {0, VASTAI_TWO_DIE, VASTAI_DIE0},
	{3, VASTAI_TWO_DIE, VASTAI_DIE1},    {4, VASTAI_FOUR_DIE, VASTAI_DIE0},
	{1, VASTAI_FOUR_DIE, VASTAI_DIE1},   {5, VASTAI_FOUR_DIE, VASTAI_DIE2},
	{7, VASTAI_FOUR_DIE, VASTAI_DIE3}};
#else
struct vastai_die_cfg die_cfg[VASTAI_DIE_CONFIG_TABLE_NUM] = {
	{0, VASTAI_SINGLE_DIE, VASTAI_DIE0, VASTAI_PKG0, {VASTAI_FN0, VASTAI_FN1, VASTAI_FN2, VASTAI_FN3}},   // Single pkg(die) per board No DP enabled
	{1, VASTAI_TWO_DIE,    VASTAI_DIE0, VASTAI_PKG0, {VASTAI_FN0, VASTAI_FN1, VASTAI_FN2, VASTAI_FN3}},   // Master-slave pkg per board’s master pkg.
	{2, VASTAI_TWO_DIE,    VASTAI_DIE1, VASTAI_PKG1, {VASTAI_FN4, VASTAI_FN5, VASTAI_FN6, VASTAI_FN7}},   // Master-slave pkg per board’s slave pkg.
	{7, VASTAI_SINGLE_DIE, VASTAI_DIE0, VASTAI_PKG0, {VASTAI_FN0, VASTAI_FN1, VASTAI_FN2, VASTAI_FN3}}};  // PKG_MODE 1, PCB board with DP enabled. And only single pkg for this mode.
#endif

static u8 vastai_pci_get_die_num(struct vastai_pci_info *priv)
{
#ifndef CONFIG_VASTAI_PCI_ONLY_DOWNLOAD_BL1
	union smcu_boot_stage reg;
	u64 reg_addr = VASTAI_SMCU_BOOT_STAGE_REG;
	u32 i;

	vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (VASTAI_SMCU_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(reg.val), 4);
	VASTAI_PCI_DBG_RATELIMIT("pci_get_die_num [read] addr=0x%llx, val=0x%08x\n", reg_addr, reg.val);
	/*
	 * VASTAI_SMCU_BOOT_STAGE_REG, 0x6E4
	 * bit[31:29]: die config, refer to struct vastai_die_cfg table
	 */
	for (i = 0; i < VASTAI_DIE_CONFIG_TABLE_NUM; i++) {
		if (die_cfg[i].val == reg.bit.die_cfg) {
			int j = 0;

			priv->die_num = die_cfg[i].die_num;
			for(j=0; j<8; j++) {
				if(die_cfg[i].fn_id[j] == (priv->dev->devfn & 0x7)) {
					vastai_create_fn_info_tree(priv->dev, die_cfg[i].pkg_id,
									die_cfg[i].die_id,
									die_cfg[i].fn_id[j],
									priv);
					break;
				}
			}
			break;
		}
	}
	if (i == VASTAI_DIE_CONFIG_TABLE_NUM) {
		priv->die_num = VASTAI_SINGLE_DIE;
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_get_die_num cfg=%d not in the cfg tbl, set to single-die.\n", reg.bit.die_cfg);
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "die num=%d, cfg=%d\n", priv->die_num, reg.bit.die_cfg);
#else
	priv->die_num = VASTAI_SINGLE_DIE;
#endif
	return priv->die_num;
}

static void show_smcu_boot_stage(struct vastai_pci_info *priv, u8 die_id, u32 smcu_bootstage)
{
	switch(smcu_bootstage)
	{
		case 0:
			VASTAI_PCI_INFO(priv, die_id, "SMCU: BL0 starts running\n");
			break;
		case 1:
			VASTAI_PCI_INFO(priv, die_id, "SMCU: BL0 needs notify host to download BL1\n");
			break;
		case 3:
			VASTAI_PCI_INFO(priv, die_id, "SMCU: BL1 needs to notify host to download smcu fw. Otherwise BL1 still not run.\n");
			break;
		case 5:
			VASTAI_PCI_INFO(priv, die_id, "SMCU: smcu fw doesn`t boot up.\n");
			break;
		case 6:
			VASTAI_PCI_INFO(priv, die_id, "SMCU: smcu fw needs to configure the PCIe for boot finish stage\n");
			break;
		case 7:
			VASTAI_PCI_INFO(priv, die_id, "SMCU: smcu fw needs to notify all cores ready\n");
			break;
		default:
			VASTAI_PCI_INFO(priv, die_id, "Unknown smcu stage\n");
			break;	
	}
}


int vastai_pci_poll_smcu_stage(struct vastai_pci_info *priv, u8 die_id, u32 smcu_bootstage)
{
	int ret = 0, poll_cnt = 0, err_cnt = 0;
	union smcu_boot_stage reg;
	u64 reg_addr = VASTAI_SMCU_BOOT_STAGE_REG;
	u32 bar_offset = 0, bar_num = VASTAI_PCI_BAR1;

	bar_offset = VASTAI_SMCU_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR;
	bar_num    = VASTAI_PCI_BAR1;

	VASTAI_PCI_INFO(priv, die_id,	"pci_poll_smcu_stage: %d for pkg[%d] from bar%d, bar_offset=%#x\n"
									"                                1:bl1_ready,2:bl1_done,3:fw_ready,4:fw_verify\n"
									"                                5:fw_bootup,6:pcie init finish,7: all cores ready\n",
									smcu_bootstage, die_id, bar_num, bar_offset);
	show_smcu_boot_stage(priv, die_id, smcu_bootstage);

	do {
		ret = vastai_pci_mmio_read(priv, bar_num, bar_offset, &(reg.val), 4);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, die_id, "pci_poll_smcu_stage [read] 0x%llx, reg.val=%#x, error:%d\n", reg_addr, reg.val, ret);
			break;
		}

		if(reg.val == PCIE_FAULT_CODE) {
			if(err_cnt < 2)
				VASTAI_PCI_ERR(priv, die_id, "pci_poll_smcu_stage: ERROR PCIE_FAULT_CODE, sleep then read again.\n0x%llx, len=4, reg.val=%#X\n", reg_addr, reg.val);

			err_cnt++;
			msleep(VASTAI_BOOT_POLL_SLEEP_MS);
			continue;
		}

		/*
		 * VASTAI_SMCU_BOOT_STAGE_REG, 0x6E4
		 * bit[26:24]: smcu boot stage.
		 *             3'h0: initial state
		 *             3'h1: wait for downloading bl1 (bootloader1)
		 *             3'h2: bl1 verify pass
		 *             3'h3: wait for downloading fw (firmware)
		 *             3'h4: fw verify pass 
		 *             3'h5: fw boot up successfully
		 *             3'h6: current die pcie init finish
		 *             3'h7: all mcu and dsp of current die init finish
		 */
		if (reg.bit.stage == smcu_bootstage) {
			VASTAI_PCI_INFO(priv, die_id,
				"pci_poll_smcu_stage %ds later, err_cnt=%d, smcu.bootstage=%#x is ready ",
				(poll_cnt / 1000), err_cnt, reg.bit.stage);
			break;
		}

		if (poll_cnt++ > VASTAI_BOOT_POLL_SLEEP_CNT) {
			VASTAI_PCI_ERR(priv, die_id,
				"pci_poll_smcu_stage poll stage %d, but stage is %d, %ds timeout!\n",
				smcu_bootstage, reg.bit.stage,
				(VASTAI_BOOT_POLL_SLEEP_CNT / 1000));
			return -ETIMEDOUT;
		}

		msleep(VASTAI_BOOT_POLL_SLEEP_MS);
	} while (1);

	return ret;
}

#ifdef CONFIG_ENV_EMULATOR
static int vastai_pci_poll_die_linkup(struct vastai_pci_info *priv, u8 die_id)
{
	int ret, poll_cnt = 0;
	union smcu_boot_stage reg;
	u64 reg_addr = VASTAI_SMCU_BOOT_STAGE_REG;

    VASTAI_PCI_DBG(priv, die_id, "pci_poll_die_linkup for PKG[%d] addr=0x%llx\n", die_id, reg_addr);

	do {
        ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (VASTAI_SMCU_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(reg.val), 4);
		if (ret < 0)
			VASTAI_PCI_ERR(priv, die_id, "pci_poll_die_linkup 0x%llx, error:%d\n", reg_addr, ret);
        
		/*
		 * VASTAI_SMCU_BOOT_STAGE_REG, 0x6E4
		 * bit[28]: rc_link_sts:
		 *          current die ep and front die or host  rc link up flag
		 * bit[27]: ep_link_sts: current die rc and next die ep link up flag
		 */
		if (reg.bit.rc_link_sts == 1)
		{
			VASTAI_PCI_INFO(priv, die_id, "pci_poll_die_linkup smcu.bootstage=0x%x, PKG[%d] linkup with host RC", reg.val, (die_id));
			break;
		}

		if (poll_cnt++ > VASTAI_BOOT_POLL_SLEEP_CNT) {
			VASTAI_PCI_ERR(priv, die_id,
				"pci_poll_die_linkup poll linkup, but link status is %d,"
				" %ds timeout!\n",
				reg.bit.rc_link_sts,
				(VASTAI_BOOT_POLL_SLEEP_CNT / 1000));
			return -ETIMEDOUT;
		}
		msleep(VASTAI_BOOT_POLL_SLEEP_MS);
	} while (1);

	return ret;
}
#endif


// Func: poll SMCU FW of master pkg whether switching boot slave bar atu is ready.
//       From Bar2
static int vastai_pci_poll_master_atu_state(struct vastai_pci_info *priv, u8 die_id)
{
	int ret, poll_cnt = 0;
	u32 val = 0;
	u64 reg_addr  = PWRAP_COMMON_HOST_ONLY4;
    u32 atu_stage = BIT_MASTER_ATU_BOOT_SLAVE_BAR_READY;
    u32 bar2_pwrap_offset = SPACE_SIZE_1M - SPACE_SIZE_12K + SPACE_SIZE_4K;

    VASTAI_PCI_DBG(priv, die_id, "pci_poll_master_atu_state die_id=%d\n", die_id);
    
	do {
        ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR2, bar2_pwrap_offset + (PWRAP_COMMON_HOST_ONLY4 - PCIE_WRAPPER_ADDR), &val, 4);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, die_id,
				       "pci_poll_master_atu_state [read] 0x%llx, len=4, error:%d\n",
				       reg_addr, ret);
			break;
		}

		if (val == atu_stage) {
			VASTAI_PCI_INFO(priv, die_id, "pci_poll_master_atu_state %ds later, atu_stage=0x%x ready.\n", (poll_cnt / 1000), atu_stage);
			break;
		}

		if (poll_cnt++ > VASTAI_BOOT_POLL_SLEEP_CNT) {
			VASTAI_PCI_ERR(priv, die_id,
				"pci_poll_master_atu_state poll stage=%#x, but value=%#x, %ds timeout!\n",
				atu_stage, val,
				(VASTAI_BOOT_POLL_SLEEP_CNT / 1000));
			return -ETIMEDOUT;
		}
		msleep(VASTAI_BOOT_POLL_SLEEP_MS);
       
	} while (1);
    
	return ret;
}


// Func: signal the master pkg in FW that slave pkg is init done.
//       From Bar2
static int vastai_pci_notify_master(struct vastai_pci_info *priv, u8 die_id)
{
	int ret = 0;
    u32 reg_val = 0;
    u32 bar2_pwrap_offset = SPACE_SIZE_1M - SPACE_SIZE_12K + SPACE_SIZE_4K;

    reg_val = SMCU_SLAVE_DIE_BOOT_DONE;
    ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR2, bar2_pwrap_offset + (PWRAP_COMMON_HOST_ONLY0 - PCIE_WRAPPER_ADDR), &reg_val, 4);

    VASTAI_PCI_INFO(priv, die_id, "pci_notify_master: die_id=%d, reg_val=%#x\n", die_id, reg_val);

    return ret;
}


int vastai_pci_set_host_stage(struct vastai_pci_info *priv, u8 die_id, u32 host_bootstage)
{
	int ret;
	union host_boot_stage reg;
	u64 reg_addr = VASTAI_HOST_BOOT_STAGE_REG;

    ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (VASTAI_HOST_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_set_host_stage [read] 0x%llx, len=4, error:%d\n",
			       reg_addr, ret);
	/*
	 * VASTAI_HOST_BOOT_STAGE_REG, 0x6E0
	 * bit[26:24]: host boot stage.
	 *             3'h0: initial state
	 *             3'h1: bl1 download finish
	 *             3'h2: fw download finish
	 *             3'h3: all dies download finish, host ready
	 */
	reg.bit.stage = host_bootstage;
    ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (VASTAI_HOST_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_set_host_stage [write] 0x%llx, len=4, val=0x%x, error:%d\n",
			       reg_addr, reg.val, ret);

    priv->dies[die_id].pkg_boot_stage = host_bootstage;
    
	return ret;
}

static int vastai_pci_set_current_die(struct vastai_pci_info *priv, u8 die_id)
{
	int ret = 0;
	union host_boot_stage reg;
	u64 reg_addr = VASTAI_HOST_BOOT_STAGE_REG;

    ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (VASTAI_HOST_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_set_current_die [read] 0x%llx, len=4, error:%d\n",
			       reg_addr, ret);
	/*
	 * VASTAI_HOST_BOOT_STAGE_REG, 0x6E0
	 * bit[31:30]: current die id.
	 * 	       2'h0:die0, 2'h1:die1, 2'h2:die2, 2'h3:die3
	 */
	reg.bit.die_id = die_id;
#ifdef CONFIG_VASTAI_PCI_FORCE_SINGLE_DIE
	reg.bit.force_single_die = 1;
#else
	reg.bit.force_single_die = 0;
#endif

    VASTAI_PCI_INFO(priv, die_id,
               "pci_set_current_die [read] 0x%llx, len=4, val=0x%x, force_single_die:%d\n",
               reg_addr, reg.val, reg.bit.force_single_die);

    ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (VASTAI_HOST_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(reg.val), 4);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_set_current_die [write] 0x%llx, len=4, val=0x%x, error:%d\n",
			       reg_addr, reg.val, ret);

	return ret;
}

static int vastai_pci_request_fw(struct vastai_fw **vfw_p,
				 const u8 *fw_hex_buf, size_t fw_hex_size,
				 u64 fw_addr)
{
	struct vastai_fw *vfw;

	*vfw_p = NULL;
	vfw = kmalloc(sizeof(struct vastai_fw), GFP_KERNEL);
	if (!vfw) {
		VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID,
			       "pci_request_fw kmalloc fail!");
		return -ENOMEM;
	}

#ifdef CONFIG_VASTAI_DOWNLOAD_FW_FROM_HEX
	vfw->data = fw_hex_buf;
	vfw->size = fw_hex_size;
	vfw->priv = fw_hex_buf;
	vfw->addr = fw_addr;
#else
	/* request firmware from file-system. */
	/* NTD */
#endif /* CONFIG_VASTAI_DOWNLOAD_FW_FROM_HEX */

	*vfw_p = vfw;

	return 0;
}


static int vastai_pci_write_fw(struct vastai_pci_info *priv,
			       unsigned int die_id,
			       struct vastai_fw *vfw)
{
	int ret;

	/* NTD: judge length of fw */
    ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR4, vfw->addr - VASTAI_PCI_BAR4_ADDR_R, (void *)(vfw->data), vfw->size);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_write_fw [write]MCU FW: 0x%llx, len=%d, error:%d\n",
			       vfw->addr, (int)(vfw->size), ret);
	return ret;
}

static int vastai_pci_write_bl1(struct vastai_pci_info *priv,
			       unsigned int die_id,
			       struct vastai_fw *vfw)
{
	int ret;

	/* NTD: judge length of fw */
    ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR2, vfw->addr - VASTAI_PCI_BAR2_ADDR, (void *)(vfw->data), vfw->size);
	if (ret < 0)
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_write_bl1 [write]BL1: 0x%llx, len=%d, error:%d\n",
			       vfw->addr, (int)(vfw->size), ret);
	return ret;
}


static void vastai_pci_release_fw(struct vastai_fw *vfw)
{
	if (vfw)
		kfree(vfw);
}

static int vastai_pci_download_fw(struct vastai_pci_info *priv,
				  unsigned int die_id,
				  const u8 *fw_hex_buf, 
				  size_t fw_hex_size,
				  u64 fw_addr)
{
	int ret;
	struct vastai_fw *vfw;
   
       VASTAI_PCI_DBG(priv, die_id, "pci_download_fw,VA_BL1_HEX_SIZE=%lu Bytes, VA_FW_HEX_SIZE=%lu Bytes\n", VASTAI_BL1_HEX_SIZE, VASTAI_FW_HEX_SIZE);	
	ret = vastai_pci_request_fw(&vfw, fw_hex_buf, fw_hex_size, fw_addr);
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_download_fw request fw bin error\n");
		goto OUT;
	}

	if(fw_addr < VASTAI_PCI_BAR4_ADDR_R) {
		ret = vastai_pci_write_bl1(priv, die_id, vfw);
		VASTAI_PCI_INFO(priv, die_id, "pci_download_fw write BL1, len=%lu Bytes\n", fw_hex_size);
	} else {
		ret = vastai_pci_write_fw(priv, die_id, vfw);
		VASTAI_PCI_INFO(priv, die_id, "pci_download_fw write MCU&DSP FW, len=%lu Bytes\n", fw_hex_size);
	}
        
	if (ret) {
		VASTAI_PCI_ERR(priv, die_id, "pci_download_fw write fw bin error\n");
		goto OUT;
	}

OUT:
	vastai_pci_release_fw(vfw);

	return ret;
}

/**
 * @brief download fw. If path_name is given, will try to
 * 	load fw from /lib/firmware first.
 *
 * @param pci_info: vastai sv100 pci base struct.
 * @param die_id: the destinat die id. not the die_index.
 * @param path_name: fw path at /lib/firmware/. If not NULL, this is most priority.
 * @param data: fw.bin data.
 * @param size: fw.bin data's size.
 * @param fw_addr: fw destinat addr.
 * @return int: 0 is success. other is failed.
 */
static int vastai_process_download_request(struct vastai_pci_info *pci_info,
				    u8 die_id, u8 *path_name, const u8 *data,
				    size_t size, u64 fw_addr)
{
	int request_ret = 0, ret = 0;
	const struct firmware *fw_entry = NULL;
	const u8 *__data = data;
	size_t __size = size;

	/* most priority */
	if (path_name) {
//		request_ret = request_firmware(&fw_entry, path_name, &(pci_info->dev->dev));
        request_ret = -1;
		if (request_ret != 0)
			VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID,
					"process_download_request ret[%d]: can't find fw/bl from path[/lib/firmware/%s], it will download fw/bl from ko file \n",
					request_ret, path_name);
	}

	/* get fw from /lib/firmarw success */
	if (!request_ret && fw_entry) {
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID,
				"\x1b[31m \nprocess_download_request get fw/bl from path[/lib/firmware/%s] success, to download fw/bl from this path \x1b[0m \n",
				path_name);
		__data = fw_entry->data;
		__size = fw_entry->size;
	}

	/* Not get any fw */
	if (!__data) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				"process_download_request get fw/bl failed! path_name %s, data_addr %p\n",
				path_name, data);
		return -ENOENT;
	}

	/* load fw */
	ret = vastai_pci_download_fw(pci_info, die_id, __data, __size, fw_addr);
	/* we need free fw_entry */
//	if (fw_entry)
//		release_firmware(fw_entry);
	return ret;
}


// Func: should be called before SMCU FW boots up
static void init_boot_hw_cfg(struct vastai_pci_info *priv)
{
	int ret = 0;
	struct boot_hw_cfg *boot_hw_cfg  = kzalloc(sizeof(struct boot_hw_cfg), GFP_ATOMIC);;

#ifdef CONFIG_POWER_CFG
	boot_hw_cfg->power_cfg = CONFIG_POWER_CFG;
#else
	boot_hw_cfg->power_cfg = 0;
#endif

	if(priv->gart_en && CONFIG_ATUO_GART_EN)
		boot_hw_cfg->d2h_by_atu_or_gart = 1;
	else
		boot_hw_cfg->d2h_by_atu_or_gart = 0;

	boot_hw_cfg->outbound_supprt_sriov   = outbound_sriov_en;
	boot_hw_cfg->cfg_gfx_mpu_ddr_rgn	 = CFG_GFX_MPU_DDR_RGN;
	boot_hw_cfg->cfg_power_clk_en        = CFG_BBOX_EN;
	boot_hw_cfg->power_clk[EVB_CARD].gfx.GCLK   		=  EVB_GFX_GCLK;
	boot_hw_cfg->power_clk[EVB_CARD].video.ECLK 		=  EVB_Video_ECLK;
	boot_hw_cfg->power_clk[EVB_CARD].video.DCLK 		=  EVB_Video_DCLK;
	boot_hw_cfg->power_clk[EVB_CARD].video.VDSPCLK 		=  EVB_Video_VDSPCLK;
	boot_hw_cfg->power_clk[EVB_CARD].video.VCLK	 		=  EVB_Video_VCLK;
	boot_hw_cfg->power_clk[EVB_CARD].ddr.RUCLK	 		=  EVB_DDR_RUCLK;
	boot_hw_cfg->power_clk[EVB_CARD].ddr.LUCLK	 		=  EVB_DDR_LUCLK;
	boot_hw_cfg->power_clk[EVB_CARD].soc.CEDARCLK		=  EVB_SOC_CEDARCLK;
	boot_hw_cfg->power_clk[EVB_CARD].soc.LCCLK	    	=  EVB_SOC_LCCLK;
	boot_hw_cfg->power_clk[EVB_CARD].soc.RCCLK	    	=  EVB_SOC_RCCLK;
	boot_hw_cfg->power_clk[EVB_CARD].soc.BCCLK	    	=  EVB_SOC_BCCLK;
	boot_hw_cfg->power_clk[EVB_CARD].ai.OCLK	    	=  EVB_AI_OCLK;
	boot_hw_cfg->power_clk[EVB_CARD].ai.ODSPCLK	    	=  EVB_AI_ODSPCLK;
	boot_hw_cfg->power_clk[EVB_CARD].dp.PIXCLK_DP		=  EVB_DP_PIXCLK_DP;
	boot_hw_cfg->power_clk[EVB_CARD].power_rail.VDD_GFX	=  EVB_POWER_VDD_GFX;
	boot_hw_cfg->power_clk[EVB_CARD].power_rail.VDD_VID	=  EVB_POWER_VDD_VID;
	boot_hw_cfg->power_clk[EVB_CARD].power_rail.VDD_SOC	=  EVB_POWER_VDD_SOC;

	boot_hw_cfg->power_clk[AIC_CARD].gfx.GCLK   		=  AIC_GFX_GCLK;
	boot_hw_cfg->power_clk[AIC_CARD].video.ECLK 		=  AIC_Video_ECLK;
	boot_hw_cfg->power_clk[AIC_CARD].video.DCLK 		=  AIC_Video_DCLK;
	boot_hw_cfg->power_clk[AIC_CARD].video.VDSPCLK 		=  AIC_Video_VDSPCLK;
	boot_hw_cfg->power_clk[AIC_CARD].video.VCLK	 		=  AIC_Video_VCLK;
	boot_hw_cfg->power_clk[AIC_CARD].ddr.RUCLK	 		=  AIC_DDR_RUCLK;
	boot_hw_cfg->power_clk[AIC_CARD].ddr.LUCLK	 		=  AIC_DDR_LUCLK;
	boot_hw_cfg->power_clk[AIC_CARD].soc.CEDARCLK		=  AIC_SOC_CEDARCLK;
	boot_hw_cfg->power_clk[AIC_CARD].soc.LCCLK	    	=  AIC_SOC_LCCLK;
	boot_hw_cfg->power_clk[AIC_CARD].soc.RCCLK	    	=  AIC_SOC_RCCLK;
	boot_hw_cfg->power_clk[AIC_CARD].soc.BCCLK	    	=  AIC_SOC_BCCLK;
	boot_hw_cfg->power_clk[AIC_CARD].ai.OCLK	    	=  AIC_AI_OCLK;
	boot_hw_cfg->power_clk[AIC_CARD].ai.ODSPCLK	    	=  AIC_AI_ODSPCLK;
	boot_hw_cfg->power_clk[AIC_CARD].dp.PIXCLK_DP		=  AIC_DP_PIXCLK_DP;
	boot_hw_cfg->power_clk[AIC_CARD].power_rail.VDD_GFX	=  AIC_POWER_VDD_GFX;
	boot_hw_cfg->power_clk[AIC_CARD].power_rail.VDD_VID	=  AIC_POWER_VDD_VID;
	boot_hw_cfg->power_clk[AIC_CARD].power_rail.VDD_SOC	=  AIC_POWER_VDD_SOC;

	boot_hw_cfg->power_clk[ZG_VG1000].gfx.GCLK   			=  ZHAOGE_GFX_GCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].video.ECLK 			=  ZHAOGE_Video_ECLK;
	boot_hw_cfg->power_clk[ZG_VG1000].video.DCLK 			=  ZHAOGE_Video_DCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].video.VDSPCLK 		=  ZHAOGE_Video_VDSPCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].video.VCLK	 		=  ZHAOGE_Video_VCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].ddr.RUCLK	 			=  ZHAOGE_DDR_RUCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].ddr.LUCLK	 			=  ZHAOGE_DDR_LUCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].soc.CEDARCLK			=  ZHAOGE_SOC_CEDARCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].soc.LCCLK	    		=  ZHAOGE_SOC_LCCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].soc.RCCLK	    		=  ZHAOGE_SOC_RCCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].soc.BCCLK	    		=  ZHAOGE_SOC_BCCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].ai.OCLK	    		=  ZHAOGE_AI_OCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].ai.ODSPCLK	    	=  ZHAOGE_AI_ODSPCLK;
	boot_hw_cfg->power_clk[ZG_VG1000].dp.PIXCLK_DP			=  ZHAOGE_DP_PIXCLK_DP;
	boot_hw_cfg->power_clk[ZG_VG1000].power_rail.VDD_GFX	=  ZHAOGE_POWER_VDD_GFX;
	boot_hw_cfg->power_clk[ZG_VG1000].power_rail.VDD_VID	=  ZHAOGE_POWER_VDD_VID;
	boot_hw_cfg->power_clk[ZG_VG1000].power_rail.VDD_SOC	=  ZHAOGE_POWER_VDD_SOC;

	boot_hw_cfg->power_clk[ZG_VG600].gfx.GCLK   			=  ZHAOGE_VG600_GFX_GCLK;
	boot_hw_cfg->power_clk[ZG_VG600].video.ECLK 			=  ZHAOGE_VG600_Video_ECLK;
	boot_hw_cfg->power_clk[ZG_VG600].video.DCLK 			=  ZHAOGE_VG600_Video_DCLK;
	boot_hw_cfg->power_clk[ZG_VG600].video.VDSPCLK 			=  ZHAOGE_VG600_Video_VDSPCLK;
	boot_hw_cfg->power_clk[ZG_VG600].video.VCLK	 			=  ZHAOGE_VG600_Video_VCLK;
	boot_hw_cfg->power_clk[ZG_VG600].ddr.RUCLK	 			=  ZHAOGE_VG600_DDR_RUCLK;
	boot_hw_cfg->power_clk[ZG_VG600].ddr.LUCLK	 			=  ZHAOGE_VG600_DDR_LUCLK;
	boot_hw_cfg->power_clk[ZG_VG600].soc.CEDARCLK			=  ZHAOGE_VG600_SOC_CEDARCLK;
	boot_hw_cfg->power_clk[ZG_VG600].soc.LCCLK	    		=  ZHAOGE_VG600_SOC_LCCLK;
	boot_hw_cfg->power_clk[ZG_VG600].soc.RCCLK	    		=  ZHAOGE_VG600_SOC_RCCLK;
	boot_hw_cfg->power_clk[ZG_VG600].soc.BCCLK	    		=  ZHAOGE_VG600_SOC_BCCLK;
	boot_hw_cfg->power_clk[ZG_VG600].ai.OCLK	    		=  ZHAOGE_VG600_AI_OCLK;
	boot_hw_cfg->power_clk[ZG_VG600].ai.ODSPCLK	    		=  ZHAOGE_VG600_AI_ODSPCLK;
	boot_hw_cfg->power_clk[ZG_VG600].dp.PIXCLK_DP			=  ZHAOGE_VG600_DP_PIXCLK_DP;
	boot_hw_cfg->power_clk[ZG_VG600].power_rail.VDD_GFX		=  ZHAOGE_VG600_POWER_VDD_GFX;
	boot_hw_cfg->power_clk[ZG_VG600].power_rail.VDD_VID		=  ZHAOGE_VG600_POWER_VDD_VID;
	boot_hw_cfg->power_clk[ZG_VG600].power_rail.VDD_SOC		=  ZHAOGE_VG600_POWER_VDD_SOC;


	VASTAI_PCI_DBG(priv, 0,"%s boot_hw_cfg->power_cfg=%d\n",__func__,boot_hw_cfg->power_cfg);
	VASTAI_PCI_DBG(priv, 0,"%s boot_hw_cfg->bar4_5_size_cfg=%d\n",__func__,boot_hw_cfg->bar4_5_size_cfg);
	VASTAI_PCI_DBG(priv, 0,"%s boot_hw_cfg->d2h_by_atu_or_gart=%d\n",__func__,boot_hw_cfg->d2h_by_atu_or_gart);
	ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR2, (BOOT_HW_CFG_ADDR - 0x800000), boot_hw_cfg, sizeof(struct boot_hw_cfg));
	if (ret < 0)
		VASTAI_PCI_ERR(priv, 0,
				"%s [write] 0x%x, len=%ld, error:%d\n",
				__func__,(BOOT_HW_CFG_ADDR - 0x800000) , sizeof(struct boot_hw_cfg), ret);
	kfree(boot_hw_cfg);
}

static int vastai_pci_boot_die(struct vastai_pci_info *priv, u8 die_id)
{
	int ret = 0;
	u32 reg_val;

	VASTAI_PCI_DBG(priv, die_id, "pci_boot_die To download SW FW, size=%#lx for die_id=%d\n", VASTAI_FW_HEX_SIZE, die_id);

	ret = vastai_pci_set_current_die(priv, die_id);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id, "pci_boot_die set current die error\n");
		return ret;
	}

	/* polling smcu.bootstage=1, smcu is ready for downloading bl1. */
	ret = vastai_pci_poll_smcu_stage(priv, die_id, VASTAI_SMCU_BOOT_STAGE_BL1_READY);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
					"pci_boot_die poll smcu stage[%d] error\n",
					VASTAI_SMCU_BOOT_STAGE_BL1_READY);
		return ret;
	}

	/* check ddr training status */
	ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON51 - PCIE_WRAPPER_ADDR), &reg_val, 4);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_boot_die read PWRAP_COMMON51 failed.\n");
		return ret;
	}

	if (0 != (reg_val & 0b1111000)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "DDR training failed, reg addr=%#X, reg val=%#X\n", PWRAP_COMMON51, reg_val);
		return -EIO;
	}

	VASTAI_PCI_DBG(priv, die_id, "pci_boot_die, die_id=%d, to download bl1_hex.\n", die_id);

	/* download bl1 code to csram. */
	ret = vastai_process_download_request(priv, die_id,
					      BL1_PATH, vastai_bl1_hex_buf,
					      VASTAI_BL1_HEX_SIZE, VASTAI_BL1_DL_ADDR);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_boot_die download bl1 error, size=%ld, addr=%d\n",
			       VASTAI_BL1_HEX_SIZE,
			       VASTAI_BL1_DL_ADDR);
		return ret;
	}

	/* PCIe host driver set host.bootstage=1 and addr of bl1 run. */
	ret = vastai_pci_set_host_stage(priv, die_id, VASTAI_HOST_BOOT_STAGE_BL1_DONE);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_boot_die set host stage[%d] error\n",
			       VASTAI_HOST_BOOT_STAGE_BL1_DONE);
		return ret;
	}

	VASTAI_PCI_INFO(priv, die_id, "pci_boot_die BL1 downloaded finish!\n");

	/* polling smcu.bootstage=3, smcu is ready for downloading fw. */
	ret = vastai_pci_poll_smcu_stage(priv, die_id, VASTAI_SMCU_BOOT_STAGE_FW_READY);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_boot_die poll smcu stage[%d] error\n",
			       VASTAI_SMCU_BOOT_STAGE_FW_READY);
		return ret;
	}

	init_boot_hw_cfg(priv);

#ifdef CONFIG_VASTAI_PCI_ONLY_DOWNLOAD_BL1
	return ret;
#endif

    ret = vastai_process_download_request(priv, die_id,
                          FW_PATH, vastai_fw_hex_buf,
                          VASTAI_FW_HEX_SIZE, VASTAI_FW_DL_ADDR);
    if (ret < 0) {
        VASTAI_PCI_ERR(
            priv, die_id,
            "pci_boot_die download fw error, size=%ld, addr=%lld\n",
            VASTAI_FW_HEX_SIZE,
            VASTAI_FW_DL_ADDR);
        return ret;
    }

	/* PCIe host driver set host.bootstage=2 and addr of fw download. */
	ret = vastai_pci_set_host_stage(priv, die_id, VASTAI_HOST_BOOT_STAGE_FW_DONE);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_boot_die set host stage[%d] error\n",
			       VASTAI_HOST_BOOT_STAGE_FW_DONE);
		return ret;
	}

    VASTAI_PCI_DBG(priv, die_id, "pci_boot_die FW downloaded done. Wait for VA_SMCU_BOOT_STAGE_FW_BOOT_UP!!!!\n");

	/* polling smcu.bootstage=4, to check smcu fw boot up successfully */
	ret = vastai_pci_poll_smcu_stage(priv, die_id, VASTAI_SMCU_BOOT_STAGE_FW_BOOT_UP);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_boot_die poll smcu stage[%d] error\n",
			       VASTAI_SMCU_BOOT_STAGE_FW_BOOT_UP);
		return ret;
	}

	/* PCIe host driver set host.bootstage and addr of fw download */
	ret = vastai_pci_set_host_stage(priv, die_id, VASTAI_HOST_BOOT_STAGE_RCV_FW_BOOT_UP_ACK);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, die_id,
			       "pci_boot_die set host stage[%d] error\n",
			       VASTAI_HOST_BOOT_STAGE_RCV_FW_BOOT_UP_ACK);
		return ret;
	}

	VASTAI_PCI_INFO(priv, die_id, "pci_boot_die PKG[%d]: all the FW downloaded & booted up.\n", die_id);

	return ret;
}

int vastai_pci_boot_fw(struct vastai_pci_info *priv)
{
	int ret = 0;
	struct pci_dev *pdev = priv->dev;
	u32 reg_val = 0;

	if(priv->is_vf) {
		VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_boot_fw devfn=%d is a VF. No ability to reboot.\n", pdev->devfn);
		return -EIO;
	}

	VASTAI_PCI_INFO(priv, VASTAI_DIE0, "pci_boot_fw start. To wait for NOC being ready with sleeping 2s.\n");
	ssleep(2);

	ret = va_config_initial_bars(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_boot_fw, ret=%d\n", ret);
		return -EIO;
	}

	priv->boot_stage = VASTAI_HOST_BOOT_STAGE_INIT;

	ret = vastai_pci_poll_smcu_stage(priv, VASTAI_DIE0, VASTAI_SMCU_BOOT_STAGE_BL1_READY);
	if (ret) {
		VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_boot_fw poll stage[%d] error, %d\n", VASTAI_SMCU_BOOT_STAGE_BL1_READY, ret);
		return -EIO;
	}

	/* check ddr training status */
	ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON51 - PCIE_WRAPPER_ADDR), &reg_val, 4);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_boot_fw read PWRAP_COMMON51 failed.\n");
		return ret;
	}
	if (0 != (reg_val & 0b1111000)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "DDR training failed, reg addr=%#X, reg val=%#X\n", PWRAP_COMMON51, reg_val);
		return -EIO;
	}

	ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON_HOST_ONLY1 - PCIE_WRAPPER_ADDR), &reg_val, 4);
	reg_val &= 0xFFFFFFFE;
	ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON_HOST_ONLY1 - PCIE_WRAPPER_ADDR), &reg_val, 4);

	ret = vastai_pci_boot(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_boot_fw, ret=%d\n", ret);
		return -EIO;
	}

	ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON_HOST_ONLY1 - PCIE_WRAPPER_ADDR), &reg_val, 4);
	reg_val |= 0x1;
	ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON_HOST_ONLY1 - PCIE_WRAPPER_ADDR), &reg_val, 4);

	vastai_pci_state_boot_done(priv);
	vastai_pci_sync_state(priv, VASTAI_NORMAL_STATE);

	VASTAI_PCI_INFO(priv, VASTAI_DIE0, "pci_boot_fw  done!!\n");

	return ret;
}


int vastai_pci_reboot(struct vastai_pci_info *priv)
{
	int ret = 0;
	struct pci_dev *pdev = priv->dev;

	if(priv->is_vf) {
		VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_reboot devfn=%d is a VF. No ability to reboot.\n", pdev->devfn);
		return -EIO;
	}

	vastai_pcie_bw_test_cancel(priv);

	VASTAI_PCI_INFO(priv, VASTAI_DIE0, "pci_reboot Start. To wait for NOC being ready with sleeping 2s.\n");
	vastai_pci_reinitial(priv, true);
	ssleep(2);

	sdma_reset(priv);

	vastai_pci_boot_fw(priv);

	VASTAI_PCI_INFO(priv, VASTAI_DIE0, "pci_reboot  done!!\n");

	return ret;
}

// To get the boot lock named PWRAP_COMMON_LOCK0 for current PF. If non 0, succeed to get the lock. otherwise 0: failed.
static int vastai_poll_boot_lock(struct vastai_pci_info *priv)
{
	int ret = 0, poll_cnt = 0, err_cnt = 0;
	u32 val = 0;
	u64 reg_addr  = PWRAP_COMMON_LOCK0;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "host boot drv stage:%d. PF[%d] will try getting the common boot lock.\n",  priv->boot_stage, priv->dev->devfn);

	do {
        ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON_LOCK0 - PCIE_WRAPPER_ADDR), &val, 4);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "poll_boot_lock [read] 0x%llx, error:%d, failed to get the COMMON_LOCK0.\n",
				       reg_addr, ret);
			break;
		}

        if(val == PCIE_FAULT_CODE) {
            if(err_cnt < 2)
			    VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "poll_boot_lock: ERROR PCIE_FAULT_CODE, sleep then read again.\n0x%llx, reg.val=%#X\n", reg_addr, val);

            err_cnt++;
            msleep(VASTAI_BOOT_POLL_SLEEP_MS);
            continue;
        }

		if (val > 0) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%ds later, PF%d gets common boot lock successfully. lock val=%#x, err_cnt=%d\n", (poll_cnt / 1000), priv->dev->devfn, val, err_cnt);
			break;
		}

		if (poll_cnt++ > VASTAI_BOOT_LOCK_POLL_SLEEP_CNT) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"poll_boot_lock poll common lock0[%#llx], but lock still %d. %ds timeout!\n",
				reg_addr, val,
				(VASTAI_BOOT_POLL_SLEEP_CNT / 1000));
			return -ETIMEDOUT;
		}

		msleep(VASTAI_BOOT_POLL_SLEEP_MS);
	} while (1);
    
    return ret;
}


// To release the common boot lock by writing a initial val to the lock.
// Make sure current PF gets the boot lock before calling the func.
static int vastai_release_boot_lock(struct vastai_pci_info *priv)
{
    int ret = 0;
    u32 reg_val  = PWRAP_COMMON_LOCK0_VAL;
    
    ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON_LOCK0 - PCIE_WRAPPER_ADDR), &reg_val, 4);

    VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "release_boot_lock ret:%d\n", ret);
    
    return ret;
}


int vastai_pci_is_boot_done(struct vastai_pci_info *priv)
{
	union smcu_boot_stage smcu_reg;
	union host_boot_stage host_reg;
	int err_cnt = 0;

	do {
		vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (VASTAI_SMCU_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(smcu_reg.val), 4);
		vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (VASTAI_HOST_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &(host_reg.val), 4);
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				"pci_is_boot_done smcu boot stage[0x%x], host boot stage[0x%x], stage:%d, err_cnt:%d\n",
				smcu_reg.val, host_reg.val, priv->boot_stage, err_cnt);

		if(smcu_reg.val == PCIE_FAULT_CODE || host_reg.val == PCIE_FAULT_CODE) {
			if(err_cnt < 2)
				VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_is_boot_done: ERROR PCIE_FAULT_CODE, sleep then read again.\nhost_bval:%#x, smcu_bval:%#x\n", host_reg.val, smcu_reg.val);

			err_cnt++;
			msleep(VASTAI_BOOT_POLL_SLEEP_MS);
			continue;
		}

		if (smcu_reg.bit.stage == VASTAI_SMCU_BOOT_STAGE_ALLCORE_READY && 
			host_reg.bit.stage >= VASTAI_HOST_BOOT_STAGE_RCV_PCIE_INIT_READY_ACK) {
			return TRUE;
		}

		break;
	} while (1);

	return FALSE;
}

int vastai_pci_boot(struct vastai_pci_info *priv)
{
	u8 die_num, i, pre_die;
	int ret = 0;
	union die_index_data die_index_temp = {0};
	struct priv_hw_cfg *priv_hw_cfg = NULL;

	/* To get the common boot lock for current PF */
	vastai_poll_boot_lock(priv);

	if (vastai_pci_is_boot_done(priv))
	{
		/* reinit software manage source */
		priv->die_num = vastai_pci_get_die_num(priv);

		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_boot WARNING boot already done. Drv will not boot again for die_num=%d\n", priv->die_num);

		for (i = VASTAI_DIE0; i < priv->die_num; i++) {
			union die_index_data temp = {.dev_id = priv->dev_id,
										.die_id = i,
										.RESV = 0};
			priv->dies[i].die_index = temp.val;
		}

		goto UNLOCK;
	}

	/* set this PF as boot PF of the card */
	priv->is_boot_pf = 1;

	die_index_temp.dev_id = priv->dev_id;
	die_index_temp.die_id = 0;
	die_index_temp.RESV = 0;
	priv->dies[0].die_index = die_index_temp.val;
	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "die0: die_index=%d, sizeof(va_pci_info)=%ld Bytes\n", priv->dies[0].die_index, sizeof(struct vastai_pci_info));

	/* boot die0 */
	ret = vastai_pci_boot_die(priv, VASTAI_DIE0);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_boot die[%d] boot error\n", VASTAI_DIE0);
		goto UNLOCK;
	}

	// get die number from die0
	die_num = vastai_pci_get_die_num(priv);
	if (die_num == 1)
		goto OUT;

	// only boot slave pkg when reinit rst slave
	if(priv->priv_hw_cfg != NULL) {
		priv_hw_cfg = priv->priv_hw_cfg;
		if(priv_hw_cfg->sys_cfg.pkg_id == SLAVE_PKG) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s\n", "Boot slavepkg separately...");
			goto OUT;
		}
	}

#ifdef CONFIG_VASTAI_PCI_FORCE_SINGLE_DIE
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "Force Single Die Mode\n");
	goto OUT;
#endif

	/* boot other dies */
	for (i = VASTAI_DIE1; i < die_num; i++) {
		pre_die = i - 1;

		die_index_temp.dev_id = priv->dev_id;
		die_index_temp.die_id = i;
		die_index_temp.RESV = 0;
		priv->dies[i].die_index = die_index_temp.val;

#ifdef CONFIG_ENV_EMULATOR
		// poll master pkg linkup info with host RC
		vastai_pci_poll_die_linkup(priv, pre_die);
#endif

#ifdef CONFIG_ENV_SILICON
		msleep(2000);
#endif

		// poll SMCU FW of master pkg whether switching boot slave bar atu is ready.
		ret = vastai_pci_poll_master_atu_state(priv, pre_die);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					"pci_boot failed to boot slave PKG[%d] due to getting master pkg switch boot slave ATU TIMEOUT.\n", i);
			goto UNLOCK;
		}

		ret = vastai_pci_boot_die(priv, i);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_boot die[%d] boot error\n", i);

			goto UNLOCK;
		}

		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "pci_boot wait for PKG[%d] VA_SMCU_BOOT_STAGE_PCIE_INIT_READY!!!!\n", i);
		/* wait & check Slave PKG smcu PCIe init done */
		ret = vastai_pci_poll_smcu_stage(priv, i, VASTAI_SMCU_BOOT_STAGE_PCIE_INIT_READY);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, pre_die, "pci_boot poll PKG[%d] smcu stage[%d] error\n", i, VASTAI_SMCU_BOOT_STAGE_PCIE_INIT_READY);
		}
		// notifiy slave die: ACK
		vastai_pci_set_host_stage(priv, i,  VASTAI_HOST_BOOT_STAGE_RCV_PCIE_INIT_READY_ACK);

		// notify master SMCU that slave pkg init done.  //FIXME: To check here.
		vastai_pci_notify_master(priv, pre_die);
	}

OUT:
    // 1st stage wait: be sure to msleep, master die switch ATU to PCIE_INIT_READY done if have 2 pkgs.
    msleep(100);
    /* wait & check Mask PKG smcu PCIe init done */
    ret = vastai_pci_poll_smcu_stage(priv, VASTAI_DIE0, VASTAI_SMCU_BOOT_STAGE_PCIE_INIT_READY);
    if (ret < 0) {
        VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_boot poll smcu stage[%d] error\n", VASTAI_SMCU_BOOT_STAGE_PCIE_INIT_READY);
        goto UNLOCK;
    }

	vastai_pci_set_host_stage(priv, VASTAI_DIE0, VASTAI_HOST_BOOT_STAGE_RCV_PCIE_INIT_READY_ACK);
//	priv->boot_stage = VASTAI_HOST_BOOT_STAGE_READY;

    // 2nd stage wait: Block & wait for all mcus being ready
    ret = vastai_pci_poll_smcu_stage(priv, VASTAI_DIE0, VASTAI_SMCU_BOOT_STAGE_ALLCORE_READY);
    if (ret < 0) {
        VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_boot poll smcu stage[%d] error\n", VASTAI_SMCU_BOOT_STAGE_ALLCORE_READY);
    }

UNLOCK:
	// Release the boot lock for current PF
	vastai_release_boot_lock(priv);

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "----pci_boot done ----\n");

	return ret;
}

void vastai_pci_boot_wq(struct work_struct *work)
{
	struct delayed_work *delayed_work = to_delayed_work(work);
	struct vastai_pci_info *priv = container_of(delayed_work, struct vastai_pci_info, boot_wq);
	vastai_pci_boot(priv);
}

#define XSPI_RW_UNIT	(0x100000U)	// 1MB
#define VASTAI_PCIE_WAIT_XSPI_MSG1_TIME (350000)
static int vastai_wait_rw_xspi(struct vastai_pci_info *priv)
{
	int ret = 0;
	int i;

	/* wait completion int from smcu */
	if(wait_for_completion_timeout(&(priv->rw_xspi), msecs_to_jiffies(VASTAI_PCIE_WAIT_XSPI_MSG1_TIME)) <=0) {
		VASTAI_PCI_ERR(priv, 0,
				"wait_rw_xspi completion_timeout\n");
		return -9;
	}

	reinit_completion(&(priv->rw_xspi));
	if (priv->rw_xspi_msg != 0) {
		/* get error index */
		for (i = 0; i < 32; i++) {
			if ((priv->rw_xspi_msg >> i) & 0x1)
				break;
		}
		VASTAI_PCI_ERR(
			priv, DUMMY_DIE_ID,
			"rw xspi result is 0x%x (1: erase err; 2: read err; 3: write err; 4: unlock err; 5: lock err; 6: para err; 7: chksum err; 8: chktype err)\n",
			i+1);
		return -(i+1);
	}

	return ret;
}

static int send_xspi_flash_cmd(struct vastai_pci_info *priv, struct pcie_transfer_cmd *ptrans)
{
	int ret;

	ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF, ptrans, 0);
	if (ret != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s (2/5) write flash cmd err=%d\n",
			       __func__, ret);
		return ret;
	}

	ret = vastai_wait_rw_xspi(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s wait flash compltion failed %d\n",
				__func__, ret);
	}
	return ret;
}

int xspi_dma_read(struct vastai_pci_info *priv, u8 *data,
				size_t size, u64 fw_addr)
{
    struct vastai_dmadesc desc;
    union core_bitmap core_id = {.val = 0}; /* no core need be tirgger */
    int ret = 0;

    if (size > 0x100000) // 1M
    {
        VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID,"The transfer data size cannot exceed %#X!\n", 0x100000);
        return -EINVAL;
    }

    desc.is_src_dma_addr        = 0;
    desc.is_host_to_dev         = 0; // read
    desc.is_src_not_user_mem    = 1;
    desc.host_addr.dma_addr     = 0;
    desc.host_addr.vir_addr     = (void *)data;
    desc.dev_addr               = fw_addr;
    desc.dma_lenth              = size;
    ret = vastai_pci_dma_transfer_sync(priv, 0, core_id, &desc, 1);
    if (ret) {
        VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID, "%s: dma transfer fail, ret = %d\n", __func__, ret);
        return -ENXIO;
    }

    return ret;
}

int xspi_dma_write(struct vastai_pci_info *priv, const u8 *data,
				size_t size, u64 fw_addr)
{
	struct vastai_dmadesc desc;
	union core_bitmap core_id = {.val = 0}; /* no core need be tirgger */
	int ret = 0;

	if (size > 0x100000) // 1M
	{
		VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID,"The transfer data size cannot exceed %#X!\n", 0x100000);
				return -EINVAL;
	}

	desc.is_src_dma_addr		= 0;
	desc.is_host_to_dev 		= 1;	// write
	desc.is_src_not_user_mem	= 1;
	desc.host_addr.dma_addr 	= 0;
	desc.host_addr.vir_addr		= (void *)data;
	desc.dev_addr				= fw_addr;
	desc.dma_lenth				= size;
	ret = vastai_pci_dma_transfer_sync(priv, 0, core_id, &desc, 1);
	if (ret) {
		VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID, "%s: dma transfer fail, ret = %d\n", __func__, ret);
		return -ENXIO;
	}

	return ret;
}

int vastai_pci_xspi_get_bin_mode(const u8 *__data, size_t size, u8 partition, u8 *mode)
{
	u32 pos1 = 0;			// partition2
	u32 pos2 = 1024*1024;	// partition3 & 4
	u32 appver_offset = 64 + 64 + 0x1ffa0;
	u64 addr_ver = 0;
	u64 addr_vernum = 0;
	u32 tmp1 = 0;
	u32 tmp2 = 0;
	int ret = 0;

	if (partition == 1 || partition == 6) {
		// vbios & pmcu bin
		*mode = 2;
		return ret;
	} else if (partition == 2) {
		// bl0 bin
		addr_ver = ((u64)__data)+pos1+appver_offset;
		addr_vernum = addr_ver + 56;
	} else if (partition == 3 || partition == 4 || partition == 5) {
		// base & full & all bin
		addr_ver = ((u64)__data)+pos2+appver_offset;
		addr_vernum = addr_ver + 56;
	}
	tmp1 = *(u32 *)addr_ver;
	tmp2 = *(u32 *)addr_vernum;

	if (tmp1 == 0)
		*mode = 1;
	else if (tmp2 >= 0x02000000)
		*mode = 2;
	else
		ret = -ENOENT;

	return ret;
}

int vastai_pci_flash_xspi_bin(struct vastai_pci_info *priv, u8 *path_name, const void *bin_buf,
			      size_t buf_size, u8 partition, u8 *mode)
{
	int ret = 0;
	int request_ret = 0;
	const struct firmware *fw_entry = NULL;
	const u8 *__data = bin_buf;
	size_t __size = buf_size;
	size_t __size_bk = 0;
	const u8 *__tail_data = NULL;
	size_t bytes = 0;
	size_t offset = 0;
	struct pcie_transfer_cmd trans;
	u8 bin_mode = 0;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_bin (1/5) start.\n");

	if (path_name)
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_bin path_name=%s, partition=%d\n",
				path_name, partition);
	else
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_bin partition=%d\n",
				partition);

	/* most priority */
	if (path_name) {
		request_ret = request_firmware(&fw_entry, path_name, &(priv->dev->dev));
		if (request_ret != 0)
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"pci_flash_xspi_bin ret[%d]: can't find fw/bl from path[/lib/firmware/%s], it will download fw/bl from ko file\n",
					request_ret, path_name);
	}
	/* get fw from /lib/firmarw success */
	if (!request_ret && fw_entry) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"pci_flash_xspi_bin get fw/bl from path[/lib/firmware/%s] success, to download fw/bl from this path\n",
					path_name);
		__data = fw_entry->data;
		__size = fw_entry->size;
	}
	/* Not get any fw */
	if (!__data) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_flash_xspi_bin get fw/bl failed! path_name %s, data_addr %p\n",
				path_name, __data);
		return -ENOENT;
	}
	/* get max size */
	__size_bk = __size;
	__tail_data = __data + __size;

	ret = vastai_pci_xspi_get_bin_mode(__data, __size, partition, &bin_mode);
	if (ret != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_flash_xspi_bin get bin mode failed\n");
		return ret;
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_bin: bin_mode=[%d]\n", bin_mode);

	while (__data < __tail_data) {
		/* download xspi bin to ddr as wr_unit each time */
		bytes = (__size > XSPI_RW_UNIT) ? XSPI_RW_UNIT : __size;
		ret = xspi_dma_write(priv, __data, bytes, XSPI_FW_DDR_DL_ADDR);
		if (ret < 0) {
			VASTAI_PCI_ERR(
				priv, DUMMY_DIE_ID,
				"pci_flash_xspi_bin (2/5) download xspi error, offset=0x%lx, size=%ld, ddr_addr=0x%lx\n",
				offset, bytes, XSPI_FW_DDR_DL_ADDR);
			goto OUT;
		}

		/* send flash xspi cmd to smcu */
		trans.w0.s_data0.optcode = SMCU_XSPI_FW_FLASH;
		trans.w0.s_data0.rev0 = partition;
		trans.w0.s_data0.rev1 = bin_mode;
		trans.w1.data1 = bytes;
		trans.w2.data2 = offset;
		trans.w3.data3 = __size_bk;

		ret = send_xspi_flash_cmd(priv, &trans);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					"pci_flash_xspi_bin (2/5) write flash cmd err=%d\n",
					ret);
			goto OUT;
		}

		/* prepare for the next flash */
		__data += bytes;
		__size -= bytes;
		offset += bytes;
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_bin (2/5) write successfully.\n");
	*mode = bin_mode;

	dmi_set_device(priv);

OUT:
	/* we need free fw_entry */
	if (fw_entry)
		release_firmware(fw_entry);

	return ret;
}

int vastai_pci_read_xspi(struct vastai_pci_info *priv, unsigned long addr,
                        void *buf, u32 len, u8 partition, u8 mode)
{
	int ret = 0;
	u8 *__data = (u8 *)buf;
	size_t __size = len;
	size_t __size_bk = len;
	const u8 *__tail_data = __data + __size;
	size_t bytes = 0;
	size_t offset = 0;
	struct pcie_transfer_cmd trans;

	while (__data < __tail_data) {
		bytes = (__size > XSPI_RW_UNIT) ? XSPI_RW_UNIT : __size;
		/* send read xspi cmd to smcu */
		trans.w0.s_data0.optcode = SMCU_XSPI_FW_READ;
		trans.w0.s_data0.rev0 = partition;
		trans.w0.s_data0.rev1 = mode;
		trans.w1.data1 = bytes;
		trans.w2.data2 = offset;
		trans.w3.data3 = __size_bk;

		ret = send_xspi_flash_cmd(priv, &trans);
		if (ret != 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_read_xspi send cmd failed%d\n",
				ret);
			return ret;
		}

		ret = xspi_dma_read(priv, __data, bytes, XSPI_FW_DDR_DL_ADDR);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					"pci_read_xspi [read] 0x%lx, len=%ld, error:%d\n",
					offset, bytes, ret);
			return ret;
		}

		/* prepare for the next read */
		__data += bytes;
		__size -= bytes;
		offset += bytes;
	}

	return ret;
}

int vastai_pci_update_flag_xspi(struct vastai_pci_info *priv, u8 partition, u8 mode)
{
	int ret;
	struct pcie_transfer_cmd trans;

	/* send update xspi flag cmd to smcu */
	trans.w0.s_data0.optcode = SMCU_XSPI_FW_FLAG;
	trans.w0.s_data0.rev0 = partition;
	trans.w0.s_data0.rev1 = mode;

	ret = send_xspi_flash_cmd(priv, &trans);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_update_flag_xspi send cmd failed%d\n",
				ret);
	}
	return ret;
}

int vastai_pci_flash_xspi_fw(struct vastai_pci_info *priv, u8 *path_name, const void *bin_buf,
			      size_t buf_size, u8 partition)
{
	int ret = 0;
	int i = 0;
	const u8 *buf = bin_buf;
	u8 *buf_temp = 0;
	u8 mode = 0;

	if ((atomic_read(&priv->is_fw_updating) != 0)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_flash_xspi_fw priv->is_fw_updating=0x%x\n", atomic_read(&priv->is_fw_updating));
		return -ENODEV;
	}

	atomic_cmpxchg(&priv->is_fw_updating, 0, 1);

	/* a. update xspi fw */
	ret = vastai_pci_flash_xspi_bin(priv, path_name, bin_buf,
					buf_size, partition, &mode);
	if (ret) {
		goto failed;
	}

	/* read xspi fw */
	buf_temp = (u8 *)kvmalloc(buf_size, GFP_KERNEL);
	if (!buf_temp) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                        "pci_flash_xspi_fw malloc buf_temp failed!\n");
		ret = -ENOMEM;
		goto failed;
	}

	ret = vastai_pci_read_xspi(priv, 0, buf_temp,
					buf_size, partition, mode);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_flash_xspi_fw (3/5) read xspi flash err=%d\n",
			       ret);
		goto failed_1;
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_fw (3/5) read successfully.\n");

	/* c. verify data */
	for (i = 0; i < buf_size; i++) {
		if (buf[i] != *(buf_temp + i)) {
			VASTAI_PCI_ERR(
				priv, DUMMY_DIE_ID,
				"pci_flash_xspi_fw (4/5) verify xspi flash errpos=0x%x\n",
				i);
			goto failed_1;
		}
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_fw (4/5) verify successfully.\n");

	/* c. update flag */
	ret = vastai_pci_update_flag_xspi(priv, partition, mode);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_flash_xspi_fw (5/5) update flag xspi flash err=%d\n",
			       ret);
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_xspi_fw (5/5) update flag successfully, finished!\n");

failed_1:
	kvfree(buf_temp);
failed:
	atomic_set(&priv->is_fw_updating, 0);
	return ret;

}

int vastai_pci_flash_xspi(struct vastai_pci_info *priv, struct char_drv_info *dev, u8 partition)
{
	int ret = 0;
	char *path_name;
	const char *buf;
	size_t buf_size;

	if (partition == 1) {
		path_name = XSPI_FW_VBIOS_PATH;
		buf = vastai_xspi_fw_vbios_hex_buf;
		buf_size = VASTAI_XSPI_FW_VBIOS_HEX_SIZE;
	} else if (partition == 2) {
		path_name = XSPI_FW_BL0_PATH;
		buf = vastai_xspi_fw_bl0_hex_buf;
		buf_size = VASTAI_XSPI_FW_BL0_HEX_SIZE;
	} else if (partition == 3) {
		path_name = XSPI_FW_BASE_PATH;
		buf = vastai_xspi_fw_base_hex_buf;
		buf_size = VASTAI_XSPI_FW_BASE_HEX_SIZE;
	} else if (partition == 4) {
		path_name = XSPI_FW_FULL_PATH;
		buf = vastai_xspi_fw_full_hex_buf;
		buf_size = VASTAI_XSPI_FW_FULL_HEX_SIZE;
	} else if (partition == 5) {
		path_name = XSPI_FW_ALL_PATH;
		buf = vastai_xspi_fw_all_hex_buf;
		buf_size = VASTAI_XSPI_FW_ALL_HEX_SIZE;
	} else if (partition == 6) {
		path_name = XSPI_FW_PMCU_PATH;
		buf = vastai_xspi_fw_pmcu_hex_buf;
		buf_size = VASTAI_XSPI_FW_PMCU_HEX_SIZE;
	} else
		return ret;

	return vastai_pci_flash_xspi_fw(priv, path_name, buf, buf_size, partition);

}

int vastai_pci_flash_bmcu_bin(struct vastai_pci_info *priv, u8 *path_name, const void *bin_buf,
			      size_t buf_size)
{
	int ret = 0;
	int request_ret = 0;
	const struct firmware *fw_entry = NULL;
	const u8 *__data = bin_buf;
	size_t __size = buf_size;
	size_t __align = 0;
	u8 *buf_temp = 0;
	struct pcie_transfer_cmd trans;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_bmcu_bin (1/2) start.\n");

	/* most priority */
	if (path_name) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_bmcu_bin path_name=%s\n",
			path_name);
		request_ret = request_firmware(&fw_entry, path_name, &(priv->dev->dev));
		if (request_ret != 0)
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"pci_flash_bmcu_bin ret[%d]: can't find fw/bl from path[/lib/firmware/%s], it will download fw/bl from ko file\n",
					request_ret, path_name);
	}
	/* get fw from /lib/firmarw success */
	if (!request_ret && fw_entry) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"pci_flash_bmcu_bin get fw/bl from path[/lib/firmware/%s] success, to download fw/bl from this path\n",
					path_name);
		__data = fw_entry->data;
		__size = fw_entry->size;
	}
	/* Not get any fw */
	if (!__data) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_flash_bmcu_bin get fw/bl failed! path_name %s, data_addr %p\n",
				path_name, __data);
		return -ENOENT;
	}

	/* check data size as 128B align */
	__align = __size % 128;
	if (__align) {
		__align = 128 - __align;

		buf_temp = (u8 *)kvmalloc((__size + __align), GFP_KERNEL);
		if (!buf_temp) {
                VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                        "pci_flash_bmcu_bin malloc buf_temp failed!\n");
			return -ENOMEM;
		}
		memcpy(buf_temp, __data, __size);
		ret = xspi_dma_write(priv, buf_temp, (__size+__align), BMCU_FW_DDR_DL_ADDR);
	}
	else
		ret = xspi_dma_write(priv, __data, __size, BMCU_FW_DDR_DL_ADDR);

	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_flash_bmcu_bin (2/2) download bmcu error, size=%ld, ddr_addr=0x%lx\n",
			__size, BMCU_FW_DDR_DL_ADDR);
		goto OUT;
	}

	/* send flash bmcu cmd to smcu */
	trans.w0.s_data0.optcode = SMCU_BMCU_FW_FLASH;
	trans.w1.data1 = __size;

	ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF, &trans, 0);
	if (ret != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_flash_bmcu_bin flash bmcu cmd err=%d\n", ret);
		goto OUT;
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pci_flash_bmcu_bin (2/2) download bmcu successfully.\n");

OUT:
	/* we need free buf_temp */
	if (buf_temp)
		kvfree(buf_temp);
	/* we need free fw_entry */
	if (fw_entry)
		release_firmware(fw_entry);

	return ret;
}

int vastai_pci_flash_bmcu(struct vastai_pci_info *priv, struct char_drv_info *dev)
{
	int ret = 0;
	char *path_name;
	const char *buf;
	size_t buf_size;

	path_name = BMCU_FW_PATH;
	buf = vastai_bmcu_fw_hex_buf;
	buf_size = VASTAI_BMCU_FW_HEX_SIZE;

	if ((atomic_read(&priv->is_fw_updating) != 0)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_flash_bmcu priv->is_fw_updating=0x%x\n", atomic_read(&priv->is_fw_updating));
		return -ENODEV;
	}

	atomic_cmpxchg(&priv->is_fw_updating, 0, 1);

	/* a. update xspi fw */
	ret = vastai_pci_flash_bmcu_bin(priv, path_name, buf, buf_size);

	ssleep(6);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "upgrade BMCU by pcie finished!\n");

	atomic_set(&priv->is_fw_updating, 0);
	return ret;
}

int vastai_pci_switch_bmcu(struct vastai_pci_info *priv, struct char_drv_info *dev)
{
    int ret = 0;
	struct pcie_transfer_cmd trans;

	if ((atomic_read(&priv->is_fw_updating) != 0)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_switch_bmcu priv->is_fw_updating=0x%x\n", atomic_read(&priv->is_fw_updating));
		return -ENODEV;
	}

	atomic_cmpxchg(&priv->is_fw_updating, 0, 1);

	/* send switch bmcu cmd to smcu */
	trans.w0.s_data0.optcode = SMCU_BMCU_FW_SWITCH;
	ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF, &trans, 0);
	if (ret != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_switch_bmcu switch bmcu cmd err=%d\n", ret);
	}

	ssleep(6);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "degrade BMCU by pcie finished!\n");

	atomic_set(&priv->is_fw_updating, 0);
	return ret;
}


int vastai_pci_flash_bmcu_v2(struct vastai_pci_info *priv, const void *bin_buf, size_t buf_size)
{
	int ret = 0;
	const char *buf = bin_buf;
	char *buf_temp = 0;
	int i = 0;

	if ((atomic_read(&priv->is_fw_updating) != 0)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_flash_bmcu_v2 priv->is_fw_updating=0x%x\n", atomic_read(&priv->is_fw_updating));
		return -ENODEV;
	}

	atomic_cmpxchg(&priv->is_fw_updating, 0, 1);

	/* a. update xspi fw */
	ret = vastai_pci_flash_bmcu_bin(priv, 0, bin_buf, buf_size);

	ssleep(6);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "upgrade BMCU by pcie finished!\n");

	atomic_set(&priv->is_fw_updating, 0);

	/* b. download data verify */
	buf_temp = (char *)kvmalloc(buf_size, GFP_KERNEL);
	if (!buf_temp) {
                VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                        "pci_flash_bmcu_v2 malloc buf_temp failed!\n");
		return -ENOMEM;
	}

	ret = xspi_dma_read(priv, buf_temp, buf_size, BMCU_FW_DDR_DL_ADDR);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                        "pci_flash_bmcu_v2 [read] len=%ld, error:%d\n", buf_size, ret);
		goto out;
    }
	else
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "read back bmcu download data successfully.\n");

	/* c. verify data */
	for (i = 0; i < buf_size; i++) {
		if (buf[i] != *(buf_temp + i)) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_flash_bmcu_v2 data compare errpos=0x%x\n", i);
			break;
		}
	}
	if (i == buf_size) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "compare data successfully.\n");
	}

out:
	kvfree(buf_temp);
	return ret;

}

int vastai_pci_switch_bmcu_v2(struct vastai_pci_info *priv, const void *bin_buf, size_t buf_size)
{
	int ret = 0;
	struct pcie_transfer_cmd trans;

	if ((atomic_read(&priv->is_fw_updating) != 0)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_switch_bmcu_v2 priv->is_fw_updating=0x%x\n", atomic_read(&priv->is_fw_updating));
		return -ENODEV;
	}

	atomic_cmpxchg(&priv->is_fw_updating, 0, 1);

	/* send switch bmcu cmd to smcu */
	trans.w0.s_data0.optcode = SMCU_BMCU_FW_SWITCH;
	ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF, &trans, 0);
	if (ret != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_switch_bmcu_v2 switch bmcu cmd err=%d\n", ret);
	}

	ssleep(6);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "degrade BMCU by pcie finished!\n");

	atomic_set(&priv->is_fw_updating, 0);
	return ret;
}

static int vastai_wait_read_pmcu_boot_mode(struct vastai_pci_info *priv)
{
    int ret = 0;

    /* wait completion int from smcu */
    ret = wait_for_completion_timeout(&(priv->pmcu_boot_mode_done), msecs_to_jiffies(1000)); // 1000ms
    reinit_completion(&(priv->pmcu_boot_mode_done));
    if (ret <= 0) {
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "WARNING: wait_read_pmcu_boot_mode: wait completion is timeout\n");
        return -ETIMEDOUT;
    }

    if (priv->pmcu_boot_mode_msg != 0) {
        /* get pmcu boot mode which from drvier or flash */
        if (priv->pmcu_boot_mode_msg & BIT(2))
            VASTAI_PCI_INFO( priv, DUMMY_DIE_ID,
                    "PMCU firmware is loaded from flash\n");
        else if ((priv->pmcu_boot_mode_msg & BIT(1)) && (priv->pmcu_boot_mode_msg & BIT(0)))
            VASTAI_PCI_INFO( priv, DUMMY_DIE_ID,
                    "PMCU firmware is loaded from driver, and PMCU firmware in flash is bad and need to update\n");
        else
            VASTAI_PCI_INFO( priv, DUMMY_DIE_ID,
                    "PMCU firmware is loaded from driver\n");
        return 0;
    }

    return -EINVAL;
}

int vastai_pci_read_pmcu_boot_mode(struct vastai_pci_info *priv)
{
    int ret = 0;
    struct pcie_transfer_cmd trans;

    /* send cmd to get pmcu boot mode */
    trans.w0.s_data0.optcode = PMCU_BOOT_MODE_REQ;

    ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF, &trans, 0);
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_read_pmcu_boot_mode send cmd failed %d\n",
                ret);
        goto out;
    }

    ret = vastai_wait_read_pmcu_boot_mode(priv);
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_read_pmcu_boot_mode wait cmd failed %d\n",
            ret);
    }

out:
    return ret;
}
