#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/random.h>
#include <linux/dma-buf.h>
#include <linux/poll.h>

#include "vastai_pci.h"
#include "vastai_pci_test.h"
#include "vastai_udma_engine.h"
#include "vastai_dmabuf.h"
#include "vastai_state.h"

#define TCGETS         0x5401
#define TCSETS         0x5402
#define TCSETSW        0x5403
#define TCSETSF        0x5404
#define TIOCSERGSTRUCT 0x5458 /* For debugging only */
#define TIOCSERGETLSR  0x5459 /* Get line status register */
#define TIOCSERCONFIG  0x545

static int ctl_open(struct inode *inode, struct file *filp)
{
	struct vastai_file_info *file_info;
	struct vastai_ctl_file *ctl_file;

	file_info = to_vastai_file_info(inode->i_cdev);
	ctl_file = container_of(file_info, struct vastai_ctl_file, file_info);

	filp->private_data = ctl_file;
	return 0;
}

static int ctl_release(struct inode *inode, struct file *filp)
{
	return 0;
}

typedef int(*set_desc_func)(struct vastai_dmadesc*, struct dma_buf*,
				  struct kchar_cmd*,
				  int, u32, u32,
				  struct vastai_pci_info*);

int vastai_ioctl_dma_start_set_desc(struct vastai_dmadesc* desc, struct dma_buf *dmabuf,
						   struct kchar_cmd* kcmd,
						   int i, u32 done_size, u32 temp_size,
						   struct vastai_pci_info *pci_info)
{
	int ret = 0;
	union core_bitmap core_id = { .val = 0 };
	struct vastai_dma_buf *dm = get_vastai_dma_buf(dmabuf, i);

	desc->is_host_to_dev = !kcmd->dma_start_cmd.is_dev_to_host;
	desc->is_src_not_user_mem = 1;
	desc->is_src_dma_addr = 1;
	desc->dev_addr = kcmd->dma_start_cmd.axi_addr + done_size;
	desc->dma_lenth = temp_size;
	desc->host_addr.dma_addr = dm->dma_bus_addr;
	ret = vastai_pci_dma_transfer_sync(pci_info, kcmd->dma_start_cmd.die_index,
					   core_id, desc, 1);

	return ret;
}


int vastai_ioctl_dma_trans_set_desc(struct vastai_dmadesc* desc, struct dma_buf *dmabuf,
						   struct kchar_cmd* kcmd, int i, u32 done_size,
						   u32 temp_size, struct vastai_pci_info *pci_info)
{
	int ret = 0;

	desc[i].is_host_to_dev = !kcmd->dma_trans_cmd.is_dev_to_host;
	desc[i].is_src_not_user_mem = 0;
	desc[i].is_src_dma_addr = 0;
	desc[i].dev_addr = kcmd->dma_trans_cmd.axi_addr + done_size;
	desc[i].dma_lenth = temp_size;
	desc[i].host_addr.vir_addr = (void *)(kcmd->dma_trans_cmd.vir_addr + done_size);

	return ret;
}

int vastai_common_cut_dma_buf(u32 size, struct kchar_cmd* kcmd, struct vastai_dmadesc* desc,
                                        struct vastai_pci_info *pci_info, struct dma_buf *dmabuf,
                                        set_desc_func pSetDescFunc)
{
	int ret = 0;
	u32 done_size = 0;
	u32 i = 0;

	while (done_size < size) {
	u32 temp_size = (size - done_size) < VASTAI_MAX_DMA_BUF ?
			(size - done_size) : VASTAI_MAX_DMA_BUF;

	ret = pSetDescFunc(desc, dmabuf, kcmd, i, done_size, temp_size, pci_info);
	if (ret)
		return ret;

	i++;
	done_size += temp_size;
	}

	return ret;
}

static long ctl_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct vastai_ctl_file *ctl_info = (struct vastai_ctl_file *)filp->private_data;
	struct vastai_pci_info *pci_info = ctl_info->pci_info;
	struct kchar_cmd kcmd;
	int ret = 0;
	int state = 0;

	//printk("ctl_ioctl: id %d\n", cmd);

	state = vastai_get_pci_state(pci_info);
	if ((state != VASTAI_NORMAL_STATE) && (state != VASTAI_DEBUG_STATE))
	{
		VASTAI_PCI_ERR(pci_info, 0, "%s:pcie state error[%s]\n", __func__, state_name(state));
		return -ENODEV;
	}

	ret = vastai_pci_copy_from_user(&kcmd, (void *)arg, sizeof(kcmd));
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				"%s pci_copy_from_user is error\n", __func__);
		return -EIO;
	}

	switch (cmd) {

	case VASTAI_PCI_IOCTL_ALLOC: {
		struct dma_buf *dmabuf;
		struct vastai_buf_list *va_buf_list;
		if (iommu_is_enable(pci_info))
			dmabuf = vastai_alloc_dma_buf_sg(
					pci_info,
					kcmd.alloc_cmd.size);
		else
			dmabuf = vastai_alloc_dma_buf(
					pci_info,
					kcmd.alloc_cmd.size);
		if (dmabuf) {
			int fd = dma_buf_fd(dmabuf, O_CLOEXEC);
			kcmd.alloc_cmd.dma_buf_fd = fd;
			va_buf_list = (struct vastai_buf_list *)dmabuf->priv;
			kcmd.alloc_cmd.dma_addr_t = va_buf_list->buf_list[0]->dma_bus_addr;
		} else {
			kcmd.alloc_cmd.dma_buf_fd = -ENOMEM;
		}
		break;
	}
	case VASTAI_PCI_IOCTL_DMA_START: {
		struct dma_buf *dmabuf =
			dma_buf_get(kcmd.dma_start_cmd.dma_buf_fd);
		struct vastai_dmadesc desc;

		if(IS_ERR(dmabuf)){
			VASTAI_PCI_ERR(pci_info, 0,"%s, VA_PCI_IOCTL_DMA_START DMA_BUF is invalid\n", __func__);
			return PTR_ERR(dmabuf);
		}
		if (iommu_is_enable(pci_info)) {
			union core_bitmap core_id = { .val = 0 };
			struct vastai_dma_buf *dm =
				get_vastai_dma_buf(dmabuf, 0);
			u32 is_src_dma_addr = 1;
			u32 is_src_not_user_mem = 1;
			if (kcmd.dma_start_cmd.size > dm->size) {
				ret = -EINVAL;
				goto free_buf;
			}
			ret = vastai_sdma_transfer_sg(
				pci_info,
				kcmd.dma_start_cmd.die_index,
				core_id,
				dm,
				is_src_dma_addr,
				!kcmd.dma_start_cmd.is_dev_to_host,
				is_src_not_user_mem,
				kcmd.dma_start_cmd.axi_addr,
				kcmd.dma_start_cmd.size);
		} else {
			ret = vastai_common_cut_dma_buf(
				kcmd.dma_start_cmd.size,
				&kcmd,
				&desc,
				pci_info,
				dmabuf,
				vastai_ioctl_dma_start_set_desc);
		}
free_buf:
		dma_buf_put(dmabuf);
		break;
	}

	case VASTAI_PCI_IOCTL_GFX_VIDEO_SHARE: {
		struct dma_buf *dmabuf =
			dma_buf_get(kcmd.dma_gfx_video_cmd.dma_buf_fd);

		if(IS_ERR(dmabuf)){
			VASTAI_PCI_ERR(pci_info, 0,"%s, VA_PCI_IOCTL_GFX_VIDEO_SHARE DMA_BUF is invalid\n", __func__);
			return PTR_ERR(dmabuf);
		}

		ret = vastai_gfx_video_share_sync_pid(pci_info, kcmd.dma_gfx_video_cmd.pid, 
					dmabuf, kcmd.dma_gfx_video_cmd.dir, kcmd.dma_gfx_video_cmd.dev_addr);

		dma_buf_put(dmabuf);
		break;
	}

	case VASTAI_PCI_IOCTL_DMA_TRANS: {
		union core_bitmap core_id = { .val = 0 };
		u32 size = kcmd.dma_trans_cmd.length;
#if 0
#ifdef CONFIG_VASTAI_SDMA_CHECK
#ifndef DIAG_TEST_MODE
		if (!kcmd.dma_trans_cmd.is_dev_to_host && size & 0x7F)
		{
			VASTAI_PCI_ERR(pci_info, 0,"%s, cur size(%#x) is unaligned with 128B, return\n", __func__, size);
			return -EINVAL;
		}
#endif
#endif
#endif
		if (iommu_is_enable(pci_info) && !((u64)kcmd.dma_trans_cmd.vir_addr & 0x7F)) {
			ret =  vastai_sdma_transfer_by_uaddr(
				pci_info,
				kcmd.dma_trans_cmd.die_index,
				core_id,
				!kcmd.dma_trans_cmd.is_dev_to_host,
				kcmd.dma_trans_cmd.axi_addr,
				(u64)kcmd.dma_trans_cmd.vir_addr,
				size);
		} else {
			struct vastai_dmadesc *desc;
			u32 entry_count = DIV_ROUND_UP(size, VASTAI_MAX_DMA_BUF);

			if(!entry_count){
				VASTAI_PCI_ERR(pci_info, 0,"transfer size(0x%x) is invalid, return\n", size);
				return -EINVAL;
			}

			desc = kvmalloc(sizeof(struct vastai_dmadesc) * entry_count,
					GFP_KERNEL);
			if (!desc){
				VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
						"%s: desc alloc failed\n", __func__);
				return -ENOMEM;
			}

			vastai_common_cut_dma_buf(
				size,
				&kcmd,
				desc,
				pci_info,
				NULL,
				vastai_ioctl_dma_trans_set_desc);

			ret = vastai_sdma_transfer_sync_pid(
				pci_info,
				kcmd.dma_trans_cmd.die_index,
				core_id,
				desc,
				entry_count,
				kcmd.dma_trans_cmd.pid);

			kvfree(desc);
		}

		break;
	}

	case VASTAI_PCI_IOCTL_DMA_TRANS_SG: {
		union core_bitmap core_id = { .val = 0 };
		struct vastai_channel_buf *channel_buf;
		int i;
		int channel_num = kcmd.dma_trans_sg_cmd.channel_num;
		channel_buf = kvmalloc(sizeof(struct vastai_channel_buf) * channel_num, GFP_KERNEL);
		if (!channel_buf) {
			VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
					"%s: channel buf alloc failed\n", __func__);
			return -ENOMEM;
		}

		ret = vastai_pci_copy_from_user(
			channel_buf, (void*)kcmd.dma_trans_sg_cmd.channel_buf,
			sizeof(struct vastai_channel_buf) * channel_num);
		if (ret) {
			kvfree(channel_buf);
			return -1;
		}

		for (i = 0; i < channel_num; i++) {
			VASTAI_PCI_DBG(
				pci_info, kcmd.dma_trans_sg_cmd.die_index,
				"channel[%d], w:%d, h:%d, dst_w_off:%d, dst_h_off:%d, src_w_off:%d, src_h_off:%d\n",
				i, channel_buf[i].width, channel_buf[i].high,
				channel_buf[i].dst_width_offset,
				channel_buf[i].dst_high_offset,
				channel_buf[i].src_width_offset,
				channel_buf[i].src_high_offset);
		}

		vastai_dma_video(pci_info, kcmd.dma_trans_sg_cmd.die_index,
				 core_id, channel_buf, channel_num,
				 kcmd.dma_trans_sg_cmd.axi_addr,
				 kcmd.dma_trans_sg_cmd.pid);

		kvfree(channel_buf);
		break;
	}

	case VASTAI_PCI_IOCTL_DMA_FILL: {
		union core_bitmap core_id = { .val = 0 };

		ret = vastai_sdma_fill_data(
				pci_info,
				kcmd.dma_fill_cmd.die_index,
				core_id,
				kcmd.dma_fill_cmd.axi_addr,
				kcmd.dma_fill_cmd.data,
				kcmd.dma_fill_cmd.length);
		break;
	}

	case VASTAI_PCI_IOCTL_DMA_D2D: {
		u32 length = kcmd.dma_d2d_cmd.length;
		u64 src_addr = kcmd.dma_d2d_cmd.src_addr;
		u64 dst_addr = kcmd.dma_d2d_cmd.dst_addr;
		int pid = kcmd.dma_d2d_cmd.pid;
		ret = vastai_dev2dev_sdma_sync_pid(
				pci_info,
				kcmd.dma_d2d_cmd.die_index,
				src_addr,
				dst_addr,
				length,
				pid);
		break;
	}

	case VASTAI_PCI_IOCTL_DMA_TRANS_BY_VA: {
		union core_bitmap core_id = { .val = 0 };
		u32 size = kcmd.dma_trans_cmd.length;

		ret =  vastai_sdma_transfer_by_uaddr_withmmu(
			pci_info,
			kcmd.dma_trans_cmd.die_index,
			core_id,
			!kcmd.dma_trans_cmd.is_dev_to_host,
			kcmd.dma_trans_cmd.axi_addr,
			(u64)kcmd.dma_trans_cmd.vir_addr,
			size,
			kcmd.dma_trans_cmd.pid);
		break;
	}
	
	case VASTAI_PCI_IOCTL_DMA_TRANS_BY_SG: {
		union core_bitmap core_id = { .val = 0 };
		u32 size = kcmd.dma_trans_cmd.length;

		ret = vastai_sdma_transfer_by_sg(
			pci_info,
			kcmd.dma_trans_cmd.die_index,
			core_id,
			!kcmd.dma_trans_cmd.is_dev_to_host,
			kcmd.dma_trans_cmd.axi_addr,
			(u64)kcmd.dma_trans_cmd.vir_addr,
			size);
		break;
	}

	case VASTAI_PCI_IOCTL_DMA_S2M: {
		u32 length = kcmd.dma_s2m_cmd.length;
		u64 src_addr = kcmd.dma_s2m_cmd.src_addr;
		u64 dst_addr = kcmd.dma_s2m_cmd.dst_addr;
		int dir = kcmd.dma_s2m_cmd.direction;
		ret = vastai_salve2master_sdma_sync_pid(
				pci_info,
				dir,
				src_addr,
				dst_addr,
				length);

		break;
	}

	case VASTAI_PCI_IOCTL_PCIE_DMA_RAW_TRANS: {
		union core_bitmap core_id = { .val = 0 };
		u32 size = kcmd.dma_trans_cmd.length;
		ret =  vastai_pcie_dma_raw_transfer_by_uaddr(
			pci_info,
			kcmd.dma_trans_cmd.die_index,
			core_id,
			!kcmd.dma_trans_cmd.is_dev_to_host,
			kcmd.dma_trans_cmd.axi_addr,
			(u64)kcmd.dma_trans_cmd.vir_addr,
			size);

		break;
	}

	case VASTAI_PCI_IOCTL_PCIE_DMA_TRANS: {
		union core_bitmap core_id = { .val = 0 };
		u32 size = kcmd.dma_trans_cmd.length;
		if (iommu_is_enable(pci_info) && ((u64)kcmd.dma_trans_cmd.vir_addr%PAGE_SIZE == 0)) {
			ret =  vastai_sdma_transfer_by_uaddr(
				pci_info,
				kcmd.dma_trans_cmd.die_index,
				core_id,
				!kcmd.dma_trans_cmd.is_dev_to_host,
				kcmd.dma_trans_cmd.axi_addr,
				(u64)kcmd.dma_trans_cmd.vir_addr,
				size);
		} else {
			struct vastai_dmadesc *desc;
			u32 entry_count = DIV_ROUND_UP(size, VASTAI_MAX_DMA_BUF);

			if (entry_count){
				desc = kvmalloc(sizeof(struct vastai_dmadesc) * entry_count,
						GFP_KERNEL);
				if (!desc){
					VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
							"%s: desc alloc failed\n", __func__);
					return -ENOMEM;
				}
			}
			else{
				VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
					"%s: size is 0, return!\n", __func__);
				return -ENOMEM;
			}

			vastai_common_cut_dma_buf(
				size,
				&kcmd,
				desc,
				pci_info,
				NULL,
				vastai_ioctl_dma_trans_set_desc);

			ret = vastai_pci_dma_transfer_sync_pid(
				pci_info,
				kcmd.dma_trans_cmd.die_index,
				core_id,
				desc,
				entry_count,
				kcmd.dma_trans_cmd.pid);

			kvfree(desc);
		}

		break;
	}
	case VASTAI_PCI_IOCTL_GET_VA_PID: {
		kcmd.free_va_pid_cmd.va_pid = get_vastai_pid(pci_info);
		break;
	}

	case VASTAI_PCI_IOCTL_FREE_VA_PID: {
		ret = free_vastai_pid(pci_info, kcmd.free_va_pid_cmd.va_pid);
		break;
	}
	case VASTAI_PCI_IOCTL_MEMSET: {
		struct dma_buf *dmabuf = dma_buf_get(kcmd.dma_memset_cmd.fd);
		struct vastai_dma_buf *dm = get_vastai_dma_buf(dmabuf, 0);

		if(!dm){
			return -ENOMEM;
		}

		if(!dm->vir || !dm->dma_bus_addr){
			VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
				"%s: dma_alloc_coherent fail\n",__func__);
			goto ERROR_FREE_DM;
		}

		memset(dm->vir, kcmd.dma_memset_cmd.val, kcmd.dma_memset_cmd.size);

		ERROR_FREE_DM:
		dma_buf_put(dmabuf);
		break;
	}

	case VASTAI_PCI_IOCTL_DMA_CLOSE: {
		struct dma_buf *dmabuf = dma_buf_get(kcmd.dma_memset_cmd.fd);
		dma_buf_put(dmabuf);
		break;
	}

	case VASTAI_PCI_IOCTL_MMIO_WRITE: {
		struct dma_buf *dmabuf =
			dma_buf_get(kcmd.mmio_trans_cmd.dma_buf_fd);
		struct vastai_dma_buf *dm = get_vastai_dma_buf(dmabuf, 0);

		if(!dm){
			return -ENOMEM;
		}

		if(!dm->vir || !dm->dma_bus_addr){
			VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
				"%s: dma_alloc_coherent fail\n",__func__);
			return -ENOMEM;
		}

		ret = copy_from_user(dm->vir, (char __user *)kcmd.mmio_trans_cmd.vir_addr,
                       kcmd.mmio_trans_cmd.length);
		if (ret)
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
                       		"%s copy_from_user is error\n", __func__);
		ret = vastai_pci_mem_write(pci_info, pci_info->dies[0].die_index,
					kcmd.mmio_trans_cmd.axi_addr, dm->vir, kcmd.mmio_trans_cmd.length);
		if (ret < 0)
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
							"%s [mmio_write] 0x%llx, len=%d, error:%d\n",
							__func__, kcmd.mmio_trans_cmd.axi_addr, kcmd.mmio_trans_cmd.length, ret);
		else
			VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID,
							"%s [mmio_write] addr=0x%llx, "
							"len=%d\n",
							__func__, kcmd.mmio_trans_cmd.axi_addr, kcmd.mmio_trans_cmd.length);
		dma_buf_put(dmabuf);

		break;
	}

	case VASTAI_PCI_IOCTL_BOARD_TYPE:
		kcmd.board_type.type = vastai_get_board_type(pci_info);
		break;

	case VASTAI_PCI_IOCTL_CEDAR_CMD: {
		vastai_dev_cedar_send_cmd(pci_info,
				(void *)&kcmd.cedar_ioctl.cedar_info,
				pci_info->dev_id,
				0,
				(void*)kcmd.cedar_ioctl.cedar_res_info.hash);
		break;
	}

	case VASTAI_PCI_IOCTL_MMIO_READ: {
		struct dma_buf *dmabuf =
			dma_buf_get(kcmd.mmio_trans_cmd.dma_buf_fd);

		struct vastai_dma_buf *dm = get_vastai_dma_buf(dmabuf, 0);
		if(!dm){
			return -ENOMEM;
		}

		if(!dm->vir || !dm->dma_bus_addr){
			VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
				"%s: dma_alloc_coherent fail\n",__func__);
			return -ENOMEM;
		}

		ret = vastai_pci_mem_read(pci_info, pci_info->dies[0].die_index,
						kcmd.mmio_trans_cmd.axi_addr, dm->vir, kcmd.mmio_trans_cmd.length);
		if (ret < 0)
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
							"%s [mmio_read] 0x%llx, len=%d, error:%d\n",
							__func__, kcmd.mmio_trans_cmd.axi_addr, kcmd.mmio_trans_cmd.length, ret);
		else
			VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID,
							"%s [mmio_read] addr=0x%llx, "
							"len=%d\n",
							__func__, kcmd.mmio_trans_cmd.axi_addr, kcmd.mmio_trans_cmd.length);

		ret = copy_to_user((void *)kcmd.mmio_trans_cmd.vir_addr, dm->vir, kcmd.mmio_trans_cmd.length);
		if (ret)
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
							"%s [mmio_read] 0x%llx, len=%d, copy to user error:%d\n",
							__func__, kcmd.mmio_trans_cmd.axi_addr, kcmd.mmio_trans_cmd.length, ret);

		dma_buf_put(dmabuf);
		break;
	}

	case VASTAI_PCI_IOCTL_SMCU_READ:{
		ret = vastai_read_by_smcu(pci_info,kcmd.smcu_trans_cmd.axi_addr,4,&kcmd.smcu_trans_cmd.val);
		if (ret < 0)
			VASTAI_PCI_ERR(pci_info, pci_info->pkg_id,"ioctl smcu read addr[0x%llx] error:%d\n",
				kcmd.smcu_trans_cmd.axi_addr,ret);
		break;
	}

	case VASTAI_PCI_IOCTL_SMCU_WRITE:{
		ret = vastai_write_by_smcu(pci_info,kcmd.smcu_trans_cmd.axi_addr,kcmd.smcu_trans_cmd.val,4);
		if (ret < 0)
			VASTAI_PCI_ERR(pci_info, pci_info->pkg_id,"ioctl smcu write addr[0x%llx] error:%d\n",
				kcmd.smcu_trans_cmd.axi_addr,ret);
		break;
	}

#if 0
	case VASTAI_PCI_IOCTL_RENDER: {
		u32 die_index = kcmd.get_render_name.die_index;
		snprintf(kcmd.get_render_name.file_name,
			 sizeof(kcmd.get_render_name.file_name), "renderD%d",
			 pci_info->dies[vastai_pci_get_die_id(pci_info,
							      die_index)]
				 .render_id);
		printk("%s\n", kcmd.get_render_name.file_name);
		break;
	}
	case VASTAI_PCI_IOCTL_PORT: {
		strcpy(kcmd.port_name, pci_name(pci_info->dev));
		printk("%s\n", kcmd.port_name);
		break;
	}
	case VASTAI_PCI_IOCTL_GET_PERFORMANCE: {
		u32 die_index = kcmd.get_die_performance.die_index;
		memcpy(kcmd.get_die_performance.pcie_tx_vdsp_count,
		       pci_info->dies[vastai_pci_get_die_id(pci_info, die_index)]
			       .pcie_tx_vdsp_count,
		       sizeof(kcmd.get_die_performance.pcie_tx_vdsp_count));

		memcpy(kcmd.get_die_performance.pcie_rx_vdsp_count,
		       pci_info->dies[vastai_pci_get_die_id(pci_info, die_index)]
			       .pcie_rx_vdsp_count,
		       sizeof(kcmd.get_die_performance.pcie_rx_vdsp_count));
		break;
	}
	case VASTAI_PCI_IOCTL_PCI_ID:
		kcmd.get_pci_id.pci_id = 0x1ec6 | 0x0100 << 16;
		break;
	case VASTAI_PCI_IOCTL_PCI_SUB_ID:
		kcmd.get_pci_sub_id.pci_sub_id = 0xFFFF;
		break;
	case VASTAI_PCI_IOCTL_MAJOR_MINOR:
		kcmd.get_major_minor.major_minor = vastai_file_get_major_minor(&ctl_info->file_info);
		break;
	case TCGETS:
		return -ENOIOCTLCMD;

	case TCSETS:
		return -ENOIOCTLCMD;

	case TIOCSERGETLSR:
		return -ENOIOCTLCMD;

	case TIOCSERCONFIG:
		return -ENOIOCTLCMD;

	case VASTAI_PCI_IOCTL_RESET_NOTIFY: {
		int die_id = kcmd.reset_mode.die_id;
		if(kcmd.reset_mode.mode == RESET_NOTIFY_STOP) {
			init_completion(&pci_info->dies[die_id].reset_comp);
			ret = wait_for_completion_interruptible(&pci_info->dies[die_id].reset_comp);
			if (ret) {
				VASTAI_PCI_ERR(pci_info,
					       die_id,
					       "%s wait_for_completion error %d\n",
					       __FUNCTION__, ret);
			}
		} else if (kcmd.reset_mode.mode == RESET_STOP_COMPLETION) {
			//clear CMCU fifo
			union core_bitmap core_id;
			ring_buf_elem_t ring_buf_elem;

			core_id.val = 0;
			core_id.cmcu = 1;
			vastai_clear_fifo(pci_info, pci_info->dies[die_id].die_index, core_id);

			ring_buf_elem.whole = 0;
			ring_buf_elem.cmd = CORE_2_CORE_RESET_CMD;
			ring_buf_elem.use_ringbuf =
				VATOOLS_INTERRUPT_NOT_USE_RINGBUF;
			ring_buf_elem.seq_no = 0;
			ring_buf_elem.sub_cmd = CORE_2_CORE_RESET_AI_RESET_SUBCMD;

			vastai_send_pcie_cmd(pci_info, pci_info->dies[die_id].die_index,
					     VASTAI_PCIE_SUB_AI_SYS_RESET, ring_buf_elem.whole);
		} else if (kcmd.reset_mode.mode == RESET_AI_SYS_DONE) {
			init_completion(&pci_info->dies[die_id].ai_reset_done);
			ret = wait_for_completion_interruptible(&pci_info->dies[die_id].ai_reset_done);
			if (ret) {
				VASTAI_PCI_ERR(pci_info,
					       vastai_pci_get_die_id(pci_info, pci_info->dies[die_id].die_index),
					       "%s wait_for_ai_completion error %d\n",
					       __FUNCTION__, ret);
			}
		}
		break;
	}
	case VASTAI_PCI_IOCTL_SET_BBOX: {
		ret = vastai_bbox_set(pci_info,
				kcmd.bbox_cmd.die_index,
				kcmd.bbox_cmd.bbox_status);
		break;
	}
#endif
	default:
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID,
				"Unsupport cmd 0x%x\n", cmd);
		break;
	}
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID, "ioctl error %d\n", ret);
		return ret;
	}

	ret = vastai_pci_copy_to_user((void *)arg, &kcmd, sizeof(kcmd));
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				"%s pci_copy_to_user is error\n", __func__);
		return -EIO;
	}

	return 0;
}

static const struct file_operations fop = {
	.owner = THIS_MODULE,
	.open = ctl_open,
	.release = ctl_release,
	.llseek = default_llseek,
	.unlocked_ioctl = ctl_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = ctl_ioctl,
#endif

};

int vastai_ctl_file_init(struct vastai_pci_info *pci_info)
{
	struct vastai_ctl_file *ctl_file =
	vmalloc(sizeof(struct vastai_ctl_file));
	int ret;

	if (ctl_file == NULL) {
		return -ENOMEM;
	}
	ret = vastai_file_create(&ctl_file->file_info, pci_info, &fop,
			"va%d_ctl", pci_info->dev_id);


	pci_info->ctl_file = ctl_file;
	ctl_file->pci_info = pci_info;

	return ret;
}

void vastai_ctl_file_deinit(struct vastai_pci_info *pci_info)
{
	vastai_file_destroy(&pci_info->ctl_file->file_info);
	vfree(pci_info->ctl_file);
}
