/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <linux/random.h>
#include <linux/dma-buf.h>

#include "vastai_export_enum_define.h"
#include "vastai_pci.h"
#include "vastai_pci_api.h"
#include "vastai_pci_test.h"
#include "vastai_udma_test.h"
#include "vastai_udma_engine.h"
#include "vastai_dmabuf.h"
#include "vastai_cmd.h"
#include "vastai_cfg.h"
#include "vastai_ras.h"
#include "vastai_dmi_table.h"
#include "vastai_state.h"

#include "va_dma_core.h"
#include "hw_config.h"
#include "sg100_base.h"
#include "sg100_cmd.h"
#include "smmu.h"
#include "sdma.h"
#include "va_vsync.h"
#include "va_inventory.h"
#ifdef CONFIG_VASTAI_GFX
#include "../gfx/gfx_mm.h"
#endif

#include "../gfx/gfx_drv.h"

#define TCGETS			0x5401
#define TCSETS			0x5402
#define TCSETSW			0x5403
#define TCSETSF			0x5404
#define TIOCSERGSTRUCT	0x5458 /* For debugging only */
#define TIOCSERGETLSR	0x5459 /* Get line status register */
#define TIOCSERCONFIG	0x545

struct vastai_dma_buf dmb[4];
unsigned int major = 321;
unsigned char* g_host_buf_va;
atomic_t minor = ATOMIC_INIT(0);
#define PCIE_PAYLOAD_MONITOR_EN 0
extern u32 test_PCIe_Perf;
extern u32 test_smcu_RW_Perf;
extern u32 ring_buf_debug;
static struct vastai_pci_arg_t vastai_pci_args_default[VASTAI_PCI_ARG_MAX] = {
	{VASTAI_PCI_ARG_ADDR,		"addr", 		0},
	{VASTAI_PCI_ARG_VAL, 		"val", 			0},
	{VASTAI_PCI_ARG_LEN, 		"len", 			4},
	{VASTAI_PCI_ARG_BAR, 		"bar", 			0},
	{VASTAI_PCI_ARG_RUN, 		"run", 			0},
	{VASTAI_PCI_ARG_OFFSET, 	"offset", 		0},
	{VASTAI_PCI_ARG_VIR, 		"vir", 			0},
	{VASTAI_PCI_ARG_REGION, 	"region", 		0},
	{VASTAI_PCI_ARG_MODE, 		"mode", 		0},
	{VASTAI_PCI_ARG_CASE, 		"case", 		0},
	{VASTAI_PCI_ARG_USER_ADDR, 	"user_addr", 	0},
	{VASTAI_PCI_ARG_SWITCH_ATU, "atu", 			0},			// 0: swith to original atu; 1: switch to debug atu.
	{VASTAI_PCI_ARG_MMU, 		"mmu", 			0},
	{VASTAI_PCI_ARG_ADDR0, 		"addr0", 		0},
	{VASTAI_PCI_ARG_ADDR1, 		"addr1", 		0},
	{VASTAI_PCI_ARG_PA_ADDR, 	"pa", 			0},
	{VASTAI_PCI_ARG_VA_ADDR, 	"va", 			0},
	{VASTAI_PCI_ARG_PID, 		"pid", 			0},
	{VASTAI_PCI_ARG_MASKBIT, 	"maskbit", 		0},
	{VASTAI_PCI_ARG_PF, 		"pf", 			0},
	{VASTAI_PCI_ARG_CHANNEL, 	"ch", 			0},
};

static void vastai_pci_args_init(struct vastai_pci_arg_t *args)
{
	unsigned int i;

	for (i = 0; i < VASTAI_PCI_ARG_MAX; i++)
		*(args + i) = vastai_pci_args_default[i];
}

static inline unsigned long vastai_pci_args_value(struct vastai_pci_arg_t *args, unsigned char index)
{
	unsigned char i;

	for (i = 0; i < VASTAI_PCI_ARG_MAX; i++) {
		if (index == args[i].id)
			return args[i].def;
	}

	return -1;
}

int vastai_pci_malloc_udma_buf(struct vastai_pci_info *pcie_dev, struct vastai_dma_buf *dm, u32 len)
{
	int ret;

	if (dm->vir || dm->dma_bus_addr) {
		if (dm->size >= len) {
			memset((unsigned char *)(dm->vir), VASTAI_UDMA_TEST_VAL, len);
			VASTAI_PCI_INFO(pcie_dev, DUMMY_DIE_ID,
					"pci_malloc_udma_buf not malloc again! dm_dma_bus_addr:0x%lx\n",
					dm->dma_bus_addr);
			return 0;
		} else
			vastai_udma_free(pcie_dev, dm);
	}
	ret = vastai_udma_malloc(pcie_dev, dm, len);
	if (ret != 0)
		return ret;

	memset((unsigned char *)(dm->vir), VASTAI_UDMA_TEST_VAL, len);
	VASTAI_PCI_DBG(pcie_dev, DUMMY_DIE_ID, "pci_malloc_udma_buf dm_phy:0x%lx\n", dm->dma_bus_addr);
	return 0;
}

static void vastai_pci_set_udma_buf(struct vastai_dma_buf *dm, int val)
{
    if (!dm->vir || !dm->dma_bus_addr || !dm->size) {
        VASTAI_PCI_INFO(
            NULL, DUMMY_DIE_ID,
            "pci_set_udma_buf please malloc first! dm_dma_bus_addr:0x%lx\n",
            dm->dma_bus_addr);
        return;
    }
    memset((unsigned char *)(dm->vir), val, dm->size);
    VASTAI_PCI_HEX_DUMP(KERN_INFO, "dma buf: ", DUMP_PREFIX_NONE, 32, 4,
                (void *)(dm->vir),
                ((dm->size > VASTAI_UDMA_TEST_LEN) ?
                     VASTAI_UDMA_TEST_LEN :
                     dm->size),
                true);

}

static void vastai_pci_print_udma_buf(struct vastai_dma_buf *dm)
{
    if (!dm->vir || !dm->dma_bus_addr || !dm->size) {
        VASTAI_PCI_INFO(
            NULL, DUMMY_DIE_ID,
            "pci_print_udma_buf please malloc first! dm_dma_bus_addr:0x%lx\n",
            dm->dma_bus_addr);
        return;
    }
    VASTAI_PCI_HEX_DUMP(KERN_INFO, "dma buf: ", DUMP_PREFIX_NONE, 32, 4,
                (void *)(dm->vir),
                ((dm->size > VASTAI_UDMA_TEST_LEN) ?
                     VASTAI_UDMA_TEST_LEN :
                     dm->size),
                true);

}

int vastai_pci_free_udma_buf(struct vastai_pci_info *pcie_dev, struct vastai_dma_buf *dm)
{
    if (!dm->vir || !dm->dma_bus_addr || !dm->size) {
        VASTAI_PCI_INFO(
            pcie_dev, DUMMY_DIE_ID,
            "pci_free_udma_buf no malloc, no free! dm_dma_bus_addr:0x%lx\n",
            dm->dma_bus_addr);
        return 0;
    }
    return vastai_udma_free(pcie_dev, dm);

}
/* TC21: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to SRAM, as inbound transfer
 *         (from remote memory to local memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc21(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE0, VASTAI_DIR_IN, 0x100,
                    0x840000000,
                    dmb[0].phy);
    return VASTAI_TEST_PASS;
}

/* TC22: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from SRAM, as outbound transfer
 *         (from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc22(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE0, VASTAI_DIR_OUT, 0x100,
                    0x840000000,
                    dmb[0].phy);
    vastai_pci_print_udma_buf(&dmb[0]);
    return VASTAI_TEST_PASS;
}

/* TC23: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to SRAM, as inbound transfer
 *         (from remote memory to local memory), specified length.
 * param in len The length of transfer in bytes.
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc23(struct vastai_pci_info *priv, u32 len)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], len);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE0, VASTAI_DIR_IN, len,
                    VASTAI_CSRAM_ADDR(VASTAI_DIE0),
                    dmb[0].phy);
    return VASTAI_TEST_PASS;
}

/* TC24: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from SRAM, as outbound transfer
 *         (from local memory to remote memory), specified length.
 * param in len The length of transfer in bytes.
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc24(struct vastai_pci_info *priv, u32 len)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], len);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE0, VASTAI_DIR_OUT, len,
                    VASTAI_CSRAM_ADDR(VASTAI_DIE0),
                    dmb[0].phy);
    vastai_pci_print_udma_buf(&dmb[0]);
    return VASTAI_TEST_PASS;
}

/* TC25: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to DDR, as inbound transfer
 *         (from remote memory to local memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc25(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE0, VASTAI_DIR_IN, 0x100,
                    VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
                    dmb[0].phy);
    return VASTAI_TEST_PASS;
}

/* TC26: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as outbound transfer
 *         (from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc26(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE0, VASTAI_DIR_OUT, 0x100,
                    VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
                    dmb[0].phy);
    vastai_pci_print_udma_buf(&dmb[0]);
    return VASTAI_TEST_PASS;
}

/* TC27: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to DDR, as inbound transfer
 *         (from remote memory to local memory), 0x1000000 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc27(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_MALLOC_MAX);
    vastai_pci_malloc_udma_buf(priv, &dmb[1], VASTAI_UDMA_MALLOC_MAX);
    vastai_pci_malloc_udma_buf(priv, &dmb[2], VASTAI_UDMA_MALLOC_MAX);
    vastai_pci_malloc_udma_buf(priv, &dmb[3], VASTAI_UDMA_MALLOC_MAX);
    vastai_udma_start_bulk_transfer(priv,
        VASTAI_DIE0, VASTAI_DIR_IN, VASTAI_UDMA_LENGTH_MAX,
        VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0), dmb[3].phy);
    return VASTAI_TEST_PASS;
}

/* TC28: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as outbound transfer
 *         (from local memory to remote memory), 0x1000000 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc28(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_LENGTH_MAX);
    vastai_udma_start_bulk_transfer(priv,
        VASTAI_DIE0, VASTAI_DIR_OUT, VASTAI_UDMA_LENGTH_MAX,
        VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0), dmb[0].phy);
    vastai_pci_print_udma_buf(&dmb[0]);
    return VASTAI_TEST_PASS;
}

/* TC30: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as ioutbound transfer
 *         (from local memory to remote memory), specified length.
 * param in len The length of transfer in bytes.
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc30(struct vastai_pci_info *priv, u32 len)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], len);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE0, VASTAI_DIR_OUT, len,
                    VASTAI_DDR_SPACE_ADDR(VASTAI_DIE0),
                    dmb[0].phy);
    vastai_pci_print_udma_buf(&dmb[0]);
    return VASTAI_TEST_PASS;
}

/* TC52: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from SRAM, as outbound transfer
 *         (from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc52(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE1, VASTAI_DIR_OUT, 0x100,
                    VASTAI_CSRAM_ADDR(VASTAI_DIE1),
                    dmb[0].phy);
    vastai_pci_print_udma_buf(&dmb[0]);
    return VASTAI_TEST_PASS;
}

/* TC53: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to DDR, as inbound transfer
 *         (from remote memory to local memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc53(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE1, VASTAI_DIR_IN, 0x100,
                    VASTAI_DDR_SPACE_ADDR(VASTAI_DIE1),
                    dmb[0].phy);
    return VASTAI_TEST_PASS;
}

/* TC54: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer from DDR, as outbound transfer
 *         (from local memory to remote memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc54(struct vastai_pci_info *priv)
{
    vastai_pci_malloc_udma_buf(priv, &dmb[0], VASTAI_UDMA_TEST_LEN);
    vastai_udma_start_bulk_transfer(priv,
                    VASTAI_DIE1, VASTAI_DIR_OUT, 0x100,
                    VASTAI_DDR_SPACE_ADDR(VASTAI_DIE1),
                    dmb[0].phy);
    vastai_pci_print_udma_buf(&dmb[0]);
    return VASTAI_TEST_PASS;
}

/* TC101: reset test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc101(struct vastai_pci_info *priv)
{
    vastai_pci_reset_bus(priv);
    vastai_pci_tc22(priv);
    return VASTAI_TEST_PASS;
}

/* TC102: reset when UDMA transfer test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
u8 vastai_test_udma_reset_flag;
static int vastai_pci_tc102(struct vastai_pci_info *priv)
{
    vastai_test_udma_reset_flag = 1;
    vastai_pci_tc22(priv);
    vastai_test_udma_reset_flag = 0;
    vastai_pci_tc22(priv);
    return VASTAI_TEST_PASS;
}

/* TC103: reset test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
static int vastai_pci_tc103(struct vastai_pci_info *priv)
{
    vastai_pci_reset_bus(priv);
    vastai_pci_tc52(priv);
    return VASTAI_TEST_PASS;
}

/* TC104: reset when UDMA transfer test via pcie driver
 * keyword: reset
 * description: Perform a reset
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
u8 vastai_test_udma_reset_flag;
static int vastai_pci_tc104(struct vastai_pci_info *priv)
{
    vastai_test_udma_reset_flag = 1;
    vastai_pci_tc52(priv);
    vastai_test_udma_reset_flag = 0;
    vastai_pci_tc52(priv);
    return VASTAI_TEST_PASS;
}

/* TC151: udma test via pcie driver
 * keyword: udma, bulk
 * description: Perform a DMA bulk transfer to DDR, as inbound transfer
 *         (from remote memory to local memory), 0x100 bytes length.
 * param in NA
 * return VASTAI_TEST_PASS if the function completed successfully.
 * return VASTAI_TEST_FAILURE if the function performed error.
 */
#define test_yuv_hexcode "../fw/test.yuv.hex"
static unsigned char test_yuv_hex_buf[] = {
#include test_yuv_hexcode
};
#define TEST_YUV_HEX_SIZE    sizeof(test_yuv_hex_buf)
#define TEST_YUV_DDR_ADDRESS (0x80E000000UL)
static int vastai_pci_tc151(struct vastai_pci_info *priv)
{
#if 0
    vastai_pci_malloc_udma_buf(priv, &dmb[0], TEST_YUV_HEX_SIZE);
    memcpy((void *)(dmb[0].vir), test_yuv_hex_buf, TEST_YUV_HEX_SIZE);
    vastai_udma_start_bulk_transfer(VASTAI_DIE0, VASTAI_DIR_IN,
                    TEST_YUV_HEX_SIZE, TEST_YUV_DDR_ADDRESS,
                    dmb[0].phy);
#endif
    vastai_pci_mem_write(priv, priv->dies[0].die_index, TEST_YUV_DDR_ADDRESS, test_yuv_hex_buf,
                 TEST_YUV_HEX_SIZE);
    return VASTAI_TEST_PASS;
}

/* Host to fw dma sync unit test with src is virtual buf */
static int vastai_pci_tc155(struct vastai_pci_info *priv)
{
    struct vastai_dmadesc desc[2];
    int die_index = priv->dies[0].die_index;
    union core_bitmap core_id = {.val = 0}; /* no core need be tirgger */

    vastai_pci_malloc_udma_buf(priv, &dmb[0], 0x10000);
    vastai_pci_set_udma_buf(&dmb[0], 0xbc);
    vastai_pci_malloc_udma_buf(priv, &dmb[1], 0x20000);
    vastai_pci_set_udma_buf(&dmb[1], 0x5a);

    desc[0].dev_addr = 0x1040000000;
    desc[0].is_src_dma_addr = 0;
    desc[0].is_src_not_user_mem = 1;
    desc[0].host_addr.vir_addr = dmb[0].vir;
    desc[0].host_addr.dma_addr = (dma_addr_t)dmb[0].phy;
    desc[0].is_host_to_dev = 1;
    desc[0].dma_lenth = 0x10000;

    desc[1].dev_addr = 0x1040010000;
    desc[1].is_src_dma_addr = 0;
    desc[1].is_src_not_user_mem = 1;
    desc[1].host_addr.vir_addr = dmb[1].vir;
    desc[1].host_addr.dma_addr = (dma_addr_t)dmb[1].phy;
    desc[1].is_host_to_dev = 1;
    desc[1].dma_lenth = 0x20000;

    core_id.decode = 0x1;
    vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, 2);
#if 0
    VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
                desc[0].host_addr.vir_addr,
                0x100,
                true);
    VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
                desc[1].host_addr.vir_addr,
                0x100,
                true);

    vfree(desc[0].host_addr.vir_addr);
    vfree(desc[1].host_addr.vir_addr);
#endif
    return VASTAI_TEST_PASS;
}

/* Fw to Host dma sync unit test with src is virtual buf */
static int vastai_pci_tc156(struct vastai_pci_info *priv)
{
    struct vastai_dmadesc desc[2];
    int die_index = priv->dies[0].die_index;
    union core_bitmap core_id = {.val = 0}; /* no core need be tirgger */

    desc[0].dev_addr = 0x840000000;
    desc[0].is_src_dma_addr = 0;
    desc[0].is_src_not_user_mem = 1;
    desc[0].host_addr.vir_addr = vmalloc(0x100);
    desc[0].is_host_to_dev = 0;
    desc[0].dma_lenth = 0x100;

    desc[1].dev_addr = 0x840000100;
    desc[1].is_src_dma_addr = 0;
    desc[1].is_src_not_user_mem = 1;
    desc[1].host_addr.vir_addr = vmalloc(0x100);
    desc[1].is_host_to_dev = 0;
    desc[1].dma_lenth = 0x100;

    core_id.decode = 0x1;
    vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, 2);
    VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
                desc[0].host_addr.vir_addr,
                0x100,
                true);
    VASTAI_PCI_HEX_DUMP(KERN_INFO, "host buf: ", DUMP_PREFIX_NONE, 32, 4,
                desc[1].host_addr.vir_addr,
                0x100,
                true);
    vfree(desc[0].host_addr.vir_addr);
    vfree(desc[1].host_addr.vir_addr);
    return VASTAI_TEST_PASS;
}

/* Host to fw dma sync , long length */
static int vastai_pci_tc157(struct vastai_pci_info *priv)
{
    struct vastai_dmadesc desc[2];
    int die_index = priv->dies[0].die_index;
    union core_bitmap core_id = {.val = 0}; /* no core need be tirgger */
    struct vastai_dma_buf dm[2];
    int i;
    u32 desc_cnt = (get_random_u32() % 2) + 1;
    u32 offset = 0;
    u32 is_host_to_dev = get_random_u32() % 2; 

    for (i = 0; i < desc_cnt; i++) {
        u32 test_len = (get_random_u32()% 0x200000) + 900000;
        desc[i].dev_addr = ALIGN(0x840000000 + offset, 0x1000);
        desc[i].is_src_dma_addr = 1;
        desc[i].is_src_not_user_mem = 1;
        vastai_udma_malloc(priv, &dm[i], test_len);
        desc[i].host_addr.dma_addr = dm[i].phy;
        desc[i].host_addr.vir_addr = dm[i].vir;
        desc[i].is_host_to_dev = is_host_to_dev;
        desc[i].dma_lenth = test_len;
        memset(desc[i].host_addr.vir_addr, '2', test_len);
        offset += test_len;
    }

    core_id.decode = 0x1;
    vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, desc_cnt);

    for (i = 0; i < desc_cnt; i++) {
        vastai_udma_free(priv, &dm[i]);
    }
    return VASTAI_TEST_PASS;
}

// static int vastai_pci_tc158(struct vastai_pci_info *priv)
// {
//     union core_bitmap core_id = {.vdsp = BIT(0)};
//     int i = 0;
//     static int index = 0;

//     for (i = 0; i < 10; i++) {
//         u32 msg_buf = 0x80000000 + index;
//         index += 1;
//         vastai_pci_send_msg(priv, 0, core_id, &msg_buf, sizeof(msg_buf));
//     }
//     return 0;
// }

static int vastai_pci_tc159(struct vastai_pci_info *priv)
{
    struct vastai_dmadesc desc[2];
    int die_index = priv->dies[1].die_index;
    union core_bitmap core_id = {.val = 0}; /* no core need be tirgger */
    struct vastai_dma_buf dm[2];
    int i;
    u32 desc_cnt = (get_random_u32() % 2) + 1;
    u32 offset = 0;
    u32 is_host_to_dev = get_random_u32() % 2; 

    for (i = 0; i < desc_cnt; i++) {
        u32 test_len = (get_random_u32()% 0x200000) + 900000;
        desc[i].dev_addr = ALIGN(0x840000000 + offset, 0x1000);
        desc[i].is_src_dma_addr = 1;
        desc[i].is_src_not_user_mem = 1;
        vastai_udma_malloc(priv, &dm[i], test_len);
        desc[i].host_addr.dma_addr = dm[i].phy;
        desc[i].host_addr.vir_addr = dm[i].vir;
        desc[i].is_host_to_dev = is_host_to_dev;
        desc[i].dma_lenth = test_len;
        memset(desc[i].host_addr.vir_addr, '2', test_len);
        offset += test_len;
    }

    core_id.decode = 0x1;
    vastai_pci_dma_transfer_sync(priv, die_index, core_id, desc, desc_cnt);

    for (i = 0; i < desc_cnt; i++) {
        vastai_udma_free(priv, &dm[i]);
    }
    return VASTAI_TEST_PASS;
}

static int vastai_pci_test_case(struct vastai_pci_info *priv,
                struct char_drv_info *dev,
                u32 case_no)
{
    int ret = 0;
    u32 len;

    VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            "******** VA PCI TEST[%d] start ********\n",
            case_no);
    len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
    switch (case_no) {
    case 21:
        ret = vastai_pci_tc21(priv);
        break;
    case 22:
        ret = vastai_pci_tc22(priv);
        break;
    case 23:
        ret = vastai_pci_tc23(priv, len);
        break;
    case 24:
        ret = vastai_pci_tc24(priv, len);
        break;
    case 25:
        ret = vastai_pci_tc25(priv);
        break;
    case 26:
        ret = vastai_pci_tc26(priv);
        break;
    case 27:
        ret = vastai_pci_tc27(priv);
        break;
    case 28:
        ret = vastai_pci_tc28(priv);
        break;
    case 30:
        ret = vastai_pci_tc30(priv, len);
        break;
    case 52:
        ret = vastai_pci_tc52(priv);
        break;
    case 53:
        ret = vastai_pci_tc53(priv);
        break;
    case 54:
        ret = vastai_pci_tc54(priv);
        break;
    case 101:
        ret = vastai_pci_tc101(priv);
        break;
    case 102:
        ret = vastai_pci_tc102(priv);
        break;
    case 103:
        ret = vastai_pci_tc103(priv);
        break;
    case 104:
        ret = vastai_pci_tc104(priv);
        break;
    case 151:
        ret = vastai_pci_tc151(priv);
        break;
    case 155:
        ret = vastai_pci_tc155(priv);
        break;
    case 156:
        ret = vastai_pci_tc156(priv);
        break;
    case 157:
        ret = vastai_pci_tc157(priv);
        break;
    case 158:
        //ret = vastai_pci_tc158(priv);
        break;
    case 159:
        ret = vastai_pci_tc159(priv);
        break;
    default:
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "Unsupported test "
                   "case_no = %d\n",
                   case_no);
        break;
    }
    if (ret == VASTAI_TEST_PASS)
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "******** VA PCI TEST[%d] done "
                "********\n",
                case_no);
    else
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "******** VA PCI TEST[%d] fail "
                   "ret = %d ********\n",
                   case_no, ret);
    return ret;
}

/* flash xspi bootcode, "xspi_vbios val=xxx" */
static void vastai_pci_test_flashxspi_vbios(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_flash_xspi(priv, dev, 1);
		vastai_pci_get_fw_ver(priv, 0x1800);
	}
}

/* flash xspi bootcode, "xspi_bl0 val=xxx" */
static void vastai_pci_test_flashxspi_bl0(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_flash_xspi(priv, dev, 2);
		vastai_pci_get_fw_ver(priv, 0x1800);
	}
}

/* flash xspi bootcode, "xspi_base val=xxx" */
static void vastai_pci_test_flashxspi_base(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_flash_xspi(priv, dev, 3);
		vastai_pci_get_fw_ver(priv, 0x1800);
	}
}

/* flash xspi bootcode, "xspi_full val=xxx" */
static void vastai_pci_test_flashxspi_full(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_flash_xspi(priv, dev, 4);
		vastai_pci_get_fw_ver(priv, 0x1800);
	}
}

/* flash xspi bootcode, "xspi_all val=xxx" */
static void vastai_pci_test_flashxspi_all(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_flash_xspi(priv, dev, 5);
		vastai_pci_get_fw_ver(priv, 0x1800);
	}
}

/* flash xspi bootcode, "xspi_pmcu val=xxx" */
static void vastai_pci_test_flashxspi_pmcu(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_flash_xspi(priv, dev, 6);
		vastai_pci_get_fw_ver(priv, 0x1800);
	}
}

/* upgrade bmcu fw, "bmcu_upgrade val=xxx" */
static void vastai_pci_test_flashbmcu(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.pkg_id == 0 &&
		priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_flash_bmcu(priv, dev);
		vastai_pci_get_fw_ver(priv, 0x600);
	}
}

/* degrade bmcu fw, "bmcu_degrade val=xxx" */
static void vastai_pci_test_switchbmcu(struct char_drv_info *dev, struct vastai_pci_info *priv)
{
	if (priv->priv_hw_cfg->sys_cfg.pkg_id == 0 &&
		priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf) {
		vastai_pci_switch_bmcu(priv, dev);
		vastai_pci_get_fw_ver(priv, 0x600);
	}
}

int vastai_ddr_bw_test(struct vastai_pci_info *priv)
{
	int ret;
	struct pcie_transfer_cmd trans;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "ddr_bw_test begin\n");

	trans.w0.s_data0.optcode = GMCU_OPEN_DDR_BW_TEST;
	if (ONLY_1_PF == priv->fn_mode) {
		ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF,  &trans, 0);
	} else if (ONLY_2_PF == priv->fn_mode){
		ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
	} else {
		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
	}
	if (ret) 
		VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID, "ddr_bw_test:push element fail\n");

	return ret;
}

#if (defined(__aarch64__) || defined(__arm__))
static void vastai_bandwidth_monitor_work(struct work_struct *work)
{
	struct vastai_pci_info *priv = NULL;

	priv = container_of(work, struct vastai_pci_info, bandwidth_monitor.bandwidth_test_work.work);

	if (priv->pcie_bw_buf.vir) {
		memset((unsigned char *)(priv->pcie_bw_buf.vir), VASTAI_UDMA_TEST_VAL, priv->pcie_bw_buf.size);
	}
	
	schedule_delayed_work(&priv->bandwidth_monitor.bandwidth_test_work, priv->bandwidth_monitor.timeout_threshold);

	return;
}
#endif

int vastai_pcie_bw_test(struct vastai_pci_info *priv)
{
	unsigned long  dma_bus_addr;
	int ret = 0;
	u32 pcie_bw_test_len = 2*1024*1024;
	struct pcie_transfer_cmd trans;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pcie_bw_test begin\n");

	vastai_pci_malloc_udma_buf(priv, &priv->pcie_bw_buf, pcie_bw_test_len);
	if(!priv->pcie_bw_buf.vir) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pcie_bw_test:malloc dma buf fail\n");
		return -1;
	}
	dma_bus_addr = priv->pcie_bw_buf.dma_bus_addr;
	
	trans.w0.s_data0.optcode	= GMCU_OPEN_PCIE_BW_TEST;
	trans.w1.data1				= (u32)((dma_bus_addr >> 32) & 0xFFFFFFFF);
	trans.w2.data2				= (u32)(dma_bus_addr & 0xFFFFFFFF);
	trans.w3.data3				= pcie_bw_test_len;
	if (ONLY_1_PF == priv->fn_mode) {
		ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF,  &trans, 0);
	} else if (ONLY_2_PF == priv->fn_mode){
		ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
	} else {
		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
	}
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pcie_bw_test:push element fail\n");
		vastai_pci_free_udma_buf(priv, &priv->pcie_bw_buf);
		return ret;
	}

	/* add work queue for BUG-1329 */
#if (defined(__aarch64__) || defined(__arm__))
	if (!priv->bandwidth_monitor.is_begin) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "bandwidth monitor init\n");
		priv->bandwidth_monitor.is_begin = 1;
		priv->bandwidth_monitor.timeout_threshold = msecs_to_jiffies(3000); /* 3s */
		INIT_DELAYED_WORK(&priv->bandwidth_monitor.bandwidth_test_work, vastai_bandwidth_monitor_work);
		schedule_delayed_work(&priv->bandwidth_monitor.bandwidth_test_work, priv->bandwidth_monitor.timeout_threshold);
	}
#endif

	return ret;
}

void vastai_pcie_bw_test_cancel(struct vastai_pci_info *priv)
{
	int ret = 0;
	struct pcie_transfer_cmd trans;

	if (!priv->bandwidth_monitor.is_begin)
		return;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "pcie_bw_test_cancel\n");
	
	trans.w0.s_data0.optcode = GMCU_CLOSE_PCIE_BW_TEST;
	if (ONLY_1_PF == priv->fn_mode) {
		ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF,  &trans, 0);
	} else if (ONLY_2_PF == priv->fn_mode){
		ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
	} else {
		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
	}
	if (ret)
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pcie_bw_test_cancel:push element fail\n");

	cancel_delayed_work_sync(&priv->bandwidth_monitor.bandwidth_test_work);
	priv->bandwidth_monitor.is_begin = 0;

	return;
}

static int vastai_pci_cmdline_args(char *cmdline, char *cmd, struct vastai_pci_arg_t *args, int argc)
{
    int status = 0, DONE, FOUND = 0;
    char *end, *argname, *cp;
    unsigned short base;
    unsigned long result, value, val, i, j;

    for (i = 0; i < argc; i++)
        args[i].def = 0;

    /* get cmd */
    while (*cmdline == ' ' || *cmdline == '\t')
        cmdline++;

    while (*cmdline != ' ' && *cmdline != '\t' && *cmdline != '\0') {
        *cmd = *cmdline;
        cmd++;
        cmdline++;
    }
    *cmd = '\0';
    if (*cmdline == '\0')
        goto WEDONE;

    *cmdline = '\0';
    cmdline++;
    while (*cmdline == ' ' || *cmdline == '\t')
        cmdline++;

    end = cmdline;
    while (*end == ' ' || *end == '\t')
        end++;

    /*
     * Parse cmdline
     */
    DONE = (*end == '\0') ? 1 : 0;
    while (!DONE) {
        /* get the register name */
        while (*end != '=' && *end != '\0')
            end++;
        if (*end == '\0') {
            status = 1;
            goto WEDONE;
        }
        *end = '\0';
        argname = cmdline;
        /* now get value to write to register */
        cmdline = ++end;
        /* if there's whitespace after the '=', exit with an error */
        if (*end == ' ' || *end == '\t' || *end == '\n') {
            status = 1;
            goto WEDONE;
        }
        while (*end != ' ' && *end != '\t' && *end != '\n' &&
               *end != '\0')
            end++;
        if (*end == '\0')
            DONE = 1;
        else
            *end = '\0';

        if (!strcmp(argname, "file") || !strcmp(argname, "filec")) {
            val = 1;
        } else {
            /* get the base, convert value to base-10 if necessary
             */
            val = 0;
            result = 0;
            cp = cmdline;
            if (cp[0] == '0' && (cp[1] == 'x' || cp[1] == 'X')) {
                base = 16;
                cp += 2;

            } else {
                base = 10;
            }
            while (isxdigit(*cp)) {
                value = isdigit(*cp) ?
                        (*cp - '0') :
                        ((islower(*cp) ? toupper(*cp) :
                                 *cp) -
                         'A' + 10);

                result = result * base + value;
                cp++;
            }

            val = result;
        }

        FOUND = 0;
        /*
         * verify the register arg is valid, and if the value is not
         * too big, write it to the corresponding location in arg_vals
         */
        for (j = 0; j < argc && !FOUND; j++) {
            if (!strcmp(argname, args[j].name)) {
                args[j].def = val;
                FOUND = 1;
            }
        }
        if (!FOUND) {
            VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID,
                       "arg %s err\n", argname);
            status = 0;
            goto WEDONE;
        }

        /*
         * point cmdline and end to next non-whitespace
         * (next argument)
         */
        cmdline = ++end;
        while (*cmdline == ' ' || *cmdline == '\t' || *cmdline == '\n')
            cmdline++;
        end = cmdline;
        /*
         * if, after skipping whitespace, we hit end of line or EOF,
         * we're done
         */
        if (*end == '\0')
            DONE = 1;
    }

WEDONE:
    return status;
}

typedef union logsys_msg_info_s
{
    struct
    {
        u16 msg_level   : 3;
        u16 msg_type    : 3;
        u16 msg_len     : 10;
    };
    u16 n_val;
}logsys_msg_info_t;

typedef struct logsys_header_s
{
    u8 head_flag;
    u8 log_sn;
    logsys_msg_info_t msg_info;
    u32 timestamp;
} logsys_header_t;

static int vastai_get_log(struct vastai_pci_info *priv, u64 addr, u32 len)
{
    int ret;
    uint32_t log_rp = 0;
    uint32_t log_cnt = 0;
    char print_log_buf[256];
    uint32_t print_log_len;
    logsys_header_t *log_header = NULL;
    uint32_t block_end_flag;
    char *log_buf;

    log_buf = kmalloc(len * 2, GFP_KERNEL);
    if (NULL == log_buf) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "get_log kmalloc fail\n");
        return -1;
    }
#ifdef DIAG_TEST_MODE
    ret = vastai_pci_mem_read(priv, 0, addr, log_buf, len);
#else
    ret = vastai_read_by_smcu(priv, addr, len ,log_buf);
#endif    
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "read_by_smcu fail\n");
        kfree(log_buf);
        return -1;
    }

    while(1)
    {
        log_header = (logsys_header_t *)(log_buf + log_rp);
        if (log_header->head_flag == '#') {
            print_log_len = sprintf(print_log_buf, "[%d][%d]", log_cnt, log_header->timestamp);
            log_cnt++;
            memcpy((void *)(print_log_buf + print_log_len), (void *)(log_buf + log_rp + 8), log_header->msg_info.msg_len);
            print_log_len += log_header->msg_info.msg_len;
            print_log_buf[print_log_len] = '\0';
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s", print_log_buf);
            log_rp = log_rp + 8 + ((log_header->msg_info.msg_len + 3) / 4) * 4;
            if (log_rp >= len)
                break;
        }
        else {
            block_end_flag = *(uint32_t *)(log_buf + log_rp);
            if (block_end_flag == 0x5A3B96D2) {
                log_rp = (log_rp / 0x1000 + 1) * 0x1000;
                if (log_rp >= len)
                    break;
            }
            else
                break;
        }
    }
    kfree(log_buf);
    return 0;
}

extern void va_pcie_dma_tranfer_test(struct vastai_pci_info *priv,
                 u8 is_mem_to_dev,
                 u64 mem_offset,
                 u64 dev_addr,
                 u32 len,
                 int val);

extern void va_sdma_op_cp_test(struct vastai_pci_info *priv, u8 loc_dir, u8 mmu_en, u64 addr0, u64 addr1, u32 trans_len, u32 pid);
extern void va_sdma_op_fence_test(struct vastai_pci_info *priv, u8 dir, u8 mmu_en, u64 dev_addr, u32 trans_len, u32 desc_cnt, u64 wr_data);
extern void va_sdma_op_fill_test(struct vastai_pci_info *priv, u8 dir, u8 dst_mmu, u64 dst_addr, u32 trans_len, u64 trans_val);
extern void sdma_reset(struct vastai_pci_info *priv);

u64 tx_packet = 0;
u64 rx_packet = 0;
u64 tx_payload = 0;
u64 rx_payload = 0;
extern struct mutex vastai_card_info_list_lock;
extern struct list_head vastai_card_info_list;

static int vastai_pci_proc_cmd(unsigned char *input,
                   int input_len,
                   struct char_drv_info *dev)
{
    int ret, i, j;
    unsigned long addr = 0, val, len, ch, pf;
    u64 upstream_count = 0, downstream_count = 0;
    u64 read_bw_cont = 0, write_bw_cont = 0;
    u64 total_read_count = 0, total_write_count = 0;
    unsigned char cmd[64], print_str[64];
    struct vastai_pci_info *priv = dev->pcie_dev_info;

    ret = vastai_pci_cmdline_args(input, cmd, dev->args,
                      VASTAI_PCI_ARG_MAX);
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "pci_proc_cmd cmdline_args err=%d\n", ret);
        return -1;
    }
    /* NTD: printing help information */
    else if (!strncmp("help", cmd, strlen("help"))) {
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "pci_proc_cmd help\n");
    }
    else if (!strncmp("card", cmd, strlen("card"))) {
        va_get_card_info(priv);
    }
    /* read configuration space */
    else if ((!strncmp("rcfg", cmd, strlen("rcfg"))) ||
         (!strncmp("config", cmd, strlen("config")))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "pci_proc_cmd [read config] len=%d bytes\n",
                dev->valid_data_length);
        if (dev->valid_data_length)
            ret = vastai_pci_config_read(priv, addr, dev->dm.vir,
                             dev->valid_data_length);
        if (!ret) {
            for (i = 0; i < dev->valid_data_length; i += 16) {
                sprintf(print_str,
                    "%s config "
                    "space "
                    "0x%03lx: ",
                    VASTAI_PCI_DEBUG_PREFIX, (i + addr));
                VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
                            DUMP_PREFIX_NONE, 16, 4,
                            dev->dm.vir + i, 16, true);
            }
        }
    }
    /* write configuration space */
    else if ((!strncmp("wcfg", cmd, strlen("wcfg"))) ||
         (!strncmp("wconfig", cmd, strlen("wconfig")))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "pci_proc_cmd [write config] addr=0x%lx, "
                "len=%d, val=0x%lx\n",
                addr, dev->valid_data_length, val);
        if (dev->valid_data_length)
            ret = vastai_pci_config_write(priv, addr, (char *)&val,
                              dev->valid_data_length);
    }
    /* read reg:
     * cmd format e.g. "read addr=0x4a000000 len=100"
     */
    else if (!strncmp("read", cmd, sizeof("read"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        ret = vastai_pci_mem_read(priv, priv->dies[0].die_index, addr, dev->dm.vir, dev->valid_data_length);
        {
            u64 offset        = 0;
	        int bar_num       = 0;
	        translate_noc_addr(priv,priv->pkg_id,addr, &bar_num, &offset);
            VASTAI_PCI_INFO(priv, priv->pkg_id,"[read]noc_addr= 0x%lx, bar[%d] offset[0x%llx] len=0x%d\n",
										addr, bar_num, offset, dev->valid_data_length);
        }
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
        else {
            for (i = 0; i < dev->valid_data_length; i += 16) {
                sprintf(print_str,
                    "%s [read] 0x%010lx: ",
                    VASTAI_PCI_DEBUG_PREFIX, (addr + i));
                VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
                            DUMP_PREFIX_NONE, 16, 4,
                            dev->dm.vir + i, 16, true);
            }
        }
    }
    /* read buf */
    else if (!strncmp("read_buf", cmd, sizeof("read_buf"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        ret = vastai_pci_mem_read(priv, priv->dies[0].die_index, addr, dev->dm.vir, dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }

    /* read buf */
    else if (!strncmp("modify_csr_bit", cmd, sizeof("modify_csr_bit"))) {
        u32 mode = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
        u32 mask_bit = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MASKBIT);
        u64 addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        ret = vastai_pci_modify_reg32(priv,addr,mask_bit,mode);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read] 0x%llx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }

    /* read buf */
    else if (!strncmp("smcu_modify_csr_bit", cmd, sizeof("smcu_modify_csr_bit"))) {
        u32 mode = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
        u32 mask_bit = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MASKBIT);
        u64 addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        ret = vastai_smcu_modify_reg32(priv,addr,mask_bit,mode);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read] 0x%llx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }
   
    /* write reg:
     * cmd format e.g. "write addr=0x4a000000 val=0x0 len=100"
     */
    else if (!strncmp("write", cmd, sizeof("write"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd write len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd write len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        for (i = 0; i < dev->valid_data_length; i += 4) {
            ((char *)dev->dm.vir)[i] = val & 0xFF;
            if ((i + 1) == dev->valid_data_length)
                break;
            ((char *)dev->dm.vir)[i + 1] = (val >> 8) & 0xFF;
            if ((i + 2) == dev->valid_data_length)
                break;
            ((char *)dev->dm.vir)[i + 2] = (val >> 16) & 0xFF;
            if ((i + 3) == dev->valid_data_length)
                break;
            ((char *)dev->dm.vir)[i + 3] = (val >> 24) & 0xFF;
        }
        ret = vastai_pci_mem_write(priv, priv->dies[0].die_index, addr, dev->dm.vir, dev->valid_data_length);
        {
            u64 offset        = 0;
	        int bar_num       = 0;
	        translate_noc_addr(priv,priv->pkg_id,addr, &bar_num, &offset);
            VASTAI_PCI_INFO(priv, priv->pkg_id,"[read]noc_addr= 0x%lx, bar[%d] offset[0x%llx] len=0x%x val=%#lx\n",
										addr, bar_num, offset, dev->valid_data_length, val);
        }
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [write] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
        else
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd [write] addr=0x%lx, "
                    "len=%d, val=0x%lx\n",
                    addr, dev->valid_data_length, val);
    }
    /* write buf */
    else if (!strncmp("write_buf", cmd, strlen("write_buf"))) {
        unsigned long user_addr;

        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        user_addr = vastai_pci_args_value(dev->args,
                          VASTAI_PCI_ARG_USER_ADDR);
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd write_buf len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }

        ret = vastai_pci_copy_from_user(dev->dm.vir, (char __user *)user_addr,
                       dev->valid_data_length);
        if (ret)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd pci_copy_from_user is error\n");
        ret = vastai_pci_mem_write(priv, priv->dies[0].die_index,
                addr, dev->dm.vir, dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [write_buf] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
        else
            VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd [write_buf] addr=0x%lx, "
                    "len=%d, val=0x%lx\n",
                    addr, dev->valid_data_length, val);
    }

    /* read cfg buf */
    else if (!strncmp("rdcfg_buf", cmd, sizeof("rdcfg_buf"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        dev->valid_data_length = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        ret = vastai_pci_config_read(priv, addr, dev->dm.vir,
                             dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }

    /* write cfg buf */
    else if (!strncmp("wrcfg_buf", cmd, strlen("wrcfg_buf"))) {
        unsigned long user_addr;

        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        user_addr = vastai_pci_args_value(dev->args,VASTAI_PCI_ARG_USER_ADDR);
        dev->valid_data_length = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd wrcfg_buf len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }

        ret = vastai_pci_copy_from_user(dev->dm.vir, (char __user *)user_addr,
                       dev->valid_data_length);
        if (ret)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd pci_copy_from_user is error\n");
        
        ret = vastai_pci_config_write(priv, addr, (char *)dev->dm.vir,
                              dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [wrcfg_buf] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);        
    }

    else if (!strncmp("smcu_read_buf", cmd, sizeof("smcu_read_buf"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        dev->valid_data_length = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        ret = vastai_read_by_smcu(priv, addr, dev->valid_data_length ,dev->dm.vir);;
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [smcu_read] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }

    else if (!strncmp("smcu_write_buf", cmd, sizeof("smcu_write_buf"))) {
        unsigned long user_addr;
        unsigned long val;
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        user_addr = vastai_pci_args_value(dev->args,VASTAI_PCI_ARG_USER_ADDR);
        dev->valid_data_length = 4;
        
        if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd smcu_write_buf len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }

        ret = vastai_pci_copy_from_user(dev->dm.vir, (char __user *)user_addr,
                       dev->valid_data_length);
        if (ret)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd pci_copy_from_user is error\n");        
        val = *(unsigned long*)dev->dm.vir;
        ret = vastai_write_by_smcu(priv, addr, val, 4);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [smcu_write_buf] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }

    else if (!strncmp("mmio_read_buf", cmd, sizeof("smcu_read_buf"))) {
        int bar_num=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_BAR);
        int offset=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_OFFSET);
        dev->valid_data_length = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, ""change to %d bytes\n",
                       dev->valid_data_length);
        }
        ret = vastai_pci_mmio_read(priv, bar_num, offset, dev->dm.vir, dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [mmio_read_buf] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }

    else if (!strncmp("mmio_write_buf", cmd, sizeof("mmio_write_buf"))) {
        unsigned long user_addr;
        unsigned long val;
        int bar_num=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_BAR);
        int offset=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_OFFSET);
        user_addr = vastai_pci_args_value(dev->args,VASTAI_PCI_ARG_USER_ADDR);
        dev->valid_data_length = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
            
        ret = vastai_pci_copy_from_user(dev->dm.vir, (char __user *)user_addr,
                       dev->valid_data_length);
        if (ret)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,"pci_proc_cmd pci_copy_from_user is error\n");

        val = *(unsigned long*)dev->dm.vir;
        ret = vastai_pci_mmio_write(priv, bar_num, offset , dev->dm.vir, dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [smcu_write_buf] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
    }

    else if (!strncmp("get_log", cmd, sizeof("get_log"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (len <= 0) {
            len = 4096;
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "get_log read len uses default 0x%lx bytes\n", len);
            
        } else if (len > 0x100000) {
            len = 0x100000;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "get_log read len overflow, change to 0x%lx bytes\n", len);
        }
        ret = vastai_get_log(priv, addr, len);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [get_log] 0x%lx, len=%ld\n",
                       addr, len);
    }

    /* show fw version information */
    else if (!strncmp("getver", cmd, strlen("getver"))) {
        if (priv->priv_hw_cfg->sys_cfg.devfn == 0 && priv->is_pf)
            vastai_pci_show_fw_ver(priv, 0x7FFF);
    }
    /* PCIe boot */
    else if (!strncmp("boot", cmd, strlen("boot"))) {
#ifdef CONFIG_VASTAI_PCI_BOOT
        vastai_pci_boot_fw(priv);
#endif
    }
    /* FLR */
    else if (!strncmp("flr", cmd, strlen("flr"))) {
        vastai_pci_flr(priv);
    }
    /* reset pcie bus & reinitial reset, send a cmd to smcu to enable hotR, then do hotR reset */
    else if (!strncmp("hotrset", cmd, strlen("hotrset"))) {
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "hotR reset is supported now\n");
        vastai_pci_reset_bus(priv);
        msleep(2000);
        vastai_pci_reinitial(priv, true);
    }
    /* reset pcie bus, hot reset */
    else if (!strncmp("hotset", cmd, strlen("hotset"))) {
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "hot reset is supported now\n");
        vastai_pci_reset_bus(priv);
    }
    /* reset soc */
    else if (!strncmp("uncond_reset", cmd, strlen("uncond_reset"))) {
        vastai_pci_reinitial(priv, false);
    }
    /* reinitial reset mcu and dsp */
    else if (!strncmp("reset", cmd, strlen("reset"))) {
        /* send msg_irq trigger a re-initial reset. */
        //vastai_device_reset(priv, VASTAI_RESET_MSG_TRIGER);
        vastai_pci_reinitial(priv, true);
    }

    /* reinitial reset mcu and dsp */
    else if (!strncmp("gfx_reset", cmd, strlen("gfx_reset"))) {  
        vastai_gfx_reset(priv);
    }

    /* reset and boot device */
    else if (!strncmp("reboot", cmd, strlen("reboot"))) {
        /* trigger a re-initial reset and reboot dev . */
#ifdef CONFIG_VASTAI_PCI_BOOT
        vastai_pci_reboot(priv);
#endif

#ifdef EMULATOR_ENV
        vastai_video_reset();
#endif
    }
    /* select pcie generation */
    else if (!strncmp("selgen", cmd, strlen("selgen"))) {
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        vastai_pci_generation_sel(priv, val);
    }
    /* get log level */
    else if (!strncmp("getloglevel", cmd, strlen("getloglevel"))) {
	struct va_gfx_render_info render_info;
       	struct va_gfx_display_info display_info;
        vastai_get_render_info(priv, &render_info);
        vastai_get_display_info(priv, &display_info);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "current loglevel=[0x%x]\n",
                va_pci_log_level);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "(bits correspond to levels in turn: perf, "
                "data, debug, normal)\n");
    }
    /* set log level, "setloglevel val=xxx" */
    else if (!strncmp("setloglevel", cmd, strlen("setloglevel"))) {
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        va_pci_log_level = val;
        if (va_pci_log_level == 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                "current loglevel=[0x%x], printing is closed\n",
                va_pci_log_level);
        else
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "set level=0x%lx, current loglevel=[0x%x]\n",
                val, va_pci_log_level);
    }

    else if (!strncmp("set_test_perf", cmd, strlen("set_test_perf"))) {
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        if (0 == val) {
            test_PCIe_Perf    = 0;
            test_smcu_RW_Perf = 0;
        } else {
            test_PCIe_Perf    = 1;
            test_smcu_RW_Perf = 1;
        }
        
    }

    else if (!strncmp("ring_buf_debug", cmd, strlen("ring_buf_debug"))) {
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        ring_buf_debug = val;
        
    }


    else if (!strncmp("malloclma", cmd, strlen("malloclma"))) {
#ifdef CONFIG_VASTAI_GFX
        u32 length_lma = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        IMG_UINT64 physlma_addr = 0;
        if (length_lma <= 0) {
            length_lma = VASTAI_UDMA_TEST_LEN;
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd malloclma default len=%d\n",
                    length_lma);
        }
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd malloclma len=%d\n",
                    length_lma);
        physlma_addr = alloc_physmem_by_ddk(priv, length_lma);
        if (physlma_addr == 0)
        {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd malloclma failed len=%d\n",
                    length_lma);
        }
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd malloclma addr = 0x%llx len=%d\n", physlma_addr, length_lma);

        free_physmem_by_ddk(priv, physlma_addr);
#endif
    }
    /* malloc a dma buf with specified addr 0/1 for testing. Requirement from DV xiaoru,li for SDMA test. May 6th,2022 */
    /* Usage e.g. mallocbuf addr=0 len=0 */
    else if (!strncmp("mallocbuf", cmd, strlen("mallocbuf"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        addr = (addr < 0 ? 0 : addr);
        
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            dev->valid_data_length = VASTAI_UDMA_TEST_LEN;
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd malloc default len=%d\n",
                    dev->valid_data_length);
        }
	if (priv->alloc) {
               uint64_t physheap_addr = priv->alloc(priv->ddk_node, dev->valid_data_length, 0);
               if (physheap_addr != 0) {
		       VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"pci_proc_cmd malloclma addr = 0x%llx \n", physheap_addr);
                       priv->free(priv->ddk_node, physheap_addr);
               }
        }
        if(addr == 0)
            vastai_pci_malloc_udma_buf(priv, &dmb[0], dev->valid_data_length);
        else
            vastai_pci_malloc_udma_buf(priv, &dmb[1], dev->valid_data_length);
    }
    /* set a dma buf for testing */
    else if (!strncmp("setbuf", cmd, strlen("setbuf"))) {        
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        addr = (addr < 0 ? 0 : addr);
        
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        
        if(addr == 0)
            vastai_pci_set_udma_buf(&dmb[0], val);
        else
            vastai_pci_set_udma_buf(&dmb[1], val);
    }
    /* print a specified dma buf for testing */
    else if (!strncmp("printbuf", cmd, strlen("printbuf"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        addr = (addr < 0 ? 0 : addr);
        
        if(addr == 0)
            vastai_pci_print_udma_buf(&dmb[0]);
        else
            vastai_pci_print_udma_buf(&dmb[1]);
    }
    /* free a specified dma buf for testing */
    else if (!strncmp("freebuf", cmd, strlen("freebuf"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        addr = (addr < 0 ? 0 : addr);

        if(addr == 0)
            vastai_pci_free_udma_buf(priv, &dmb[0]);
        else
            vastai_pci_free_udma_buf(priv, &dmb[1]);
    }
    /* test case */
    else if (!strncmp("test", cmd, strlen("test"))) {
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        ret = vastai_pci_test_case(priv, dev, val);
    }
    /**
     * disable smcu heartbeat counter for testing heartbeat
     * 0 - hang timer counter of smcu
     * 1 - hang thread counter of smcu
     */
    else if (!strncmp("hang_smcu", cmd, strlen("hang_smcu"))) {
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        if (val == 0)
            vastai_send_pcie_cmd(priv,
                    priv->dies[0].die_index,
                    VASTAI_PCIE_SUB_HANG_SMCU);
        else if (val == 1)
            vastai_send_pcie_cmd(priv,
                    priv->dies[0].die_index,
                    VASTAI_PCIE_SUB_HANG_SMCU_THREAD);
    }
    else if (!strncmp("toggle_dma_callback", cmd, strlen("toggle_dma_callback"))) {
        int i = 0;
        u32 flage = 0;

        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        if (val == 0) {
            flage = VASTAI_PCIE_SUB_CLOSE_DMA_CALLBACK;

        } else {
            flage = VASTAI_PCIE_SUB_OPEN_DMA_CALLBACK;
        }
        for (i = 0; i < priv->die_num; i++) {
            vastai_send_pcie_cmd(priv, priv->dies[i].die_index, flage);
        }
    }

    /**
     * cmd format e.g. "dma_test addr=0x4a000000 len=100 val=0x55 offset=0x33 mode=0"
     * addr: dev_addr
     * len: dma transfer size
     * val: byte
     * offset: mem_addr offset based on 4K
     * mode: 0 - d2m; 1 - m2d
     */
    else if (!strncmp("dma_test", cmd, strlen("dma_test"))) {
        unsigned long is_mem_to_dev = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
        unsigned long mem_offset = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_OFFSET);
        unsigned long dev_addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        unsigned long len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        unsigned long val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        va_pcie_dma_tranfer_test(priv,
                    is_mem_to_dev,
                    mem_offset,
                    dev_addr,
                    len,
                    val);
    }
    
    // Func:
    // Params: dir: 0 host-->dev ddr; 1 dev ddr --> host; 2 host -->host; 3 dev ddr --> dev ddr
    //              By default: dmb[0]: src addr;  dmb[1]: dest addr, if we need test MEM to MEM.
    //              dev_addr: dev DDR address.
    /*
     *
     * echo "sdma_cp mode=0,mmu=0,addr=0x1000000000,len=0x10000,case=10">/dev/kchar\:0 
     *
     */

	else if (!strncmp("sdma_rst", cmd, strlen("sdma_rst"))) {

        sdma_reset(priv);
    }
    else if (!strncmp("sdma_cp", cmd, strlen("sdma_cp"))) {
        u8 dir = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);  // val
        u8 mmu = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MMU);
        u64 addr0 = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR0);
		u64 addr1 = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR1);
        u32 trans_len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        u32 pid        = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);

        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
            "pci_proc_cmd, dir=%d, mmu_en=%d, addr0:%#llx,addr1:%#llx, trans_len=%#x\n", 
            dir, mmu, addr0, addr1, trans_len);

        va_sdma_op_cp_test(priv, dir, mmu, addr0, addr1, trans_len, pid);
    }

	else if (!strncmp("check_sdma", cmd, strlen("check_sdma"))) {
		check_cur_sdma(priv);
    }

    else if (!strncmp("sdma_fence", cmd, strlen("sdma_fence"))) {
        unsigned long dir = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
        unsigned long mmu = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MMU);
        unsigned long dev_addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        unsigned long trans_len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        unsigned long desc_cnt = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_CASE);
        unsigned long wr_data = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        
        va_sdma_op_fence_test(priv, dir, mmu, dev_addr, trans_len, desc_cnt, wr_data);
    }
    else if (!strncmp("sdma_fill", cmd, strlen("sdma_fill"))) {
        unsigned long dir = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
        unsigned long mmu = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MMU);
        unsigned long dst_addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        unsigned long trans_len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        unsigned long wr_data = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        va_sdma_op_fill_test(priv, dir, mmu, dst_addr, trans_len, wr_data);
    } 

    else if (!strncmp("page_entry", cmd, strlen("page_entry"))) {
        unsigned long dir = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
        unsigned long pa = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PA_ADDR);
        unsigned long va = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VA_ADDR);
        unsigned long size = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        unsigned long pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
		u64 start,end, va_pid;

        va_pid = va_create_pid(priv, pid);
		start = ktime_get();
        va_create_page_mapping(priv, pa, va, size, va_pid, (int)dir);
		end = ktime_get();

		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
			"create_page time cnt:%06llu(us)\n", (end - start)/1000);
    }

        else if (!strncmp("free_page", cmd, strlen("free_page"))) {
        u64 va = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VA_ADDR);
        int pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
		u32 size = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
		u64 start,end;
		
		start = ktime_get();
        va_free_page_maping(priv, va, size, pid);
		end = ktime_get();

		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
			"free_page time cnt:%06llu(us)\n", (end - start)/1000);
    }

		else if (!strncmp("get_pa", cmd, strlen("get_page"))) {
		u64 va = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VA_ADDR);
		int pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
		u64 pa;
	
		pa = va_virt_to_phys(priv, va, pid);
		if (pa)
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
			"The VA[0x%llx] mapping to PA[0x%llx]\n", va, pa);
		}

		else if (!strncmp("get_va", cmd, strlen("get_page"))) {
		u64 pa = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PA_ADDR);
		int pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
		u64 va;

		va = va_phys_to_virt(priv, pa, pid);
		if (va)
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
			"The PA[0x%llx] mapping to VA[0x%llx]\n", pa, va);
		}

		else if (!strncmp("kill_pid", cmd, strlen("kill_pid"))) {
			int pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
			va_kill_pid(priv, pid);
    	}

		else if (!strncmp("walk_page", cmd, strlen("walk_page"))) {
			int pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
			//walk_translation_info(priv, pid);
			walk_page_from_dev(priv, pid);
		}

		else if (!strncmp("get_page", cmd, strlen("get_page"))) {
			u64 va = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VA_ADDR);
			int pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
			get_page_entry_info(priv, va, pid);
		}

		else if (!strncmp("tlb_flush", cmd, strlen("tlb_flush"))) {
			unsigned long pid = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PID);
			unsigned long va = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VA_ADDR);
			unsigned long len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);

			/* TLB Flush */
			ret = smmu_fill_tlb_invalid_cmd_descriptor(priv, 0x30, pid, 1, va, len);
			if (ret)
			{
				VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "TLB Flush failed for va[%#lx]\n", va);
			}

		}


	/**
	 * cmd format e.g. "gmcu_cmd mode=0x10 addr=0x920000 val=0x55 len=16"
	 * val: sub_cmd
	 */
	else if (!strncmp("gmcu_cmd", cmd, strlen("gmcu_cmd"))) {
		int ret;
		struct pcie_transfer_cmd trans;
		unsigned long sub_cmd = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
		unsigned long addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
		unsigned long val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
		unsigned long len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);

		trans.w0.s_data0.optcode	= sub_cmd;
		trans.w0.s_data0.rev0		= len;
		trans.w0.s_data0.rev1		= 0;
		trans.w0.s_data0.rev2		= 0;
		trans.w1.data1				= addr;
		trans.w2.data2				= val;
		trans.w3.data3				= 0;

		if (SMCU_SWITCH_LOG_LEVEL == sub_cmd) {
			trans.w1.data1			= val;
		}
		
		if ( ONLY_1_PF==priv->fn_mode) {
			ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF,  &trans, 0);
		} else if (ONLY_2_PF==priv->fn_mode){
			ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
		} else {
			ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF,  &trans, 0);
		}

		if (ret) {
			VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "push element fail\n");
		}

	}

    /**
     * cmd format e.g. "vsync_en val=60"
     * val: freq_hz
     */
    else if (!strncmp("vsync_en", cmd, strlen("vsync_en"))) {
        int ret;
        unsigned long val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        ret = va_vsync_timer_enable(priv, val);
        if (ret) {
            VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "vsync_en fail\n");
        }
    }

    /**
     * cmd format e.g. "vsync_dis val=60"
     * val: freq_hz
     */
    else if (!strncmp("vsync_dis", cmd, strlen("vsync_dis"))) {
        int ret;

        ret = va_vsync_timer_disable(priv);
        if (ret) {
            VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "vsync_dis fail\n");
        }
    }
    
    /**
     * cmd format e.g. "memmap" > /dev/kchar:\0
     * val: sub_cmd
     */
    else if (!strncmp("memmap", cmd, strlen("memmap"))) {
        dump_bar_addr_map(priv);
    }

	/**
	 * cmd format e.g. "hw_cfg" > /dev/kchar:\0
	 * val: sub_cmd
	 */
	else if (!strncmp("hw_cfg", cmd, strlen("hw_cfg"))) {
		dump_current_fn_hw_cfg_info(priv);
	}
	
	/**
	 * cmd format e.g. "dump_atu" > /dev/kchar:\0
	 * val: sub_cmd
	 */
	else if (!strncmp("dump_atu", cmd, strlen("dump_atu"))) {
		dump_atu(priv);
	}

    else if (!strncmp("mpu_cfg", cmd, strlen("mpu_cfg"))) {
        unsigned long mode = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
        if(0 == mode) {
            smmu_mpu_cfg(priv);
            gfx_mpu_cfg(priv);
        } else if (1 == mode) {
            smmu_mpu_cfg(priv);
        } else if(2 == mode){
            gfx_mpu_cfg(priv);
        }       
	}

    else if (!strncmp("ddr_cfg", cmd, strlen("ddr_cfg"))) {
		dump_ddr_info(priv);
        //ddr_partition_info(priv);
	}
	
    /**
     * cmd format e.g. "kchar_map" > /dev/kchar:\0
     */
    else if (!strncmp("kchar_map", cmd, strlen("kchar_map"))) {
        dump_pf_vf_map_kchar(priv);
    }
    else if (!strncmp("ste_info", cmd, strlen("ste_info"))) {
        int ret;

        ret = device_smmu_ste_table_init(priv);
        if(ret!=0){
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"update ste info failed\n");
        }        
    }

    else if (!strncmp("cmdq_test", cmd, strlen("cmdq_test"))) {
        int ret;

        ret = smmu_fill_cfg_invalid_cmd_descriptor(priv, 0x6, 0, 0);
        if(ret!=0){
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"cmdq test failed\n");
        }        
    }

    else if(!strncmp("check_soc", cmd, strlen("check_soc"))) {
        check_soc(priv);
    }

	/**
	 * cmd format e.g. "smcu_cmd mode=0x10 addr=0x920000 val=0x55 len=16"
	 * mode: sub_cmd
	 */
	else if (!strncmp("smcu_cmd", cmd, strlen("smcu_cmd"))) {
		int ret;
		struct pcie_transfer_cmd trans;
		unsigned long sub_cmd		= vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_MODE);
		u64 addr					= vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
		unsigned long len			= vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
		unsigned long val			= vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

		trans.w0.s_data0.optcode    = sub_cmd;
		if (SMCU_CONFIG_ENTRY == sub_cmd) {
			trans.w1.data1			= addr >> 32;
			trans.w2.data2			= addr & 0xFFFFFFFF;
			trans.w3.data3			= len;
		}
		else if (SMCU_SWITCH_LOG_LEVEL == sub_cmd) {
			trans.w1.data1			= val;
		}
		else {
			trans.w0.s_data0.rev0	= len;
			trans.w0.s_data0.rev1	= 0;
			trans.w0.s_data0.rev2	= 0;
			trans.w1.data1			= addr&0XFFFFFFFF;
			trans.w2.data2			= addr>>32;
			trans.w3.data3			= val;
		}

		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF,  &trans, 0);
		if(ret!=0){
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"smcu_cmd error\n");
		}
	}

    /**
     * cmd format e.g. "mmio_read bar=0x2 offset=0x0 len=0x100"
     * val: sub_cmd
     */
    else if (!strncmp("mmio_read", cmd, strlen("mmio_read"))) {
        int ret=0;
        int bar_num=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_BAR);
        int offset=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_OFFSET);
        int len=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (len <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            len = 4;
        } else if (len> dev->dm.size) {
            len = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       len);
        }
        ret = vastai_pci_mmio_read(priv, bar_num, offset, dev->dm.vir, len);

        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [mmio_read] 0x%x, len=%d, error:%d\n",
                       offset, len, ret);
        else {
            for (i = 0; i < len; i += 16) {
                sprintf(print_str,
                    "%s [mmio_read] 0x%010x: ",
                    VASTAI_PCI_DEBUG_PREFIX, (offset + i));
                    VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
                            DUMP_PREFIX_NONE, 16, 4,
                            dev->dm.vir + i, 16, true);
            }
        }

    }

    /* mmio  write reg:
     * cmd format e.g. "mmio_write bar=0x2 offset=0x0 len=4"
     */
    else if (!strncmp("mmio_write", cmd, sizeof("mmio_write"))) {
        int ret=0;
        int bar_num=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_BAR);
        int offset=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_OFFSET);
        int len=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        int val=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        if (len <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd write len uses default 4 bytes\n");
            len = 4;
        } else if (len > dev->dm.size) {
            len = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd write len overflow, "
                       "change to %d bytes\n",
                       len);
        }
        for (i = 0; i < len; i += 4) {
            ((char *)dev->dm.vir)[i] = val & 0xFF;
            if ((i + 1) == len)
                break;
            ((char *)dev->dm.vir)[i + 1] = (val >> 8) & 0xFF;
            if ((i + 2) == len)
                break;
            ((char *)dev->dm.vir)[i + 2] = (val >> 16) & 0xFF;
            if ((i + 3) == len)
                break;
            ((char *)dev->dm.vir)[i + 3] = (val >> 24) & 0xFF;
        }
        ret = vastai_pci_mmio_write(priv, bar_num, offset , dev->dm.vir, len);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [mmio_write] 0x%x, len=%d, error:%d\n",
                       offset, len, ret);
        else
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd [mmio_write] offset=0x%x, "
                    "len=%d, val=0x%xbarx\n",
                    offset, len, val);
    }

    /**
     * cmd format e.g. "malloc len=0x1000 val=0x55aaa "
     */
    else if (!strncmp("malloc", cmd, strlen("malloc"))) {
        u64 host_buf_pa;
        int len  = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        int val  = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        if(0 == len) {
            len = 0x100;
        }

        if(0 == val){
            val = 0x12345678;
        }
        g_host_buf_va = kzalloc(len, GFP_KERNEL);
        host_buf_pa   = virt_to_phys(g_host_buf_va);
        VASTAI_PCI_INFO(priv, priv->pkg_id,"g_host_buf_va=0x%p host_buf_pa=%#llx\n", g_host_buf_va, host_buf_pa);
        for(i = 0;i < (len/4); i++){
            *(u32*)g_host_buf_va  = 0x12345678;
            g_host_buf_va        += 4;
        }

    }

    /**
     * cmd format e.g. "free"
     */
    else if (!strncmp("free", cmd, strlen("free"))) {
        kfree(g_host_buf_va);
    }

    
    /**
     * cmd format e.g. "smcu_read addr=0x800000 len=0x4"
     */
    else if (!strncmp("smcu_read", cmd, strlen("smcu_read"))) {
        int ret  = 0;
        u64 addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        int len  = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (len <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            len = 4;
        } else if (len> dev->dm.size) {
            len = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       len);
        }
        ret = vastai_read_by_smcu(priv, addr, len ,dev->dm.vir);

        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [smcu_read] 0x%llx, len=%d, error:%d\n",
                       addr, len, ret);
        else {
            for (i = 0; i < len; i += 16) {
                sprintf(print_str,
                    "%s [smcu_read] 0x%010llx: ",
                    VASTAI_PCI_DEBUG_PREFIX, (addr + i));
                    VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
                            DUMP_PREFIX_NONE, 16, 4,
                            dev->dm.vir + i, 16, true);
            }
        }
    }

    /* smcu  write reg:
     * cmd format e.g. "smcu_write addr=0x800000 val=0x55aa"
     */
    else if (!strncmp("smcu_write", cmd, sizeof("smcu_write"))) {
        int ret     = 0;
        u64 addr    = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        unsigned long val   = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        u32 len     = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        ret         = vastai_write_by_smcu(priv, addr, val, len);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [smcu_write] 0x%llx, error:%d\n",
                       addr, ret);
        else
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd [smcu_write] addr=0x%llx, "
                    "val=0x%lx\n",
                    addr,  val);
    }

    /**
     * cmd format e.g. "read_slave_cfg offset=0x0 len=0x100"
     * val: sub_cmd
     */
    else if (!strncmp("read_slave_cfg", cmd, strlen("read_slave_cfg"))) {
        int ret=0;
        int offset=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_OFFSET);
        int len=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (len <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            len = 4;
        } else if (len> dev->dm.size) {
            len = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       len);
        }
        ret = vastai_slave_config_read(priv, offset, dev->dm.vir, len);

        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read_slave_cfg] 0x%x, len=%d, error:%d\n",
                       offset, len, ret);
        else {
            for (i = 0; i < len; i += 16) {
                sprintf(print_str,
                    "%s [read_slave_cfg] 0x%010x: ",
                    VASTAI_PCI_DEBUG_PREFIX, (offset + i));
                    VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
                            DUMP_PREFIX_NONE, 16, 4,
                            dev->dm.vir + i, 16, true);
            }
        }
    }

    /**
     * cmd format e.g. "fread addr=0x0 len=0x100"
     * val: sub_cmd
     */
    else if (!strncmp("fread", cmd, sizeof("fread"))) {
        int ret         = 0 ;
        u64 fn_addr     = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        int len         = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        u64 noc_addr    = 0;
        if (len <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd fread len uses default 4 bytes\n");
            len = 4;
        } else if (len> dev->dm.size) {
            len = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd fread len overflow, "
                       "change to %d bytes\n",
                       len);
        }
        
        ret = vastai_pci_mem_fread(priv, fn_addr, dev->dm.vir ,len);

        {
            u64 offset      = 0;
            int bar_num     = 0;
            int region_num  = 0;
            translate_fn_addr(priv, fn_addr , &bar_num, &offset, &region_num);
            if (priv->bar[bar_num].region_cnt > 0) {
			    noc_addr 	  = fn_addr - priv->bar[bar_num].regions[region_num].fn_base_addr + priv->bar[bar_num].regions[region_num].noc_addr;
            } else {
                noc_addr 	  = fn_addr - priv->bar[bar_num].fn_base_addr + priv->bar[bar_num].noc_addr;			
            }

            {
                VASTAI_PCI_INFO(priv, priv->pkg_id,"[fread]fn_addr=0x%llx noc_addr=0x%llx, bar[%d] offset[%#llx] region[%d] len=0x%d \n", 
                                                fn_addr, noc_addr, bar_num, offset, region_num, len);
            }
        }

        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [fread] 0x%llx, len=%d, error:%d\n",
                       fn_addr, len, ret);
        else {
            for (i = 0; i < len; i += 16) {
                sprintf(print_str,
                    "%s [fread] faddr:0x%010llx noc:0x%010llx: ",
                    VASTAI_PCI_DEBUG_PREFIX, (fn_addr + i),( noc_addr + i));
                    VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
                            DUMP_PREFIX_NONE, 16, 4,
                            dev->dm.vir + i, 16, true);
            }
        }

    }
    /* fread buf */
    else if (!strncmp("fread_buf", cmd, sizeof("fread_buf"))) {
        int ret                = 0 ;
        u64 fn_addr            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        dev->valid_data_length = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);

        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd fread len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length> dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd fread len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        
        ret = vastai_pci_mem_fread(priv, fn_addr, dev->dm.vir ,dev->valid_data_length);

        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [fread] 0x%llx, len=%d, error:%d\n",
                       fn_addr, dev->valid_data_length, ret);

    }
    /* fwrite reg:
     * cmd format e.g. "fwrite addr=0x0 val=0x55 len=4"
     */
    else if (!strncmp("fwrite", cmd, sizeof("fwrite"))) {
        int ret=0;
        u64 fn_addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        int len     = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        int val     = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        if (len <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd fwrite len uses default 4 bytes\n");
            len = 4;
        } else if (len > dev->dm.size) {
            len = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd fwrite len overflow, "
                       "change to %d bytes\n",
                       len);
        }
        for (i = 0; i < len; i += 4) {
            ((char *)dev->dm.vir)[i] = val & 0xFF;
            if ((i + 1) == len)
                break;
            ((char *)dev->dm.vir)[i + 1] = (val >> 8) & 0xFF;
            if ((i + 2) == len)
                break;
            ((char *)dev->dm.vir)[i + 2] = (val >> 16) & 0xFF;
            if ((i + 3) == len)
                break;
            ((char *)dev->dm.vir)[i + 3] = (val >> 24) & 0xFF;
        }
        ret = vastai_pci_mem_fwrite(priv, fn_addr, dev->dm.vir, len);
        //this is for debug
        {
            u64 offset      = 0;
            int bar_num     = 0;
            int region_num  = 0;
            u64 noc_addr    = 0;
             if (priv->bar[bar_num].region_cnt > 0) {
			    noc_addr = fn_addr - priv->bar[bar_num].regions[region_num].fn_base_addr + priv->bar[bar_num].regions[region_num].noc_addr;
            } else {
                noc_addr = fn_addr - priv->bar[bar_num].fn_base_addr + priv->bar[bar_num].noc_addr;			
            }

            {
                VASTAI_PCI_INFO(priv, priv->pkg_id,"[fwrite]fn_addr=0x%llx noc_addr=0x%llx, bar[%x] offset[%llx] region[%x] len=0x%d \n", 
                                                fn_addr, noc_addr, bar_num, offset, region_num, len);
            }
        }

        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [fwrite] 0x%llx, len=%d, error:%d\n",
                       fn_addr, len, ret);
        else
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd [fwrite] fn_addr=0x%llx, "
                    "len=%d, val=0x%xbarx\n",
                    fn_addr, len, val);
    }
    /* fwrite buf */
    else if (!strncmp("fwrite_buf", cmd, sizeof("fwrite_buf"))) {
        int ret                 = 0;
        unsigned long user_addr = 0;
        u64 fn_addr             = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        user_addr               = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_USER_ADDR);
        dev->valid_data_length  = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);

        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd fwrite len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd fwrite len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }

        ret = vastai_pci_copy_from_user(dev->dm.vir, (char __user *)user_addr,
                       dev->valid_data_length);
        if (ret)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd pci_copy_from_user is error\n");

        ret = vastai_pci_mem_fwrite(priv, fn_addr, dev->dm.vir, dev->valid_data_length);

        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [fwrite] 0x%llx, len=%d, error:%d\n",
                       fn_addr, dev->valid_data_length, ret);
    }

     /* write_slave_cfg reg:
     * cmd format e.g. "write_slave_cfg offset=0x0 val=0x55 len=4"
     */
    else if (!strncmp("write_slave_cfg", cmd, sizeof("write_slave_cfg"))) {
        int ret=0;
        int offset=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_OFFSET);
        int len=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        int val=vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        if (len <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd write len uses default 4 bytes\n");
            len = 4;
        } else if (len > dev->dm.size) {
            len = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd write len overflow, "
                       "change to %d bytes\n",
                       len);
        }
        for (i = 0; i < len; i += 4) {
            ((char *)dev->dm.vir)[i] = val & 0xFF;
            if ((i + 1) == len)
                break;
            ((char *)dev->dm.vir)[i + 1] = (val >> 8) & 0xFF;
            if ((i + 2) == len)
                break;
            ((char *)dev->dm.vir)[i + 2] = (val >> 16) & 0xFF;
            if ((i + 3) == len)
                break;
            ((char *)dev->dm.vir)[i + 3] = (val >> 24) & 0xFF;
        }
        ret = vastai_slave_config_write(priv, offset , dev->dm.vir, len);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [write_slave_cfg] 0x%x, len=%d, error:%d\n",
                       offset, len, ret);
        else
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd [write_slave_cfg] offset=0x%x, "
                    "len=%d, val=0x%xbarx\n",
                    offset, len, val);
    }

#ifdef DIAG_TEST_MODE
    /**
     * cmd format e.g. "diag_read addr=0x800000 len=0x100"
     * val: sub_cmd
     */
    else if (!strncmp("diag_read", cmd, strlen("diag_read"))) {
        addr  = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        if (addr>0 && addr<0x10000000) {
            addr += 0xFF0000000;
        }

        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd read len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd read len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        ret = vastai_pci_mem_read(priv, priv->dies[0].die_index, addr, dev->dm.vir, dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
        else {
            for (i = 0; i < dev->valid_data_length; i += 16) {
                sprintf(print_str,
                    "%s [read] 0x%010lx: ",
                    VASTAI_PCI_DEBUG_PREFIX, (addr + i));
                VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
                            DUMP_PREFIX_NONE, 16, 4,
                            dev->dm.vir + i, 16, true);
            }
        }

    }

    /* write reg:
     * cmd format e.g. "diag_write addr=0x800000 val=0x0 len=100"
     */
    else if (!strncmp("diag_write", cmd, sizeof("diag_write"))) {
        addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        if (addr>0 && addr<0x10000000) {
            addr += 0xFF0000000;
        }
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        dev->valid_data_length
            = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        if (dev->valid_data_length <= 0) {
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd write len uses default 4 bytes\n");
            dev->valid_data_length = 4;
        } else if (dev->valid_data_length > dev->dm.size) {
            dev->valid_data_length = dev->dm.size;
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd write len overflow, "
                       "change to %d bytes\n",
                       dev->valid_data_length);
        }
        for (i = 0; i < dev->valid_data_length; i += 4) {
            ((char *)dev->dm.vir)[i] = val & 0xFF;
            if ((i + 1) == dev->valid_data_length)
                break;
            ((char *)dev->dm.vir)[i + 1] = (val >> 8) & 0xFF;
            if ((i + 2) == dev->valid_data_length)
                break;
            ((char *)dev->dm.vir)[i + 2] = (val >> 16) & 0xFF;
            if ((i + 3) == dev->valid_data_length)
                break;
            ((char *)dev->dm.vir)[i + 3] = (val >> 24) & 0xFF;
        }
        ret = vastai_pci_mem_write(priv, priv->dies[0].die_index, addr, dev->dm.vir, dev->valid_data_length);
        if (ret < 0)
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [write] 0x%lx, len=%d, error:%d\n",
                       addr, dev->valid_data_length, ret);
        else
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd [write] addr=0x%lx, "
                    "len=%d, val=0x%lx\n",
                    addr, dev->valid_data_length, val);
    }

#endif
    /**
     * cmd format e.g. "mem_test addr=0x4a000000 len=100 val=0x55"
     * write val, then read to compare
     */
    else if (!strncmp("mem_test", cmd, strlen("mem_test"))) {
        unsigned long addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
        unsigned long len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
        unsigned long val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
        char *w_buf;
        char *r_buf;
        int ret;
        int i;

        w_buf = kmalloc(len * 2, GFP_KERNEL);
        if (!w_buf) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                    "pci_proc_cmd buf alloc fail, 0x%lx\n",
                    len * 2);
            return -ENOMEM;
        }
        r_buf = w_buf + len;
        memset(w_buf, val, len);

        for (i = 0; i < len; i += 4) {
            vastai_pci_mem_write(priv, 0, addr + i, w_buf, 4);
            vastai_pci_mem_read(priv, 0, addr + i, r_buf, 4);
            if(memcmp(w_buf, r_buf, 4)) {
                VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                    "mem_test don't pass, error_addr=%lx\n",
                    addr + i);
                goto err_ret;
            }
        }

        ret = vastai_pci_mem_write(priv, 0, addr, w_buf, len);
        if (ret < 0) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [write] 0x%lx, len=%ld, error:%d\n",
                       addr, len, ret);
            goto err_ret;
        }
        ret = vastai_pci_mem_read(priv, 0, addr, r_buf, len);
        if (ret < 0) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "pci_proc_cmd [read] 0x%lx, len=%ld, error:%d\n",
                       addr, len, ret);
            goto err_ret;
        }
        if(memcmp(w_buf, r_buf, len)) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                    "mem_test don't pass\n");
            goto err_ret;
        }

        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "mem_test pass\n");

        return 0;
err_ret:
        kfree(w_buf);
        return -EIO;
    }
    else if (!strncmp("dma_payload", cmd, strlen("dma_payload"))) {
        va_dma_t *dma;
        dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV |
                        VA_DMA_ATTR_DIR_DEV_TO_MEM);

        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                "dma paylaod: tx=0x%llxB/s, rx=0x%llxB/s\n",
                dma->monitor.tx_byte, dma->monitor.rx_byte);

        va_dma_free(priv, dma);
    }

	else if (!strncmp("iommu_en", cmd, strlen("iommu_en"))) 
	{
		priv->iommu_en = true;
	}

    /* spi fw update */
    else if (!strncmp("xspi_base", cmd, strlen("xspi_base")))
    {
        /* update 1.5MB */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---base\n");
        vastai_pci_test_flashxspi_base(dev, priv);
    }
    else if (!strncmp("xspi_bl0", cmd, strlen("xspi_bl0")))
    {
        /* update bl0 508KB */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---bl0\n");
        vastai_pci_test_flashxspi_bl0(dev, priv);
    }
    else if (!strncmp("xspi_vbios", cmd, strlen("xspi_vbios")))
    {
        /* update vbios 512KB */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---vbios\n");
        vastai_pci_test_flashxspi_vbios(dev, priv);
    }
    else if (!strncmp("xspi_full", cmd, strlen("xspi_full")))
    {
        /* update 2MB */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---full\n");
        vastai_pci_test_flashxspi_full(dev, priv);
    }
    else if (!strncmp("xspi_all", cmd, strlen("xspi_all")))
    {
        /* update 2.5MB */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---all\n");
        vastai_pci_test_flashxspi_all(dev, priv);
    }
    else if (!strncmp("xspi_pmcu", cmd, strlen("xspi_pmcu")))
    {
        /* update 512KB */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---pmcu\n");
        vastai_pci_test_flashxspi_pmcu(dev, priv);
    }
    else if (!strncmp("bmcu_upgrade", cmd, strlen("bmcu_upgrade")))
    {
        /* upgrade bmcu */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---bmcu_upgrade\n");
        vastai_pci_test_flashbmcu(dev, priv);
    }
    else if (!strncmp("bmcu_degrade", cmd, strlen("bmcu_degrade")))
    {
        /* degrade bmcu */
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"pci_proc_cmd ---bmcu_degrade\n");
        vastai_pci_test_switchbmcu(dev, priv);
    }
	else if (!strncmp("pcie_perf_begin", cmd, strlen("pcie_perf_begin"))) {
		vastai_pcie_bw_test(priv);
	}
	/*
	 * cmd format e.g. "pcie_perf val=1000"
	 * val : monitor us tick
	 */
	else if (!strncmp("pcie_perf", cmd, strlen("pcie_perf"))) {
		val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "==================pcie_perf==================\n");
		/* UPSTREAM */
		ret = vastai_pcie_monitor(priv, val, 1, &read_bw_cont);
		if (ret < 0)
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pcie_monitor : upstream fail\n");

		/* DOWNSTREAM */
		ret = vastai_pcie_monitor(priv, val, 0, &write_bw_cont);
		if (ret < 0)
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pcie_monitor : downstream fail\n");

		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"PCIE  | upstream bandwidth   |  %lld MB/s\n",
						read_bw_cont * 1000000 / val / 1024 / 1024);

		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"PCIE  | downstream bandwidth |  %lld MB/s\n",
						write_bw_cont * 1000000 / val / 1024 / 1024);

		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"PCIE total bandwidth         |  %lld MB/s\n",
						(read_bw_cont + write_bw_cont) * 1000000 / val / 1024 / 1024);
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "---------------------------------------------\n");
	}
	else if (!strncmp("ddr_perf_begin", cmd, strlen("ddr_perf_begin"))) {
		vastai_ddr_bw_test(priv);
	}
	/*
	 * cmd format e.g. "ddr_perf val=1000 ch=2"
	 * val : monitor us tick
	 * ch: ddr channel, when 4 ddr use channel 2 6 3 7
	 */
	else if (!strncmp("ddr_perf", cmd, strlen("ddr_perf"))) {
		val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
		ch = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_CHANNEL);

        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "==================ddr_perf==================\n");
        for (i = 0; i < 4; i++) {
            /* UPSTREAM */
            ret = vastai_ddr_monitor(priv, val, 1, ch, i, &read_bw_cont);
            if (ret < 0)
                VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "ddr_monitor : upstream fail\n");
            upstream_count += read_bw_cont;

            /* DOWNSTREAM */
            ret = vastai_ddr_monitor(priv, val, 0, ch, i, &write_bw_cont);
            if (ret < 0)
                VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "ddr_monitor : downstream fail\n");
            downstream_count += write_bw_cont;

            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                				"DDR%lu port%d    |  read bandwidth    |  %lld MB/s\n",
                				ch, i, read_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                				"DDR%lu port%d    |  write bandwidth   |  %lld MB/s\n",
                				ch, i, write_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "---------------------------------------------\n");

        }
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"DDR%lu all port |  read bandwidth    |  %lld MB/s\n",
            				ch, upstream_count * 1000000 / val / 1024 / 1024);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"DDR%lu all port |  write bandwidth   |  %lld MB/s\n",
            				ch, downstream_count * 1000000 / val / 1024 / 1024);
		
	}
    /*
	 * cmd format e.g. "sdma_perf val=1000"
	 * val: monitor us tick
	 */
    else if (!strncmp("sdma_perf", cmd, strlen("sdma_perf"))){
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "==================sdma_perf==================\n");
        for (i = 0; i < 2; i++) {
            /* UPSTREAM */
            ret = vastai_sdma_monitor(priv, val, 1, i, &read_bw_cont);
            if (ret < 0)
            	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "sdma_monitor : upstream fail\n");
            total_read_count += read_bw_cont;

            /* DOWNSTREAM */
            ret = vastai_sdma_monitor(priv, val, 0, i, &write_bw_cont);
            if (ret < 0)
            	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "sdma_monitor : downstream fail\n");
            total_write_count += write_bw_cont;

            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"SDMA%d  |  read bandwidth    |  %lld MB/s\n",
            				i, read_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"SDMA%d  |  write bandwidth   |  %lld MB/s\n",
            				i, write_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "---------------------------------------------\n");
        }
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"SDMA total read bandwidth   |  %lld MB/s\n",
						total_read_count * 1000000 / val / 1024 / 1024);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"SDMA total write bandwidth  |  %lld MB/s\n",
						total_write_count * 1000000 / val / 1024 / 1024);
    }
    /*
	 * cmd format e.g. "venc_perf val=1000"
	 * val: monitor us tick
	 */
    else if (!strncmp("venc_perf", cmd, strlen("venc_perf"))){

        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);

        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "==================venc_perf==================\n");
        for (i = 0; i < 4; i++) {
            /* UPSTREAM */
            ret = vastai_venc_monitor(priv, val, 1, i, &read_bw_cont);
            if (ret < 0)
            	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "venc_monitor : upstream fail\n");
            total_read_count += read_bw_cont;

            /* DOWNSTREAM */
            ret = vastai_venc_monitor(priv, val, 0, i, &write_bw_cont);
            if (ret < 0)
            	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "venc_monitor : downstream fail\n");
            total_write_count += write_bw_cont;

            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"VENC%d  |  read bandwidth    |  %lld MB/s\n",
            				i, read_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"VENC%d  |  write bandwidth   |  %lld MB/s\n",
            				i, write_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "---------------------------------------------\n");
        }
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"VENC total read bandwidth   |  %lld MB/s\n",
						total_read_count * 1000000 / val / 1024 / 1024);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"VENC total write bandwidth  |  %lld MB/s\n",
						total_write_count * 1000000 / val / 1024 / 1024);
    }
    /*
	 * cmd format e.g. "vdec_perf val=1000"
	 * val: monitor us tick
	 */
    else if (!strncmp("vdec_perf", cmd, strlen("vdec_perf"))){
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
    
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "==================vdec_perf==================\n");
        for (i = 0; i < 2; i++) {
            /* UPSTREAM */
            ret = vastai_vdec_monitor(priv, val, 1, i, &read_bw_cont);
            if (ret < 0)
            	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "vdec_monitor : upstream fail\n");
            total_read_count += read_bw_cont;

            /* DOWNSTREAM */
            ret = vastai_vdec_monitor(priv, val, 0, i, &write_bw_cont);
            if (ret < 0)
            	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "vdec_monitor : downstream fail\n");
            total_write_count += write_bw_cont;

            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"VDEC%d  |  read bandwidth    |  %lld MB/s\n",
            				i, read_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"VDEC%d  |  write bandwidth   |  %lld MB/s\n",
            				i, write_bw_cont * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "---------------------------------------------\n");
        }
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"VDEC total read bandwidth   |  %lld MB/s\n",
						total_read_count * 1000000 / val / 1024 / 1024);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"VDEC total write bandwidth  |  %lld MB/s\n",
						total_write_count * 1000000 / val / 1024 / 1024);
    }
    /*
	 * cmd format e.g. "gfx_perf val=1000"
	 * val: monitor us tick
	 */
    else if (!strncmp("gfx_perf", cmd, strlen("gfx_perf"))){
        val = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_VAL);
    
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "===================gfx_perf==================\n");
        for (i = 0; i < 4; i++) {
            upstream_count = 0;
            downstream_count = 0;
            for (j = 0; j < 2; j++) {
                /* UPSTREAM */
                ret = vastai_gfx_monitor(priv, val, 1, i, j, &read_bw_cont);
                if (ret < 0)
                	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "gfx_monitor : upstream fail\n");
                upstream_count += read_bw_cont;

                /* DOWNSTREAM */
                ret = vastai_gfx_monitor(priv, val, 0, i, j, &write_bw_cont);
                if (ret < 0)
                	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "gfx_monitor : downstream fail\n");
                downstream_count += write_bw_cont;

                VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                				"GFX%d port%d    |  read bandwidth    |  %lld MB/s\n",
                				i, j, read_bw_cont * 1000000 / val / 1024 / 1024);
                VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
                				"GFX%d port%d    |  write bandwidth   |  %lld MB/s\n",
                				i, j, write_bw_cont * 1000000 / val / 1024 / 1024);
                VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "---------------------------------------------\n");
            }
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"GFX%d all port |  read bandwidth    |  %lld MB/s\n",
            				i, upstream_count * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            				"GFX%d all port |  write bandwidth   |  %lld MB/s\n",
            				i, downstream_count * 1000000 / val / 1024 / 1024);
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "---------------------------------------------\n");
            total_read_count += upstream_count;
            total_write_count += downstream_count;
        }
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"GFX total read bandwidth    |  %lld MB/s\n",
						total_read_count * 1000000 / val / 1024 / 1024);
        VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, 
						"GFX total write bandwidth   |  %lld MB/s\n",
						total_write_count * 1000000 / val / 1024 / 1024);
    }
	/*
	 * cmd format e.g. "set_cii addr=0x78 pf=1 ch=3"
	 */
	else if (!strncmp("set_cii", cmd, strlen("set_cii"))) {
		addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
		pf = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_PF);
		ch = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_CHANNEL);

		ret = vastai_cii_set(priv, ch, addr, pf);
		if (ret < 0)
				VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "cii_set fail\n");
	}
	else if (!strncmp("inventory_info", cmd, strlen("inventory_info"))) {
		show_inventory_info(priv);
	}
	/*
	 * cmd format e.g. "cedar addr=0x104a000000 len=0x100 addr0=0x104b000000"
	 */
	else if (!strncmp("cedar", cmd, strlen("cedar"))) {
		u64 hash[8] = {0};
		u64 addr0 = 0;
		struct vastai_cedar_img_info alg_info = {0};

		addr = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR);
		len = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_LEN);
		addr0 = vastai_pci_args_value(dev->args, VASTAI_PCI_ARG_ADDR0);

		alg_info.alg_type = SHA3_224_IMG_1088,
		alg_info.input_addr = (uint64_t)addr,
		alg_info.output_addr = (uint64_t)addr0,
		alg_info.blen = len;

		ret = vastai_dev_cedar_send_cmd(priv, &alg_info, 0, 0, hash);
		if (ret < 0)
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "dev_cedar_send_cmd fail\n");
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "hash value: 0x%llx 0x%llx 0x%llx 0x%llx\r\n", hash[0], hash[1], hash[2], hash[3]);
	}
	else if (!strncmp("print_tree_info", cmd, strlen("print_tree_info"))) {
		struct vastai_card_info *card_loop = NULL;
		int card = 0;

		printk("global:");
		mutex_lock(&vastai_card_info_list_lock);
		list_for_each_entry(card_loop, &vastai_card_info_list, card_node) {
			struct vastai_pkg_info *pkg_loop = NULL;
			u8 *card_string = list_is_last(&card_loop->card_node, &vastai_card_info_list) ? " " : "│";
			u8 *curr_card_string = list_is_last(&card_loop->card_node, &vastai_card_info_list) ? "└──" : "├──";

			printk("%s card%d\n", curr_card_string, card++);
			list_for_each_entry(pkg_loop, &card_loop->vastai_pkg_info_list, pkg_node) {
				struct vastai_die_info *die_loop = NULL;
				u8 *pkg_string = list_is_last(&pkg_loop->pkg_node, &card_loop->vastai_pkg_info_list) ? " " : "│";
				u8 *curr_pkg_string = list_is_last(&pkg_loop->pkg_node, &card_loop->vastai_pkg_info_list) ? "└──" : "├──";

				printk("%s   %s pkg%d\n", card_string, curr_pkg_string, pkg_loop->pkg_id);

				list_for_each_entry(die_loop, &pkg_loop->vastai_die_info_list, die_node) {
					struct vastai_pci_info *fn_loop = NULL;
					u8 *die_string = list_is_last(&die_loop->die_node, &pkg_loop->vastai_die_info_list) ? " " : "│";
					u8 *curr_die_string = list_is_last(&die_loop->die_node, &pkg_loop->vastai_die_info_list) ? "└──" : "├──";

					printk("%s   %s   %s die%d\n", card_string, pkg_string, curr_die_string, die_loop->die_id);

					list_for_each_entry(fn_loop, &die_loop->vastai_fn_info_list, fn_node[die_loop->die_id]) {
						u8 *curr_fn_string = list_is_last(&fn_loop->fn_node[die_loop->die_id], &die_loop->vastai_fn_info_list) ? "└──" : "├──";

						printk("%s   %s   %s   %s fn%d\n", card_string, pkg_string, die_string, curr_fn_string, fn_loop->dev->devfn & 0x7);
					}
				}
			}
		}
		mutex_unlock(&vastai_card_info_list_lock);
	}
	/* unknown cmd */
	else
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"pci_proc_cmd unknown cmd %s\n", cmd);

    /* return error code for user */
    return ret;
}

static int vastai_pci_char_open(struct inode *inode, struct file *filp)
{
    struct char_drv_info *dev;
    struct char_drv_info *anonymous_dev; /* we need a anonymous_dev for multi thread open kchar */
    int ret = 0;

    /* VASTAI_PCI_FUNC_ENTERY; */
    dev = container_of(inode->i_cdev, struct char_drv_info, testcdev);
    anonymous_dev = kmalloc(sizeof(*dev), GFP_KERNEL);
    if (!anonymous_dev) {
        return -ENOMEM;
    }
    memcpy(anonymous_dev, dev, sizeof(*dev));
    anonymous_dev->valid_data_length = 0;
#ifdef DIAG_TEST_GFX_PERF
    ret = vastai_udma_malloc(anonymous_dev->pcie_dev_info, &(anonymous_dev->dm), 4096);
#else
    ret = vastai_udma_malloc(anonymous_dev->pcie_dev_info, &(anonymous_dev->dm), VASTAI_PCI_TLV_BUF_LEN);
#endif
    if  (ret) {
        VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
                   "malloc buf failed %d\n", ret);
        return ret;
    }

    vastai_pci_args_init(anonymous_dev->args);
    filp->private_data = anonymous_dev;

    return 0;
}

static ssize_t vastai_pci_char_read(struct file *filp, char __user *buffer,
				size_t count, loff_t *offset)
{
	int ret;
	struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
	char *read_buf = (char *)(dev->dm.vir) + *offset;

	VASTAI_PCI_FUNC_ENTERY;
	VASTAI_PCI_DBG(dev->pcie_dev_info, DUMMY_DIE_ID,
			"pci_char_read count[%ld] offset[%lld] dev->l[%d]!\n",
			count,
			*offset,
			dev->valid_data_length);

	/* ***********************************
	 * arg check:
	 * 1. count != 0 && count <= remain_size (remain_size = dev->l - *offset)
	 * 2. dev->l < dm.size
	 * 3. remain_size >= 0
	 * ***********************************/
	if ((count == 0) || ((*offset) >= (dev->valid_data_length)) || (dev->valid_data_length > dev->dm.size)) {
		return 0;
	}
	if (count > (dev->valid_data_length - *offset)) {
		count = dev->valid_data_length - *offset;
	}

	ret = vastai_pci_copy_to_user((void *)buffer, read_buf, count);
	if (ret != 0) {
		VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
				"pci_char_read pci_copy_to_user get error!\n");
		return -ret;
	}

	*offset += count;

	return count;
}

static int vastai_cmd_dma(struct file *filp,
                const char __user *buffer,
                size_t count,
                loff_t *offset);
static int vastai_cmd_rc(struct file *filp,
                const char __user *buffer,
                size_t count,
                loff_t *offset);
static int vastai_diag_get_die_info(struct file *filp,
                   const char __user *buffer,
                   size_t count,
                   loff_t *offset);
static int vastai_send_smi(struct file *filp,
               const char __user *buffer,
               size_t count,
               loff_t *offset);
static int vastai_get_fw_version(struct file *filp,
                   const char __user *buffer,
                   size_t count,
                   loff_t *offset);
struct {
    char *cmd_name;
    char *usage;
    int (*cmd_fn)(struct file *filp,
              const char __user *buffer,
              size_t count,
              loff_t *offset);
} cmd_table[] = {
    {.cmd_name = "DMA", .cmd_fn = vastai_cmd_dma, .usage = NULL},
    {.cmd_name = "RC", .cmd_fn = vastai_cmd_rc, .usage = NULL},
    {.cmd_name = "get_die_info", .cmd_fn = vastai_diag_get_die_info, .usage = NULL},
    {.cmd_name = "smi", .cmd_fn = vastai_send_smi, .usage = NULL},
    {.cmd_name = "get_fw_version", .cmd_fn = vastai_get_fw_version, .usage = NULL},
};

/* TODO: use sys/bus/pci/ dir struction to characterize die information */
static int vastai_diag_get_die_info(struct file *filp,
                   const char __user *buffer,
                   size_t count,
                   loff_t *offset)
{
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
    struct vastai_pci_info *priv = dev->pcie_dev_info;
    int ret;
    int die_index[4]={0};
    int die_num;
    int i = 0;

    ret = get_vastai_die_info(priv, die_index, &die_num);
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "diag_get_die_info get_va_die_info error %d\n", ret);
        return ret;
    }

    ((int *)dev->dm.vir)[0] = die_num;
    dev->valid_data_length += 4;
    VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
            "va device_id [%d] die_num[%d]\n", priv->dev_id, die_num);
    for (i = 0; i < die_num; i++) {
        ((int *)dev->dm.vir)[i + 1] = die_index[i];
        VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
                "[%02d] ", die_index[i]);
        dev->valid_data_length += 4;
    }

    return 0;
}

static int vastai_get_fw_version(struct file *filp,
                   const char __user *buffer,
                   size_t count,
                   loff_t *offset)
{
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
    struct vastai_pci_info *priv = dev->pcie_dev_info;

    memcpy(dev->dm.vir, &priv->fw_ver, sizeof(struct vastai_fw_version));
    dev->valid_data_length += sizeof(struct vastai_fw_version);
    return 0;
}

static int vastai_send_smi(struct file *filp,
               const char __user *buffer,
               size_t count,
               loff_t *offset)
{
    struct pcie_transfer_info trans = {.type = VASTAI_SMI};
    int ret;
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
    struct vastai_pci_info *priv = dev->pcie_dev_info;
    char temp_buf[64] = {0};
    int die_index;
    u32 die_id;

    ret = vastai_pci_copy_from_user(temp_buf, buffer, 32);
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "send_smi pci_copy_from_user error\n");
        return -EINVAL;
    }
    sscanf(temp_buf, "smi %d", &die_index);

    die_id = vastai_pci_get_die_id(priv, die_index);
    ret = vastai_pci_msgq_wr(priv, die_index, &(priv->dies[die_id].msgq[1]), &trans, sizeof(trans));
    if (ret)
        return -EINVAL;
    return 0;
}

static size_t get_user_dma_desc_size(u32 desc_num)
{
    struct user_dma_desc temp;
    return sizeof(temp) + sizeof(temp.desc_tab[0]) * desc_num;
}

static int vastai_cmd_rc(struct file *filp,
                const char __user *buffer,
                size_t count,
                loff_t *offset)
{
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
    struct vastai_pci_info *priv = dev->pcie_dev_info;
    struct user_dma_desc *dma_desc = NULL;
    u32 desc_num = 1;
    int i;
    int ret;
    u64 desc_tab[64];
    int die_index;

    dma_desc = kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);

REFILL_DMA_DESC:
    if (!dma_desc) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "cmd_rc kmalloc failed size %ld\n",
                   get_user_dma_desc_size(desc_num));
        return -ENOMEM;
    }

    ret = vastai_pci_copy_from_user(dma_desc, buffer, get_user_dma_desc_size(desc_num));
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "cmd_rc pci_copy_from_user error\n");
        goto FREE_DMA_DESC;
    }

    if (strncmp(dma_desc->cmd_str, "RC ", sizeof(dma_desc->cmd_str))) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "cmd_rc get err cmd line %s\n",
                   dma_desc->cmd_str);
        /* continue */
    }

    if (dma_desc->num != desc_num && desc_num == 1) {
        desc_num = dma_desc->num;
        if (desc_num == 0) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "desc_num less 1!\n");
            ret = -EINVAL;
            goto FREE_DMA_DESC;
        }
        kfree(dma_desc);
        dma_desc = kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);
        goto REFILL_DMA_DESC;
    }
    die_index = priv->dies[dma_desc->die_id].die_index;

    for (i = 0; i < desc_num; i++) {
        u8 ctr_byte = vastai_udma_ctr_byte(i, desc_num, dma_desc->mode);

        desc_tab[i] = vastai_udma_push_desc(priv,
                                  die_index,
                                  dma_desc->ch,
                                  dma_desc->desc_tab[i].dev_addr,
                                  dma_desc->desc_tab[i].host_addr,
                                  dma_desc->desc_tab[i].length,
                                  ctr_byte);
    }
    vastai_udma_trigger(priv, die_index, dma_desc->ch, dma_desc->is_host_to_dev, desc_tab, desc_num);
    ret = vastai_udma_polling_done(priv, die_index, dma_desc->ch, desc_tab, desc_num);
FREE_DMA_DESC:
    kfree(dma_desc);
    return ret;
}

static int vastai_cmd_dma(struct file *filp,
                const char __user *buffer,
                size_t count,
                loff_t *offset)
{
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
    struct vastai_pci_info *priv = dev->pcie_dev_info;
    struct user_dma_desc *dma_desc = NULL;
    u32 desc_num = 1;
    int i;
    int ret;
    u64 desc_tab[64];
    u64 dma_length = 0;
    int die_index;

    dma_desc = kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);

REFILL_DMA_DESC:
    if (!dma_desc) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "cmd_dma kmalloc failed size %ld\n",
                   get_user_dma_desc_size(desc_num));
        return -ENOMEM;
    }

    ret = vastai_pci_copy_from_user(dma_desc, buffer, get_user_dma_desc_size(desc_num));
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "cmd_dma pci_copy_from_user error\n");
        goto FREE_DMA_DESC;
    }

    if (strncmp(dma_desc->cmd_str, "DMA ", sizeof(dma_desc->cmd_str))) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "cmd_dma get err cmd line %s\n",
                   dma_desc->cmd_str);
        /* continue */
    }

    if (dma_desc->num != desc_num && desc_num == 1) {
        desc_num = dma_desc->num;
        if (desc_num == 0) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "desc_num less 1!\n");
            ret = -EINVAL;
            goto FREE_DMA_DESC;
        }
        kfree(dma_desc);
        dma_desc = kzalloc(get_user_dma_desc_size(desc_num), GFP_KERNEL);
        goto REFILL_DMA_DESC;
    }
    die_index = priv->dies[dma_desc->die_id].die_index;

    for (i = 0; i < desc_num; i++) {
        u8 ctr_byte = vastai_udma_ctr_byte(i, desc_num, dma_desc->mode);
        bool is_post_write = ctr_byte & BIT(2);
        bool is_pre_fetch = ctr_byte & BIT(1);

        if (dma_length + dma_desc->desc_tab[i].length > dev->dm.size) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "transfer too long 0x%llx!\n", dma_length);
            return -EINVAL;
        }

        if (dma_desc->is_host_to_dev && !is_post_write) {
            ret = vastai_pci_copy_from_user(dev->dm.vir + dma_length,
                    (void *)dma_desc->desc_tab[i].host_addr,
                    dma_desc->desc_tab[i].length);
        if (ret != 0) {
            VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                       "cmd_dma %d pci_copy_from_user get error!\n",
                       __LINE__);
                return ret;
            }
        }
        desc_tab[i] = vastai_udma_push_desc(priv,
                                  die_index,
                                  dma_desc->ch,
                                  dma_desc->desc_tab[i].dev_addr,
                                  (dev->dm.phy + dma_length) | ((dma_desc->die_id) ? 0xb000000000 : 0x0),
                                  dma_desc->desc_tab[i].length,
                                  ctr_byte);
        if (!(dma_desc->is_host_to_dev && is_post_write) && !(!(dma_desc->is_host_to_dev) && is_pre_fetch))
            dma_length += dma_desc->desc_tab[i].length;
    }
    vastai_udma_trigger(priv, die_index, dma_desc->ch, dma_desc->is_host_to_dev, desc_tab, desc_num);
    ret = vastai_udma_polling_done(priv, die_index, dma_desc->ch, desc_tab, desc_num);
    dev->valid_data_length = dma_length;
FREE_DMA_DESC:
    kfree(dma_desc);
    return ret;
}


static ssize_t vastai_pci_char_write(struct file *filp,
                     const char __user *buffer, size_t count,
                     loff_t *offset)
{
    int ret;
    unsigned char *input_kernel_buf;
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
    int i;
    size_t real_count = min(count, (size_t)VASTAI_CMD_READ_LEN);

    /* VASTAI_PCI_FUNC_ENTERY; */
    if (count == 0)
        return count;

    if (!dev->dm.vir) {
        VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
                   "pci_char_write dev->replay is Empty!\n");
        return count;
    }

    input_kernel_buf = kzalloc(VASTAI_CMD_READ_LEN, GFP_KERNEL);
    if (!input_kernel_buf) {
        return -ENOMEM;
    }
    ret = vastai_pci_copy_from_user(input_kernel_buf,
                 (char __user *)buffer,
                 real_count);
    if (ret != 0) {
        VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
                   "pci_char_write pci_copy_from_user get error!\n");
        goto FREE_INPUT_BUF;
    }

    ret = -1;
    for (i = 0; i < ARRAY_SIZE(cmd_table); i++) {
        char *cmd = cmd_table[i].cmd_name;
        if (strncmp(cmd, input_kernel_buf, strlen(cmd)) == 0) {
            ret = cmd_table[i].cmd_fn(filp, buffer, count, offset);
            if (ret < 0) {
                VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
                           "get error code %d\n", ret);
                goto FREE_INPUT_BUF;
            }
        }
    }
    input_kernel_buf[real_count - 1] = 0; /* set \0 for string */

    if (ret != 0) { /* Is this cmd handle by cmd_table ? */
        VASTAI_PCI_DBG(dev->pcie_dev_info, DUMMY_DIE_ID,
                "write:%s\n", input_kernel_buf);
        ret = vastai_pci_proc_cmd(input_kernel_buf, VASTAI_CMD_READ_LEN, dev);
    }

FREE_INPUT_BUF:
    kfree(input_kernel_buf);

    if (ret != 0) {
        VASTAI_PCI_ERR(dev->pcie_dev_info, DUMMY_DIE_ID,
                   "pci_char_write we get error code %d!\n", ret);
        return ret;

    }
    return count;
}

static long vastai_pci_char_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;
    struct vastai_pci_info *priv = dev->pcie_dev_info;
    struct kchar_cmd kcmd;
    int ret;

    ret = vastai_pci_copy_from_user(&kcmd, (void *)arg, sizeof(kcmd));
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "pci_char_ioctl pci_copy_from_user is error\n");
        return -EIO;
    }

    switch (cmd)  {
        case VASTAI_PCI_IOCTL_ALLOC: {
            struct dma_buf *dmabuf = vastai_alloc_dma_buf(priv, kcmd.alloc_cmd.size);
            if (dmabuf) {
                int fd = dma_buf_fd(dmabuf, O_CLOEXEC);
                kcmd.alloc_cmd.dma_buf_fd = fd;
            } else {
                kcmd.alloc_cmd.dma_buf_fd = -ENOMEM;
            }
            break;
        }
        case  VASTAI_PCI_IOCTL_DMA_START: {
            struct dma_buf *dmabuf = dma_buf_get(kcmd.dma_start_cmd.dma_buf_fd);
            struct vastai_dma_buf *dm = get_vastai_dma_buf (dmabuf, 0);
            struct vastai_dmadesc desc;
            union core_bitmap core_id = {.val = 0};

            desc.dev_addr = kcmd.dma_start_cmd.axi_addr;
            desc.dma_lenth = kcmd.dma_start_cmd.size;
            desc.is_host_to_dev = !kcmd.dma_start_cmd.is_dev_to_host;
            desc.is_src_not_user_mem = 1;
            desc.is_src_dma_addr = 1;
            desc.host_addr.dma_addr = dm->phy;

            vastai_pci_dma_transfer_sync(priv, kcmd.dma_start_cmd.die_index, core_id, &desc, 1);
            dma_buf_put(dmabuf);
            break;
        }
        case VASTAI_PCI_IOCTL_POLL_LOG: {
            /* trigger smcu to poll log buf */
            u32 flag = 0;
            int ret;
            u32 timeout_water = kcmd.log_poll_cmd.timeout_water != 0x0 ?
                        kcmd.log_poll_cmd.timeout_water : 500;
            u32 die_index = kcmd.log_poll_cmd.die_index;
            struct vastai_sv100_die *die =
                &(((struct vastai_pci_info *)priv)->dies[vastai_pci_get_die_id(priv, die_index)]);

            vastai_pci_mem_write(priv, die_index, LOG_MSG_FLAG, &flag, sizeof(flag));
            /* wait complete */
            ret = wait_for_completion_interruptible_timeout(&(die->log_comp), HZ * timeout_water / 1000);
            if (ret == 0) {
                kcmd.log_poll_cmd.is_timeout = 1;
            }

            /* get log ring ctl */
            /* we need size == 0x1D0 buf */
            vastai_pci_mem_read(priv, kcmd.log_poll_cmd.die_index, LOG_CTRL_INFO, dev->dm.vir, 0x1D0);
            ret = vastai_pci_copy_to_user(kcmd.log_poll_cmd.log_ring_ctl, dev->dm.vir, 0x1D0);
            if (ret) {
                VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                    "pci_char_ioctl pci_copy_to_user is error\n");
                return -EIO;
            }
            break;
        }
        case TCGETS:
            return -ENOIOCTLCMD;

        case TCSETS:
            return -ENOIOCTLCMD;

        case TIOCSERGETLSR:
            return -ENOIOCTLCMD;

        case TIOCSERCONFIG:
            return -ENOIOCTLCMD;
	case VASTAI_PCI_IOCTL_BOARD_TYPE:
		kcmd.board_type.type = vastai_get_board_type(dev->pcie_dev_info);
		break;

        default:
            VASTAI_PCI_INFO(dev->pcie_dev_info, DUMMY_DIE_ID,
                    "Unsupport cmd\n");
            break;

    }
    ret = vastai_pci_copy_to_user((void *)arg, &kcmd, sizeof(kcmd));
    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "pci_char_ioctl pci_copy_to_user is error\n");
        return -EIO;
    }

    return 0;
}

static int vastai_pci_char_release(struct inode *inode, struct file *filp)
{
    struct char_drv_info *dev = (struct char_drv_info *)filp->private_data;

    /* VASTAI_PCI_FUNC_ENTERY; */

    vastai_udma_free(dev->pcie_dev_info, &(dev->dm));
    /* free anonymous_dev */
    kfree(dev);

    return 0;
}

static const struct file_operations fop = {
	.owner = THIS_MODULE,
	.open = vastai_pci_char_open,
	.release = vastai_pci_char_release,
	.write = vastai_pci_char_write,
	.read = vastai_pci_char_read,
	.llseek = default_llseek,
	.unlocked_ioctl = vastai_pci_char_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = vastai_pci_char_ioctl,
#endif
};
struct class *myclass;

static int vastai_pci_ioctl_cmd_init(struct vastai_pci_info *bus)
{
    int ret;
    dev_t dev;
    struct char_drv_info *drv;

    drv = kmalloc(sizeof(struct char_drv_info), GFP_KERNEL);
    if (!drv) {
        ret = -ENOMEM;
        return ret;
    }

    bus->p_char = drv;
    drv->pcie_dev_info = bus;
    drv->major = 321;
    dev = MKDEV(drv->major, 0);
    ret = register_chrdev_region(dev, 1, "char");
    if (ret) {
        alloc_chrdev_region(&dev, 0, 1, "char");
        drv->major = MAJOR(dev);
    }
    drv->testcdev.owner = THIS_MODULE;
    cdev_init(&(drv->testcdev), &fop);
    cdev_add(&(drv->testcdev), dev, 1);
    
    drv->mydev = device_create(myclass,
                   NULL,
                   dev,
                   NULL,
                   "kchar:%d",
                   bus->dev_id);        
    VASTAI_PCI_INFO(bus, DUMMY_DIE_ID, "init ioctl node /dev/kchar:%d ok.\n", bus->dev_id);

    return 0;
}

static int vastai_pci_ioctl_cmd_deinit(struct vastai_pci_info *bus)
{
    dev_t dev;
    struct char_drv_info *drv = bus->p_char;

    vastai_pci_free_udma_buf(bus, &dmb[0]);
    dev = MKDEV(drv->major, 0);
    device_destroy(myclass, dev);
    cdev_del(&(drv->testcdev));
    unregister_chrdev_region(dev, 1);
    VASTAI_PCI_INFO(bus, DUMMY_DIE_ID,
            "deinit ioctl node /dev/kchar:%d ok.\n", bus->dev_id);

    return 0;
}

static void vastai_payload_work(struct work_struct *work)
{
    struct vastai_pci_info *priv = NULL;
    int ret;
    u32 buf[4] = {0};

    priv = container_of(work, struct vastai_pci_info, payload_work.work);
    ret = vastai_pci_mem_read(priv, 0, 0x9000028, buf, sizeof(buf));

    if (ret) {
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
                   "payload_work: read payload reg failed, %d\n", ret);
        memset(buf, 0, sizeof(buf));
        return;
    }

    if (buf[0] < (rx_packet & 0xffffffff))
        rx_packet += 0x100000000;

    if (buf[1] < (rx_payload & 0xffffffff))
        rx_payload += 0x100000000;

    if (buf[2] < (tx_packet & 0xffffffff))
        tx_packet += 0x1000000001;

    if (buf[3] < (tx_payload & 0xffffffff))
        tx_payload += 0x100000000;

    rx_packet  = (rx_packet  & 0xffffffff00000000) + buf[0];
    rx_payload = (rx_payload & 0xffffffff00000000) + buf[1];
    tx_packet  = (tx_packet  & 0xffffffff00000000) + buf[2];
    tx_payload = (tx_payload & 0xffffffff00000000) + buf[3];

    schedule_delayed_work(&priv->payload_work, msecs_to_jiffies(1000)/* 10s */);
}


int vastai_pci_ioctl_init(struct vastai_pci_info *bus)
{
#if PCIE_PAYLOAD_MONITOR_EN
	INIT_DELAYED_WORK(&bus->payload_work, vastai_payload_work);
	schedule_delayed_work(&bus->payload_work, msecs_to_jiffies(1000)/* 1s */);
#else
	(void)vastai_payload_work;
#endif
    // create vastai_ctl node
    vastai_ctl_file_init(bus);

    // create kchar node
    return vastai_pci_ioctl_cmd_init(bus);
}

int vastai_file_create(struct vastai_file_info *file_info,
               struct vastai_pci_info *pci_info,
               const struct file_operations *fop, const char *fmt, ...)
{
    va_list vargs;
    int ret = 0;

    file_info->dev_num = MKDEV(major, atomic_inc_return(&minor));
    ret = register_chrdev_region(file_info->dev_num, 1, "vastai");
    if (ret) {
        VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
                   "register device num failed! alloc one [%d]\n",
                   file_info->dev_num);
        alloc_chrdev_region(&file_info->dev_num, 0, 1, "char");
    }
    file_info->dev.owner = THIS_MODULE;
    cdev_init(&file_info->dev, fop);
    ret = cdev_add(&file_info->dev, file_info->dev_num, 1);
    if (ret < 0) {
        VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID, "cdev_add failed %d\n",
                   ret);
        return ret;
    }

    va_start(vargs, fmt);
    vsnprintf(file_info->file_name, 32, fmt, vargs);
    va_end(vargs);
    file_info->device = device_create(myclass, NULL, file_info->dev_num,
                      file_info, file_info->file_name);

    if (!file_info->device) {
        ret = PTR_ERR(file_info->device);
        VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
                   "device_create_vargs failed %d\n", ret);
        return ret;
    }
    return ret;
}

void vastai_file_destroy(struct vastai_file_info *file_info)
{
    device_destroy(myclass, file_info->dev_num);
    cdev_del(&file_info->dev);
    unregister_chrdev_region(file_info->dev_num, 1);
    atomic_dec_return(&minor);
}

int vastai_pci_ioctl_deinit(struct vastai_pci_info *bus)
{
#if PCIE_PAYLOAD_MONITOR_EN
    cancel_delayed_work_sync(&bus->payload_work);
#endif

    vastai_ctl_file_deinit(bus);

    return vastai_pci_ioctl_cmd_deinit(bus);
}
#define DEVICE_NUMBER 1
#define DEVICE_SNAME "static_vastai_chardevice"
#define DEVICE_DNAME "dynamic_vastai_chardevice"
#define DEVICE_MINOR_NUMBER 0
#define DEVICE_CLASS_NAME "chrdev_class"
#define DEVICE_NODE_NAME "va_dev"

static int vastai_dev_char_open(struct inode *ind, struct file *fp)
{
	printk("demo open\n");
	return 0;
}

static ssize_t vastai_dev_char_read(struct file *filp, char __user *buffer,
				    size_t count, loff_t *offset)
{
	printk("va_dev_char_read:ok!\n");
	return 0;
}
static const struct file_operations fop_dev = {
	.owner = THIS_MODULE,
	.open = vastai_dev_char_open,
	.read = vastai_dev_char_read,
};
struct cdev char_dev;
struct class *vastai_class;
struct device *device;
dev_t dev_num;
static int dev_major_num,dev_minor_num;

int vastai_dev_node_init(void){
	int ret;
	ret = alloc_chrdev_region(&dev_num,DEVICE_MINOR_NUMBER,DEVICE_NUMBER,DEVICE_DNAME);
	if (ret < 0) {
		printk("va_dev:alloc_chrdev_region error!\n");
	}
	dev_major_num = MAJOR(dev_num);
	dev_minor_num = MINOR(dev_num);
	char_dev.owner = THIS_MODULE;
	cdev_init(&char_dev,&fop_dev);
	cdev_add(&char_dev,dev_num,DEVICE_NUMBER);

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 4, 0)
	vastai_class = class_create(THIS_MODULE, DEVICE_CLASS_NAME);
#else
	vastai_class = class_create(DEVICE_CLASS_NAME);
#endif

	device = device_create(vastai_class,NULL,dev_num,NULL,DEVICE_NODE_NAME);
	printk("va_dev init ok!\n");
	return 0;
}

void vastai_dev_node_deinit(void){
	unregister_chrdev_region(MKDEV(dev_major_num,dev_minor_num),DEVICE_NUMBER);
	cdev_del(&char_dev);
	device_destroy(vastai_class,dev_num);
	class_destroy(vastai_class);
	printk("va_dev deinit ok!\n");
}