/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/random.h>
#include <linux/dma-buf.h>

#include "vastai_pci.h"
#include "vastai_version.h"
#define TEMP_BUF_SIZE 512

static int version_open(struct inode *inode, struct file *filp)
{
	struct vastai_file_info *file_info;
	struct vastai_version_file *version_file;

	file_info = to_vastai_file_info(inode->i_cdev);
	version_file =
		container_of(file_info, struct vastai_version_file, file_info);

	filp->private_data = version_file;
	return 0;
}

static int version_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static int version_sprint(void *buf, struct vastai_pci_info *pci_info)
{
	int ret = 0;

	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.smcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.cmcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.omcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.gmcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.vemcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.odsp_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.vdsp_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.pmcu_ver);
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.vdmcu_ver);
	ret += sprintf(buf + ret, "[%s] [active]\n", pci_info->fw_ver.ver);			// bmcu active
	ret += sprintf(buf + ret, "[%s] [backup]\n", pci_info->fw_ver.ver_bk);		// bmcu backup
	ret += sprintf(buf + ret, "[%s] [active]\n", pci_info->fw_ver.bl0_ver);		// bl0 active
	ret += sprintf(buf + ret, "[%s] [backup]\n", pci_info->fw_ver.bl0_ver_bk);	// bl0 backup
	ret += sprintf(buf + ret, "[%s]\n", pci_info->fw_ver.driver_ver);

	return ret;
}

static ssize_t version_read(struct file *filp, char __user *buffer,
			    size_t count, loff_t *offset)
{
	struct vastai_version_file *version_file = filp->private_data;
	struct vastai_pci_info *pci_info = version_file->pci_info;
	void *buf = vmalloc(TEMP_BUF_SIZE);
	int ret;

	if ((count < sizeof(pci_info->dies[0].fw_ver) * 2) || (*offset != 0))
		return 0;

	ret = version_sprint(buf, pci_info);
	*offset += ret;

	if (vastai_pci_copy_to_user(buffer, buf, ret)) {
		return -EIO;
	}

	return ret;
}

static const struct file_operations fop = {
	.owner = THIS_MODULE,
	.open = version_open,
	.release = version_release,
	.read = version_read,
};

int vastai_version_file_init(struct vastai_pci_info *pci_info)
{
	struct vastai_version_file *version_file =
		vmalloc(sizeof(struct vastai_version_file));
	int ret;

	if (version_file == NULL) {
		return -ENOMEM;
	}
	ret = vastai_file_create(&version_file->file_info, pci_info, &fop,
			"va_version_%d", pci_info->dev_id);

	pci_info->version_file = version_file;
	version_file->pci_info = pci_info;

	return ret;
}

void vastai_version_file_deinit(struct vastai_pci_info *pci_info)
{
	vastai_file_destroy(&pci_info->version_file->file_info);
	vfree(pci_info->version_file);
}

/*
 * @brief gen a string include fw version information
 *
 * @param pci_info: vastai sg100 pci base struct.
 * @param mask: bitmap for which mcu's version need print
 * @return char point: string buffer. It's need free after used.
 */
void vastai_pci_show_fw_ver(struct vastai_pci_info *pci_info, u32 mask)
{
	if (mask & 0x1)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.smcu_ver);

	if (mask & 0x2)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.cmcu_ver);

	if (mask & 0x4)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.omcu_ver);

	if (mask & 0x8)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.gmcu_ver);

	if (mask & 0x10)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.vemcu_ver);

	if (mask & 0x20)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.odsp_ver);

	if (mask & 0x40)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.vdsp_ver);

	if (mask & 0x80)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.pmcu_ver);

	if (mask & 0x100)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.vdmcu_ver);

	if (mask & 0x200)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s] [active]\n", pci_info->fw_ver.ver);

	if (mask & 0x400)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s] [backup]\n", pci_info->fw_ver.ver_bk);

	if (mask & 0x800)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s] [active]\n", pci_info->fw_ver.bl0_ver);

	if (mask & 0x1000)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s] [backup]\n", pci_info->fw_ver.bl0_ver_bk);

	if (mask & 0x2000)
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.driver_ver);

	if (mask & 0x4000) {
		VASTAI_PCI_INFO(pci_info, DUMMY_DIE_ID, "[%s]\n", pci_info->fw_ver.phy_ver);
	}
}

void vastai_pci_get_fw_ver(struct vastai_pci_info *pci_info, u32 mask)
{
	int ret = 0;

#if 0
	ret = vastai_pci_mem_write(pci_info, die_index, VSERSION_INFO_BASE, VASTAI_VERSION, VASTAI_VER_LEN);
	if (ret < 0) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID, "pci_get_fw_ve write version error:%d\n", ret);
		break;
	}
#endif
	if (pci_info->is_pf)
		ret = vastai_pci_mem_read(pci_info, 0, VSERSION_INFO_BASE, &pci_info->fw_ver, sizeof(struct vastai_fw_version));
	else
		ret = vastai_read_by_smcu(pci_info, VSERSION_INFO_BASE, sizeof(struct vastai_fw_version), &pci_info->fw_ver);
	
	if (ret < 0) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID, "pci_get_fw_ver read firmware version error:%d\n", ret);
		return;
	}

	pci_info->fw_ver.ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.smcu_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.cmcu_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.omcu_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.gmcu_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.vemcu_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.odsp_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.vdsp_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.pmcu_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.vdmcu_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.ver_bk[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.bl0_ver[VASTAI_VER_LEN - 1] = 0;
	pci_info->fw_ver.bl0_ver_bk[VASTAI_VER_LEN - 1] = 0;
	snprintf(pci_info->fw_ver.driver_ver, VASTAI_VER_LEN,
			"Driver: %s %s %s %s", VASTAI_PCI_DRIVER_VERSION,
			VASTAI_PCI_GIT_BRANCH_NAME, VASTAI_PCI_GIT_COMMIT_HASH,
			VASTAI_PCI_MKPACKAGE_TIME);
#if 0
	/* pcie_phy version info not in csram */
	/* pcie_hpy version in xspiflash 0xd000 */
	/* we read from app partition, relative position */
	vastai_pci_read_xspi(pci_info, i, 0x6000, die->fw_ver.phy_ver,
				sizeof(die->fw_ver.phy_ver), 0);
	if (die->fw_ver.phy_ver[0] == 0xFF)
		strcpy(die->fw_ver.phy_ver, "no pcie phy version");
#endif

	if (!ret && pci_info->is_main_dev_in_os) {
		vastai_pci_show_fw_ver(pci_info, mask);
	}
}

