/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/delay.h>
#include "vastai_udma_test.h"
#include "vastai_udma_engine.h"


u8 vastai_udma_ctr_byte(int i, int sum, enum VASTAI_DMA_MODE mode)
{
	bool is_last = i == (sum - 1);
	u8 ret = 0;
	enum {
		normal = 0,
		pre_fetch,
		post_write,
	} desc_type = 0;

	switch (mode) {
	case VASTAI_DMA_MODE_BULK:
		desc_type = 0;
		break;
	case VASTAI_DMA_MODE_SCATTER:
		if (i == 0) {
			desc_type = pre_fetch;
		} else {
			desc_type = post_write;
		}
		break;
	case VASTAI_DMA_MODE_GATHER:
		if (is_last) {
			desc_type = post_write;
		} else {
			desc_type = pre_fetch;
		}
		break;
	}

	if (!is_last) {
		ret |= BIT(5);
	}
	if (desc_type == post_write) {
		ret |= BIT(2);
	} else if (desc_type == pre_fetch) {
		ret |= BIT(1);
	}
	return ret;
}

u64 vastai_udma_push_desc(struct vastai_pci_info *priv,
			  unsigned int die_index,
			  unsigned int dma_chn,
			  u64 sys_addr,
			  u64 ext_addr,
			  u32 length,
			  u8 ctr_byte)
{
	u32 desc;
	PCIE_xd_desc temp = {0};
	/* TODO: production config need a ringbuf manager */
	{
		static struct {
			u32 start;
			u32 cur;
			u32 end;
		} desc_mgr[8] = {
			{
				.start = 0x8C14000,
				.cur = 0x8C14000,
				.end = 0x8C14000 + 0x640,
			},
			{
				.start = 0x8C14000 + 0x640,
				.cur = 0x8C14000 + 0x640,
				.end = 0x8C14000 + 0x640 * 2,
			},
			{
				.start = 0x8C14000 + 0x640 * 2,
				.cur = 0x8C14000 + 0x640 * 2,
				.end = 0x8C14000 + 0x640 * 3,
			},
			{
				.start = 0x8C14000 + 0x640 * 3,
				.cur = 0x8C14000 + 0x640 * 3,
				.end = 0x8C14000 + 0x640 * 4,
			},
			{
				.start = 0x8C23000,
				.cur = 0x8C23000,
				.end = 0x8C23000 + 0x640,
			},
			{
				.start = 0x8C23000 + 0x640,
				.cur = 0x8C23000 + 0x640,
				.end = 0x8C23000 + 0x640 * 2,
			},
			{
				.start = 0x8C23000 + 0x640 * 2,
				.cur = 0x8C23000 + 0x640 * 2,
				.end = 0x8C23000 + 0x640 * 3,
			},
			{
				.start = 0x8C23000 + 0x640 * 3,
				.cur = 0x8C23000 + 0x640 * 3,
				.end = 0x8C23000 + 0x640 * 4,
			},

		};
		desc = desc_mgr[dma_chn].cur;
		desc_mgr[dma_chn].cur += sizeof(PCIE_xd_desc);
		if (desc_mgr[dma_chn].cur >= desc_mgr[dma_chn].end)
			desc_mgr[dma_chn].cur = desc_mgr[dma_chn].start;
	}
	if (!desc)
		return 0;

	temp.sys_lo_addr = ((uint32_t)(sys_addr & VASTAI_VAL_DW_MASK));
	temp.sys_hi_addr = ((uint32_t)((sys_addr >> 32) & VASTAI_VAL_DW_MASK));

	temp.ext_lo_addr = ((uint32_t)(ext_addr & VASTAI_VAL_DW_MASK));
	temp.ext_hi_addr = ((uint32_t)((ext_addr >> 32) & VASTAI_VAL_DW_MASK));

	temp.size_and_ctrl.size = (length & VASTAI_UDMA_LENGTH_MASK);
	temp.size_and_ctrl.ctrl_bits.control_bits = (ctr_byte);

	vastai_pci_mem_write(priv, die_index, (u64)desc, &temp, sizeof(temp));
	return (u64)desc;
}

void vastai_udma_trigger(struct vastai_pci_info *priv,
			 unsigned int die_index,
			 unsigned int dma_chn,
			 unsigned int is_host_to_dev,
			 u64 *desc_tab,
			 int num)
{
	/* change this function as vastai_func */

	int i = 0;
	u32 zero = 0;
	u32 val = 0;
	u32 dma_base = dma_chn < 4 ? 0x9500000 : 0x9e00000;
	dma_chn = dma_chn < 4 ? dma_chn : dma_chn - 4;

	/* set desc_tab list */
	for (i = 0; i < num - 1; i++) {
		u32 next_desc = desc_tab[i+1] & VASTAI_VAL_DW_MASK;
		u64 next_desc_hi = (desc_tab[i+1] >> 32) & VASTAI_VAL_DW_MASK;
		vastai_pci_mem_write(priv,
				     die_index,
				     desc_tab[i] + offsetof(PCIE_xd_desc, next),
				     &next_desc,
				     sizeof(u32));
		vastai_pci_mem_write(priv,
				     die_index,
				     desc_tab[i] + offsetof(PCIE_xd_desc, next_hi_addr),
				     &next_desc_hi,
				     sizeof(u32));
	}
	vastai_pci_mem_write(priv,
		   	     die_index,
		   	     desc_tab[num - 1] + offsetof(PCIE_xd_desc, next),
			     &zero,
		   	     sizeof(u32));
	vastai_pci_mem_write(priv,
			     die_index,
			     desc_tab[num - 1] + offsetof(PCIE_xd_desc, next_hi_addr),
			     &zero,
			     sizeof(u32));

	val = desc_tab[0] & 0xFFFFFFFF;
	vastai_pci_mem_write(priv,
			     die_index,
			     dma_base + dma_chn * 0x14 + 0x4,
			     &val,
			     sizeof(u32));

	val = desc_tab[0] >> 32;
	vastai_pci_mem_write(priv,
			     die_index,
			     dma_base + dma_chn * 0x14 + 0x8,
			     &val,
			     sizeof(u32));
	val = BIT(dma_chn) | BIT(dma_chn+8);
	vastai_pci_mem_write(priv,
			     die_index,
			     dma_base + 0xa4,
			     &val,
			     sizeof(u32));
	mb();
	val = 0x1; /* set go */
	if (!is_host_to_dev)
		val |= BIT(1);
	vastai_pci_mem_write(priv,
			     die_index,
			     dma_base + dma_chn * 0x14,
			     &val,
			     sizeof(u32));
}

int vastai_udma_polling_done(struct vastai_pci_info *priv,
			     unsigned int die_index,
			     unsigned int dma_chn,
			     u64 *desc_tab,
			     int num)
{
	u32 done = 0;
	u32 err = 0;
	int patience = 50;

	do {
		int i;
		u32 status = 0;

		for (i = 0; i < num; i++) {

			vastai_pci_mem_read(priv,
					    die_index,
					    desc_tab[i] + offsetof(PCIE_xd_desc, status),
					    &status,
					    sizeof(u32));
			if (status != 0x10000)
				break;
		}

		if (i == num) {
			done = 1;
			break;
		} else if (status != 0) {
			err = 1;
			break;
		}

		msleep(1);

	} while (!done && patience--);

	if (done && !err) {
		VASTAI_PCI_INFO(priv, vastai_pci_get_die_id(priv, die_index),
				"DMA chn %d DONE\n", dma_chn);
		return 0;
	}
	return err;
}

/*-----------------------------------------------------------------------------
 * Perform a DMA bulk transfer
 *-----------------------------------------------------------------------------
 */
int vastai_udma_start_bulk_transfer(struct vastai_pci_info *priv,
				    u8 die_id, uint32_t obDirection,
				    uint32_t len, uintptr_t localAddr,
				    uintptr_t remoteAddr)
{
	u64 desc[1];
	int ret;

	desc[0] = vastai_udma_push_desc(priv,
					priv->dies[die_id].die_index,
					0,
					localAddr,
					remoteAddr,
					len,
					0x1);

	vastai_udma_trigger(priv,
			    priv->dies[die_id].die_index,
			    0,
			    !obDirection,
			    desc,
			    1);

	ret = vastai_udma_polling_done(priv,
				       priv->dies[die_id].die_index,
				       0,
				       desc,
				       1);
	return ret;
}
