#include <linux/aer.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/version.h>
#include <linux/rtc.h>
#include <linux/timer.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include "vastai_pci.h"
#include "vastai_pcie_aer.h"

static const char *vastai_aer_uncorrectable_error_string[] = {
    "Undefined",            /* Bit Position 0   */
    NULL,
    NULL,
    NULL,
    "DLP",              /* Bit Position 4   */
    "SDES",             /* Bit Position 5   */
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    "TLP",              /* Bit Position 12  */
    "FCP",              /* Bit Position 13  */
    "CmpltTO",          /* Bit Position 14  */
    "CmpltAbrt",            /* Bit Position 15  */
    "UnxCmplt",         /* Bit Position 16  */
    "RxOF",             /* Bit Position 17  */
    "MalfTLP",          /* Bit Position 18  */
    "ECRC",             /* Bit Position 19  */
    "UnsupReq",         /* Bit Position 20  */
    "ACSViol",          /* Bit Position 21  */
    "UncorrIntErr",         /* Bit Position 22  */
    "BlockedTLP",           /* Bit Position 23  */
    "AtomicOpBlocked",      /* Bit Position 24  */
    "TLPBlockedErr",        /* Bit Position 25  */
    "PoisonTLPBlocked",     /* Bit Position 26  */
    NULL,               /* Bit Position 27  */
    NULL,               /* Bit Position 28  */
    NULL,               /* Bit Position 29  */
    NULL,               /* Bit Position 30  */
    NULL,               /* Bit Position 31  */
};

static const char *vastai_aer_correctable_error_string[] = {
    "RxErr",            /* Bit Position 0   */
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    "BadTLP",           /* Bit Position 6   */
    "BadDLLP",          /* Bit Position 7   */
    "Rollover",         /* Bit Position 8   */
    NULL,
    NULL,
    NULL,
    "Timeout",          /* Bit Position 12  */
    "NonFatalErr",          /* Bit Position 13  */
    "CorrIntErr",           /* Bit Position 14  */
    "HeaderOF",         /* Bit Position 15  */
    NULL,               /* Bit Position 16  */
    NULL,               /* Bit Position 17  */
    NULL,               /* Bit Position 18  */
    NULL,               /* Bit Position 19  */
    NULL,               /* Bit Position 20  */
    NULL,               /* Bit Position 21  */
    NULL,               /* Bit Position 22  */
    NULL,               /* Bit Position 23  */
    NULL,               /* Bit Position 24  */
    NULL,               /* Bit Position 25  */
    NULL,               /* Bit Position 26  */
    NULL,               /* Bit Position 27  */
    NULL,               /* Bit Position 28  */
    NULL,               /* Bit Position 29  */
    NULL,               /* Bit Position 30  */
    NULL,               /* Bit Position 31  */
};

void dump_ep_aer_regs(struct pci_dev *pdev)
{
	int ret = 0;
	int bit = 0;
	struct vastai_aer_regs aer_regs = {0};
	struct vastai_pci_info *priv = pci_get_drvdata(pdev);
	if(!priv)
		return;
	if(priv->is_vf)
		return;
	ret = vastai_pci_config_read(priv, SG100_EP_AER_OFFSET, (char*)&aer_regs,44);
	if(!ret) {
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] UNCORR_ERR_STATUS  = 0x%08x\n",aer_regs.UNCORR_ERR_STATUS);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] UNCORR_ERR_MASK    = 0x%08x\n",aer_regs.UNCORR_ERR_MASK);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] UNCORR_ERR_SEV     = 0x%08x\n",aer_regs.UNCORR_ERR_SEV);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] CORR_ERR_STATUS    = 0x%08x\n",aer_regs.CORR_ERR_STATUS);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] CORR_ERR_MASK      = 0x%08x\n",aer_regs.CORR_ERR_MASK);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] ADV_ERR_CAP_CTRL   = 0x%08x\n",aer_regs.ADV_ERR_CAP_CTRL);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] HDR_LOG0           = 0x%08x\n",aer_regs.HDR_LOG0);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] HDR_LOG1           = 0x%08x\n",aer_regs.HDR_LOG1);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] HDR_LOG2           = 0x%08x\n",aer_regs.HDR_LOG2);
		VASTAI_PCI_INFO(priv, priv->pkg_id,"[va-aer] HDR_LOG3           = 0x%08x\n",aer_regs.HDR_LOG3);
		for(bit = 0; bit < 32; bit ++) {			
			if((aer_regs.UNCORR_ERR_STATUS >> bit & 1) && (!(aer_regs.UNCORR_ERR_MASK>>bit & 1))){
                if(NULL != vastai_aer_uncorrectable_error_string[bit])
                    VASTAI_PCI_ERR(priv, priv->pkg_id,"[va-aer]UNCORR_ERR_STATUS[%02d] %s\n",
                    bit,vastai_aer_uncorrectable_error_string[bit]);
			} 

			if((aer_regs.CORR_ERR_STATUS >> bit & 1) && (!(aer_regs.CORR_ERR_MASK>>bit & 1))){
                if(NULL != vastai_aer_correctable_error_string[bit])
                    VASTAI_PCI_ERR(priv, priv->pkg_id,"[va-aer]CORR_ERR_STATUS[%02d] %s\n",
                    bit,vastai_aer_correctable_error_string[bit]);
			} 
		}
	}
	return;
}
