/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2022 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2022/07/16
 * Author: jmshi
 */

#include "vastai_pci.h"
#include "smmu.h"
#include "va_dma_core.h"
#include "sg100_gmcu.h"
#include "hw_queue.h"
#include "sg100_cmd.h"
#include "vastai_pcie_public.h"
#include "vastai_cmd.h"
#include "hw_config.h"
#include "vastai_udma_engine.h"
#ifdef CONFIG_VASTAI_GFX
#include "../../gfx/gfx_mm.h"
#endif

int level0_cnt = 0;
int level1_cnt = 0;
int level2_cnt = 0;
int level3_cnt = 0;

extern unsigned char gart_en;

void is_need_to_enable_gart(struct vastai_pci_info *priv)
{
    int nid;
    unsigned long start_pa;
    unsigned long start_pfn = 0;
    unsigned long end_pa;
    unsigned long end_pfn = 0;
    const unsigned long atu_support_pa_max = 0x8000000000;

    if (gart_en < GART_NOPUT)
    {
    	VASTAI_PCI_INFO(priv, 0, "%s gart by module parameter\n", gart_en ? "open" : "close");
        priv->gart_en = gart_en;
		return;
    }

    for_each_online_node(nid)
    {
        pg_data_t *pgdat = NODE_DATA(nid);
        start_pfn = pgdat->node_start_pfn;
        end_pfn = pgdat->node_start_pfn + pgdat->node_spanned_pages;
        start_pa = (u64)start_pfn << PAGE_SHIFT;
        end_pa = end_pfn ? ((u64)end_pfn << PAGE_SHIFT) - 1 : 0;

        /* Enable gart mmu when system physical address is over 512GB. */
        if((end_pa >= atu_support_pa_max) || (start_pa >= atu_support_pa_max))
        {
			VASTAI_PCI_INFO(priv, 0, "open gart by system check\n");
			priv->gart_en = GART_ENABLE;
            return;
        }
    }

	VASTAI_PCI_INFO(priv, 0, "close gart by system check\n");
	priv->gart_en = GART_DISABLE;
    return;
}

#if 0
static unsigned char *va_smmu_pci_bar_vmem(struct vastai_pci_info *priv, int bar)
{
	if (!priv) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pcie_dev is NULL\n");
		return NULL;
	}

	return priv->bar[bar].vmem;
}
#endif

static void smmu_read_page_entry(struct vastai_pci_info *priv, u64 page_off, u64 *page)
{
	*page = *(u64 *)((page_off - priv->bar[VASTAI_PCI_BAR4].noc_addr) + priv->bar[VASTAI_PCI_BAR4].vmem);

	return;
}

static void smmu_write_page_entry(struct vastai_pci_info *priv, u64 page_off, u64 *page)
{
	u64 *page_addr;

	page_addr = (u64 *)((page_off - priv->bar[VASTAI_PCI_BAR4].noc_addr) + priv->bar[VASTAI_PCI_BAR4].vmem);
	*page_addr = *page;

	return;
}

static inline int va_smmu_insert_cmd(int optcode, struct vastai_pci_info *priv)
{
	int ret = 0;

	return ret;
}

int smmu_fill_prefetch_cmd_descriptor(struct vastai_pci_info *priv, int optcode, int stream_id, int substream_id, int size, int stride, u64 address)
{
	u8 command[16];
	int addr_high, addr_low;
	int ret;
	struct pcie_transfer_cmd *cmd;

	if (!(optcode == CMD_PREFETCH_CONFIG || optcode ==CMD_PREFETCH_ADDR))
	{
		VASTAI_PCI_ERR(priv, 0,
				"Invalid prefetch cmd!\n");
	}

	address = address >> 12;
	addr_high = (u32)((u64)address >> 32) & 0xFFFFF;
	addr_low  = (u32)((u64)address & 0xFFFFFFFF);

	command[0]  = (u8)optcode;
	command[1]	= (u8)(((substream_id & 0xF) << 4) | 0x8);
	command[2]  = (u8)(((substream_id >> 4) & 0xFF));
	command[3]  = (u8)(((substream_id >> 12) & 0xFF));

	command[4]  = (u8)( stream_id 	   & 0xFF);
	command[5]  = (u8)((stream_id >> 8)  & 0xFF);
	command[6]  = (u8)((stream_id >> 16) & 0xFF);
	command[7]  = (u8)((stream_id >> 24) & 0xFF);

	if (optcode == CMD_PREFETCH_ADDR)
	{
		command[8]  = (u8)((size & 0x1f) || stride & 0x7);
		command[9]  = (u8)((addr_high & 0xf) << 4) || ((stride >> 3) & 0x3);
		command[10] = (u8)((addr_high >> 4) & 0xFF);
		command[11] = (u8)((addr_high >> 12) & 0xFF);
		command[12] = (u8)((addr_low)       & 0xFF);
		command[13] = (u8)((addr_low >> 8)  & 0xFF);
		command[14] = (u8)((addr_low >> 16) & 0xFF);
		command[15] = (u8)((addr_low >> 24) & 0xFF);
	}
	else
	{
		command[8]  = 0;
		command[9]  = 0;
		command[10] = 0;
		command[11] = 0;
		command[12] = 0;
		command[13] = 0;
		command[14] = 0;
		command[15] = 0;
	}

	cmd = (struct pcie_transfer_cmd *)command;
	if ( ONLY_1_PF==priv->fn_mode) {
		ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF,  cmd, 0);
	} else if (ONLY_2_PF==priv->fn_mode){
		ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF,  cmd, 0);
	} else {
		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF,  cmd, 0);
	}

	if(ret!=0){
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"h2d_cmd_test error\n");
	}

	return 0;

//	return smmu_add_command(command);
}

int smmu_fill_cfg_invalid_cmd_descriptor(struct vastai_pci_info *priv, int optcode, int substream_id, int leaf)
{
	u8 command[16];
	int ret;
	struct pcie_transfer_cmd *cmd;

	if (!(optcode == CMD_SET_INVALID || optcode == CMD_SET_RANGE_INVALID || 
				optcode == CMD_CD_INVALID || optcode == CMD_CD_INVALID_ALL))
	{
		VASTAI_PCI_ERR(priv, 0,
				"Invalid cfg cmd!\n");
	}

	command[0]  = (u8)optcode;
	if (CMD_CD_INVALID == optcode)
	{
		command[1]	= (u8)(((substream_id & 0xF) << 4) | 0x8);
		command[2]  = (u8)(((substream_id >> 4) & 0xFF));
		command[3]  = (u8)(((substream_id >> 12) & 0xFF));
	}
	else
	{
		command[1]  = 0;
		command[2]  = 0;
		command[3]  = 0;
	}

	/* get  streamid from fw */
	command[4] = 0;
	command[5] = 0;
	command[6] = 0;
	command[7] = 0;

	if (CMD_SET_RANGE_INVALID == optcode)
	{
		command[8]  = 31;
	}
	else if (CMD_CD_INVALID_ALL == optcode)
	{
		command[8]  = 0;
	}
	else
	{
		command[8]  = leaf;
	}

	command[9]  = 0;
	command[10] = 0;
	command[11] = 0;
	command[12] = 0;
	command[13] = 0;
	command[14] = 0;
	command[15] = 0;

	cmd = (struct pcie_transfer_cmd *)command;
	if ( ONLY_1_PF==priv->fn_mode) {
		ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF, cmd, 0);
	} else if (ONLY_2_PF==priv->fn_mode){
		ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF, cmd, 0);
	} else {
		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF, cmd, 0);
	}
	if(ret!=0){
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"h2d_cmd_test error\n");
	}


	return 0;
//	return smmu_add_command(command);
}

int smmu_fill_tlb_invalid_cmd_descriptor(struct vastai_pci_info *priv, int optcode, int asid, int leaf, u64 address, u32 range)
{
	u8 command[16];
	int addr_high, addr_low;
	int ret;
	struct pcie_transfer_cmd *cmd;
	int num,scale,tg,ttl;

	if (!(optcode == CMD_TLBI_FLUSH_VMID || optcode == CMD_TLBI_FLUSH_ASID || optcode == CMD_TLBI_FLUSH_VA || 
				optcode == CMD_TLBI_FLUSH_VA_RANGE || optcode == CMD_TLBI_FLUSH_ALL))
	{
		VASTAI_PCI_ERR(priv, 0,
				"Invalid cfg cmd!\n");
	}

	if (CMD_TLBI_FLUSH_VA_RANGE == optcode)
	{
		scale = ffs(range >> 12) - 1;
		num = range/(1<<(scale + VA_PTE_SHIFT)) - 1;
		tg = 1; // 4k page_siez
		ttl = 3;
		VASTAI_PCI_DBG(priv, 0,
				"va:%#llx, range:%#x, scale:%d, num:%d\n", address, range, scale, num);
	}

	addr_high = (u32)((u64)address >> 32) & 0xFFFFFFFF;
	addr_low  = (u32)((u64)address & 0xFFFFFFFF) >> 12;

	command[0]  = (u8)optcode;

	if (CMD_TLBI_FLUSH_VA_RANGE == optcode)
	{
		command[1]  = (u8)((num & 0xf) << 4);
		command[2]  = (u8)((num >> 4) | ((scale & 0xf) << 4));
		command[3]  = (u8)(scale >> 4);
	}
	else
	{
		command[1]  = 0;
		command[2]  = 0;
		command[3]  = 0;
	}

	command[4]  = 0;
	command[5]  = 0;

	if (CMD_TLBI_FLUSH_ASID == optcode || CMD_TLBI_FLUSH_VA_RANGE == optcode)
	{
		command[6]	= (u8)(asid         & 0xFF);
		command[7]	= (u8)((asid >> 24) & 0xFF);
	}
	else
	{
		command[6]  = 0;
		command[7]  = 0;
	}

	if (CMD_TLBI_FLUSH_VA_RANGE == optcode || CMD_TLBI_FLUSH_VA == optcode)
	{
		command[8]  = leaf;
		command[9]  = (u8)((addr_low & 0xF) << 4 | (tg << 2) | ttl);
#if 0
		command[10] = (u8)((addr_high >> 4)  & 0xFF);
		command[11] = (u8)((addr_high >> 12) & 0xFF);
		command[12] = (u8)((addr_low)       & 0xFF);
		command[13] = (u8)((addr_low >> 8)  & 0xFF);
		command[14] = (u8)((addr_low >> 16) & 0xFF);
		command[15] = (u8)((addr_low >> 24) & 0xFF);
#endif
		command[10] = (u8)((addr_low >> 4)  & 0xFF);
		command[11] = (u8)((addr_low >> 12) & 0xFF);
		command[12] = (u8)((addr_high)       & 0xFF);
		command[13] = (u8)((addr_high >> 8)  & 0xFF);
		command[14] = (u8)((addr_high >> 16) & 0xFF);
		command[15] = (u8)((addr_high >> 24) & 0xFF);
	}
	else
	{
		command[8]  = 0;
		command[9]  = 0;
		command[10] = 0;
		command[11] = 0;
		command[12] = 0;
		command[13] = 0;
		command[14] = 0;
		command[15] = 0;
	}

	cmd = (struct pcie_transfer_cmd *)command;
	if (ONLY_1_PF==priv->fn_mode) {
		ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF, cmd, 0);
	} else if (ONLY_2_PF==priv->fn_mode){
		ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF, cmd, 0);
	} else {
		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF, cmd, 0);
	}

	if(ret!=0) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"h2d_cmd_test error\n");
	}

	return 0;

//	return smmu_add_command(command);
}

void smmu_fill_tlb_invalid_cmd_descriptor_port(void *devInfo, u64 address, u32 range)
{
#if 0
       struct vastai_pci_info *priv =(struct vastai_pci_info *)devInfo;
       mutex_lock(&priv->vastai_smmu_info->pid_lock);
       smmu_fill_tlb_invalid_cmd_descriptor(priv, 0x30, 0, 1, address, range);
       mutex_unlock(&priv->vastai_smmu_info->pid_lock);
#endif
}

EXPORT_SYMBOL(smmu_fill_tlb_invalid_cmd_descriptor_port);

int smmu_fill_resum_cmd_descriptor(struct vastai_pci_info *priv, int stream_id, int stag, int action)
{
	u8 command[16];
	int ret;
	struct pcie_transfer_cmd *cmd;

	command[0]	= (u8)CMD_RESUME;
	
	command[1]  = ((u8)((action & 0x3) << 4));;
	command[2]	= 0;
	command[3]	= 0;
	
	command[4]  = (u8)( stream_id 	   & 0xFF);
	command[5]  = (u8)((stream_id >> 8)  & 0xFF);
	command[6]  = (u8)((stream_id >> 16) & 0xFF);
	command[7]  = (u8)((stream_id >> 24) & 0xFF);

	command[8]  = (u8)( stag	& 0xFF);
	command[9]  = (u8)((stag	>> 8)  & 0xFF);

	command[10] = 0;
	command[11] = 0;
	command[12] = 0;
	command[13] = 0;
	command[14] = 0;
	command[15] = 0;

	cmd = (struct pcie_transfer_cmd *)command;
	if (ONLY_1_PF==priv->fn_mode) {
		ret = vastai_pci_send_msg(priv, 0, ONLY_1PF_HOST_TO_GMCU1_CMD_BUF, cmd, 0);
	} else if (ONLY_2_PF==priv->fn_mode){
		ret = vastai_pci_send_msg(priv, 0, ONLY_2PF_HOST_TO_GMCU_CMD_BUF, cmd, 0);
	} else {
		ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_GMCU_CMD_BUF, cmd, 0);
	}
	if(ret!=0){
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"h2d_cmd_test error\n");
	}


	return 0;

//	return smmu_add_command(command);
}


int smmu_fill_sync_cmd_descriptor(struct vastai_pci_info *priv)
{
	u32 msi_data, data, msi_addr;
	int pf = priv->priv_hw_cfg->sys_cfg.pfn;
	int vf_active = priv->priv_hw_cfg->sys_cfg.vf_active;
	int vf = priv->priv_hw_cfg->sys_cfg.vfn, vector = 15;
#ifdef VASTAI_PERFORMANCE_TEST
	static unsigned long max_time = 0;
	static unsigned long times = 0;
	unsigned long start, end;
#endif

	msi_data = (pf << 24) | (vf_active ? (1 << 15 | vf << 16) : 0) | vector;
	msi_addr = priv->priv_hw_cfg->sys_cfg.smmu_sync_01;

#ifdef VASTAI_PERFORMANCE_TEST
	printk("%s start(%06d)\n", __func__, priv->vastai_smmu_info->sync_cnt++);
	start = ktime_get();
#endif
	do
	{
		vastai_pci_mem_read(priv, 0, msi_addr, &data, 4);
	}while(msi_data!=data);

#ifdef VASTAI_PERFORMANCE_TEST
	end = ktime_get();

	max_time = (max_time > (end - start)) ? max_time : (end - start);
	times++;
	if(times % 100 == 0)
	{
		printk("%s max_sync=%06lu", __func__, max_time / 1000);
	}
	printk("%s sync time=%06lu us, end(%06d)", __func__, (end - start)/1000, priv->vastai_smmu_info->sync_cnt);
#endif
	data = 0;
	vastai_pci_mem_write(priv, 0, msi_addr, &data, 4);

	return 0;
}

static u64 alloc_mem_from_ddk(struct vastai_pci_info *priv, int size, int level)
{
	u64 real_addr, alloc_addr = 0;
	u64 ddr_base = ((u64)((u64)priv->priv_hw_cfg->header.hi32_ddr_base << 32) | priv->priv_hw_cfg->header.lo32_ddr_base);
	u64 *page_addr;
	int i = 0;
	u64 data = 0;

#ifdef CONFIG_DDK_ALLOC_MEM

	if (size != PAGE_MAX_INDEX * PAGE_ENTRY_SIZE)
		return 0;

	if(priv->hasInitGfxMemAlloc)
	{
		alloc_addr = priv->AllocMemFromGfxHeap(priv->heapPtr, size);
	}
	else
#endif
	{
		switch (level)
		{

			case LEVEL0: {
				alloc_addr = ddr_base + LEVEL0_BASE + level0_cnt * size;
				if(alloc_addr >= ddr_base + LEVEL1_BASE)
				{
					alloc_addr = ddr_base + LEVEL0_BASE;
					level0_cnt = 0;
				}
				level0_cnt++;
				break;
			}
			case LEVEL1: {
				alloc_addr = ddr_base + LEVEL1_BASE + level1_cnt * size;
				if(alloc_addr >= ddr_base + LEVEL2_BASE)
				{
					alloc_addr = ddr_base + LEVEL1_BASE;
					level1_cnt = 0;
				}
				level1_cnt++;
				break;
			}
			case LEVEL2: {
				alloc_addr = ddr_base + LEVEL2_BASE + level2_cnt * size;
				if(alloc_addr >= ddr_base + LEVEL3_BASE)
				{
					alloc_addr = ddr_base + LEVEL2_BASE;
					level2_cnt = 0;
				}
				level2_cnt++;
				break;
			}
			case LEVEL3: {
				alloc_addr = ddr_base + LEVEL3_BASE + level3_cnt * size;
				if(alloc_addr >= ddr_base + 0xA000000)
				{
					alloc_addr = ddr_base + LEVEL3_BASE;
					level3_cnt = 0;
				}
				level3_cnt++;
				break;
			}
			default:
				VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
						"Unsupport level 0x%x\n", level);
			break;
		}
	}

#ifdef CONFIG_DDK_ALLOC_MEM
	real_addr = ddr_base + alloc_addr - 0x1040000000ULL;
#else
	real_addr = alloc_addr;
#endif

	while(i < PAGE_MAX_INDEX)
	{
		/*Note:when enable DIAG_TEST_WC priv->bar[VASTAI_PCI_BAR4].vmem =NULL*/
		page_addr = (u64 *)((real_addr + (i * PAGE_ENTRY_SIZE) - priv->bar[VASTAI_PCI_BAR4].noc_addr) + priv->bar[VASTAI_PCI_BAR4].vmem);
		*page_addr = data;
		i++;
	}

	return real_addr;
}

static void free_mem_from_ddk(struct vastai_pci_info *priv, u64 phy)
{
	if (phy > 0x1040000000 && phy < 0x1040040000)
	{
		phy = 0;
		phy = phy;
		return;
	}

#ifdef CONFIG_DDK_ALLOC_MEM
	if (priv->hasInitGfxMemAlloc)
	{
		u64 ddk_addr;
		ddk_addr = phy - ((u64)((u64)priv->priv_hw_cfg->header.hi32_ddr_base << 32) | priv->priv_hw_cfg->header.lo32_ddr_base)
							+ 0x1040000000;
		priv->FreeMemToGfxHeap(priv->heapPtr, ddk_addr);
	}
	else
#endif
	{
		phy = 0;
		phy = phy;
	}

	return;
}

static int get_pid_info(struct vastai_pci_info *priv, int pid, struct vastai_pid_pt_entry_info **pid_info)
{
	int ret = 0;
	struct vastai_pid_pt_entry_info *cur_pid;

	int cur_pid_num = priv->vastai_smmu_info->pid_cnt;

	if (!cur_pid_num)
	{
		ret = VASTAI_EMPTY;
		return ret;
	}

	list_for_each_entry (cur_pid, &priv->vastai_smmu_info->pid_head, node) {
		if (cur_pid->pid == pid) {
			*pid_info = cur_pid;
			ret = VASTAI_MATCH;
			return ret;
		}
	}

	ret = VASTAI_NULL;
	return ret;
}

static int va_is_already_map(u64 va, struct va_translation_info *cur_tl)
{
	/*0: don't map 1: already map*/
	return ((va >= cur_tl->va) && (va <= (cur_tl->va + cur_tl->size - 1))
				&& (ENTRY_MAGIC_NUM == cur_tl->magic_number));
}

static int pa_is_already_map(u64 pa, struct va_translation_info *cur_tl)
{
	/*0: don't map 1: already map*/
	return ((pa >= cur_tl->pa) && (pa <= (cur_tl->pa + cur_tl->size - 1))
				&& (ENTRY_MAGIC_NUM == cur_tl->magic_number));
}

static int check_page_is_mapped(struct vastai_pci_info *priv, u64 va, struct vastai_pid_pt_entry_info *pid_info, struct va_translation_info **tl_info)
{
	int ret = 0;
	struct va_translation_info *cur_tl;
	int cur_tl_num = pid_info->tl_info_cnt;

	if (!cur_tl_num)
	{
		//VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "entry list empty! #######\n");
		return ret;
	}

	list_for_each_entry (cur_tl, &pid_info->entry_head, node) {
		if (va_is_already_map(va, cur_tl))
		{
			*tl_info = cur_tl;
			return 1;
		}
	}

	return ret;
}

u64 va_virt_to_phys(struct vastai_pci_info *priv, u64 va, int pid)
{
	int ret = 0;
	struct va_translation_info *cur_tl;
	struct vastai_pid_pt_entry_info *pt_info;
	int cur_tl_num = 0;

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Can't find PID(%d) info\n", pid);
		return 0;
	}

	cur_tl_num = pt_info->tl_info_cnt;

	if (!cur_tl_num)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "entry list empty! #######\n");
		return 0;
	}

	list_for_each_entry (cur_tl, &pt_info->entry_head, node) {
		if (va_is_already_map(va, cur_tl))
		{
			return (va - cur_tl->va + cur_tl->pa);
		}
	}

	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "don't have matched entry\n");
	return 0;

}

u64 va_phys_to_virt(struct vastai_pci_info *priv, u64 pa, int pid)
{
	int ret = 0;
	struct va_translation_info *cur_tl;
	struct vastai_pid_pt_entry_info *pt_info;
	int cur_tl_num = 0;

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Can't find PID(%d) info\n", pid);
		return 0;
	}

	cur_tl_num = pt_info->tl_info_cnt;

	if (!cur_tl_num)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "entry list empty! #######\n");
		return 0;
	}

	list_for_each_entry (cur_tl, &pt_info->entry_head, node) {
		if (pa_is_already_map(pa, cur_tl))
		{
			return (pa - cur_tl->pa + cur_tl->va);
		}
	}

	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "don't have matched entry\n");
	return 0;

}

int get_page_entry_info(struct vastai_pci_info *priv, u64 va, int pid)
{
	int ret = 0;
	struct vastai_pid_pt_entry_info *pt_info;
	u64 page;
	u64 pgd_base, addr;
	u64 entry_off;

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Can't find PID(%d) info\n", pid);
		return ret;
	}

	if ((va >> 48) & VA_TYPE_CHECK)
	{
		pgd_base = pt_info->ttb1_pgd_base;
	}
	else
	{
		pgd_base = pt_info->ttb0_pgd_base;
	}

	entry_off = pgd_base + va_pgd_index(va) * PAGE_ENTRY_SIZE;
//	ret = vastai_read_by_smcu(priv, entry_off, PAGE_ENTRY_SIZE, &page);
	ret = vastai_pci_mem_read(priv, 0, entry_off, &page, PAGE_ENTRY_SIZE);
	if (ret) {
		VASTAI_PCI_ERR(priv, 0,"%s Invalidate pgd entry fail!, %d\n", __func__, ret);
		ret = -EPERM;
		return ret;
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "PGD Entry 0x%llx:[0x%llx]\n", entry_off, page);

	addr = page & PAGE_ADDR_MASK;
	entry_off = addr + va_pud_index(va) * PAGE_ENTRY_SIZE;
//	ret = vastai_read_by_smcu(priv, entry_off, PAGE_ENTRY_SIZE, &page);
	ret = vastai_pci_mem_read(priv, 0, entry_off, &page, PAGE_ENTRY_SIZE);
	if (ret) {
		VASTAI_PCI_ERR(priv, 0,"%s Invalidate pud entry fail!, %d\n", __func__, ret);
		ret = -EPERM;
		return ret;
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "PUD Entry 0x%llx:[0x%llx]\n", entry_off, page);

	addr = page & PAGE_ADDR_MASK;
	entry_off = addr + va_pmd_index(va) * PAGE_ENTRY_SIZE;
//	ret = vastai_read_by_smcu(priv, entry_off, PAGE_ENTRY_SIZE, &page);
	ret = vastai_pci_mem_read(priv, 0, entry_off, &page, PAGE_ENTRY_SIZE);
	if (ret) {
		VASTAI_PCI_ERR(priv, 0,"%s Invalidate pmd entry fail!, %d\n", __func__, ret);
		ret = -EPERM;
		return ret;
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "PMD Entry 0x%llx:[0x%llx]\n", entry_off, page);

	addr = page & PAGE_ADDR_MASK;
	entry_off = addr + va_pte_index(va) * PAGE_ENTRY_SIZE;
//	ret = vastai_read_by_smcu(priv, entry_off, PAGE_ENTRY_SIZE, &page);
	ret = vastai_pci_mem_read(priv, 0, entry_off, &page, PAGE_ENTRY_SIZE);
	if (ret) {
		VASTAI_PCI_ERR(priv, 0,"%s Invalidate pte entry fail!, %d\n", __func__, ret);
		ret = -EPERM;
		return ret;
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "PTE Entry 0x%llx:[0x%llx]\n", entry_off, page);

	return ret;
}

void walk_translation_info(struct vastai_pci_info *priv, int pid)
{
	int i = 0, ret = 0;
	struct va_translation_info *cur_tl;
	struct vastai_pid_pt_entry_info *pt_info;
	int cur_tl_num = 0;

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Can't find PID(%d) info\n", pid);
		return;
	}

	cur_tl_num = pt_info->tl_info_cnt;

	if (!cur_tl_num)
	{
		ret = 1;
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "entry list empty! #######\n");
		return;
	}

	list_for_each_entry (cur_tl, &pt_info->entry_head, node) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "Entry[%4d] va:0x%016llx, pa:0x%016llx, maping_count:%2d\n", i, cur_tl->va, cur_tl->pa, cur_tl->map_cnt);
		get_page_entry_info(priv, cur_tl->va, pid);
		i++;
	}

	return;
}

static u64 translate_idx_to_va(int pgd_idx, int pud_idx, int pmd_idx, int pte_idx)
{
	return ((u64)pgd_idx << VA_PGD_SHIFT) | ((u64)pud_idx << VA_PUD_SHIFT) | ((u64)pmd_idx << VA_PMD_SHIFT) |  ((u64)pte_idx << VA_PTE_SHIFT);
}

static int page_entry_valid_check(struct vastai_pci_info *priv, void *entry)
{
	PAGE_INFO *page = (PAGE_INFO *)entry;

	/* page entry valid return 0, otherwise return 1*/
	return ((page->attr0 != 0x0) || (page->attr1 != 0x0) || (page->attr2 != 0x0) || (page->attr3 != 0x0) || 
		(page->attr4 != 0x3));//|| (page->outaddr_47_32 != (u32)(priv->vastai_smmu_info->cd_base >> 32)));
}

static u64 get_next_level_entry_phys(struct vastai_pci_info *priv, void *entry)
{
	PAGE_INFO *page = (PAGE_INFO *)entry;
	u64 hi_addr, lo_addr;

	if (page_entry_valid_check(priv, entry))
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Invalid page entry!, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x, 0x%x \n", page->attr0, page->attr1, page->attr2, page->attr3, page->attr4, page->outaddr_31_12, page->outaddr_47_32);
		return 0;
	}

	hi_addr = (u64)page->outaddr_47_32 << 32;
	lo_addr = (u64)page->outaddr_31_12 << 12;

	return hi_addr | lo_addr;
}

void walk_page_from_dev(struct vastai_pci_info *priv, int pid)
{
	int ret = 0;
	int pgd_idx = 0;
	u64 pgd_base, va_base;
	struct vastai_pid_pt_entry_info *pt_info;

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, 0, "Can't find PID(%d) info\n", pid);
		return;
	}

	mutex_lock(&pt_info->entry_lock);
	pgd_base = pt_info->ttb0_pgd_base;
	/* gpu addr*/
	va_base = 0x0;
free_page:
	while(pgd_idx < PAGE_MAX_INDEX)
	{
		u64 pgd, pgd_off;

		pgd_off = pgd_base + pgd_idx * PAGE_ENTRY_SIZE;
		smmu_read_page_entry(priv, pgd_off, &pgd);

		//if pgd is valid, goto free pud
		if (!page_entry_valid_check(priv, &pgd))
		{
			int pud_idx = 0;
			while(pud_idx < PAGE_MAX_INDEX)
			{
				u64 pud, pud_off, pud_base = get_next_level_entry_phys(priv, &pgd);
				pud_off = pud_base + pud_idx * PAGE_ENTRY_SIZE;
				smmu_read_page_entry(priv, pud_off, &pud);

				//if pud is valid, goto free pmd
				if (!page_entry_valid_check(priv, &pud))
				{
					int pmd_idx = 0;
					while(pmd_idx < PAGE_MAX_INDEX)
					{
						u64 pmd, pmd_off, pmd_base = get_next_level_entry_phys(priv, &pud);
						pmd_off = pmd_base + pmd_idx * PAGE_ENTRY_SIZE;
						smmu_read_page_entry(priv, pmd_off, &pmd);

						//if pmd is valid, free pmd
						if (!page_entry_valid_check(priv, &pmd))
						{
							int pte_idx = 0;
							while(pte_idx < PAGE_MAX_INDEX)
							{
								u64 pte, pte_off, pte_base = get_next_level_entry_phys(priv, &pmd);
								pte_off = pte_base + pte_idx * PAGE_ENTRY_SIZE;
								smmu_read_page_entry(priv, pte_off, &pte);

								//if pmd is valid, free pmd
								if ((pte & PTE_VALIDE_CHECK) == PTE_VALIDE_CHECK)
								{
									VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "[walk_page]va:0x%016llx, pa:0x%016llx!\n", 
											va_base + translate_idx_to_va(pgd_idx, pud_idx, pmd_idx, pte_idx), pte & PAGE_ADDR_MASK);
								}
								pte_idx++;
							}
						}
						pmd_idx++;
					}
				}
				pud_idx++;
			}
		}
		pgd_idx++;
	}

	if (pgd_base != pt_info->ttb1_pgd_base)
	{
		pgd_idx = 0;
		/* cpu addr */
		va_base = 0xFFFF000000000000;
		pgd_base = pt_info->ttb1_pgd_base;
		goto free_page;
	}
	mutex_unlock(&pt_info->entry_lock);

	return;
}

#if 0
static int add_page_info(struct vastai_pci_info *priv, u64 va, u64 pa, u32 page_count, struct vastai_pid_pt_entry_info *pt_info, int s_bit)
{
	int ret = 0;
	int i;
	struct va_translation_info *tl_info;

	for (i = 0; i < page_count; i++)
	{
		tl_info = kzalloc(sizeof(struct va_translation_info), GFP_KERNEL);
		if (!tl_info) {
			VASTAI_PCI_ERR(priv, 0,
					"tl_info kmalloc fail, %d\n", i);
			ret = 1;
			return ret;
		}

		tl_info->pa = pa;
		tl_info->va = va;
		tl_info->size = VASTAI_PAGE_SIZE;
		tl_info->s_bit = s_bit;
		tl_info->magic_number = ENTRY_MAGIC_NUM;
		tl_info->map_cnt = 1;

		list_add(&(tl_info->node) , &(pt_info->entry_head));
		pt_info->tl_info_cnt += 1;

		tl_info = NULL;
		pa += VASTAI_PAGE_SIZE;
		va += VASTAI_PAGE_SIZE;
	}


#if 0
	for (i = 0; i < page_count; i++)
	{
		tl_info->pa = pa;
		tl_info->va = va;
		tl_info->size = VASTAI_PAGE_SIZE;
		tl_info->s_bit = s_bit;
		tl_info->magic_number = ENTRY_MAGIC_NUM;
		tl_info->map_cnt = 1;
		list_add(&(tl_info->node) , &(pt_info->entry_head));
		pt_info->tl_info_cnt += 1;
		tl_info++;
		pa += VASTAI_PAGE_SIZE;
		va += VASTAI_PAGE_SIZE;
	}
#endif

	return ret;
}
#endif

static int init_pte(struct vastai_pci_info *priv, pmd *pmdp, 
								u64 addr, u64 end, u64 pa, int s_bit)
{
	pte ptep;
	u64 pte_phys, pte_idx;
	int pa_low, pa_high;
	int ret = 0;
	u64 pte_off;

	pte_phys = get_next_level_entry_phys(priv, pmdp);
	if(!pte_phys)
	{
		ret = -EPERM;
		return ret;
	}

	do {
		pte_idx = va_pte_index(addr);

		pa_high = (int)(((u64)(pa) >> 32) & 0xFFFFFFFF);
		pa_low  = (int)((u64)(pa) & 0xFFFFFFFF);

		ptep.attr0 = 0;
		ptep.attr1 = s_bit;
		ptep.attr2 = 0;
		ptep.outaddr_47_32 = pa_high & 0xFFFF;
		ptep.outaddr_31_12 = pa_low >> 12;
		ptep.attr3 = 0x11f;
		ptep.attr4 = 3;

		pte_off = pte_phys + (va_pte_index(addr) * PAGE_ENTRY_SIZE);
		VASTAI_PCI_DBG(priv, 0,"create pte(%#llx) in %#llx\n", addr, pte_off);

		smmu_write_page_entry(priv, pte_off, (u64 *)&ptep);

		pa += VASTAI_PAGE_SIZE;
	} while (addr += VASTAI_PAGE_SIZE, addr < end);

	return ret;
}

static int alloc_init_cont_pte(struct vastai_pci_info *priv, pud *pudp, 
										pmd *pmdp, u64 addr, u64 end, u64 pa, int s_bit)
{
	u64 next;
	u64 pte_phys, pmd_off;
	int ret = 0;

	if (page_entry_valid_check(priv, pmdp)) {
		u64 pmd_phys;
		pte_phys = alloc_mem_from_ddk(priv, PAGE_ENTRY_SIZE * PAGE_MAX_INDEX, LEVEL3);

		pmdp->attr0 = 0;
		pmdp->attr1 = 0;
		pmdp->attr2 = 0;
		pmdp->outaddr_47_32 = (u32)(((u64)(pte_phys) >> 32)) & 0xFFFF;
		pmdp->outaddr_31_12 = (u32)((u64)(pte_phys) & 0xFFFFFFFF) >> 12;
		pmdp->attr3 = 0;
		pmdp->attr4 = 3;

		pmd_phys = get_next_level_entry_phys(priv, pudp);
		if(!pmd_phys)
		{
			ret = INIT_PTE_FAIL;
			goto err;
		}
		pmd_off = pmd_phys + PAGE_ENTRY_SIZE * va_pmd_index(addr);
		smmu_write_page_entry(priv, pmd_off, (u64 *)pmdp);
	}

	do {
		next = va_pte_addr_end(addr, end);
		ret = init_pte(priv, pmdp, addr, next, pa, s_bit);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s init pte failed! %d\n",
					__func__, ret);
			ret = INIT_PTE_FAIL;
			goto err;
		}

		pa += next - addr;
	} while (addr = next, addr < end);

	return ret;

err:
	free_mem_from_ddk(priv, pte_phys);
	return ret;
}

static int init_pmd(struct vastai_pci_info *priv, pud *pudp, u64 addr, 
								u64 end, u64 pa, int s_bit)
{
	u64 next, pmd_phys, pmd_idx;
	pmd pmdp;
	int ret = 0;
	pmd old_pmd;

	pmd_phys = get_next_level_entry_phys(priv, pudp);
	if(!pmd_phys)
	{
		ret = INIT_PMD_FAIL;
		return ret;
	}

	do {
		pmd_idx = va_pmd_index(addr);

		smmu_read_page_entry(priv, pmd_phys + pmd_idx *  PAGE_ENTRY_SIZE, (u64 *)&pmdp);
		old_pmd = pmdp;

		next = va_pmd_addr_end(addr, end);
		ret = alloc_init_cont_pte(priv, pudp, &pmdp, addr, next, pa, s_bit);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s init pte failed! %d\n",
					__func__, ret);
			ret = INIT_PMD_FAIL;
			return ret;
		}

		if ((old_pmd.attr4 != 0) && 
				((old_pmd.outaddr_31_12 != pmdp.outaddr_31_12) || (old_pmd.outaddr_47_32 != pmdp.outaddr_47_32)))
		{
			VASTAI_PCI_ERR(priv, 0,
				"current pud entry(%d) have already init!\n", (u32)va_pmd_index(addr));
			return ret;
		}

		pa += next - addr;
	} while (addr = next, addr < end);

	return ret;
}

static int alloc_init_cont_pmd(struct vastai_pci_info *priv, pgd *pgdp, 
											pud *pudp, u64 addr, u64 end, u64 pa, int s_bit)
{
	u64 next;
	u64 pmd_phys;
	int ret = 0;

	if (page_entry_valid_check(priv, pudp)) {
		u64 pud_off;
		u64 pud_phys;
		pmd_phys = alloc_mem_from_ddk(priv, PAGE_ENTRY_SIZE * PAGE_MAX_INDEX, LEVEL2);

		pudp->attr0 = 0;
		pudp->attr1 = 0;
		pudp->attr2 = 0;
		pudp->outaddr_47_32 = (u32)(((u64)(pmd_phys) >> 32)) & 0xFFFF;
		pudp->outaddr_31_12 = (u32)((u64)(pmd_phys) & 0xFFFFFFFF) >> 12;
		pudp->attr3 = 0;
		pudp->attr4 = 3;

		pud_phys = get_next_level_entry_phys(priv, pgdp);
		if(!pud_phys)
		{
			ret = INIT_PUD_FAIL;
			goto err;
		}
		pud_off =pud_phys + PAGE_ENTRY_SIZE * va_pud_index(addr);

		smmu_write_page_entry(priv, pud_off, (u64 *)pudp);
	}

	do {
		next = va_pmd_addr_end(addr, end);
		ret = init_pmd(priv, pudp, addr, next, pa, s_bit);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s init pmd failed! %d\n",
					__func__, ret);
			ret = INIT_PMD_FAIL;
			goto err;
		}

		pa += next - addr;
	} while (addr = next, addr < end);

	return ret;
err:
	free_mem_from_ddk(priv, pmd_phys);
	return ret;
}

static int alloc_init_pud(struct vastai_pci_info *priv, pgd *pgdp, 
									u64 addr, u64 end, u64 pa, u64 pgd_base, int s_bit)
{
	int ret = 0;
	u64 next, pud_idx;
	pud pudp;
	u64 pud_phys;
	pud old_pud;

	if (page_entry_valid_check(priv, pgdp))
	{

		u64 pgd_off;
		pud_phys = alloc_mem_from_ddk(priv, PAGE_ENTRY_SIZE * PAGE_MAX_INDEX, LEVEL1); //alloc 4k pud entry

		pgdp->attr0 = 0;
		pgdp->attr1 = 0;
		pgdp->attr2 = 0;
		pgdp->outaddr_47_32 = ((u32)(((u64)(pud_phys >> 32)) & 0xFFFFFFFF) & 0xFFFF);
		pgdp->outaddr_31_12 = (u32)((u64)(pud_phys) & 0xFFFFFFFF) >> 12;
		pgdp->attr3 = 0;
		pgdp->attr4 = 3;

		pgd_off = pgd_base + (PAGE_ENTRY_SIZE * va_pgd_index(addr));

		smmu_write_page_entry(priv, pgd_off, (u64 *)pgdp);
	}

	pud_phys = get_next_level_entry_phys(priv, pgdp);
	do {
		pud_idx = va_pud_index(addr);

		smmu_read_page_entry(priv, pud_phys + pud_idx * PAGE_ENTRY_SIZE, (u64 *)&pudp);

		old_pud = pudp;

		next = va_pud_addr_end(addr, end);
		ret = alloc_init_cont_pmd(priv, pgdp, &pudp, addr, next, pa, s_bit);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s alloc init pmd fail!\n",__func__);
			ret = INIT_PUD_FAIL;
			goto err;
		}
		if ((old_pud.attr4 == 0x3) && 
				((old_pud.outaddr_31_12 != pudp.outaddr_31_12) || (old_pud.outaddr_47_32 != pudp.outaddr_47_32)))
		{
			VASTAI_PCI_ERR(priv, 0,
				"current pud entry(%d) have already init!\n", (int)va_pud_index(addr));
			ret = INIT_PUD_FAIL;
			goto err;
		}

		pa += next - addr;
	} while (addr = next, addr < end);

	return ret;

err:
	free_mem_from_ddk(priv, pud_phys);
	return ret;
}

static int init_smmu_page_entry(struct vastai_pci_info *priv, u64 pa, u64 va, u64 pgd_base,
				int page_number, int s_bit)
{
	u32 ret = 0; 
	u64 addr, end, next, pgd_idx;
	u32 length; 
	pgd pgdp;

	pa &= VA_PTE_MASK;
	addr = va & VA_PTE_MASK;
	length = page_number * VASTAI_PAGE_SIZE;
	end = addr + length;

	do {
		pgd_idx = va_pgd_index(addr);
		smmu_read_page_entry(priv, pgd_base + pgd_idx * PAGE_ENTRY_SIZE, (u64 *)&pgdp);
		next = va_pgd_addr_end(addr, end);
		ret = alloc_init_pud(priv, &pgdp, addr, next, pa, pgd_base, s_bit);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s alloc init pud in %#llx fail!",__func__, pgd_base);
			ret = INIT_PGD_FAIL;
			return ret;
		}
		pa += next - addr;

		addr = next;
	} while (addr != end);

	VASTAI_PCI_DBG(priv, 0, "%s(), Mappint va[0x%llx->0x%llx] page entry [%lld]-[%lld]-[%lld]-[%lld] to (0x%llx)[%lld]-[%lld]-[%lld]-[%lld] have been create\n", 
					__func__, va, end, va_pgd_index(va), va_pud_index(va), va_pmd_index(va), va_pte_index(va), 
					end, va_pgd_index(end), va_pud_index(end), va_pmd_index(end), va_pte_index(end));

	return ret;
}

int va_create_page_mapping(struct vastai_pci_info *priv, u64 pa, 
									u64 va, int size, int pid, int s_bit)
{
	u32 i, ret = 0; 
	u64 addr, pgd_base;
	u32 length, page_cnt; 
	struct vastai_pid_pt_entry_info *pt_info;

	pa &= VA_PTE_MASK;
	addr = va & VA_PTE_MASK;
	length = VA_PAGE_ALIGN(size + (va & ~VASTAI_PAGE_MASK));
	page_cnt = length >> VA_PTE_SHIFT;

	if (!length)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "the length is at least 4k\n");
		return -ENOMEM;
	}

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Can't find PID(%d) info\n", pid);
		return ret;
	}
	ret = 0;

	if ((addr >> 48) & VA_TYPE_CHECK)
	{
		pgd_base = pt_info->ttb1_pgd_base;
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "Pid(%d) create CPUVA entry,va[0x%llx] to pa[0x%llx], length[0x%x]\n", pid, addr, pa, length);
	}
	else
	{
		pgd_base = pt_info->ttb0_pgd_base;
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "Pid(%d) create GPUVA entry,va[0x%llx] to pa[0x%llx], length[0x%x]\n", pid, addr, pa, length);
	}

	mutex_lock(&pt_info->entry_lock);

	if(GART_VA_PID == pid)
	{
		int gart_va_off = addr/VASTAI_PAGE_SIZE;
		pte ptep;
		int pa_low, pa_high;
		struct smmu_info *va_smmu_info;
		u64 pte_base, pte_off;

		va_smmu_info = priv->vastai_smmu_info;
		pte_base = va_smmu_info->va_gart_info.gart_entry_base;
		pte_off = pte_base + gart_va_off * PAGE_ENTRY_SIZE;

		for (i = 0; i < page_cnt; i++)
		{
			pa_high = (int)(((u64)(pa) >> 32) & 0xFFFFFFFF);
			pa_low  = (int)((u64)(pa) & 0xFFFFFFFF);

			ptep.attr0 = 0;
			ptep.attr1 = s_bit;
			ptep.attr2 = 0;
			ptep.outaddr_47_32 = pa_high & 0xFFFF;
			ptep.outaddr_31_12 = pa_low >> 12;
			ptep.attr3 = 0x11f;
			ptep.attr4 = 3;

			smmu_write_page_entry(priv, pte_off + i * PAGE_ENTRY_SIZE, (u64 *)&ptep);
			pa += VASTAI_PAGE_SIZE;
		}
	}
	else
	{
		ret = init_smmu_page_entry(priv, pa, addr, pgd_base, page_cnt, s_bit);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s() create page mapping failed!%d", __func__, ret);
			mutex_unlock(&pt_info->entry_lock);
			return ret;
		}
	}
	mutex_unlock(&pt_info->entry_lock);

	return ret;
}

int va_create_page_mapping_port(void *priv, u64 pa, u64 va, int size, int pid, int s_bit)
{
	return va_create_page_mapping((struct vastai_pci_info *)priv, pa, va, size, pid, s_bit);
}
EXPORT_SYMBOL(va_create_page_mapping_port);

int va_fixed_mapping_in_gart(struct vastai_pci_info *priv, struct smmu_info *va_smmu_info)
{
	u64 length, pte_base, pmd_base, pud_base, pgd_base;
	int ret = 0;
	pgd pgdp;
	pud pudp;
	pmd pmdp;
	int i, pmd_cnt, pud_cnt, pgd_cnt;

	length = VA_PAGE_ALIGN(va_smmu_info->va_gart_info.gart_size);

	if((length < MIN_GATR_SIZE) || (length > MAX_GATR_SIZE))
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "length (0x%llx) is invalid\n", length);

	pmd_cnt = DIV_ROUND_UP(length, PMD_VA_RANGE);
	pud_cnt = DIV_ROUND_UP(length, PUD_VA_RANGE);
	pgd_cnt = DIV_ROUND_UP(length, PGD_VA_RANGE);

	pte_base = va_smmu_info->va_gart_info.gart_entry_base;
	pmd_base = va_smmu_info->va_gart_info.gart_pmd_base;
	pud_base = va_smmu_info->va_gart_info.gart_pud_base;
	pgd_base = va_smmu_info->va_gart_info.gart_pgd_base;
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s, pgd_base:0x%llx, pud_base:0x%llx, pmd_base:0x%llx, pte_base:0x%llx\n", 
								__func__, pgd_base, pud_base, pmd_base, pte_base);

	/* init pmd */
	for (i = 0; i < pmd_cnt; i++)
	{
		pmdp.attr0 = 0;
		pmdp.attr1 = 0;
		pmdp.attr2 = 0;
		pmdp.outaddr_47_32 = (u32)(((u64)(pte_base) >> 32)) & 0xFFFF;
		pmdp.outaddr_31_12 = (u32)((u64)(pte_base) & 0xFFFFFFFF) >> 12;
		pmdp.attr3 = 0;
		pmdp.attr4 = 3;

		smmu_write_page_entry(priv, pmd_base + i * PAGE_ENTRY_SIZE, (u64 *)&pmdp);

		pte_base += VASTAI_PAGE_SIZE;
	}

	/* init pud */
	for (i = 0; i < pud_cnt; i++)
	{
		pudp.attr0 = 0;
		pudp.attr1 = 0;
		pudp.attr2 = 0;
		pudp.outaddr_47_32 = (u32)(((u64)(pmd_base) >> 32)) & 0xFFFF;
		pudp.outaddr_31_12 = (u32)((u64)(pmd_base) & 0xFFFFFFFF) >> 12;
		pudp.attr3 = 0;
		pudp.attr4 = 3;

		smmu_write_page_entry(priv, pud_base + i * PAGE_ENTRY_SIZE, (u64 *)&pudp);

		pmd_base += VASTAI_PAGE_SIZE;
	}

	/* init pgd */
	for (i = 0; i < pgd_cnt; i++)
	{
		pgdp.attr0 = 0;
		pgdp.attr1 = 0;
		pgdp.attr2 = 0;
		pgdp.outaddr_47_32 = ((u32)(((u64)(pud_base >> 32)) & 0xFFFFFFFF) & 0xFFFF);
		pgdp.outaddr_31_12 = (u32)((u64)(pud_base) & 0xFFFFFFFF) >> 12;
		pgdp.attr3 = 0;
		pgdp.attr4 = 3;

		smmu_write_page_entry(priv, pgd_base + i * PAGE_ENTRY_SIZE, (u64 *)&pgdp);

		pud_base += VASTAI_PAGE_SIZE;
	}

	return ret;
}

static int __va_create_pid(struct vastai_pci_info *priv, int pid)
{
	int ret = 0;
	u64 ttb0_pgd_base;
	u64 ttb1_pgd_base;
	struct cd_s_info *cd_info;
	int ttbtr0_high, ttbtr0_low, ttbtr1_high, ttbtr1_low;
	struct vastai_pid_pt_entry_info *pt_info;
	struct vastai_pid_pt_entry_info *pt_check;
	struct smmu_info *va_smmu_info;
	u64 cd_off;

	va_smmu_info = priv->vastai_smmu_info;
	cd_off= va_smmu_info->cd_base + (pid * CD_SIZE);

	ret = get_pid_info(priv, pid, &pt_check);
	if(VASTAI_MATCH == ret)
	{
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "PID(%d) is already init, break\n", pid);
		return 0;
	}

	pt_info = kzalloc(sizeof(struct vastai_pid_pt_entry_info), GFP_KERNEL);
	if (NULL == pt_info)
	{
		ret = -ENOMEM;
		return ret;
	}

	cd_info = kzalloc(sizeof(struct cd_s_info), GFP_KERNEL);
	if (NULL == cd_info) {
		VASTAI_PCI_ERR(priv, 0,
				"cd_info kmalloc fail\n");
		kfree(pt_info);
		ret = -ENOMEM;
		return ret;
	}

	if (GART_VA_PID == pid){
		ttb0_pgd_base = va_smmu_info->va_gart_info.gart_pgd_base;
		ttb1_pgd_base = va_smmu_info->va_gart_info.gart_pgd_base;
	}else{
		ttb0_pgd_base = alloc_mem_from_ddk(priv, 0x1000, LEVEL0);
		ttb1_pgd_base = alloc_mem_from_ddk(priv, 0x1000, LEVEL0);
	}
	pt_info->ttb1_pgd_base = ttb1_pgd_base;
	pt_info->ttb0_pgd_base = ttb0_pgd_base;
	pt_info->pid = pid;
	pt_info->magic_number = PID_MAGIC_NUM;
	pt_info->tl_info_cnt = 0;

	ttbtr0_high = (((u32)((u64)(pt_info->ttb0_pgd_base) >> 32)) & 0xFFFFF);
	ttbtr0_low = ((u32)((u64)(pt_info->ttb0_pgd_base) & 0xFFFFFFFF)) >> 4;
	ttbtr1_high = (((u32)((u64)(pt_info->ttb1_pgd_base) >> 32)) & 0xFFFFF);
	ttbtr1_low = ((u32)((u64)(pt_info->ttb1_pgd_base) & 0xFFFFFFFF)) >> 4;

	VASTAI_PCI_DBG(priv, 0, "Init pid(%d) cd_off:0x%llx, ttb0_addr:0x%llx, ttb1_addr:0x%llx\n", 
				pid, cd_off, pt_info->ttb0_pgd_base, pt_info->ttb1_pgd_base);

	cd_info->w0.T0SZ  =  16;
	cd_info->w0.TG0   =  0;
	cd_info->w0.T1SZ  =  16;
	cd_info->w0.TG1   =  2;
	cd_info->w0.valid =  1;
	cd_info->w0.EPD1  =  0;

	cd_info->w1.IPS   =  5;
	cd_info->w1.AFFD  =  1;
	cd_info->w1.AA64  =  1;
	cd_info->w1.R     =  1;
	cd_info->w1.A     =  1;
	cd_info->w1.ASET  =  1;
	cd_info->w1.ASID  =  pt_info->pid;

	cd_info->w2.HAD0        =  1;
	cd_info->w2.TTB0_low    =  ttbtr0_low;
	
	cd_info->w3.TTB0_high   =  ttbtr0_high;
	cd_info->w3.HWU59       =  1;
	cd_info->w3.HWU60       =  1;
	cd_info->w3.HWU61       =  1;
	cd_info->w3.HWU62       =  1;

	cd_info->w4.TTB1_low    =  ttbtr1_low;
	cd_info->w5.TTB1_high   =  ttbtr1_high;

	ret = vastai_pci_mem_write(priv, 0, cd_off, cd_info, CD_SIZE);
	if (ret) {
		VASTAI_PCI_ERR(priv, 0,
				"%s write smmu ste information to dev fail!, %d\n",
				__func__, ret);
		ret = -EPERM;
		goto err;
	}

	list_add(&(pt_info->node) , &(va_smmu_info->pid_head));
	va_smmu_info->pid_cnt += 1;

	kfree(cd_info);

	// init entry list for current pid
	INIT_LIST_HEAD(&(pt_info->entry_head));
	// init entry lock for current pid
	mutex_init(&pt_info->entry_lock);
	return ret;

err:
	free_mem_from_ddk(priv, ttb0_pgd_base);
	free_mem_from_ddk(priv, ttb1_pgd_base);
	kfree(cd_info);
	kfree(pt_info);
	return ret;
}

int va_create_pid(struct vastai_pci_info *priv, int pid)
{
	int va_pid = 0;
	int cur_pid = 0;
	struct smmu_info *va_smmu_info = priv->vastai_smmu_info;;

	mutex_lock(&va_smmu_info->pid_lock);

	if (pid == GART_VA_PID)
	{
		if (priv->vastai_smmu_info->vastai_pid[0] == 0)
		{
			va_pid = 0;
			priv->vastai_smmu_info->vastai_pid[0] = PID_MAGIC_NUM;
			__va_create_pid(priv, va_pid);
			mutex_unlock(&va_smmu_info->pid_lock);
			return va_pid;
		}
		else if (priv->vastai_smmu_info->vastai_pid[0] == PID_MAGIC_NUM)
		{
			va_pid = 0;
			mutex_unlock(&va_smmu_info->pid_lock);
			return va_pid;
		}
		else
		{
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s(), current state of va_pid is invalid, %d\n", 
							__func__, priv->vastai_smmu_info->vastai_pid[0]);
			mutex_unlock(&va_smmu_info->pid_lock);
			return -1;
		}
	}

	if (priv->vastai_smmu_info->pid_cnt == MAX_VASTAI_PID)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s(), pid cnt is full, can't add a new pid%d, break\n", 
								__func__, pid);
		mutex_unlock(&va_smmu_info->pid_lock);
		return -1;
	}
	// pid0 is reserved for gart mmu
	for (cur_pid = 1; cur_pid < MAX_VASTAI_PID; cur_pid++)
	{
		if(0 == priv->vastai_smmu_info->vastai_pid[cur_pid])
		{
			va_pid = cur_pid;
		}

		if (pid == priv->vastai_smmu_info->vastai_pid[cur_pid])
		{
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s(), pid(%d) have mapped to va pid(%d)break\n", 
								__func__, pid, cur_pid);
			mutex_unlock(&va_smmu_info->pid_lock);
			return cur_pid;
		}
	}

	priv->vastai_smmu_info->vastai_pid[va_pid] = pid;
	__va_create_pid(priv, va_pid);

	mutex_unlock(&va_smmu_info->pid_lock);

	return va_pid;
}
int va_create_pid_port(struct vastai_pci_info *priv, int pid)
{
	return va_create_pid((struct vastai_pci_info *)priv, pid);
}
EXPORT_SYMBOL(va_create_pid_port);

#if 0
static int remove_page_info(struct vastai_pci_info *priv, u64 va, struct vastai_pid_pt_entry_info *pid_info)
{
	int ret = 0;
	struct va_translation_info *cur_tl;
	int cur_tl_num = pid_info->tl_info_cnt;

	if (!cur_tl_num)
	{
		ret = 1;
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "entry list empty! #######\n");
		return ret;
	}

	list_for_each_entry (cur_tl, &pid_info->entry_head, node) {
		if (va_is_already_map(va, cur_tl))
		{
			list_del(&(cur_tl->node));
			kfree(cur_tl);
			//VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "free entry ok!\n");
			pid_info->tl_info_cnt -= 1;
			return ret;
		}
	}
	

	ret = 1;
	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "don't have matched entry, 0x%llx\n", va);

	return ret;
}

static int remove_page_info_repeatedly(struct vastai_pci_info *priv, u64 addr, u32 page_number, struct vastai_pid_pt_entry_info *pid_info)
{
	int i=0;
	int ret=0;
	u64 va = addr;
	for (i = 0; i < page_number; i++)
	{
		remove_page_info(priv, va, pid_info);
		va += VASTAI_PAGE_SIZE;
	}

	return ret;
}
#endif

#if 0
static inline int remove_pid_info(struct vastai_pci_info *priv, struct vastai_pid_pt_entry_info *pt_info)
{
	int ret = 0;

	mutex_lock(&priv->vastai_smmu_info->pid_lock);
	list_del(&(pt_info->node));
	priv->vastai_smmu_info->pid_cnt -= 1;
	mutex_unlock(&priv->vastai_smmu_info->pid_lock);

	kfree(pt_info);

	return ret;
}

static int compare_entry_index(struct vastai_pci_info *priv, u64 addr, struct va_translation_info *tl_info, u64 ENTRY_MASK)
{

	//VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "srv_va:0x%llx, dst_va:0x%llx, entry_mask:0x%llx\n", addr, tl_info->va, ENTRY_MASK);
	if (((tl_info->va & ENTRY_MASK) == (addr & ENTRY_MASK))
			|| (((tl_info->va + tl_info->size) & ENTRY_MASK) == (addr & ENTRY_MASK)))
	{
		return 1;
	}

	return 0;
}

static int check_cur_entry_is_stil_used(struct vastai_pci_info *priv, struct vastai_pid_pt_entry_info *pt_info, u64 addr, u64 ENTRY_MASK)
{
	int ret = 0;
	struct va_translation_info *cur_tl;
	int total_tl_num = pt_info->tl_info_cnt;

	if (!total_tl_num)
	{
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "entry list empty\n");
		return ret;
	}

	list_for_each_entry (cur_tl, &pt_info->entry_head, node){
		if (!((cur_tl->va <= addr) && ((cur_tl->va + cur_tl->size - 1) >= addr)))
		{
			if (compare_entry_index(priv, addr, cur_tl, ENTRY_MASK))
			{
				//VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "The current entry is still used by entry[0x%llx]\n", cur_tl->va);
				ret = 1;
				return ret;
			}
		}
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s(), total %d of the entries don't matched with the VA[0x%llx]\n", __func__, total_tl_num, addr);
	return ret;
}
#endif
#if 0
static int free_pte_entry(struct vastai_pci_info *priv, struct vastai_pid_pt_entry_info *pt_info, pmd *pmdp, u64 addr, u64 end)
{
	u64 pte_phys, pte_idx;
	int ret = 0;
	u64 pte_off;
	pte ptep;

	pte_phys = get_next_level_entry_phys(priv, pmdp);
	if(!pte_phys)
	{
		ret = -EPERM;
		return ret;
	}

	ptep.attr4 = 0;
	ptep.attr3 = 0;
	ptep.outaddr_31_12 = 0;
	ptep.outaddr_47_32 = 0;
	ptep.attr2 = 0;
	ptep.attr1 = 0;
	ptep.attr0 = 0;

	do {
		pte_idx = va_pte_index(addr);
		pte_off = pte_phys + (va_pte_index(addr) * PAGE_ENTRY_SIZE);

		smmu_write_page_entry(priv, pte_off, (u64 *)&ptep);
		} while (addr += VASTAI_PAGE_SIZE, addr < end);

	return ret;
}

static int free_pmd_entry(struct vastai_pci_info *priv, struct vastai_pid_pt_entry_info *pt_info, pud *pudp, u64 addr, u64 end)
{
	u64 next, pmd_phys, pmd_idx;
	pmd pmdp;
	int ret = 0;

	pmd_phys = get_next_level_entry_phys(priv, pudp);
	if(!pmd_phys)
	{
		ret = -EPERM;
		return ret;
	}

	do {
		next = va_pmd_addr_end(addr, end);
		pmd_idx = va_pmd_index(addr);

		smmu_read_page_entry(priv, pmd_phys + pmd_idx *  PAGE_ENTRY_SIZE, (u64 *)&pmdp);

		ret = free_pte_entry(priv, pt_info, &pmdp, addr, next);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s free pte failed! %d\n",
					__func__, ret);
			ret = FREE_PMD_FAIL;
			return ret;
		}

	} while (addr = next, addr < end);

	return ret;

}

static int free_pud_entry(struct vastai_pci_info *priv, struct vastai_pid_pt_entry_info *pt_info, pgd *pgdp, u64 addr, u64 end)
{
	int ret = 0;
	u64 next, pud_idx;
	pud pudp;
	u64 pud_phys;

	pud_phys = get_next_level_entry_phys(priv, pgdp);
	do {
		next = va_pud_addr_end(addr, end);
		pud_idx = va_pud_index(addr);

		smmu_read_page_entry(priv, pud_phys + pud_idx * PAGE_ENTRY_SIZE, (u64 *)&pudp);

		ret = free_pmd_entry(priv, pt_info, &pudp, addr, end);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s free pmd fail!\n",__func__);
			ret = FREE_PUD_FAIL;
			return ret;
		}

	} while (addr = next, addr < end);

	return ret;
}

static int free_smmu_page_entry(struct vastai_pci_info *priv, u64 va, u64 pgd_base,
				u32 page_number, struct vastai_pid_pt_entry_info *pt_info)
{
	u32 ret = 0; 
	u64 addr, length, end, next, pgd_idx;
	pgd pgdp;

	addr = va & VA_PTE_MASK;
	length = page_number * VASTAI_PAGE_SIZE;
	end = addr + length;

	do {
		next = va_pgd_addr_end(addr, end);

		pgd_idx = va_pgd_index(addr);

		smmu_read_page_entry(priv, pgd_base + pgd_idx * PAGE_ENTRY_SIZE, (u64 *)&pgdp);

		ret = free_pud_entry(priv, pt_info, &pgdp, addr, end);
		if (ret) {
			VASTAI_PCI_ERR(priv, 0,
					"%s free pud fail!",__func__);
			ret = FREE_PGD_FAIL;
			return ret;
		}

		addr = next;
	} while (addr < end);

	VASTAI_PCI_DBG(priv, 0, "%s(),page entry(0x%llx-0x%llx) [%lld]-[%lld]-[%lld]-[%lld] to [%lld]-[%lld]-[%lld]-[%lld] have been cleared\n", 
				__func__, va, end, va_pgd_index(va), va_pud_index(va), va_pmd_index(va), va_pte_index(va), 
				va_pgd_index(end), va_pud_index(end), va_pmd_index(end), va_pte_index(end));

	return ret;
}
#endif
int va_free_page_maping(struct vastai_pci_info *priv, u64 va, u32 size, int pid)
{
	u32 page_cnt, length, ret = 0; 
	u64 addr, end, pgd_base;
	struct vastai_pid_pt_entry_info *pt_info;

	addr = va & VA_PTE_MASK;
	length = VA_PAGE_ALIGN(size + (va & ~VASTAI_PAGE_MASK));
	end = addr + length;
	page_cnt = length >> VA_PTE_SHIFT;

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Can't find PID(%d) info\n", pid);
		return ret;
	}

	if ((va >> 48) & VA_TYPE_CHECK)
	{
		pgd_base = pt_info->ttb1_pgd_base;
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "free CPUVA entry va[0x%llx], length:0x%x\n", va, length);
	}
	else
	{
		pgd_base = pt_info->ttb0_pgd_base;
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "free GPUVA entry va[0x%llx], length:0x%x\n", va, length);
	}

	if(GART_VA_PID == pid)
	{
		VASTAI_PCI_DBG(priv, 0, "%s, cur_pid is gart pid, nothing to do\n", __func__);
		goto tlb_va_invalid;
	}

#if 0
	mutex_lock(&pt_info->entry_lock);
	for (i = 0; i < page_cnt; i++)
	{
		ret = free_smmu_page_entry(priv, addr, pgd_base, 1, pt_info);
		if(ret)
		{
			VASTAI_PCI_ERR(priv, 0,"free Page entry[0x%llx] failed!\n", addr);
			mutex_unlock(&pt_info->entry_lock);
			return ret;
		}
		addr += VASTAI_PAGE_SIZE;
	}
	mutex_unlock(&pt_info->entry_lock);
#endif

tlb_va_invalid:
	mutex_lock(&priv->vastai_smmu_info->pid_lock);
	ret = smmu_fill_tlb_invalid_cmd_descriptor(priv, CMD_TLBI_FLUSH_VMID, pid, 0, addr, length);
	if (ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "TLB Flush va(0x%llx)failed!\n", addr);
		return ret;
	}
	smmu_fill_sync_cmd_descriptor(priv);
	mutex_unlock(&priv->vastai_smmu_info->pid_lock);

	return ret;
}

int va_free_page_maping_port(void *priv, u64 va, u32 size, int pid)
{
	return va_free_page_maping((struct vastai_pci_info *)priv, va, size, pid);
}
EXPORT_SYMBOL(va_free_page_maping_port);

static int __va_kill_pid(struct vastai_pci_info *priv, int pid)
{
	int ret = 0;
	int pgd_idx = 0;
	u64 pgd_base;
	struct vastai_pid_pt_entry_info *pt_info;

	ret = get_pid_info(priv, pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, 0, "Can't find PID(%d) info\n", pid);
		return ret;
	}

	if(GART_VA_PID == pid)
	{
		VASTAI_PCI_DBG(priv, 0, "%s, cur_pid is gart pid, nothing to do\n", __func__);
		goto tlb_all_invalid;
	}

	pgd_base = pt_info->ttb0_pgd_base;

free_page:
	while(pgd_idx < PAGE_MAX_INDEX)
	{
		u64 pgd, pgd_off;

		pgd_off = pgd_base + pgd_idx * PAGE_ENTRY_SIZE;
		smmu_read_page_entry(priv, pgd_off, &pgd);

		//if pgd is valid, goto free pud
		if (!page_entry_valid_check(priv, &pgd))
		{
			int pud_idx = 0;
			while(pud_idx < PAGE_MAX_INDEX)
			{
				u64 pud, pud_off, pud_base = get_next_level_entry_phys(priv, &pgd);
				pud_off = pud_base + pud_idx * PAGE_ENTRY_SIZE;
				smmu_read_page_entry(priv, pud_off, &pud);

				//if pud is valid, goto free pmd
				if (!page_entry_valid_check(priv, &pud))
				{
					int pmd_idx = 0;
					while(pmd_idx < PAGE_MAX_INDEX)
					{
						u64 pmd, pmd_off, pmd_base = get_next_level_entry_phys(priv, &pud);
						pmd_off = pmd_base + pmd_idx * PAGE_ENTRY_SIZE;
						smmu_read_page_entry(priv, pmd_off, &pmd);

						//if pmd is valid, free pmd
						if (!page_entry_valid_check(priv, &pmd))
						{
							free_mem_from_ddk(priv, get_next_level_entry_phys(priv, &pmd));
						}
						pmd_idx++;
					}

					free_mem_from_ddk(priv, get_next_level_entry_phys(priv, &pud));
				}
				pud_idx++;
			}

			free_mem_from_ddk(priv, get_next_level_entry_phys(priv, &pgd));
		}

		pgd_idx++;
	}
	free_mem_from_ddk(priv, pgd_base);

	if (pgd_base != pt_info->ttb1_pgd_base)
	{
		pgd_idx = 0;
		pgd_base = pt_info->ttb1_pgd_base;
		goto free_page;
	}

tlb_all_invalid:
	/* TLB Flush all entry */
	ret = smmu_fill_tlb_invalid_cmd_descriptor(priv, CMD_TLBI_FLUSH_ASID, pid, 0, 0, 0);
	if (ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "TLB Flush all entry of pid(%d) failed!\n", pid);
		return ret;
	}

	/* Flush cfg entry */
	ret = smmu_fill_cfg_invalid_cmd_descriptor(priv, CMD_CD_INVALID, pid, 0);
	if (ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Invalidate CD entry of pid(%d) failed!\n", pid);
		return ret;
	}

	list_del(&(pt_info->node));
	mutex_destroy(&pt_info->entry_lock);
	kfree(pt_info);
	priv->vastai_smmu_info->pid_cnt -= 1;

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "Killed pid(%d) success!\n", pid);

	return ret;
}

int va_kill_pid(struct vastai_pci_info *priv, int va_pid)
{

	mutex_lock(&priv->vastai_smmu_info->pid_lock);

	if ((priv->vastai_smmu_info->pid_cnt != 0) &&
			priv->vastai_smmu_info->vastai_pid[va_pid])
	{
		priv->vastai_smmu_info->vastai_pid[va_pid] = 0;
		__va_kill_pid(priv, va_pid);
		mutex_unlock(&priv->vastai_smmu_info->pid_lock);
		return 0;
	}
	mutex_unlock(&priv->vastai_smmu_info->pid_lock);

	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s(),can't delete pid%d\n", __func__, va_pid);
	return -1;
}
int va_kill_pid_port(void *priv, int pid)
{
	return va_kill_pid(priv, pid);
}
EXPORT_SYMBOL(va_kill_pid_port);

int device_smmu_ste_table_init(struct vastai_pci_info *priv)
{
	int ret;
	struct pcie_transfer_cmd trans;

	if (!priv->vastai_smmu_info)
		return -EINVAL;

	trans.w0.s_data0.optcode = SMCU_SMMU_STE_ENTRY;
	trans.w0.s_data0.rev0 = 0;
	trans.w0.s_data0.rev1 = 0;
	trans.w0.s_data0.rev2 = 0;
	trans.w1.data1 = 0;
	trans.w2.data2 = ((int)((priv->vastai_smmu_info->cd_base) & 0xFFFFFFFF)) >> 6;
	trans.w3.data3 = (((int)((priv->vastai_smmu_info->cd_base) >> 32)) & 0xFFFFF);

	ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF,  &trans, 0);

	if(ret!=0) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"send smmu cmd error\n");
	}

	return 0;
}

void va_get_smmu_err(struct vastai_pci_info *priv, struct pcie_transfer_cmd *event_err)
{
	u32 va_low,va_high, ret;
	u64 va;
	struct va_smmu_err_info *err_info = priv->vastai_smmu_info->err_info;
	struct vastai_pid_pt_entry_info *pt_info;
	struct va_translation_info *tl_info;

	VASTAI_PCI_ERR(priv, 0,"get smmu err\n");
	err_info->err_code = event_err->w0.s_data0.rev0;
	err_info->vmid     = event_err->w1.data1;
	err_info->pid      = event_err->w0.data0 >> 16;
	va_low   = event_err->w2.data2;
	va_high  = event_err->w3.data3;
	va = (u64)(((u64)va_high << 32) | va_low);
	err_info->va = va;

	VASTAI_PCI_ERR(priv, 0, "err_code[0x%x], sid[0x%x],ssid[0x%x],va[0x%llx]\n", 
				err_info->err_code, err_info->vmid, err_info->pid, va);

	ret = get_pid_info(priv, err_info->pid, &pt_info);
	if(VASTAI_MATCH != ret)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "Can't find PID(%d) info\n", err_info->pid);
		return;
	}

	if(check_page_is_mapped(priv, va & VA_PTE_MASK, pt_info, &tl_info))
		get_page_entry_info(priv, va, err_info->pid);

	VASTAI_PCI_ERR(priv, 0, "######The current page information is as follows\n");
//	walk_translation_info(priv, err_info->pid);
	walk_page_from_dev(priv, err_info->pid);
}

static int va_get_smmu_info(struct vastai_pci_info *priv, struct smmu_info *vastai_smmu_info)
{
	vastai_smmu_info->ddr_base = ((u64)((u64)priv->priv_hw_cfg->header.hi32_ddr_base << 32) | priv->priv_hw_cfg->header.lo32_ddr_base);

	if(priv->priv_hw_cfg->sys_cfg.vf_active)
		vastai_smmu_info->cd_base = ((u64)((u64)priv->priv_hw_cfg->header.hi32_ddr_base << 32) | priv->priv_hw_cfg->header.lo32_ddr_base) + CD_VF_OFFSET;
	else
		vastai_smmu_info->cd_base = ((u64)((u64)priv->priv_hw_cfg->header.hi32_ddr_base << 32) | priv->priv_hw_cfg->header.lo32_ddr_base) + CD_PF_OFFSET;

	vastai_smmu_info->ste_base = STE_BASE_ADDR;
	vastai_smmu_info->pid_cnt  = 0;

	VASTAI_PCI_INFO(priv, 0, "cd_base:0x%llx, pte_base:0x%llx\n", vastai_smmu_info->cd_base, vastai_smmu_info->ste_base);

	return 0;
}

static void smmu_cmd_sync_done_callback(void *arg)
{
	// nothing to do
	return;
}

void smmu_handle_for_hotr(struct vastai_pci_info *priv)
{
	int cur_pid;

	for (cur_pid = 0; cur_pid < MAX_VASTAI_PID; cur_pid++)
	{
		if(cur_pid == 0 && PID_MAGIC_NUM == priv->vastai_smmu_info->vastai_pid[cur_pid])
		{
			VASTAI_PCI_INFO(priv, 0, "kill gart va_pid[%d]\n", cur_pid);
			va_kill_pid(priv, cur_pid);
			priv->vastai_smmu_info->vastai_pid[cur_pid] = 0;
		}

		if(priv->vastai_smmu_info->vastai_pid[cur_pid])
		{
			VASTAI_PCI_INFO(priv, 0, "kill va_pid[%d]->usr_pid[%d]\n", 
				cur_pid, priv->vastai_smmu_info->vastai_pid[cur_pid]);
			va_kill_pid(priv, cur_pid);
			priv->vastai_smmu_info->vastai_pid[cur_pid] = 0;
		}
	}

	if (priv->vastai_smmu_info->pid_cnt != 0)
	{
		VASTAI_PCI_INFO(priv, 0, "kill pid failed\n");
	}

	return;
}

void smmu_handle_for_flr(struct vastai_pci_info *priv)
{
	return;
}


#if 0
static void smmu_sync_work_handle(struct kthread_work *work)
{
	struct vastai_pci_info *priv =
		container_of(work, struct vastai_pci_info, smmu_sync_kwork);

	smmu_fill_sync_cmd_descriptor(priv);
}

static void smmu_sync_kthread_queue(struct vastai_pci_info *priv)
{
	//struct sched_param param = { .sched_priority = MAX_RT_PRIO - 10 };

	kthread_init_worker(&priv->smmu_sync_kworker);
	priv->smmu_sync_kworker_task = kthread_run(kthread_worker_fn,
					   &priv->smmu_sync_kworker, "smmu sync kthread worker");
	if (IS_ERR(priv->smmu_sync_kworker_task)) {
		printk("failed to create smmu sync task\n");
		return;
	}
	kthread_init_work(&priv->smmu_sync_kwork, smmu_sync_work_handle);

	//sched_setscheduler(priv->smmu_sync_kworker_task, SCHED_FIFO, &param);
}
#endif

int get_gart_page_size(struct vastai_pci_info *priv, u64 gart_size)
{
	u32 pte_num_total_need = 0;
	u32 pmd_num_total_need = 0;
	u32 pud_num_total_need = 1;
	u32 pgd_num_total_need = 1;
	u32 gart_page_size_total_need = 0;

	if (gart_size > GART_SIZE_MAX || gart_size < GART_SIZE_MIN)
	{
		VASTAI_PCI_ERR(priv, 0, "gart size is invalid(0x%llx)\n", gart_size);
		return 0;
	}

	pte_num_total_need = gart_size >> VA_PTE_SHIFT;
	pmd_num_total_need = DIV_ROUND_UP(pte_num_total_need, PAGE_MAX_INDEX);

	gart_page_size_total_need = (pte_num_total_need + pmd_num_total_need + pud_num_total_need + pgd_num_total_need) * PAGE_ENTRY_SIZE;

	gart_page_size_total_need = round_up(gart_page_size_total_need, DDK_BASE_ALIGN);

	return gart_page_size_total_need;
}

int vastai_smmu_init(struct vastai_pci_info *priv)
{
	int ret;
	struct smmu_info *va_smmu_info;
	struct va_smmu_err_info *va_err_info;

	if (!priv)
		return -EINVAL;

	va_smmu_info = kzalloc(sizeof(*va_smmu_info), GFP_KERNEL);
	if (!va_smmu_info) {
		VASTAI_PCI_ERR(priv, 0,
				"smmu_info kmalloc fail\n");
		return -ENOMEM;
	}

	va_err_info = kzalloc(sizeof(*va_err_info), GFP_KERNEL);
	if (!va_err_info) {
		VASTAI_PCI_ERR(priv, 0,
				"smmu_err_info kmalloc fail\n");
		return -ENOMEM;
	}

	ret = va_get_smmu_info(priv, va_smmu_info);
	if (ret) {
		VASTAI_PCI_ERR(priv, 0,
				"smmu_init get smmu information fail!, %d\n",
				ret);
		ret = -EPERM;
		goto err_free;
	}
	init_completion(&va_smmu_info->cmd_sync_done);

	va_smmu_info->sync_cnt = 0;
	INIT_LIST_HEAD(&(va_smmu_info->pid_head));
	va_smmu_info->err_info = va_err_info;
	mutex_init(&va_smmu_info->pid_lock);

	/* get gart info */
	if (priv->gart_en && CONFIG_ATUO_GART_EN)
	{
		if ((PF_4_WITH_SRIOV == priv->priv_hw_cfg->sys_cfg.fn_mode && !priv->priv_hw_cfg->sys_cfg.vf_active))
		{
			VASTAI_PCI_INFO(priv, 0, "PF don't open gart in 4PF_SRIOV mode\n");
			goto GART_DISABLE_IN_PF;
		}

		va_smmu_info->va_gart_info.gart_size = (u64)priv->priv_hw_cfg->sys_cfg.gart_size << 20;
		va_smmu_info->va_gart_info.gart_entry_total_size = get_gart_page_size(priv, va_smmu_info->va_gart_info.gart_size);

		/* page addr shoule be real addr */
		va_smmu_info->va_gart_info.gart_entry_base = priv->ddr_info.ddr_entry[GPU_HEAP].noc_addr;

		/* reserved 1M for pdg/pud/pmd entry, pte base = va_smmu_info->va_gart_info.gart_entry_base  */
		va_smmu_info->va_gart_info.gart_pgd_base = va_smmu_info->va_gart_info.gart_entry_base + va_smmu_info->va_gart_info.gart_entry_total_size - (PAGE_MAX_INDEX * PAGE_ENTRY_SIZE);
		va_smmu_info->va_gart_info.gart_pud_base = va_smmu_info->va_gart_info.gart_pgd_base - (PAGE_MAX_INDEX * PAGE_ENTRY_SIZE);
		va_smmu_info->va_gart_info.gart_pmd_base = va_smmu_info->va_gart_info.gart_entry_base + va_smmu_info->va_gart_info.gart_entry_total_size - (1 << 20);

		VASTAI_PCI_INFO(priv, 0, "Current mode:GART(%dM), page_total_size:0x%x, pte_base:0x%llx, pmd_base:0x%llx, pud_base:0x%llx, pgd_base:0x%llx\n", 
					priv->priv_hw_cfg->sys_cfg.gart_size, va_smmu_info->va_gart_info.gart_entry_total_size, 
					va_smmu_info->va_gart_info.gart_entry_base, va_smmu_info->va_gart_info.gart_pmd_base,
					va_smmu_info->va_gart_info.gart_pud_base, va_smmu_info->va_gart_info.gart_pgd_base);

		va_fixed_mapping_in_gart(priv, va_smmu_info);
		
	}
	else
		VASTAI_PCI_INFO(priv, 0, "Current mode:ATU\n");

GART_DISABLE_IN_PF:
	priv->vastai_smmu_info = va_smmu_info;

	ret = va_register_pcie_interrupt(priv, COMMON_MSIX_SMMU_ERR_VECTOR,
			 smmu_cmd_sync_done_callback,
			 &(priv->vastai_smmu_info->cmd_sync_done));

	return 0;

err_free:
	kfree(va_smmu_info);
	return ret;
}
