/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/dma-mapping.h>
#include <linux/io.h>
#include <linux/types.h>

#include "vastai_pci.h"

static volatile uint32_t *CPS_transform_addr_to_vmem(volatile void *address)
{
	struct vastai_pci_info *priv = get_vastai_pci_device_info(0);
	uint64_t addr = (uint64_t)address;

	if ((addr < priv->bar[VASTAI_PCI_BAR0].noc_addr) ||
	    (addr >= priv->bar[VASTAI_PCI_BAR0].noc_addr +
			     priv->bar[VASTAI_PCI_BAR0].mmio_len)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "%s addr 0x%llx is err!\n", __func__, addr);
		return NULL;
	} else
		return (volatile uint32_t
				*)(priv->bar[VASTAI_PCI_BAR0].vmem + addr -
				   priv->bar[VASTAI_PCI_BAR0].noc_addr);
}

uint32_t CPS_ReadReg32(volatile uint32_t *address)
{
	return readl(CPS_transform_addr_to_vmem(address));
}

void CPS_WriteReg32(volatile uint32_t *address, uint32_t value)
{
	writel(value, CPS_transform_addr_to_vmem(address));
}

uint8_t CPS_UncachedRead8(volatile uint8_t *address)
{
	return readb(CPS_transform_addr_to_vmem(address));
}

uint16_t CPS_UncachedRead16(volatile uint16_t *address)
{
	return readw(CPS_transform_addr_to_vmem(address));
}

uint32_t CPS_UncachedRead32(volatile uint32_t *address)
{
	return readl(CPS_transform_addr_to_vmem(address));
}

void CPS_UncachedWrite8(volatile uint8_t *address, uint8_t value)
{
	writeb(value, CPS_transform_addr_to_vmem(address));
}

void CPS_UncachedWrite16(volatile uint16_t *address, uint16_t value)
{
	writew(value, CPS_transform_addr_to_vmem(address));
}

void CPS_UncachedWrite32(volatile uint32_t *address, uint32_t value)
{
	writel(value, CPS_transform_addr_to_vmem(address));
}

void CPS_BufferCopy(volatile uint8_t *dst, volatile const uint8_t *src,
		    uint32_t size)
{
	memcpy((void *)dst, (void *)src, size);
}

/**
 * Invalidate cache using Linux DMA sync routines
 * @param[in] address Virtual address of buffer to sync. Buffer must
 * be allocated using dma_map_single() function
 * @param[in] size Size of buffer
 * @param[in] device Pointer to struct device* which owns buffer
 */
void CPS_CacheInvalidate(void *address, size_t size, uintptr_t device)
{
	uintptr_t phys_address = __pa(address);
	struct device *dev = (struct device *)device;

	dma_sync_single_for_cpu(dev, phys_address, size, DMA_BIDIRECTIONAL);
}

/**
 * Flush cache using Linux DMA sync routines
 * @param[in] address Virtual address of buffer to sync. Buffer must
 * be allocated using dma_map_single() function
 * @param[in] size Size of buffer
 * @param[in] device Pointer to struct device* which owns buffer
 */
void CPS_CacheFlush(void *address, size_t size, uintptr_t device)
{
	uintptr_t phys_address = __pa(address);
	struct device *dev = (struct device *)device;

	dma_sync_single_for_device(dev, phys_address, size, DMA_BIDIRECTIONAL);
}
