/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/aer.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/version.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/timer.h>

#include "vastai_state.h"
#include "vastai_dmabuf.h"
#include "vastai_pci.h"
#include "vastai_pci_api.h"
#ifdef CONFIG_VASTAI_PCI_BOOT
#include "vastai_pci_boot.h"
#endif
#include "vastai_pci_test.h"
#include "vastai_udma_engine.h"
#include "vastai_version.h"
#include "sg100_smcu.h"
#include "vastai_ras.h"
#include "vastai_dmi_table.h"
#include "vastai_pcie_public.h"

#ifdef CONFIG_VASTAI_AI
#include "vastai_ai.h"
#endif

#ifdef CONFIG_VASTAI_VIDEO
#include "vastai_render.h"
#endif

#ifdef CONFIG_VASTAI_TOOLS_SUITE
#include "vatools_driver.h"
#endif

#include "hw/va_hw_init.h"
#include "hw/pcie_2_reg.h"
#include "sg100_smu.h"
#include "sg100_base.h"
#include "vastai_cmd.h"
#include "hw/smmu.h"
#include "hw/va_dma_core.h"
#include "../common/hw_config.h"
#include "vastai_ai_info.h"
#include "logsys.h"
#include "vastai_fifo.h"
#include "sg100_cmd.h"
#include "sg100_pcie_resource.h"
#include "sg100_pcie_wrap.h"
#include "vastai_ai_info.h"
#include "config.h"
#ifdef CONFIG_VASTAI_GFX
#include "../gfx/services/server/env/linux/pvr_drv.h"
#endif
#include "vastai_pcie_aer.h"

#include "va_inventory.h"

/* PCI_VENDOR_ID_VASTAI */
#define PF_DEVICE_ID					0x0200
#define PF_DEVICE_L_ID					0x0201
#define PF_DEVICE_M_ID					0x0202
#define VF_DEVICE_ID					0x0203 // need be removed if BL0 is updated by zouyang in Jun 2023
#define VF_DEVICE_ID2					0x020F

#define GFX_PF_DEVICE_ID				0x0204
#define AUD_PF_DEVICE_ID				0x02FF //0x0205  // disable Audio probe

#define VASTAI_PCIE_2_SMCU				0
#define VASTAI_PCIE_2_VDSP				1
#define VASTAI_PCIE_2_CMCU				1

static bool msi_enabled = 1;
module_param(msi_enabled, bool, 0);

unsigned char va_pci_log_level = VASTAI_PCI_NORMAL_LEVEL;
// unsigned char va_pci_log_level = VASTAI_PCI_NORMAL_LEVEL | VASTAI_PCI_DEBUG_LEVEL | VASTAI_PCI_DATA_LEVEL | VASTAI_PCI_PERF_LEVEL;
module_param(va_pci_log_level, byte, 0);

unsigned char gart_en = GART_NOPUT;
module_param(gart_en, byte, 0);

static bool va_bar4_wc_en = 0;
module_param(va_bar4_wc_en, bool, 0);

/*
0:not support,gpu access host ddr size 512GB 
1:support,gpu access host ddr size 512GB/PF num
*/
bool outbound_sriov_en = 0;
module_param(outbound_sriov_en, bool, 0);

struct list_head vastai_pci_info_list;
struct list_head vastai_pci_card_list;
struct mutex vastai_pci_modify_lock;
struct mutex vastai_smcu_modify_lock;

struct list_head vastai_card_info_list;
struct mutex vastai_card_info_list_lock;


#ifdef CONFIG_PROBE_RESET
u32 g_pf_reinit_done = 0;
#endif

/* If a pci_dev with the card_id exists in the OS, the pci_dev is not added */
static int vastai_add_card_id(struct vastai_pci_info *priv)
{
	struct list_head *ptr;
	struct vastai_pci_info *pci_info;

	if (list_empty(&vastai_pci_card_list)) {
		list_add_tail(&(priv->card_list), &vastai_pci_card_list);
		return 0;
	}

	list_for_each(ptr, &vastai_pci_card_list) {
		pci_info = list_entry(ptr, struct vastai_pci_info, card_list);
		if (pci_info->card_id == priv->card_id) {
			return -EINVAL;
		}
	}

	list_add_tail(&(priv->card_list), &vastai_pci_card_list);

	return 0;
}

static void vastai_del_card_id(struct vastai_pci_info *priv)
{
	struct list_head *ptr;

	if (!priv->is_pf)
		return;

	if (list_empty(&vastai_pci_card_list))
		return;

	list_for_each(ptr, &vastai_pci_card_list) {
		if (ptr == &(priv->card_list)) {
			list_del(&(priv->card_list));
			return;
		}
	}

	return;
}

int get_vastai_die_info(struct vastai_pci_info *priv, int die_index[], int *die_num)
{
	int i = 0;

	if (!die_index || !priv) {
		return -EINVAL;
	}
	for (i = 0; i < priv->die_num; i++) {
		die_index[i] = priv->dies[i].die_index;
	}
	*die_num = priv->die_num;
	return 0;
}

u64 vastai_get_host_time_ns(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,6,0)
	struct timespec time;
#else
	u64 time;
#endif
	u64 time_ns;


#if LINUX_VERSION_CODE < KERNEL_VERSION(5,6,0)
	/* getnstimeofday function is removed from kernel5.6 */
	getnstimeofday(&time);
#else
	time = ktime_get_real_ns();
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(5,6,0)
	time_ns = timespec_to_ns(&time);
#else
	time_ns = time;
#endif

	return time_ns;
}

struct vastai_pci_info *get_vastai_pci_device_info(char dev_id)
{
	struct list_head *ptr;
	struct vastai_pci_info *info;
	if (list_empty(&vastai_pci_info_list)) {
		return NULL;
	}

	/* for now, we only get first entry */
	list_for_each(ptr, &vastai_pci_info_list) {
		info = list_entry(ptr, struct vastai_pci_info, dev_list);
		if (info->dev_id == dev_id) {
			return info;
		}
	}
	return NULL;
}

#ifdef CONFIG_VASTAI_TOOLS_SUITE
int get_vastai_pci_device_number(void)
{
	struct list_head *ptr;
	int device_num = 0;

	if (list_empty(&vastai_pci_info_list)) {
		return 0;
	}

	list_for_each(ptr, &vastai_pci_info_list) { device_num++; }
	return device_num;
}
#endif

int vastai_pci_get_die_id(struct vastai_pci_info *pcie_dev, u32 die_index)
{
	return ((union die_index_data)die_index).die_id;
}

int vastai_pci_generation_sel(struct vastai_pci_info *priv, u8 gen)
{
	u64 reg_addr;
	u32 reg_val;
	u8 lnksta, lnkcap, lnkcnt = 0;
	unsigned int die_index = priv->dies[0].die_index;
	int ret;

	if ((gen < VASTAI_PCI_GEN1) || (gen > VASTAI_PCI_GEN4)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_generation_sel cann't switch to GEN%d\n", gen);
		return -EINVAL;
	}

	/* try to change generation up to 10 times */
	while (lnkcnt < 10) {
		/* get negotiated link speed */
		vastai_pci_config_read(priv,
				       PCIE_LINK_CONTROL_STATUS_REG_OFFSET,
				       (char *)&reg_val, 4);
		lnksta = (reg_val & 0x000F0000) >> 16;
		if (gen == lnksta) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"pci_generation_sel lnksta=GEN%d, return.\n",
					gen);
			return 0;
		}

		/* configurate pcie_generation_sel(bit18:17) to GEN4 */
		reg_addr = VASTAI_PCI_CONFIGURATION_IO_0;
		ret = vastai_pci_mem_read(priv, die_index, reg_addr, &reg_val, 4);
		if (ret < 0) {
			return -EIO;
		}
		lnkcap = ((reg_val & 0x00060000) >> 17) + 1;
		if (gen > lnkcap) {
			reg_val = (reg_val & 0xFFF9FFFF) | (3 << 17);
			vastai_pci_mem_write(priv, 0, reg_addr, &reg_val, 4);
			VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				"configurate pcie_generation_sel to GEN4\n");
		}

		/* configurate target link speed (LnkCtl2) */
		vastai_pci_config_read(priv,
				       PCIE_LINK_CONTROL_STATUS_REG2_OFFSET,
				       (char *)&reg_val, 4);
		reg_val = (reg_val & 0xFFFFFFF0) | gen;
		vastai_pci_config_write(priv,
					PCIE_LINK_CONTROL_STATUS_REG2_OFFSET,
					(char *)&reg_val, 4);

		/* trigger link retraining */
		reg_addr = VASTAI_PCI_CDN_LINKWIDTH_CONTROL_REG;
		while (1) {
			ret = vastai_pci_mem_read(priv, die_index, reg_addr, &reg_val, 4);
			if (ret < 0) {
				return -EIO;
			}
			/* sw must wait for retrain link bit to be clear before
			 * setting it again. */
			if ((reg_val & 0x80000000) == 0)
				break;
		}
		reg_val &= 0xFCFFFFFF;
		reg_val |= ((gen - 1) << 24) | (1 << 31);
		vastai_pci_mem_write(priv, die_index, reg_addr, &reg_val, 4);

		lnkcnt++;
	}
	return 0;
}


static int vastai_pci_legacy_irq(int irq, void *arg)
{
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID,
			"legacy irq called %d\n", irq);

	// legacy_irq_handle(irq);

	return IRQ_HANDLED;
}

int va_register_pcie_interrupt(struct vastai_pci_info *priv,
				  u32 interrupt_id,
				  void (*handler_function)(void *),
				  void *arg)
{
	//unsigned long flags;

	if (!handler_function) {
		VASTAI_PCI_ERR(priv, 0,
				"%s: handler_function is null\n", __func__);
		return -EINVAL;
	}

	if (interrupt_id >= VASTAI_PCIE_MSIX_NUM) {
		VASTAI_PCI_ERR(priv, 0,
				"%s: interrupt_id is invalid, %d\n",
				__func__, interrupt_id);
		return -EINVAL;
	}

	// TODO
	//spin_lock_irqsave(, flags);

	if (priv->irq_handles[interrupt_id].handler_function) {
		VASTAI_PCI_ERR(priv, 0,
			"%s: handler_function has beed registered, %p\n",
			__func__,
			priv->irq_handles[interrupt_id].handler_function);
		return -EPERM;
	}
	priv->irq_handles[interrupt_id].arg = arg;
	priv->irq_handles[interrupt_id].handler_function = handler_function;
	// TODO
	//spin_unlock_irqrestore(, flags);

	return 0;
}

int va_unregister_pcie_interrupt(struct vastai_pci_info *priv,
				 u32 interrupt_id)
{
	if (interrupt_id >= VASTAI_PCIE_MSIX_NUM) {
		VASTAI_PCI_ERR(priv, 0,
				"%s: interrupt_id is invalid, %d\n",
				__func__, interrupt_id);
		return -EINVAL;
	}

	priv->irq_handles[interrupt_id].handler_function = NULL;
	priv->irq_handles[interrupt_id].arg = NULL;

	return 0;
}

static int vastai_pci_msi_irq(int irq, void *arg)
{
	struct vastai_pci_info *priv = arg;
	struct vastai_interrupt_handle *irq_handle;
    
	irq -= priv->irq;
	if(0 != irq){
		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,"msi irq called %d \n",irq);
	}
	
	irq_handle = &priv->irq_handles[irq];
	if (irq_handle->handler_function)
		irq_handle->handler_function(irq_handle->arg);

	return IRQ_HANDLED;
}

static int vastai_pci_disable_irq(struct vastai_pci_info *priv)
{
	int i;

	if (priv->msix_en == 1) {
		for (i = 0; i < priv->irq_num; i++) {
			va_unregister_pcie_interrupt(priv,i);
			disable_irq(priv->msix[i].vector);
			free_irq(priv->msix[i].vector, (void *)priv);
			VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				       "pci_disable_irq disable_irq(%d) ok\n",
				       priv->msix[i].vector);
		}
		pci_disable_msix(priv->dev);
	}

	if (priv->msi_en == 1) {
		for (i = 0; i < priv->irq_num; i++) {
			va_unregister_pcie_interrupt(priv,i);
			if (priv->irq) {
				disable_irq(priv->irq + i);
				if (!free_irq(priv->irq + i, (void *)priv))
					return -1;
			}
		}

		pci_disable_msi(priv->dev);
	}

	return 0;
}

int vastai_pci_msgq_wr(struct vastai_pci_info *priv,
		       unsigned int die_index,
		       struct msgq *msgq,
		       void *data,
		       u32 data_size)
{
	struct vastai_msgq msgq_temp;
	int ret;
	u64 msg_addr = msgq->addr;

	mutex_lock(&(msgq->msgq_lock));
	ret = vastai_pci_mem_read(priv, die_index, msg_addr, &msgq_temp, sizeof(msgq_temp));
	if (ret) {
		VASTAI_PCI_ERR(priv, vastai_pci_get_die_id(priv, die_index),
			       "pci_msgq_wr read err, ret=%d\n", ret);
		goto OUT;
	}

	if (msgq_temp.wr < msgq_temp.start) {
		msgq_temp.wr = msgq_temp.start;
	} else {
		msgq_temp.wr = msgq_temp.wr + data_size;
	}

	if (msgq_temp.wr + data_size >= msgq_temp.end)
		msgq_temp.wr = msgq_temp.start;
	
	if (msgq_temp.wr == msgq_temp.rd) {
		VASTAI_PCI_ERR(priv, vastai_pci_get_die_id(priv, die_index),
			       "pci_msgq_wr wr = rd = 0x%x\n", msgq_temp.wr);
		ret = -ENOMEM;
		goto OUT;
	}

	vastai_pci_mem_write(priv, die_index, (u64)msgq_temp.wr, data, data_size);
	VASTAI_PCI_DBG(priv, vastai_pci_get_die_id(priv, die_index),
		"pci_msgq_wr wr:0x%x, rd:0x%x\n", msgq_temp.wr, msgq_temp.rd);
	mb();
	ret = vastai_pci_mem_write(priv,
				   die_index,
				   msg_addr + offsetof(struct vastai_msgq, wr),
				   &(msgq_temp.wr),
				   4);
OUT:
	mutex_unlock(&(msgq->msgq_lock));
	return ret;
}

int vastai_irq_register(struct vastai_pci_info *pcie_dev,
			int die_index,
			int irq, /* reference vastai_sv100_reg.h +272 */
			vastai_intr_handle handle,
			void *private_data) /* for user keep self data */
{
	int die_id = vastai_pci_get_die_id(pcie_dev, die_index);

	if (die_id < 0) {
		return -ENODEV;
	}
	pcie_dev->dies[die_id].ivt[irq].intr_handle = handle;
	pcie_dev->dies[die_id].ivt[irq].private_data = private_data;
	return 0;
}

int vastai_irq_unregister(struct vastai_pci_info *pcie_dev,
			  int die_index,
			  int irq)/* reference vastai_sv100_reg.h +272 */
{
	int die_id = vastai_pci_get_die_id(pcie_dev, die_index);

	if (die_id < 0) {
		return -ENODEV;
	}
	pcie_dev->dies[die_id].ivt[irq].intr_handle = NULL;
	pcie_dev->dies[die_id].ivt[irq].private_data = NULL;
	return 0;

}





/* hot reset */
int vastai_pci_reset_bus(struct vastai_pci_info *priv)
{
	int ret;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "start to hot-reset pcie bus.\n");

	vastai_pci_sync_state(priv, VASTAI_RESET_STATE);
	vastai_pci_state_reset_start(priv);
	/* wait for traffic idle */
	msleep(1000);
	
	ret = pci_probe_reset_bus(priv->dev->bus);
	if (ret < 0) {
		vastai_pci_state_sync_error(priv);
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_probe_reset_bus fail, ret=%d", ret);
		return ret;
	}
	
#if defined(RHEL_RELEASE_VERSION)
#if RHEL_RELEASE_CODE < RHEL_RELEASE_VERSION(8, 0)
	ret = pci_reset_bus(priv->dev->bus);
#else
	ret = pci_reset_bus(priv->dev);
#endif
#elif KERNEL_VERSION(4, 19, 0) <= LINUX_VERSION_CODE
	/* 2018-07-19 18:04:15 -0500
	 * PCI: Rename pci_try_reset_bus() to pci_reset_bus()
	 */
	ret = pci_reset_bus(priv->dev);
#else
	ret = pci_reset_bus(priv->dev->bus);
#endif
	if (ret < 0) {
		vastai_pci_state_sync_error(priv);
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_reset_bus fail, ret=%d", ret);
		return ret;
	}

	vastai_pci_state_boot_done(priv);
	vastai_pci_sync_state(priv, VASTAI_NORMAL_STATE);

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "PCIe hot reset pcie bus finish.\n");

	return ret;
}

#if 0
// FUNC: Do re_initial reset in smcu FW
// ARGS: @is_conditon
// 1: condition reset
// 2: unconditional reset
static int vastai_mcu_do_reinitial(struct vastai_pci_info *priv, bool is_conditon)
{
	int ret = 0;
	struct pcie_transfer_cmd trans;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "mcu_do_reinitial\n");

	trans.w0.s_data0.optcode	= SMCU_CMD_REINIT_RST_PKG;
	trans.w1.data1				= is_conditon;

	ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF,  &trans, 0);
	if (ret) {
		VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID, "mcu_do_reinitial: pci_send_msg failed\n");
		return -EAGAIN;
	}

	/* wait completion int from smcu */
	if (wait_for_completion_timeout(&(priv->send_reinit_rst_done), msecs_to_jiffies(VASTAI_PCIE_WAIT_SMCU_MSG1_TIME)) <= 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "mcu_do_reinitial wait for completion timeout, failed\n");
		return -ETIMEDOUT;
	}

	reinit_completion(&(priv->send_reinit_rst_done));

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s\n", "SMCU recvd reinit_rst cmd.\n");

	return ret;
}

// Func: re_initial reset
// 1: condition reset
// 2: unconditional reset
static int vastai_pci_do_reinit_rst(struct vastai_pci_info *priv, bool is_conditon)
{
	int ret = 0;
	u32 val = 0;

	ret = vastai_mcu_do_reinitial(priv, is_conditon);

	if(ret < 0) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "Failed to do reinit in MCU. Try to re_initial SoC from Drv.\n");

		val = is_conditon ? (1 << 0) : (1 << 2);

		ret = vastai_pci_mem_write(priv, 0, PCIE_WRAPPER_REINIT_RST_HOST, &val, 4);
		if (ret < 0) {
			vastai_pci_state_sync_error(priv);
			VASTAI_PCI_ERR(priv, 0, "pci_do_reinit_rst [write] %#x, len=4, error:%d\n", PCIE_WRAPPER_REINIT_RST_HOST, ret);
			return ret;
		}
	}

	return ret;
}
#endif

// Func: re_initial reset
// 1: condition reset
// 2: unconditional reset
static int vastai_pci_mmio_do_reinit_rst(struct vastai_pci_info *priv, bool is_conditon)
{
	int ret = 0;
	u32 val = 0;
	u32 zero = 0;

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "re_initial reset SoC from MMIO by Drv.\n");

	val = is_conditon ? (1 << 0) : (1 << 2);

	vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (VASTAI_HOST_BOOT_STAGE_REG - PCIE_WRAPPER_ADDR), &zero, 4);
	ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (PCIE_WRAPPER_REINIT_RST_HOST - PCIE_WRAPPER_ADDR), &val, 4);
	if (ret < 0) {
		vastai_pci_state_sync_error(priv);
		VASTAI_PCI_ERR(priv, 0, "pci_mmio_do_reinit_rst [write] %#x, len=4, error:%d\n", PCIE_WRAPPER_REINIT_RST_HOST, ret);
		return ret;
	}

	return ret;
}

int vastai_pci_reinitial(struct vastai_pci_info *priv, bool is_conditon)
{
	int ret = 0;

	vastai_pci_sync_state(priv, VASTAI_RESET_STATE);
	vastai_pci_state_reset_start(priv);

	/* wait others func set state */
	msleep(100);

	ret = vastai_pci_mmio_do_reinit_rst(priv, is_conditon);

	vastai_pci_state_reset_done(priv);

	return ret;
}

int vastai_send_pcie_cmd(struct vastai_pci_info *priv, u32 die_index, u32 sub_cmd)
{
	struct pcie_transfer_info trans = {
		.type = VASTAI_PCIE,
		.pcie_config_addr = sub_cmd};
	int ret;
	u32 die_id = vastai_pci_get_die_id(priv, die_index);

	ret = vastai_pci_msgq_wr(priv, die_index, &(priv->dies[die_id].msgq[1]),
		 &trans, sizeof(trans));
	if (ret) {
		VASTAI_PCI_ERR(priv, die_index,
			"send_pcie_cmd sub_cmd[%d] error, ret=%d",
			sub_cmd, ret);
		return -EINVAL;
	}

	return 0;
}

void vastai_die_lock(void *pci_info)
{
	struct vastai_pci_info *priv = (struct vastai_pci_info *)pci_info;
	int ret = 0;
	u32 val = 0;
	u64 reg_addr = PWRAP_COMMON_LOCK1;

	while(1) {
		ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (reg_addr - PCIE_WRAPPER_ADDR), &val, 4);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "die_lock [read] 0x%llx failed\n", reg_addr);
		}

		if(val == PCIE_FAULT_CODE) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "die_lock: ERROR PCIE_FAULT_CODE\n");
		}

		if (val > 0)
			break;

		/* for schedule */
		usleep_range(1, 2);
	}
}
EXPORT_SYMBOL(vastai_die_lock);

void vastai_die_unlock(void *pci_info)
{
	struct vastai_pci_info *priv = (struct vastai_pci_info *)pci_info;
	int ret = 0;
	u64 reg_addr = PWRAP_COMMON_LOCK1;
	u32 reg_val  = PWRAP_COMMON_LOCK1_VAL;

	ret = vastai_pci_mmio_write(priv, VASTAI_PCI_BAR1, (reg_addr - PCIE_WRAPPER_ADDR), &reg_val, 4);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "die_unlock [write] 0x%llx failed\n", reg_addr);
	}
}
EXPORT_SYMBOL(vastai_die_unlock);

static void vastai_cedar_ddr_addr_bypass_mpu(struct vastai_pci_info *priv, struct vastai_cedar_img_info *cedar_info)
{
	u64 pf_ddr_offset;
	
	pf_ddr_offset = (((u64)priv->priv_hw_cfg->header.hi32_ddr_base<<32) | priv->priv_hw_cfg->header.lo32_ddr_base) - 0x1040000000;

	if (cedar_info->input_addr >= 0x1040000000)
		cedar_info->input_addr = cedar_info->input_addr + pf_ddr_offset;

	if (cedar_info->output_addr >= 0x1040000000)
		cedar_info->output_addr = cedar_info->output_addr + pf_ddr_offset;

	if (cedar_info->key >= 0x1040000000)
		cedar_info->key = cedar_info->key + pf_ddr_offset;

	if (cedar_info->iv >= 0x1040000000)
		cedar_info->iv = cedar_info->iv + pf_ddr_offset;

	if (cedar_info->msg >= 0x1040000000)
		cedar_info->msg = cedar_info->msg + pf_ddr_offset;

	if (cedar_info->tag >= 0x1040000000)
		cedar_info->tag = cedar_info->tag + pf_ddr_offset;

	if (cedar_info->random >= 0x1040000000)
		cedar_info->random = cedar_info->random + pf_ddr_offset;

	if (cedar_info->pubkey >= 0x1040000000)
		cedar_info->pubkey = cedar_info->pubkey + pf_ddr_offset;

	if (cedar_info->prikey >= 0x1040000000)
		cedar_info->prikey = cedar_info->prikey + pf_ddr_offset;

	if (cedar_info->sign >= 0x1040000000)
		cedar_info->sign = cedar_info->sign + pf_ddr_offset;

	return;
}

int vastai_dev_cedar_send_cmd(void *pci_info, void *cedar_info, int dev_id, int die_id, void *hash_output)
{
	int output_hash_len, type;
	int ret;
	u32 cedar_info_addr;
	u32 hash_res_addr;
	u64 zero_hash[8] = {0};
	struct vastai_pci_info *priv = (struct vastai_pci_info *)pci_info;
	struct vastai_cedar_img_info *op = (struct vastai_cedar_img_info *)cedar_info;
	struct smcu_task_info smcu_task_desc = {0};
	struct smcu_rw_ret_info smcu_ret = {0};
	struct ring_buf_entry *ch_entry = NULL;

	if(die_id != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "dev_cedar_send_cmd invalid die id %d\n", die_id);
		return die_id;
	}

	type = op->alg_type;
	if (type <= 3)
		output_hash_len = 28;
	else if (type <= 7)
		output_hash_len = 32;
	else if (type <= 11)
		output_hash_len = 48;
	else if (type <= 15)
		output_hash_len = 64;
	else if (type <= 19)
		output_hash_len = 32;
	else if (type <= 23)
		output_hash_len = 28;

	ch_entry = find_hw_cfg_acord_name(priv, COMMON_RESERVED_CSRAM_NAME);
	if (ch_entry == NULL)
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "VA get ring buf error!\n");
		return -1;
	}
	hash_res_addr  = ch_entry->ring_buf_base + 0x100;
	cedar_info_addr = ch_entry->ring_buf_base + 0x200;
	if (0 == op->output_addr)
		op->output_addr = hash_res_addr;

	vastai_cedar_ddr_addr_bypass_mpu(priv, op);

	mutex_lock(&priv->cedar_lock);
	
	vastai_pci_mem_write(priv, dev_id, op->output_addr, (void*)zero_hash, 4);
	vastai_pci_mem_write(priv, dev_id, cedar_info_addr, (void*)op, sizeof(vastai_cedar_img_info_t));

	smcu_task_desc.w0_data0 = SMCU_CEDAR_CMD_REQ;
	smcu_task_desc.w1_data1 = cedar_info_addr; /* addr */
	smcu_task_desc.w2_data2 = sizeof(vastai_cedar_img_info_t); /*length*/
	ret = vastai_submit_smcu_task_and_done(priv, &smcu_task_desc, &smcu_ret);
	if (ret != 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "dev_cedar_send_cmd write cedar cmd err=%d\n", ret);
		mutex_unlock(&priv->cedar_lock);
		return ret;
	}

	if (0xff != smcu_ret.w1_data1) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "dev_cedar_send_cmd cedar handle err, smcu_ret=0x%x\n", smcu_ret.w1_data1);
		mutex_unlock(&priv->cedar_lock);
		return -1;
	}

	if(type <= 23) {
		vastai_pci_mem_read(priv, dev_id, op->output_addr, (void*)hash_output, output_hash_len);
	}
	mutex_unlock(&priv->cedar_lock);

	return 0;
}
EXPORT_SYMBOL(vastai_dev_cedar_send_cmd);

bool vastai_dev_not_ready(struct vastai_pci_info *priv)
{
	if (atomic_read(&priv->is_reseting) || priv->disable)
		return true;

	return false;
}

int vastai_device_reset(struct vastai_pci_info *priv, u32 flags)
{
	int ret = 0;

	ret = atomic_cmpxchg(&priv->is_reseting, 0, 1);
	if (ret) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"device_reset: device have been reseting\n");
		return 0;
	}

//	if (flags & VASTAI_RESET_MSG_TRIGER) {
//		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
//				"device_reset: triger device reset by msg\n");
//		ret = vastai_send_pcie_cmd(priv, priv->dies[0].die_index, VASTAI_PCIE_SUB_RESET);
//		if (ret) {
//			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
//				"device_reset: send reset msg error, %d\n",
//				ret);
//			goto failed;
//		}
//		msleep(1);
//	}

	/* clear all completion and workqueue */
	/* TODO: kill all task */
	/* re-initial data */

	if (flags & VASTAI_RESET_BOOT_DEV) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"device_reset: reboot device\n");
#ifdef CONFIG_VASTAI_PCI_BOOT
		ret = vastai_pci_reboot(priv);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"device_reset: reboot device error, %d\n",
				ret);
			goto failed;
		}
#endif
	}

	atomic_set(&priv->is_reseting, 0);
	return 0;

failed:
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"device_reset: disable device\n");
	priv->disable = true;
	atomic_set(&priv->is_reseting, 0);

	return ret;
}



// func: dump the map of all kchar:xx ~ BDF on current PC
void dump_pf_vf_map_kchar(struct vastai_pci_info *priv)
{
	struct vastai_pci_info *vptr = NULL;
    int i = 0;

//    VASTAI_PCI_DBG(NULL, DUMMY_DIE_ID, "%s: va_pci_info_list=0x%p", __func__, &vastai_pci_info_list);

	if (list_empty(&(vastai_pci_info_list))) {
        VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "pci_info_list:0x%p is an empty list now", &vastai_pci_info_list);
	} else {
        list_for_each_entry(vptr, &(vastai_pci_info_list), dev_list) {
            if(vptr) {
                if(vptr->dev == NULL || vptr->dev->bus == NULL) {
                    VASTAI_PCI_ERR(NULL, DUMMY_DIE_ID, "vptr->dev=%p, NULL ptr found.\n", vptr->dev);
                    continue;
                }

//                VASTAI_PCI_DBG(NULL, DUMMY_DIE_ID, "va[%d]: vptr:0x%p, vptr->dev_list:0x%p, dev_id:%d, bus_number:%d\n", i, vptr, &vptr->dev_list, vptr->dev_id, vptr->dev->bus->number);
                VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "va[%d]: devfn:%d ~ kchar:%d ~ %s\n", i, vptr->dev->devfn, vptr->dev_id, pci_name(vptr->dev));
            }
            else
                VASTAI_PCI_DBG(NULL, DUMMY_DIE_ID, "LinkList: elem vptr:0x%p, NULL\n", vptr);
        }
    }

}


void dump_bar_addr_map(struct vastai_pci_info *priv)
{
	struct pci_dev *pdev = priv->dev;
	int i = 0, r = 0;
	u64 region_offset = 0;
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "S<==================memmap fn:%d===================>S\n", pdev->devfn);

	for (i = 0; i < VASTAI_PCI_BAR_NUM; i++) {
		int flag = pci_resource_flags(pdev, i);
		if (!(flag & IORESOURCE_MEM))
			continue;

        if(i == VASTAI_PCI_BAR0){
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)  pa:0x%lx~0x%lx, va:0x%lx\n", i,
               					(unsigned long)priv->bar[i].mmio_start,
               					(unsigned long)priv->bar[i].mmio_end,
                			    (unsigned long)priv->bar[i].vmem);
        
        	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d) noc:bar0 is not mapping, len:0x%lx(%ldK), \n", i,							
           					    (unsigned long)priv->bar[i].mmio_len,
         					    (unsigned long)priv->bar[i].mmio_len / 1024);
		}

		 if(i == VASTAI_PCI_BAR1){
            VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)  pa:0x%lx~0x%lx, va:0x%lx\n", i,
               					(unsigned long)priv->bar[i].mmio_start,
               					(unsigned long)priv->bar[i].mmio_end,
                			    (unsigned long)priv->bar[i].vmem);

			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)  fa:0x%llx~0x%llx \n", i,
								priv->bar[i].fn_base_addr,
								priv->bar[i].fn_base_addr + priv->bar[i].mmio_len - 1);
								
        	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d) noc:0x%llx~0x%llx len:0x%lx(%ldK), \n", i,
								priv->bar[i].noc_addr,
								priv->bar[i].noc_addr + priv->bar[i].mmio_len - 1,
           					    (unsigned long)priv->bar[i].mmio_len,
         					    (unsigned long)priv->bar[i].mmio_len /1024);
        }

        if(i == VASTAI_PCI_BAR2){
            for(r = 0; r < priv->bar[VASTAI_PCI_BAR2].region_cnt; r++) {
				region_offset = priv->bar[i].regions[r].bar_offset;
                VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)-R%d  pa:0x%llx~0x%llx, va:0x%llx)\n", i, r,
                        		(unsigned long)priv->bar[i].mmio_start + region_offset,
                       		    (unsigned long)priv->bar[i].mmio_start + region_offset + priv->bar[i].regions[r].region_size - 1,
                        		(unsigned long)priv->bar[i].vmem + region_offset);

				VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)-R%d  fa:0x%llx~0x%llx\n", i, r,
								priv->bar[i].regions[r].fn_base_addr,
								priv->bar[i].regions[r].fn_base_addr + priv->bar[i].regions[r].region_size - 1);
                
                VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)-R%d noc:0x%llx~0x%llx, len:0x%llx(%lldK), \n", i, r,
								priv->bar[i].regions[r].noc_addr,
								priv->bar[i].regions[r].noc_addr + priv->bar[i].regions[r].region_size - 1,
                    			priv->bar[i].regions[r].region_size,
                    			priv->bar[i].regions[r].region_size / 1024);
            }
        }

		if(i == VASTAI_PCI_BAR4){
			if(priv->bar[VASTAI_PCI_BAR4].region_cnt > 0) {
				for(r = 0; r < priv->bar[VASTAI_PCI_BAR4].region_cnt; r++) {
					region_offset = priv->bar[i].regions[r].bar_offset;
					if(priv->bar[i].regions[r].region_size){
						VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)-R%d  pa:0x%llx~0x%llx, va:0x%llx)\n", i, r,
									(unsigned long)priv->bar[i].mmio_start + region_offset,
									(unsigned long)priv->bar[i].mmio_start + region_offset + priv->bar[i].regions[r].region_size - 1,
									(unsigned long)priv->bar[i].vmem + region_offset);
						
						VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)-R%d  fa:0x%llx~0x%llx \n", i, r,
										priv->bar[i].regions[r].fn_base_addr,
										priv->bar[i].regions[r].fn_base_addr + priv->bar[i].regions[r].region_size - 1);

						VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)-R%d noc:0x%llx~0x%llx, len:0x%llx(%lldM), \n", i, r,
										priv->bar[i].regions[r].noc_addr,
										priv->bar[i].regions[r].noc_addr + priv->bar[i].regions[r].region_size - 1,
										priv->bar[i].regions[r].region_size,
										priv->bar[i].regions[r].region_size /1024/1024);
					}					
				}
			} else {
				VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)  pa:0x%lx~0x%lx, va:0x%lx\n", i,
               					(unsigned long)priv->bar[i].mmio_start,
               					(unsigned long)priv->bar[i].mmio_end,
                			    (unsigned long)priv->bar[i].vmem);
				
				VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d)  fa:0x%llx~0x%llx \n", i,
									priv->bar[i].fn_base_addr,
									priv->bar[i].fn_base_addr + priv->bar[i].mmio_len - 1);

				VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,"BAR(%d) noc:0x%llx~0x%llx len:0x%lx(%ldM), \n", i,
									priv->bar[i].noc_addr,
									priv->bar[i].noc_addr + priv->bar[i].mmio_len - 1,
									(unsigned long)priv->bar[i].mmio_len,
									(unsigned long)priv->bar[i].mmio_len /1024/1024);
			}
		
		}
    
    }
	
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "E<==================memmap fn:%d===================>E\n", pdev->devfn);
}


// Func: init bar for device at romcode stage
int va_config_initial_bars(struct vastai_pci_info *priv)
{
	struct pci_dev *pdev = priv->dev;
	int ret = 0, i = 0;
	for (i = 0; i < VASTAI_PCI_BAR_NUM; i++) {
		int flag = pci_resource_flags(pdev, i);
		if (!(flag & IORESOURCE_MEM))
			continue;

		priv->bar[i].mmio_start = pci_resource_start(pdev, i);
		priv->bar[i].mmio_end = pci_resource_end(pdev, i);
		priv->bar[i].mmio_flags = pci_resource_flags(pdev, i);
		priv->bar[i].mmio_len = pci_resource_len(pdev, i);
		if ((VASTAI_PCI_BAR4 == i) && (1 == va_bar4_wc_en)) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "BAR[%d] wc enable", VASTAI_PCI_BAR4);
			priv->bar[i].mem = ioremap_wc(priv->bar[i].mmio_start, priv->bar[i].mmio_len);
		}
		else {
			priv->bar[i].mem = ioremap(priv->bar[i].mmio_start, priv->bar[i].mmio_len);
		}
		priv->bar[i].vmem = priv->bar[i].mem;
#ifdef DIAG_TEST_WC
		if(VASTAI_PCI_BAR4 == i){
			//boot download fw need call mmio api
			priv->diag_vddr  = priv->bar[i].vmem;
		}
#endif

		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "BAR[%d] len=0x%llx pa[0x%llx] va[0x%p]", i, priv->bar[i].mmio_len,
				priv->bar[i].mmio_start, priv->bar[i].mem);
		if (priv->bar[i].vmem == NULL) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s:cannot remap bar%d mmio, aborting\n", pci_name(pdev), i);
			ret = -EIO;
		}

		priv->bar_num += 1;
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "priv->bar_num=%d",priv->bar_num);

	return ret;
}

extern int va_sdma_cp_test_for_probe(struct vastai_pci_info *priv, 
					u8 loc_dir, u8 mmu_en, u64 addr0, 
					u64 addr1, u32 trans_len, u32 pid);

// sdma连续read Host 2次,每次data size 1M
static int va_sdma_flush_after_probe(struct vastai_pci_info *priv)
{
	u8 dir = 1;
	u8 mmu = 0;
	u64 addr0  = 0x1048100000;
	u64 addr1  = 0;
	u32 pid	   = 0;
	u32 trans_len = 0x100000;
	int i = 0, ret = 0;

	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "To do SDMA check\n");

	for (i = 0; i < 2; i++) 
	{
		ret = va_sdma_cp_test_for_probe(priv, dir, mmu, addr0, addr1, trans_len, pid);
		if (ret < 0)
		{
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "%s: checked sdma flush for %d times\n", pci_name(priv->dev), i);
			// need to do hot reset
			return ret;
		}
	}

    VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "SDMA check done\n");

	return 0;
}

/*numa info description:
 * non-numa mode: numa mode:-1, node_cpus:null
 * numa mode: numa mode:0/1/.., node_cpus: [32-63,96-127]
 */
static int vastai_pci_numa_cpu_info(struct vastai_pci_info *pci_info, char *numa_cpu_info_item, int len)
{
	int	numa_node;
	const struct cpumask *node_cpumask;

	numa_node = dev_to_node(&pci_info->dev->dev);
	node_cpumask = cpumask_of_node(numa_node);
	snprintf(numa_cpu_info_item, len, "pci_name:[%s]*numa_node:[%d]*node_cpus:[%*pbl];",
			pci_name(pci_info->dev), numa_node, cpumask_pr_args(node_cpumask));

	return 0;
}

int vastai_pci_numa_cpu_info_all(char *buf, int len)
{
	struct vastai_pci_info *pci_info = NULL;
	char numa_cpu_info_item[128] = {0};
	u32 total_len = 0;
	u32 item_len = 0;

	if (list_empty(&vastai_pci_info_list))
		return -1;

	list_for_each_entry(pci_info, &vastai_pci_info_list, dev_list) {
		vastai_pci_numa_cpu_info(pci_info, numa_cpu_info_item, sizeof(numa_cpu_info_item));
		total_len = strlen(buf);
		item_len = strlen(numa_cpu_info_item);
		if (total_len + item_len > len) {
			pr_err("numa_cpu_info buffer space not enough!\n");
			return -2;
		}
		strcat(buf, numa_cpu_info_item);
	}
	pr_info("pci_numa_cpu:%s\n", buf);

	return 0;
}

// Func: get card info API only available for PF
//       Needs put it at the end of the probe.
void va_get_card_info(struct vastai_pci_info *priv)
{
	int ret = 0;
	int tmp = 0, tmp2 = 0;
	u8 card_type = 0, ddr_sz = 0;
	u8 ddr_bar_sz = 0;
	char *pstr_ddr_sz = NULL, *pstr_card_type = NULL;

	ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON77_DDR_CARD - PCIE_WRAPPER_ADDR), &tmp, 4);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s read PWRAP_COMMON77_DDR_CARD failed, ret:%d\n", __func__, ret);
		return;
	}

	ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR1, (PWRAP_COMMON51_DDR_BAR_SZ - PCIE_WRAPPER_ADDR), &tmp2, 4);
	if (ret < 0) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s read PWRAP_COMMON51_DDR_BAR_SZ failed, ret:%d\n", __func__, ret);
		return;
	}

	card_type  = (tmp >> BIT_WRAP_COMM77_CARD_TYPE) & 0x3;
	ddr_sz     = (tmp >> BIT_WRAP_COMM77_DDR_MODE)  & 0x3;
	ddr_bar_sz = tmp2 >> BIT_WRAP_COMM51_PF_DDR_BAR_SZ & 0x7;

    if(card_type == 0)
        pstr_card_type = "EVB-DALIANG";
    else if(card_type == 1)
        pstr_card_type = "ANYI-VGPU1200";
    else if(card_type == 2)
        pstr_card_type = "ZHAOGE-VGPU1000";
    else if(card_type == 3)
        pstr_card_type = "ZHAOGE-VGPU600";
    else {
        pstr_card_type = "ZHAOGE-VGPUxxx";
    }

	// total ddr sz on board
	if(ddr_sz == 0) {
		pstr_ddr_sz = "32GB";
	}
	else if(ddr_sz == 1) {
		pstr_ddr_sz = "16GB";
	}
	else if(ddr_sz == 2) {
		pstr_ddr_sz = "8GB";
	}
	else {
		pstr_ddr_sz = "64GB";
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "Card info: %s, %s DDR. ddr_bar_sz_f:%d \n", pstr_card_type, pstr_ddr_sz, ddr_bar_sz);

}

static void dump_hw_cfg(struct vastai_pci_info *priv)
{	
	int i;
	u32 len = 0x400;
	u64 addr = 0;
	void  *dump_addr;
	int ret=0;
	unsigned char print_str[64];

	dump_addr = kzalloc(len, GFP_KERNEL);


	ret = vastai_pci_mmio_read(priv, VASTAI_PCI_BAR2, 0, dump_addr,len );
    if(ret){
		VASTAI_PCI_INFO(priv, 0, "%s fail, %d,beacause mmio read fail,ret=%d\n",__func__,__LINE__,ret);
        return;
    }
	VASTAI_PCI_INFO(priv, 0, "%s, %d, dump hw_cfg\n",__func__,__LINE__);
	for (i = 0; i < len; i += 16) {
		sprintf(print_str,
			"%s [dump_hw_cfg] 0x%08llx: ",
			VASTAI_PCI_DEBUG_PREFIX, (addr + i));
			VASTAI_PCI_HEX_DUMP(KERN_WARNING, print_str,
					DUMP_PREFIX_NONE, 16, 4,
					dump_addr + i, 16, true);
	}

	kfree(dump_addr);
	return;
}

static int notify_smcu_driver_sts(struct vastai_pci_info *priv, u8 state)
{
	int ret = 0;
	struct smcu_task_info		smcu_task_desc = {0};
	struct smcu_rw_ret_info 	smcu_ret = {0};

	smcu_task_desc.w0_data0 = (state<<8)|HOST_NOTIFY_SMCU_STATUS;
	ret = vastai_submit_smcu_task_and_done(priv,&smcu_task_desc,&smcu_ret);
	if(ret)
		return ret;

	/*check smcu return data whether correcct*/
	if(HOST_NOTIFY_SMCU_STATUS_CHECK_NUM != smcu_ret.w1_data1) {
		VASTAI_PCI_ERR(priv, priv->pkg_id,"smcu_check_val[0x%x] != smcu_return_addr[0x%x]\n",
					HOST_NOTIFY_SMCU_STATUS_CHECK_NUM,smcu_ret.w0_data0);
		return -EIO;
	}

	return ret;
}


// Func: -1, if invalid cfg then fail.
//        0, valid cfg.
static int check_cfg_validty(struct vastai_pci_info *priv)
{
    // no need to enable SRIOV for current PF if gfx_mode is invalid
    if(priv->fn_mode != PF_4_WITH_SRIOV) {
        VASTAI_PCI_INFO(priv, 0, "%s, WARNING: Invalid Fn Mode:%d, no need to enable SR-IOV for PF%d \n", __func__, priv->fn_mode, priv->dev->devfn);
        return -1;
    }

	return 0;
}

static bool is_find_card_info(struct pci_dev *pdev)
{
	struct vastai_card_info *ploop = NULL;
	bool isfind = false;

	mutex_lock(&vastai_card_info_list_lock);
	list_for_each_entry(ploop, &vastai_card_info_list, card_node) {
		if((ploop->bdf_info.dev_id == (pdev->devfn >> 3)) &&
			(ploop->bdf_info.bus_id == (pdev->bus->number)) &&
			(ploop->bdf_info.domain_id == pci_domain_nr(pdev->bus))) {
			isfind = true;
			break;
		}
	}
	mutex_unlock(&vastai_card_info_list_lock);

	return isfind;
}

static int vastai_create_card_info_tree(struct pci_dev *pdev)
{
	struct vastai_card_info *card_info = NULL;
	if(is_find_card_info(pdev))
		return 0;

	card_info = kzalloc(sizeof(struct vastai_card_info), GFP_KERNEL);
	if (!card_info)
		return -ENOMEM;

	//card_info->bdf_info.fn_id     = pdev->devfn & 0x7;
	card_info->bdf_info.dev_id    = pdev->devfn >> 3;
	card_info->bdf_info.bus_id    = pdev->bus->number;
	card_info->bdf_info.domain_id = pci_domain_nr(pdev->bus);
	INIT_LIST_HEAD(&card_info->vastai_pkg_info_list);

	mutex_lock(&vastai_card_info_list_lock);
	list_add_tail(&(card_info->card_node), &(vastai_card_info_list));
	mutex_unlock(&vastai_card_info_list_lock);

	return 0;
}


#define BOOT_PF_ID 0  // 1/3/2
#define DUALPKG_SLAVE_1ST_PF_ID 4
static int vastai_pci_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	struct vastai_pci_info *priv = NULL;
	int ret, i;

	VASTAI_PCI_FUNC_ENTERY;

#ifndef VASTAI_VF_TILE_MODE_SUPPORT
	if (pdev->is_virtfn) {
		return -ENXIO;
	}
#endif

	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "****************************probe devfn=%d *******************************\n", pdev->devfn);
	VASTAI_PCI_DBG(NULL, DUMMY_DIE_ID, "pci_dev=0x%p, deviceID=%#x, at romcode stage.\n", pdev, id->device);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "Domain=%#x, bus=%#x, official name:%s\n", pci_domain_nr(pdev->bus), pdev->bus->number, pci_name(pdev));
	VASTAI_PCI_DBG(NULL, DUMMY_DIE_ID, "pci_info_list=0x%p\n", &vastai_pci_info_list);

#ifndef VASTAI_MULTI_PF_SUPPORT
	if (pdev->devfn != BOOT_PF_ID) {
		/* We only claim function 0 */
		VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "pci_probe devfn=%d, No support for multiple PF & VF.\n", pdev->devfn);
		return -EPERM;
	}
#endif

	priv = kzalloc(sizeof(struct vastai_pci_info), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	ret = vastai_create_card_info_tree(pdev);
	if (ret)
		return ret;

	INIT_LIST_HEAD(&(priv->vacc_dev_head));
	if (list_empty(&(vastai_pci_info_list))) {
		/* this is first device */
		priv->dev_id = 0;
		VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "pci_info_list:0x%p is empty now", &vastai_pci_info_list);
	} else {
		struct vastai_pci_info *last;
		last = list_last_entry(&(vastai_pci_info_list), struct vastai_pci_info, dev_list);
		priv->dev_id = last->dev_id + 1;
	}

	/* we assume only one die in this device first, and get the real die num at boot */
	priv->die_num = 1; /* this member will modify at vastai_pci_boot */
	priv->boot_stage = VASTAI_HOST_BOOT_STAGE_INIT;
	priv->dev = pdev;
	priv->is_pf = 0;
	priv->is_vf = 0;
	priv->id = id;

	/* Check function type */
	if(id->device == PF_DEVICE_ID   || id->device == GFX_PF_DEVICE_ID || id->device == AUD_PF_DEVICE_ID || 
		id->device == PF_DEVICE_L_ID || id->device == PF_DEVICE_M_ID)
		priv->is_pf = 1;

	if(id->device == VF_DEVICE_ID || id->device == VF_DEVICE_ID2) {
		priv->is_vf = 1; 
	}

	pci_set_drvdata(pdev, priv);

	if (pci_enable_device(pdev) < 0) {
		ret = -EIO;
		goto ERROR_FREE_PRIV;
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"devfn=%d, is_physfn=%d, is_virtfn=%d, is_pf=%d, is_vf=%d\n",
			pdev->devfn, pdev->is_physfn, pdev->is_virtfn, priv->is_pf, priv->is_vf);

// FIXME: for dualpkg in AIC card debug 0302.2023
#ifdef CONFIG_VASTAI_CARD_AIC_16GB
#ifdef CONFIG_VASTAI_PCI_FORCE_SINGLE_DIE
	if ((pdev->devfn >= 4) && priv->is_pf)
	{
		VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "Force single die on AIC dualpkg card.\n");
		ret = -EIO;
		goto ERROR_FREE_PRIV;
	}
#endif
#endif

	ret = vastai_dmabuf_init(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"Failed to init dmabuf. errno %d\n", ret);

		goto ERROR_DISABLE_PCIE;
	}

	/*
	 * Enabling the device will enable any BARs,
	 * device memory is now accessible.
	 */
#if 0   // FIXEDME: when debug display drv with GPUSW Dale, 12/10/2023
	if (pci_request_regions(pdev, "vastai_pci_basic") < 0) {
		ret = -1;
		goto ERROR_DEINIT_DMABUF;
	}
#endif

	priv->legacy_en = 0;
	priv->msi_en = 0;
	priv->msix_en = 1;
	priv->irq = pdev->irq;

	if (priv->legacy_en == 1)
		priv->irq = pdev->irq;

	if (priv->msi_en == 1) {
		priv->irq_num = pci_msi_vec_count(pdev);
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"pci_msi_vec_count irq_num=%d\n",
				priv->irq_num);

		ret = pci_alloc_irq_vectors(pdev, 1, priv->irq_num, PCI_IRQ_MSI);
		if (ret > 0) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
					"pci_enable_msi_range ret=%d\n", ret);
			priv->irq_num = ret;
			priv->msi_en = 1;
		} else {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_enable_msi_range err=%d\n", ret);
			priv->msi_en = 0;
		}
		priv->irq = pdev->irq;
	}

	if (priv->msix_en == 1) {
		for (i = 0; i < VASTAI_PCIE_MSIX_NUM; i++) {
			priv->msix[i].entry = i;
			priv->msix[i].vector = 0;
		}
		priv->irq_num = pci_enable_msix_range(pdev, priv->msix, 1,
			VASTAI_PCIE_MSIX_NUM);
		if (priv->irq_num > 0) {
			VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
				"pci_enable_msix_range %d ok\n",
				priv->irq_num);
			priv->msix_en = 1;
		} else {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_enable_msix_range %d err\n",
				priv->irq_num);
			priv->msix_en = 0;
		}
		priv->irq = priv->msix[0].vector;
	}
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
		"legacy %d msi_en %d, msix_en %d, dev->irq %d, priv->irq %d\n",
		priv->legacy_en, priv->msi_en, priv->msix_en,
		pdev->irq, priv->irq);

	if (priv->legacy_en == 1) {
		ret = request_irq(
			priv->irq, (irq_handler_t)(&vastai_pci_legacy_irq),
			IRQF_NO_SUSPEND | IRQF_NO_THREAD | IRQF_PERCPU,
			VASTAI_PCI_DRIVER_PREFIX, (void *)priv);
		if (ret)
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_probe request_irq(%d), error %d\n",
				priv->irq, ret);
		else
			VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				"pci_probe request_irq(%d) ok\n",
				priv->irq);
	}
	if (priv->msi_en == 1) {
		for (i = 0; i < priv->irq_num; i++) {
			ret = request_irq(priv->irq + i,
				(irq_handler_t)(&vastai_pci_msi_irq),
				IRQF_SHARED, VASTAI_PCI_DRIVER_PREFIX,
				(void *)priv);
			// IRQF_SHARED is better replaced by 0
			if (ret) {
				VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					"pci_probe request_irq(%d), error %d\n",
					priv->irq + i, ret);
				break;
			}
			VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				"pci_probe request_irq(%d) ok\n",
				priv->irq + i);
		}
	}
	if (priv->msix_en == 1) {
		for (i = 0; i < priv->irq_num; i++) {
			ret = request_irq(priv->msix[i].vector,
				(irq_handler_t)(&vastai_pci_msi_irq),
				IRQF_SHARED, VASTAI_PCI_DRIVER_PREFIX,
				(void *)priv);
			if (ret) {
				VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					"pci_probe request_irq(%d), error %d\n",
					priv->msix[i].vector,
					ret);
				break;
			}

			VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				"pci_probe request_irq(%d) ok\n",
				priv->msix[i].vector);
		}
	}

	list_add_tail(&(priv->dev_list), &(vastai_pci_info_list));

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
	/* Enable AER */
	pci_enable_pcie_error_reporting(pdev);
#endif

	/* Enable bus mastering (we don't use it though) */
	pci_set_master(pdev);
	// Set MRRS
	pcie_set_readrq(pdev, 128);

	/*Get PCIe Link Width and Speed*/
	vastai_get_pcie_attr(priv);

	// before smcu fw init done, soc is still using atu config at romcode stage. Only for PF.
	if((ret = va_config_initial_bars(priv)) < 0){
		goto ERROR_DISABLE_PCIE;
	} else if(priv->is_pf && (priv->bar_num < 4)) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"pci_probe bar_num[%d] is less than 4. Pls check bar configuration.\n",
			priv->bar_num);
		ret = -EIO;
		goto ERROR_DISABLE_PCIE;
	}

	mutex_init(&priv->trans_mutex);
	spin_lock_init(&priv->mmio_read_spin_lock);
	spin_lock_init(&priv->mmio_write_spin_lock);
	mutex_init(&vastai_pci_modify_lock);
	mutex_init(&vastai_smcu_modify_lock);	
	spin_lock_init(&priv->trans_lock);
	spin_lock_init(&priv->swtich_atu_lock);
	spin_lock_init(&priv->ring_buf_lock);
	init_completion(&priv->swtich_atu_done);
	init_completion(&priv->rw_xspi);
	init_completion(&priv->send_reinit_rst_done);
	init_completion(&priv->pmcu_boot_mode_done);
	init_completion(&priv->smcu_flr_done);

	vastai_pci_ioctl_init(priv);

	/* atuo gart enable check*/
	if (CONFIG_ATUO_GART_EN)
		is_need_to_enable_gart(priv);
	else
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"Disable atuo gart check\n");

#ifdef CONFIG_VASTAI_PCI_BOOT
	INIT_DELAYED_WORK(&priv->boot_wq, vastai_pci_boot_wq);

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
		"pci_dev:devfn=%d, cfg_size=%#x, pcie_cap=%#x, msi_cap=%#x, msi_en=%d, msix_cap=%#x, msix_en=%d\n", 
		pdev->devfn, pdev->cfg_size, pdev->pcie_cap, pdev->msi_cap, pdev->msi_enabled, pdev->msix_cap, pdev->msix_enabled);

	vastai_pci_state_boot_start(priv);

	if (priv->is_pf) {
#ifdef CONFIG_PROBE_RESET
		if (!g_pf_reinit_done)
		{
			ret = vastai_pci_mmio_do_reinit_rst(priv, TRUE);
			if (ret < 0) {
				goto ERROR_RELEASE_REGIONS;
			}
			g_pf_reinit_done++;
			msleep(1000);
		}
#endif

		ret = vastai_pci_boot(priv);
		if (ret < 0) {
			goto ERROR_RELEASE_REGIONS;
		}
	}
	priv->boot_stage = VASTAI_HOST_BOOT_STAGE_READY;
#endif

	ret = get_hw_config(priv);
	if (!ret) {
		priv->card_id = priv->priv_hw_cfg->sys_cfg.card_id;
		priv->pkg_id = priv->priv_hw_cfg->sys_cfg.pkg_id;
		priv->fn_mode = priv->priv_hw_cfg->sys_cfg.fn_mode;
	} else {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "get_hw_config failed, %d\n", ret);
		dump_hw_cfg(priv);
		goto ERROR_DEINIT_DMABUF;
	}

	ret = vastai_add_card_id(priv);
	if ((ret == 0) || priv->is_boot_pf) {
		priv->is_main_dev_in_os = 1;
	}

	if(priv->gart_en && CONFIG_ATUO_GART_EN)
		priv->hasInitGfxMemAlloc = false;

	init_vastai_pci_mem(priv);

	ret = va_hw_init(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "va_hw_init failed, %d\n", ret);
	}

	if (priv->is_main_dev_in_os) {
		if (priv->is_pf) {
			ret = logsys_init(priv);
			if (ret < 0) {
				VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "logsys_init fail ret[%d]\n", ret);
			}
		}

		vastai_version_file_init(priv);
	}

#ifdef CONFIG_PVP_HASH_CMP_EN
	mutex_init(&priv->cedar_lock);
#endif
	/* *********************************
	 * if vatools register error,
	 * Pcie driver should not return error,
	 * print a msg is enough for debug
	 * *********************************/
#ifdef CONFIG_VASTAI_TOOLS_SUITE
	init_completion(&(priv->tools_info.smi_ack[0]));
	init_completion(&(priv->tools_info.smi_ack[1]));
	init_completion(&(priv->tools_info.core_time_sync_ack));
	init_completion(&(priv->tools_info.log_comp));
	mutex_init(&(priv->tools_info.tools_dev_mutex));
	mutex_init(&(priv->tools_info.tools_die_mutex));

	ret = tools_init(priv);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"line %d, cannot do tools_init %d\n", __LINE__,
			ret);
	}
#endif

	init_all_ring_buf(priv);
	smcu_to_host_msix_init(priv);
	vastai_exception_init(priv);
	init_ai_info(priv);

	/* *********************************
	 * if ai and video register error,
	 * Pcie driver should not return error,
	 * print a msg is enough for debug
	 * *********************************/
#ifdef CONFIG_VASTAI_AI
	i = 0;
	ret = vacc_init(priv, priv->dies[i].die_index);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"%s:cannot do vacc_init %d\n",
			pci_name(pdev), ret);
	}
#endif

#ifdef CONFIG_VASTAI_VIDEO
	INIT_LIST_HEAD(&priv->render_head);
	ret = vastai_video_init(priv, priv->dies[0].die_index);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"%s:cannot do video_init %d\n",
			pci_name(pdev), ret);
	}
#endif

	device_smmu_ste_table_init(priv);

	if (priv->is_boot_pf) {
//		ret = vastai_pci_poll_smcu_stage(priv, VASTAI_DIE0, VASTAI_SMCU_BOOT_STAGE_ALLCORE_READY);
//		if (ret < 0) {
//			VASTAI_PCI_ERR(priv, VASTAI_DIE0, "pci_probe poll smcu stage[%d] error\n", VASTAI_SMCU_BOOT_STAGE_ALLCORE_READY);
//		}

		vastai_pci_set_host_stage(priv, VASTAI_DIE0, VASTAI_HOST_BOOT_STAGE_READY);
	}

	/* get fw version */
	vastai_pci_get_fw_ver(priv, 0xFFFFFFFF);

	if (priv->is_main_dev_in_os) {
		vastai_pci_read_pmcu_boot_mode(priv);
	}


#ifdef DIAG_TEST_DDK
	{
		u64 temp_addr;
		u32 csram_data;
		VASTAI_PCI_INFO(priv, 0,"switch atu start\n");
		temp_addr = (u64)priv->priv_hw_cfg->header.csram_base+0xFF0000000;
		ret = vastai_pci_mem_read(priv, 0, temp_addr, &csram_data, 0x4);
		if (ret < 0)
		VASTAI_PCI_ERR(priv, 0,"pci_probe [read] 0x%llx, len=%d, error:%d\n", temp_addr, 4, ret);
		if(0x55AA == csram_data) {
			VASTAI_PCI_INFO(priv, 0,"switch atu pass addr=0x%lx\n",priv->bar[VASTAI_PCI_BAR4].regions[0].noc_addr);
		} else {
			VASTAI_PCI_ERR(priv, 0,"switch atu error\n");
		}
	}
#endif

#ifdef CONFIG_VASTAI_GFX
	if (priv->is_boot_pf)
	{
		ret = pvr_init();
		VASTAI_PCI_INFO(priv, VASTAI_DIE0, "ddk pvr init\n");
		if (ret) {
			VASTAI_PCI_ERR(priv, VASTAI_DIE0, "ddk pvr init failed %d\n", ret);
		}
	}
#endif

#ifndef DIAG_TEST_WC
	/*Note:when enable DIAG_TEST_WC priv->bar[VASTAI_PCI_BAR4].vmem = NULL*/
	if (priv->is_boot_pf)
	{
		ret = va_create_pid(priv, 128);
		if (ret < 0) {
			VASTAI_PCI_ERR(priv, 0,"va_create_pid create failed\n");
		}
	}
#endif
	// SRIOV enable function block
#ifdef VASTAI_VF_TILE_MODE_SUPPORT
	{
		int total_vfs = 0;

		if (priv->is_vf)
			goto skip_vf_en;

		total_vfs = pci_sriov_get_totalvfs(pdev);
		if(total_vfs > 0) {
			if(check_cfg_validty(priv) < 0) {
				goto skip_vf_en;
			}

			VASTAI_PCI_INFO(priv, 0, "pci_probe PF%d to enable SRIOV\n", pdev->devfn);

			ret = pci_enable_sriov(pdev, total_vfs);
			if (ret)
				VASTAI_PCI_ERR(priv,  0, "SR-IOV enable %d VFs failed, %d\n", total_vfs, ret);
			else
				VASTAI_PCI_INFO(priv, 0, "SR-IOV enable %d VFs success\n", total_vfs);
		}
		else
			VASTAI_PCI_INFO(priv, 0, "pci_probe WARNING: SR-IOV disabled due to %d VFs\n", total_vfs);
	}
skip_vf_en:
#endif

	vastai_pci_state_boot_done(priv);

	if(priv->is_pf) {
		// Needs put it at the end of the probe
		va_get_card_info(priv);
	}

	/* heartbeat monitor init, init at the end */
#if VASTAI_HEARTBEAT_ENABLE
	if (priv->is_boot_pf) {
		vastai_heartbeat_monitor_init(priv);
	}
#endif

	if (priv->is_boot_pf) {
		ret = dmi_set_device(priv);
		if (ret < 0)
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "%s:flash dmi is fail %d\n", pci_name(pdev), ret);
	}

	// Fixme: dualpkg, error handle here.
	if (priv->is_boot_pf)
	{
		va_sdma_flush_after_probe(priv);
	}

	notify_smcu_driver_sts(priv, VASTAI_NORMAL_STATE);

	// parse Inventory Info for each function for non EVB Card
	if(priv->priv_hw_cfg->sys_cfg.card_type != EVB_DALIANG) {
		va_read_parse_inventory(priv);
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"************probe devfn=%d, va_fn=%d of card=%u, %s, Done*********\n",
			pdev->devfn, priv->priv_hw_cfg->sys_cfg.pfn, priv->card_id, va_get_product_sn(priv));

	// Important: add this below line only for FCT test. DO NOT edit below line.
	// If you want to change this, notice tools team. Thanks.
	vatools_fct_probe_done(priv, pdev->devfn, priv->card_id);
	priv->probe_done = true;

	return 0;

#ifdef CONFIG_VASTAI_PCI_BOOT
ERROR_RELEASE_REGIONS:
	pci_release_regions(pdev);
#endif

ERROR_DEINIT_DMABUF:
	vastai_dmabuf_deinit(priv);

ERROR_DISABLE_PCIE:
	pci_disable_device(pdev);
ERROR_FREE_PRIV:
	pci_set_drvdata(pdev, NULL);

	// delete the node if error happened
	list_del_init(&(priv->dev_list));

	kfree(priv);
	return ret;
}

static int vastai_delete_card_info_tree(struct pci_dev *pdev)
{
	struct vastai_card_info *ploop = NULL;

	mutex_lock(&vastai_card_info_list_lock);
	list_for_each_entry(ploop, &vastai_card_info_list, card_node) {
		if((ploop->bdf_info.dev_id == (pdev->devfn >> 3)) &&
			(ploop->bdf_info.bus_id == (pdev->bus->number)) &&
			(ploop->bdf_info.domain_id == pci_domain_nr(pdev->bus))) {
			struct vastai_pkg_info *pkg_loop = NULL;
			struct vastai_pkg_info *pkg_tmp  = NULL;
			/* TODO: should check multi fn, now is sv100*/
			list_for_each_entry_safe(pkg_loop, pkg_tmp, &ploop->vastai_pkg_info_list, pkg_node) {
				struct vastai_die_info *die_loop = NULL;
				struct vastai_die_info *die_tmp  = NULL;

				list_for_each_entry_safe(die_loop, die_tmp, &pkg_loop->vastai_die_info_list, die_node) {
					struct vastai_pci_info *fn_loop = NULL;
					struct vastai_pci_info *fn_tmp  = NULL;

					list_for_each_entry_safe(fn_loop, fn_tmp, &die_loop->vastai_fn_info_list, fn_node[die_loop->die_id]) {
						if((fn_loop->dev->devfn & 0x7) == (pdev->devfn & 7))
							list_del(&(fn_loop->fn_node[die_loop->die_id]));
					}
					if(list_empty(&die_loop->vastai_fn_info_list)) {
						list_del(&(die_loop->die_node));
						kfree(die_loop);
					}
				}
				if(list_empty(&pkg_loop->vastai_die_info_list)) {
					list_del(&(pkg_loop->pkg_node));
					kfree(pkg_loop);
				}
			}
			if(list_empty(&ploop->vastai_pkg_info_list)) {
				list_del(&(ploop->card_node));
				kfree(ploop);
			}
			break;
		}
	}
	mutex_unlock(&vastai_card_info_list_lock);

	return 0;
}

static void vastai_pci_remove_internal(struct pci_dev *pdev, bool is_in_shutdown)
{
	struct vastai_pci_info *priv = pci_get_drvdata(pdev);
	int i = 0;

	if(!priv) {
		VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "pci_remove_internal devfn=%d, NULL priv. return\n", pdev->devfn);
		return;
	}

	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "-----------%s devfn=%d, va_fn=%d-----------\n",
			(is_in_shutdown ? "shutdown" : "remove"), pdev->devfn, priv->priv_hw_cfg->sys_cfg.pfn);

	notify_smcu_driver_sts(priv,VASTAI_NO_RUN_STATE);
	vastai_pci_state_power_off(priv);

	if (priv->is_pf && pci_num_vf(pdev) > 0 && !check_cfg_validty(priv))
		pci_disable_sriov(pdev);

	vastai_pcie_bw_test_cancel(priv);

#if VASTAI_HEARTBEAT_ENABLE
	if (priv->is_boot_pf) {
		vastai_heartbeart_monitor_deinit(priv);
	}
#endif

#ifdef CONFIG_VASTAI_AI
	vacc_exit(priv);
#endif

#ifdef CONFIG_VASTAI_TOOLS_SUITE
	tools_exit(priv);
#endif

#ifdef CONFIG_VASTAI_VIDEO
	vastai_video_deinit(priv);
#endif

	if (priv->is_main_dev_in_os) {
		if (priv->is_pf)
			logsys_deinit(priv);
		
		vastai_version_file_deinit(priv);
	}
	
	deinit_all_ring_buf(priv);
	vastai_pci_ioctl_deinit(priv);
	mutex_destroy(&priv->trans_mutex);

	vastai_pci_disable_irq(priv);

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
	pci_disable_pcie_error_reporting(pdev);
#endif

	va_hw_deinit(priv, is_in_shutdown);
	free_ai_info(priv);

	if (priv->is_boot_pf) {
		vastai_pci_mmio_do_reinit_rst(priv, TRUE);
	}

#ifdef CONFIG_VASTAI_PCI_BOOT
	cancel_delayed_work_sync(&priv->boot_wq);
#endif

#ifdef CONFIG_VASTAI_GFX
	if(!is_in_shutdown)
		pvr_exit();
#endif

	//pci_release_regions(pdev);//because vbios has use bar 4/5 region
	pci_disable_device(pdev);

	for (i = 0; i < VASTAI_PCI_BAR_NUM; i++) {
		int flag = pci_resource_flags(pdev, i);
		if (!(flag & IORESOURCE_MEM))
			continue;
		if (priv->bar[i].vmem != NULL) {
			iounmap(priv->bar[i].vmem);
		}
#ifdef DIAG_TEST_WC
		if(VASTAI_PCI_BAR4 ==i){
			if(priv->diag_vddr != NULL){
				iounmap(priv->diag_vddr);
			}
			if(priv->diag_vcsr != NULL){
				iounmap(priv->diag_vcsr);
			}
		}
#endif
	}

	vastai_delete_card_info_tree(pdev);
	list_del(&(priv->dev_list));
	vastai_del_card_id(priv);
	free_hw_cfg(priv);
	pci_set_drvdata(pdev, NULL);
	if(NULL != priv->ddr_info.pcie_to_ddk_info_addr)
		kfree(priv->ddr_info.pcie_to_ddk_info_addr);
	kfree(priv);

	msleep(500);

	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "pci_remove_internal done!\n");
}

static void vastai_pci_remove(struct pci_dev *pdev)
{
	vastai_pci_remove_internal(pdev, false);
}

static void vastai_pci_shutdown(struct pci_dev *pdev)
{
	vastai_pci_remove_internal(pdev, true);
}


#if KERNEL_VERSION(5, 9, 0) <= LINUX_VERSION_CODE
/* 2020-08-07 18:48:15 -0700
 * Merge tag 'pci-v5.9-changes' of
 * git://git.kernel.org/pub/scm/linux/kernel/git/helgaas/pci
 */
static pci_ers_result_t vastai_pci_io_error_detected(struct pci_dev *pdev, pci_channel_state_t state)
#else
static pci_ers_result_t vastai_pci_io_error_detected(struct pci_dev *pdev, enum pci_channel_state state)
#endif
{
	dump_ep_aer_regs(pdev);
	switch (state) {
	case pci_channel_io_normal:
		dev_info(&pdev->dev, "Can recover basic device\n");
		return PCI_ERS_RESULT_CAN_RECOVER;
	case pci_channel_io_frozen:
		pci_disable_device(pdev);
		dev_info(&pdev->dev, "Request reset for basic device\n");
		return PCI_ERS_RESULT_NEED_RESET;
	case pci_channel_io_perm_failure:
		dev_err(&pdev->dev, "pci_io_error_detected: pci_channel_io_perm_failure.\n");
		return PCI_ERS_RESULT_DISCONNECT;
	}

	/* Request a slot reset. */
	return PCI_ERS_RESULT_NEED_RESET;
}

static pci_ers_result_t vastai_pci_io_slot_reset(struct pci_dev *pdev)
{
	pdev->error_state = pci_channel_io_normal;

	pci_restore_state(pdev);
	if (pci_enable_device(pdev)) {
		printk("can't enable device\n");
		dev_info(&pdev->dev, "Cannot re-enable basic device after reset.\n");
		return PCI_ERS_RESULT_DISCONNECT;
	}
	pci_set_master(pdev);
#if KERNEL_VERSION(4, 18, 0) <= LINUX_VERSION_CODE
	/* 2022-11-7 17:26:08 
	 * PCI/AER: Rationalize error status register clearing
	 */
	//pci_aer_clear_nonfatal_status(pdev); //fixme: 5.4.0 not support the api
#else
	pci_cleanup_aer_uncorrect_error_status(pdev);
#endif

	return PCI_ERS_RESULT_RECOVERED;
}

static void vastai_pci_io_resume(struct pci_dev *pdev)
{
	dev_info(&pdev->dev, "resuming basic device\n");
}

static int vastai_pci_sriov_configure(struct pci_dev *pdev, int num_vfs)
{
	int ret = num_vfs;
	int total_vfs = 0;
	struct vastai_pci_info *pci_info = pci_get_drvdata(pdev);
	struct pcie_transfer_cmd trans;

#ifdef DYNAMIC_SRIOV_EN
	struct smcu_task_info     smcu_task_desc = {0};
	struct smcu_rw_ret_info smcu_ret = {0};
	u8 pfn = pci_info->priv_hw_cfg->sys_cfg.pfn;
	u8 smcu_ret_pfn;
	struct pcie_to_ddk_info *ddk_trans;
#endif

	if (!pci_info->is_pf) {
		VASTAI_PCI_ERR(pci_info, 0, "pci_sriov_configure only PF can enable/disable VFs\n");
		return -EPERM;
	}

	total_vfs = pci_sriov_get_totalvfs(pdev);
	if (total_vfs == 0) {
		VASTAI_PCI_ERR(pci_info, 0, "pci_sriov_configure no available VF\n");
		return ENXIO;
	}

	if(check_cfg_validty(pci_info) < 0) {
		return -EPERM;
	}

	if (num_vfs <= 0) {
		pci_disable_sriov(pdev);
		return ret;
	}

	if (num_vfs > total_vfs)
		num_vfs = total_vfs;

#ifdef DYNAMIC_SRIOV_EN
	/*opcode:SMCU_CONFIG_VF_NUM_CHANGE */
	smcu_task_desc.w0_data0 = (num_vfs<<8)|SMCU_CONFIG_VF_NUM_CHANGE;
	ret = vastai_submit_smcu_task_and_done(pci_info,&smcu_task_desc,&smcu_ret);
	if(ret)
		return ret;
	
	smcu_ret_pfn = (smcu_ret.w0_data0>>8) & 0xFF;
	if((pfn != smcu_ret_pfn) && (DYNAMIC_SRIOV_CHECK_NUM != smcu_ret.w1_data1)) {
		VASTAI_PCI_ERR(pci_info, pci_info->pkg_id,"dynamic change vf num fail for check return invalid pfn[%d] and check_val[0x%x] \n",
					smcu_ret_pfn,smcu_ret.w1_data1);
		return -EIO;
	}
	pci_info->priv_hw_cfg->sys_cfg.cur_pf_vf_num = num_vfs;
	pci_info->ddr_info.cur_pf_vf_num		     = num_vfs;
	if(pci_info->is_pf) {
		pci_info->ddr_info.vf_ddr_size   =(pci_info->ddr_info.fn_ddr_size - META_FW_SIZE_32M - VF_MANAGEMENT_SIZE_64M)
				/pci_info->ddr_info.cur_pf_vf_num/DDR_SIZE_16M_ALIGN*DDR_SIZE_16M_ALIGN;
	}

	ddk_trans = (struct pcie_to_ddk_info*)pci_info->ddr_info.pcie_to_ddk_info_addr;

	ddk_trans->cur_pf_vf_num = pci_info->ddr_info.cur_pf_vf_num;
	ddk_trans->vf_ddr_size	 = pci_info->ddr_info.vf_ddr_size;
#endif

	if (pci_enable_sriov(pdev, num_vfs)) {
		trans.w0.s_data0.optcode	= SMCU_MODIFY_SRIOV_CAP;
		ret = vastai_pci_send_msg(pci_info, 0, COMMON_HOST_TO_SMCU_CMD_BUF,  &trans, 0);
		if (ret) {
			VASTAI_PCI_ERR(pci_info, pci_info->pkg_id, "pci_sriov_configure, %d, push element to [host_to_smcu buf] fail, ret=%d\n",
							__LINE__,ret);
		}
		mdelay(200);
		if (pci_enable_sriov(pdev, num_vfs))
			VASTAI_PCI_ERR(pci_info,  0, "enable SR-IOV %d VFs, failed. %d\n", num_vfs, ret);
		ret = 0;
	} else {
		VASTAI_PCI_INFO(pci_info, 0, "enable SR-IOV %d VFs, succeeded\n", num_vfs);
		ret = num_vfs;
	}

	return ret;
}

board_type_e  vastai_get_board_type(struct vastai_pci_info *priv)
{
	if (0x100 == (priv->id->device & 0xf00))
		return SV100;
	else if (0x200 == (priv->id->device & 0xf00))
		return SG100;

	VASTAI_PCI_ERR(priv, VASTAI_DIE0,
			"unknow device[0x%x]\n", priv->id->device);
	return TYPE_UNKNOW;
}

bool is_vastai_probe_done(struct vastai_pci_info *priv)
{
	return priv->probe_done;
}


static const struct pci_error_handlers vastai_pci_err_handler = {
	.error_detected		= vastai_pci_io_error_detected,
	.slot_reset			= vastai_pci_io_slot_reset,
	.resume				= vastai_pci_io_resume,
};

static const struct pci_device_id vastai_pci_ids[] = {
	{PCI_VDEVICE(VASTAI, PF_DEVICE_ID), 0},
	{PCI_VDEVICE(VASTAI, PF_DEVICE_L_ID), 0},
	{PCI_VDEVICE(VASTAI, PF_DEVICE_M_ID), 0},
	{PCI_VDEVICE(VASTAI, VF_DEVICE_ID), 0},
	{PCI_VDEVICE(VASTAI, VF_DEVICE_ID2), 0},
	{PCI_VDEVICE(VASTAI, GFX_PF_DEVICE_ID), 0},
	{PCI_VDEVICE(VASTAI, AUD_PF_DEVICE_ID), 0},

	{PCI_VDEVICE(DAOCLOUD, PF_DEVICE_ID), 0},
	{PCI_VDEVICE(DAOCLOUD, VF_DEVICE_ID2), 0},
	{PCI_VDEVICE(DAOCLOUD, GFX_PF_DEVICE_ID), 0},
	{PCI_VDEVICE(DAOCLOUD, AUD_PF_DEVICE_ID), 0},
	{/* terminate list */}
};

static struct pci_driver vastai_pci_driver = {
	.name				= VASTAI_PCI_DRIVER_PREFIX,
	.id_table			= &vastai_pci_ids[0],
	.probe				= vastai_pci_probe,
	.remove				= vastai_pci_remove,
	.shutdown			= vastai_pci_shutdown,
	.sriov_configure	= vastai_pci_sriov_configure,
	.err_handler		= &vastai_pci_err_handler
};

extern struct class *myclass;
#if defined(RHEL_RELEASE_VERSION) && (KERNEL_VERSION(5, 14, 0) <= LINUX_VERSION_CODE)
static char *devnode(const struct device *dev, umode_t *mode)
#elif KERNEL_VERSION(6, 2, 0) <= LINUX_VERSION_CODE
static char *devnode(const struct device *dev, umode_t *mode)
#else
static char *devnode(struct device *dev, umode_t *mode)
#endif
{
	struct vastai_file_info *file_info = dev_get_drvdata(dev);

	if (mode)
		*mode = VASTAI_FILE_MODE;
	return kstrdup(file_info->file_name, GFP_KERNEL);
}

static int __init vastai_pci_init(void)
{
	int rc;

	dmi_info_init();

	//VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "%s\n", VASTAI_PCI_DRIVER_COPYRIGHT);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[driver  ver]: %s\n", VASTAI_PCI_DRIVER_VERSION);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[branch info]: %s\n", VASTAI_PCI_GIT_BRANCH_NAME);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[commit hash]: %s\n", VASTAI_PCI_GIT_COMMIT_HASH);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[commit mesg]: %s\n", VASTAI_PCI_GIT_COMMIT_MSG);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[pack   time]: %s\n", VASTAI_PCI_MKPACKAGE_TIME);
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "[build  time]: %s %s\n", __DATE__, __TIME__);

	VASTAI_PCI_FUNC_ENTERY;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 4, 0)
 	myclass = class_create(THIS_MODULE, "char_class");
#else
	myclass = class_create("char_class");
#endif
	
	myclass->devnode = devnode;

#ifdef EMULATOR_ENV
	rc = vastai_register_device();
	if (rc)
		return rc;
#endif

	INIT_LIST_HEAD(&vastai_pci_info_list);
	INIT_LIST_HEAD(&vastai_pci_card_list);
	INIT_LIST_HEAD(&vastai_card_info_list);
	mutex_init(&vastai_card_info_list_lock);
	rc=vastai_dev_node_init();
	if(rc)
		return rc;
	rc = pci_register_driver(&vastai_pci_driver);
	if (rc)
		return rc;

	//VASTAI_PCI_FUNC_EXIT;
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "va_pci_init done!\n");

	return 0;
}

static void __exit vastai_pci_exit(void)
{
	VASTAI_PCI_FUNC_ENTERY;
	pci_unregister_driver(&vastai_pci_driver);
	vastai_dev_node_deinit();
	class_destroy(myclass);
#ifdef EMULATOR_ENV
	vastai_unregister_device();
#endif
	//VASTAI_PCI_FUNC_EXIT;
	VASTAI_PCI_INFO(NULL, DUMMY_DIE_ID, "va_pci_exit done!\n");
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,4,0)
MODULE_IMPORT_NS(DMA_BUF);
#endif

module_init(vastai_pci_init);
module_exit(vastai_pci_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Gang Wang <gang.wang>");
MODULE_DEVICE_TABLE(pci, vastai_pci_ids);
