#include <linux/aer.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/version.h>
#include <linux/rtc.h>
#include <linux/timer.h>
#include <linux/fs.h>
#include <linux/uaccess.h>

#include "vastai_pci.h"
#include "vastai_pci_api.h"
#include "vastai_dmabuf.h"
#include "vastai_fifo.h"
#include "vastai_state.h"
#include "sg100_cmd.h"

char* state_name(int state)
{
	switch(state)
	{
		case VASTAI_NO_RUN_STATE	: return "NO_RUN_STATE";
		case VASTAI_NORMAL_STATE	: return "NORMAL_STATE";
		case VASTAI_UPDATE_STATE	: return "UPDATE_STATE";
		case VASTAI_BOOT_STATE		: return "BOOT_STATE";
		case VASTAI_EXIT_STATE		: return "EXIT_STATE";
		case VASTAI_ERROR_STATE		: return "ERROR_STATE";
		case VASTAI_RESET_STATE		: return "RESET_STATE";
		case VASTAI_DEBUG_STATE		: return "DEBUG_STATE";
		default: break;
	}

	return "unknown state.";
}

int vastai_pci_state_boot_start(struct vastai_pci_info* priv)
{
	int ret = VASTAI_STATE_TRANSIT_SUCCESS;
	int old;
	old = atomic_cmpxchg(&priv->pci_state, VASTAI_NO_RUN_STATE, VASTAI_BOOT_STATE);
	if (old != VASTAI_NO_RUN_STATE) {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID,
			"pci_state_boot_start: boot device must from NO_RUN_STATE. %d\n",
			old);
		ret = -EIO;
	}
	return ret;
}

void vastai_pci_state_boot_done(struct vastai_pci_info* priv)
{
	atomic_set(&(priv->pci_state), VASTAI_NORMAL_STATE);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "device state is %s\n", state_name(VASTAI_NORMAL_STATE));
}

void vastai_pci_state_reset_start(struct vastai_pci_info* priv)
{
	atomic_set(&priv->pci_state, VASTAI_RESET_STATE);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "device state is %s\n", state_name(VASTAI_RESET_STATE));
}

void vastai_pci_state_reset_done(struct vastai_pci_info* priv)
{
	atomic_set(&priv->pci_state, VASTAI_NO_RUN_STATE);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "device state is %s\n", state_name(VASTAI_NO_RUN_STATE));
}

void vastai_pci_state_power_off(struct vastai_pci_info* priv)
{
	atomic_set(&priv->pci_state, VASTAI_EXIT_STATE);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "device state is %s\n", state_name(VASTAI_EXIT_STATE));
}

void vastai_pci_state_sync_error(struct vastai_pci_info* priv)
{
	atomic_set(&(priv->pci_state), VASTAI_ERROR_STATE);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "device state is %s\n", state_name(VASTAI_ERROR_STATE));
}

inline int vastai_get_pci_state(struct vastai_pci_info* priv)
{
	return atomic_read(&(priv->pci_state));
}

int vastai_pci_state_change(struct vastai_pci_info* priv, u32 cur_state, u32 next_state)
{
	int ret = VASTAI_STATE_TRANSIT_SUCCESS;
	int old;
	old = atomic_cmpxchg(&priv->pci_state, cur_state, next_state);
	if (old != cur_state) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			"device state is %s, but not %s, can't change to %s\n",
			state_name(old), state_name(cur_state), state_name(next_state));
		ret = -EIO;
	}
	else {
		VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "device state is %s\n", state_name(next_state));
	}
	
	return ret;
}

int vastai_pci_set_state(struct vastai_pci_info* priv, u32 state)
{
	if (VASTAI_STATE_END <= state) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "pci_set_state : err state %d\n", state);
		return -EPERM;
	}
	
	atomic_set(&(priv->pci_state), state);
	VASTAI_PCI_INFO(priv, DUMMY_DIE_ID, "device state is %s\n", state_name(state));

	return 0;
}

// Note: this func is also used in msix handler. msleep is not allowed.
int vastai_pci_sync_state(struct vastai_pci_info *priv, u32 state)
{
	int ret = 0;
	struct pcie_transfer_cmd trans;

	trans.w0.s_data0.optcode	= SMCU_SYNC_FUNC_STATE;
	trans.w1.data1				= state;
	ret = vastai_pci_send_msg(priv, 0, COMMON_HOST_TO_SMCU_CMD_BUF,  &trans, 0);
	if(ret != 0){
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,"pci_sync_state:failed\n");
		return -ENODEV;
	}

	msleep(10);

	return 0;
}

