/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/delay.h>
#include "vastai_udma_engine.h"
#include <linux/dma-buf.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/version.h>
#include <linux/bitops.h>
#include <linux/workqueue.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include "vastai_dmabuf.h"
#include "hw/va_dma_core.h"
#include "hw/pcie_2_dma.h"
#include "hw/smmu.h"
#include "common/hw_config.h"


int vastai_pci_dma_transfer_sync(void *pci_info, int die_index,
				 union core_bitmap core_id,
				 struct vastai_dmadesc desc[], int desc_num)
{
#ifdef PCIE_DMA_ENABLE
	return vastai_pci_dma_transfer_sync_pid(pci_info, die_index, core_id,
			desc, desc_num, -1);
#else
	return vastai_sdma_transfer_sync_pid(pci_info, die_index, core_id,
			desc, desc_num, 0);
#endif
}

static int fill_dma_buf(void *priv, struct va_dma_desc_elem *elem,
			 const struct vastai_dmadesc *desc)
{
	if (!desc->is_src_dma_addr) {
		 if (iommu_is_enable(priv)) {
			elem->buf = vastai_dma_buf_sg_get(priv,
						 desc->dma_lenth);
		} else {
			elem->buf = vastai_mempool_get(priv,
						 desc->dma_lenth);
		}

		if (!elem->buf) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
						"%s udma alloc failed\n", __FUNCTION__);
			return -ENOMEM;
		}
		elem->cpu_addr =
			elem->buf->dma_bus_addr;
	} else {
		elem->cpu_addr =
			 desc->host_addr.dma_addr;
	}

#ifndef DIAG_TEST_MODE
    if(desc->is_host_to_dev)
	{
        elem->size = round_up(desc->dma_lenth, VASTAI_DMA_ALIGN_SIZE);
	}
	else
#endif
	{
        elem->size = desc->dma_lenth;
	}
	elem->dev_addr = desc->dev_addr;

	return 0;
}

static void clean_dma_buf(void *pcie_dev, struct va_dma_desc_elem *elem,
			   u32 is_src_dma_addr)
{
	if (!is_src_dma_addr) {
		if (iommu_is_enable(pcie_dev))
			vastai_dma_buf_sg_put(pcie_dev, elem->buf);
		else
			vastai_mempool_put(pcie_dev, elem->buf);
	}
	memset(elem, 0, sizeof(struct va_dma_desc_elem));
}

static int vastai_sdma_common_mem_copy(bool is_host_to_dev, struct vastai_dmadesc *cur_desc,
						struct va_dma_desc_elem *elem,
						 void *pci_info,
						 int i, int die_index)
{
	int ret = 0;

	if (!(cur_desc[i].is_src_dma_addr)) {
		if (cur_desc[i].is_src_not_user_mem) {
			if(is_host_to_dev)
				memcpy(elem[i].buf->vir, cur_desc[i].host_addr.vir_addr, cur_desc[i].dma_lenth);
			else
				memcpy(cur_desc[i].host_addr.vir_addr, elem[i].buf->vir, cur_desc[i].dma_lenth);
		} else {
			if(is_host_to_dev)
				ret = vastai_pci_copy_from_user(elem[i].buf->vir, cur_desc[i].host_addr.vir_addr, cur_desc[i].dma_lenth);
			else
				ret = vastai_pci_copy_to_user(cur_desc[i].host_addr.vir_addr, elem[i].buf->vir, cur_desc[i].dma_lenth);
			if (ret) {
				VASTAI_PCI_ERR(
					pci_info,
					vastai_pci_get_die_id(
					   pci_info, die_index),
					"sdma_common_mem_copy copy %s user failed dma[0x%p] user[0x%p]\n",
					is_host_to_dev ? "from" : "to",
					elem[i].buf->vir,
					(void *)cur_desc[i].host_addr.vir_addr);
				return ret;
			}
		}
	}

   return ret;
}

static void vastai_sdma_complete(void *param)
{
	struct completion *completion = param;

	if (completion)
	complete(completion);
	return;
}

int vastai_pci_dma_transfer_sync_pid(void *priv, int die_index,
				 union core_bitmap core_id,
				 struct vastai_dmadesc desc[], int desc_num, int pid)
{
	va_dma_t *dma;
	struct va_dma_descriptor dma_desc;
	struct va_dma_desc_elem *elem;
	struct completion completion;
	bool host_to_dev = desc[0].is_host_to_dev;

	int ret = 0;
	u8 dir;
	int i;

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"pci_dma_transfer_sync_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	elem = kvmalloc(sizeof(*elem) * desc_num, GFP_KERNEL);
	if (!elem) {
		ret = -ENOMEM;
		goto dma_free;
	}

	init_completion(&completion);

	if(desc[0].is_host_to_dev)
	{
		dir = DMA_MEM_TO_DEV;
	}
	else
	{
		dir = DMA_DEV_TO_MEM;
	}

	dma_desc.direction = dir;
	INIT_LIST_HEAD(&dma_desc.elem_head);
	dma_desc.elem_num = 0;
	dma_desc.callback = vastai_sdma_complete;
	dma_desc.callback_param = &completion;

	for (i = 0; i < desc_num; i++)
	{
		elem[i].desc_type = 0;
		ret = fill_dma_buf(priv, &(elem[i]), &(desc[i]));
		if (ret) {
			desc_num = i;
			goto CLEAN_DMA_BUF;
		}

		if (host_to_dev){
			ret = vastai_sdma_common_mem_copy(host_to_dev,
				desc, elem, priv, i, die_index);
			if(ret)
				goto CLEAN_DMA_BUF;
		}
		dma_desc.elem_num = dma_desc.elem_num + 1;
		list_add_tail(&(elem[i].node), &(dma_desc.elem_head));
	}

	ret = va_dma_desc_submit(dma, &dma_desc);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_dma_transfer_sync_pid: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

	ret = wait_for_completion_interruptible(&completion);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "pci_dma_transfer_sync_pid: wait completion error %d\n", ret);
		goto dma_free;
	}

	for (i = 0; i <  desc_num; i++) {
		if (!host_to_dev){
			ret = vastai_sdma_common_mem_copy(host_to_dev,
				desc, elem, priv, i, die_index);
			if(ret)
				goto CLEAN_DMA_BUF;
		}
		clean_dma_buf(priv, &(elem[i]),
			      desc->is_src_dma_addr);
	}

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "PCIE DMA transfer completoin!\n");
	kvfree(elem);

	return ret;
CLEAN_DMA_BUF:
	for (i = 0; i < dma_desc.elem_num; i++) {
		clean_dma_buf(priv, &(elem[i]),
				 desc->is_src_dma_addr);
	}
	kvfree(elem);

dma_free:
	va_dma_free(priv, dma);
	
	return ret;
}

int vastai_gfx_video_share_sync_pid(void *priv, int pid, struct dma_buf *dmabuf, int dir, u64 dev_addr)
{
	struct sg_table *table;
	struct scatterlist *sg;
	struct dma_buf_attachment *attachment;
	int i;
	u64 dev_pa = dev_addr;
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem *elem;
	int ret, total_desc_num, direction, src_mmu = 0, dst_mmu = 0;
	struct completion completion;
	struct vastai_pci_info *pci_info = priv;

	attachment = dma_buf_attach(dmabuf, &(pci_info->dev->dev));
	table = dma_buf_map_attachment(attachment, DMA_BIDIRECTIONAL);

	sg = table->sgl;
	total_desc_num = table->nents;

	if(dir)
	{
		direction = DMA_MEM_TO_DEV;
	}
	else
	{
		direction = DMA_DEV_TO_MEM;
	}

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"gfx_video_share_sync_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	init_completion(&completion);
	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));
	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.direction = direction;
	sdma_desc.callback = vastai_sdma_complete;
	sdma_desc.callback_param = &completion;
	sdma_desc.src_mmu = src_mmu;
	sdma_desc.dest_mmu = dst_mmu;
	sdma_desc.pid = pid;
	sdma_desc.desc_loc = 0;
	sdma_desc.priv = priv;
	sdma_desc.elem_num = total_desc_num;
	sdma_desc.d2d = 0;
	sdma_desc.sync_tail = 0;

	elem = kvmalloc(sizeof(*elem) * total_desc_num, GFP_KERNEL);
	if (!elem) {
		ret = -ENOMEM;
		goto dma_free;
	}

	for (i = 0; i < total_desc_num; i++) {
		elem[i].desc_type = 0;
		elem[i].cpu_addr = sg_dma_address(sg);
		elem[i].size = sg_dma_len(sg);
		elem[i].dev_addr = dev_pa;

		dev_pa += sg_dma_len(sg);
		sg = sg_next(sg);
		list_add_tail(&(elem[i].node), &(sdma_desc.elem_head));		
	}

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "gfx_video_share_sync_pid: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

	ret = wait_for_completion_interruptible(&completion);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "gfx_video_share_sync_pid: wait completion error %d\n", ret);
		mdelay(2000);
		goto dma_free;
	}

dma_free:
	va_dma_free(priv, dma);
	kvfree(elem);
	return ret;
}

int vastai_sdma_transfer_sync_pid(void *priv, int die_index,
				union core_bitmap core_id,
				struct vastai_dmadesc desc[], int desc_num, int pid)
{
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem *elem;
	bool host_to_dev = desc[0].is_host_to_dev;
	int dma_transfer_total = 0;
	int total_desc_num = desc_num;
	int ret = 0;
	u8 dir;
	int src_mmu = 0;
	int dst_mmu = 0;
	int i, remain_num;
	struct vastai_dmadesc *cur_desc = desc;
	struct completion completion;
	init_completion(&completion);

#ifdef VASTAI_PERFORMANCE_TEST
	int size = 0;
#endif

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"sdma_transfer_sync_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	if(desc[0].is_host_to_dev)
	{
		dir = DMA_MEM_TO_DEV;
	}
	else
	{
		dir = DMA_DEV_TO_MEM;
	}

	while(total_desc_num > dma_transfer_total)
	{
		remain_num = total_desc_num - dma_transfer_total;
		
		if (remain_num > VASTAI_DMA_SUBMIT_LIMIT)
			remain_num = VASTAI_DMA_SUBMIT_LIMIT;

		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
						"sdma_transfer_sync_pid: push total %d elem, is_src_dma_addr:%d, is_not_usr_mem:%d\n", remain_num, cur_desc->is_src_dma_addr, cur_desc->is_src_not_user_mem);

		memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));
		INIT_LIST_HEAD(&sdma_desc.elem_head);
		sdma_desc.direction = dir;
		sdma_desc.elem_num = 0;
		sdma_desc.callback = vastai_sdma_complete;
		sdma_desc.callback_param = &completion;
		sdma_desc.src_mmu = src_mmu;
		sdma_desc.dest_mmu = dst_mmu;
		sdma_desc.pid = pid;
		sdma_desc.desc_loc = 0;
		sdma_desc.priv = priv;
		sdma_desc.elem_num = remain_num;
		sdma_desc.d2d = 0;
		sdma_desc.sync_tail = 0;
		sdma_desc.priority = 0;

		elem = kvmalloc(sizeof(*elem) * remain_num, GFP_KERNEL);
		if (!elem) {
			ret = -ENOMEM;
			goto dma_free;
		}

		for (i = 0; i < remain_num; i++)
		{
			elem[i].desc_type = 0;
			ret = fill_dma_buf(priv, &(elem[i]), &(cur_desc[i]));
			if (ret) {
				remain_num = i;
				goto dma_free;
			}

			if (host_to_dev){
				ret = vastai_sdma_common_mem_copy(host_to_dev,
					cur_desc, elem, priv, i, die_index);
				if(ret)
					goto CLEAN_DMA_BUF;
			}
			
			list_add_tail(&(elem[i].node), &(sdma_desc.elem_head));
#ifdef VASTAI_PERFORMANCE_TEST
			size += elem[i].size;
#endif
		}

#ifdef VASTAI_PERFORMANCE_TEST
        unsigned long old_ts, now_ts;
        old_ts= ktime_get();
#endif
		ret = va_dma_desc_submit(dma, &sdma_desc);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "sdma_transfer_sync_pid: push_dma_desc error %d\n", ret);
			goto CLEAN_DMA_BUF;
		}

		ret = wait_for_completion_interruptible(&completion);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "sdma_transfer_sync_pid: wait completion error %d\n", ret);
			mdelay(2000);
			sdma_desc.priv = NULL;
			goto CLEAN_DMA_BUF;
		}

#ifdef VASTAI_PERFORMANCE_TEST
        now_ts= ktime_get();  
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,"%dk : %06lu(us)\n", size/1024, (now_ts - old_ts)/1000);
#endif
		for (i = 0; i <  remain_num; i++) {
			if (!host_to_dev){
				ret = vastai_sdma_common_mem_copy(host_to_dev,
					cur_desc, elem, priv, i, die_index);
				if(ret)
					goto CLEAN_DMA_BUF;
			}
			clean_dma_buf(priv, &(elem[i]),
				      cur_desc->is_src_dma_addr);
		}

		reinit_completion(&completion);
		kvfree(elem);
		dma_transfer_total += remain_num;
		cur_desc += VASTAI_DMA_SUBMIT_LIMIT;
	}

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "%s\n", "**SDMA Completion**");
	va_dma_free(priv, dma);

	return ret;
CLEAN_DMA_BUF:
	for (i = 0; i < remain_num; i++) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "i:%d, dir: %d, desc_type:%d, cpu_addr:0x%llx, device_addr:0x%llx, size:0x%x\n", 
								i, dir, elem[i].desc_type, elem[i].cpu_addr, elem[i].dev_addr, elem[i].size);
	
		clean_dma_buf(priv, &(elem[i]),
				 cur_desc->is_src_dma_addr);
	}
	dump_current_fn_hw_cfg_info(priv);
	kvfree(elem);

dma_free:
	va_dma_free(priv, dma);
	
	return ret;
}

int vastai_sdma_transfer_asyn_pid(void *priv, int die_index,
				union core_bitmap core_id,
				struct vastai_dmadesc desc[], int desc_num, int pid)
{
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem *elem;
	bool host_to_dev = desc[0].is_host_to_dev;
	int dma_transfer_total = 0;
	int total_desc_num = desc_num;
	int ret = 0;
	u8 dir;
	int src_mmu = 0;
	int dst_mmu = 0;
	int i, remain_num;
	struct vastai_dmadesc *cur_desc = desc;

#ifdef VASTAI_PERFORMANCE_TEST
	int size = 0;
#endif

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"sdma_transfer_asyn_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	if(desc[0].is_host_to_dev)
	{
		dir = DMA_MEM_TO_DEV;
	}
	else
	{
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"sdma_transfer_asyn_pid: reading from chip does not support asynchrony\n");
		return -EAGAIN;
	}

	while(total_desc_num > dma_transfer_total)
	{
		remain_num = total_desc_num - dma_transfer_total;

		if (remain_num > VASTAI_DMA_SUBMIT_LIMIT)
			remain_num = VASTAI_DMA_SUBMIT_LIMIT;

		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
				       "sdma_transfer_asyn_pid: push total %d elem", remain_num);

		memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));
		INIT_LIST_HEAD(&sdma_desc.elem_head);
		sdma_desc.direction = dir;
		sdma_desc.elem_num = 0;
		sdma_desc.callback = NULL;
		sdma_desc.callback_param = NULL;
		sdma_desc.src_mmu = src_mmu;
		sdma_desc.dest_mmu = dst_mmu;
		sdma_desc.pid = pid;
		sdma_desc.desc_loc = 0;
		sdma_desc.priv = priv;
		sdma_desc.elem_num = remain_num;
		sdma_desc.d2d = 0;
		sdma_desc.sync_tail = 1;

		elem = kvmalloc(sizeof(*elem) * remain_num, GFP_KERNEL);
		if (!elem) {
			ret = -ENOMEM;
			goto dma_free;
		}

		for (i = 0; i < remain_num; i++)
		{
			elem[i].desc_type = 0;
			ret = fill_dma_buf(priv, &(elem[i]), &(cur_desc[i]));
			if (ret) {
				remain_num = i;
				goto dma_free;
			}

			ret = vastai_sdma_common_mem_copy(host_to_dev,
				cur_desc, elem, priv, i, die_index);
			if(ret)
				goto CLEAN_DMA_BUF;
	
			list_add_tail(&(elem[i].node), &(sdma_desc.elem_head));
#ifdef VASTAI_PERFORMANCE_TEST
			size += elem[i].size;
#endif
		}

#ifdef VASTAI_PERFORMANCE_TEST
        unsigned long old_ts, now_ts;
        old_ts= ktime_get();
#endif
		ret = va_dma_desc_submit(dma, &sdma_desc);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				       "sdma_transfer_asyn_pid: push_dma_desc error %d\n", ret);
			goto CLEAN_DMA_BUF;
		}

#ifdef VASTAI_PERFORMANCE_TEST
        now_ts= ktime_get();  
        VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,"%dk : %06lu(us)\n", size/1024, (now_ts - old_ts)/1000);
#endif
		for (i = 0; i <  remain_num; i++) {
			clean_dma_buf(priv, &(elem[i]),
				      cur_desc->is_src_dma_addr);
		}

		kvfree(elem);
		dma_transfer_total += remain_num;
		cur_desc += VASTAI_DMA_SUBMIT_LIMIT;
	}

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "sdma_transfer_asyn_pid() transfer completoin!\n");
	va_dma_free(priv, dma);

	return ret;
CLEAN_DMA_BUF:
	for (i = 0; i < remain_num; i++) {
		clean_dma_buf(priv, &(elem[i]),
				 cur_desc->is_src_dma_addr);
	}
	kvfree(elem);

dma_free:
	va_dma_free(priv, dma);
	
	return ret;
}

int vastai_dev2dev_sdma_sync_pid(void *priv, int die_index,
				u64 src_addr, u64 dst_addr, u32 length, int pid)
{
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem elem;
	struct completion completion;
	int ret = 0;
	int src_mmu = 0;
	int dst_mmu = 0;

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"dev2dev_sdma_sync_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	init_completion(&completion);

	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));
	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.direction = DMA_DEV_TO_DEV;
	sdma_desc.elem_num = 1;
	sdma_desc.callback = vastai_sdma_complete;
	sdma_desc.callback_param = &completion;
	sdma_desc.src_mmu = src_mmu;
	sdma_desc.dest_mmu = dst_mmu;
	sdma_desc.pid = pid;
	sdma_desc.desc_loc = 0;
	sdma_desc.d2d = 0;
	sdma_desc.priv = priv;
	sdma_desc.sync_tail = 0;

	elem.cpu_addr = src_addr;
	elem.dev_addr = dst_addr;
	elem.size = length;
	elem.desc_type = 0;
	list_add_tail(&(elem.node), &(sdma_desc.elem_head));

#ifdef VASTAI_PERFORMANCE_TEST
	unsigned long old_ts, now_ts;
	old_ts= ktime_get();
#endif

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "dev2dev_sdma_sync_pid: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

	ret = wait_for_completion_interruptible(&completion);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "dev2dev_sdma_sync_pid: wait completion error %d\n", ret);
		mdelay(2000);
		goto dma_free;
	}

#ifdef VASTAI_PERFORMANCE_TEST
	now_ts= ktime_get();  
	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,"%dk : %06lu(us)\n", length/1024, (now_ts - old_ts)/1000);
#endif

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "dev2dev_sdma_sync_pid() transfer completoin!\n");

	va_dma_free(priv, dma);
	return ret;
dma_free:
	va_dma_free(priv, dma);
	
	return ret;
}

int vastai_dev2dev_sdma_asyn_pid(void *priv, int die_index,
				u64 src_addr, u64 dst_addr, u32 length, int pid)
{
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem elem;
	int ret = 0;
	int src_mmu = 0;
	int dst_mmu = 0;

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"dev2dev_sdma_asyn_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));

	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.direction = DMA_DEV_TO_DEV;
	sdma_desc.elem_num = 1;
	sdma_desc.callback = NULL;
	sdma_desc.callback_param = NULL;
	sdma_desc.src_mmu = src_mmu;
	sdma_desc.dest_mmu = dst_mmu;
	sdma_desc.pid = pid;
	sdma_desc.desc_loc = 0;
	sdma_desc.d2d = 0;
	sdma_desc.priv = priv;
	sdma_desc.sync_tail = 1;

	elem.cpu_addr = src_addr;
	elem.dev_addr = dst_addr;
	elem.size = length;
	elem.desc_type = 0;
	list_add_tail(&(elem.node), &(sdma_desc.elem_head));

#ifdef VASTAI_PERFORMANCE_TEST
	unsigned long old_ts, now_ts;
	old_ts= ktime_get();
#endif

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "dev2dev_sdma_asyn_pid: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

#ifdef VASTAI_PERFORMANCE_TEST
	now_ts= ktime_get();  
	VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,"%dk : %06lu(us)\n", length/1024, (now_ts - old_ts)/1000);
#endif

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "dev2dev_sdma_asyn_pid() transfer completoin!\n");

	va_dma_free(priv, dma);
	return ret;
dma_free:
	va_dma_free(priv, dma);
	
	return ret;
}

int vastai_salve2master_sdma_sync_pid(void *priv, int dir,
				u64 src_addr, u64 dst_addr, u32 length)
{
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem elem;
	struct completion completion;
	int ret = 0;
	int src_mmu = 0;
	int dst_mmu = 0;

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"salve2master_sdma_sync_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	init_completion(&completion);
	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));

	// 0:read, 1:write
	if (dir)
	{
		sdma_desc.direction = DMA_DEV_TO_MEM;
		elem.cpu_addr = dst_addr;
		elem.dev_addr = src_addr;
	}
	else
	{
		sdma_desc.direction = DMA_MEM_TO_DEV;
		elem.cpu_addr = src_addr;
		elem.dev_addr = dst_addr;
	}

	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.elem_num = 1;
	sdma_desc.callback = vastai_sdma_complete;
	sdma_desc.callback_param = &completion;
	sdma_desc.src_mmu = src_mmu;
	sdma_desc.dest_mmu = dst_mmu;
	sdma_desc.desc_loc = 0;
	sdma_desc.d2d = 1;
	sdma_desc.priv = priv;
	sdma_desc.sync_tail = 0;

	elem.size = length;
	elem.desc_type = 0;
	list_add_tail(&(elem.node), &(sdma_desc.elem_head));

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "salve2master_sdma_sync_pid: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

	ret = wait_for_completion_interruptible(&completion);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "salve2master_sdma_sync_pid: wait completion error %d\n", ret);
		mdelay(2000);
		goto dma_free;
	}

	VASTAI_PCI_DBG(NULL, DUMMY_DIE_ID, "salve2master_sdma_sync_pid() transfer completoin!\n");

	va_dma_free(priv, dma);
	return ret;
dma_free:
	va_dma_free(priv, dma);
	
	return ret;
}

int vastai_salve2master_sdma_asyn_pid(void *priv, int dir,
				u64 src_addr, u64 dst_addr, u32 length)
{
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem elem;
	int ret = 0;
	int src_mmu = 0;
	int dst_mmu = 0;

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"salve2master_sdma_asyn_pid: dma alloc failed\n");
		return -EAGAIN;
	}

	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));
	// 0:read, 1:write
	if (dir)
	{
		sdma_desc.direction = DMA_DEV_TO_MEM;
		elem.cpu_addr = dst_addr;
		elem.dev_addr = src_addr;
	}
	else
	{
		sdma_desc.direction = DMA_MEM_TO_DEV;
		elem.cpu_addr = src_addr;
		elem.dev_addr = dst_addr;
	}

	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.elem_num = 1;
	sdma_desc.callback = NULL;
	sdma_desc.callback_param = NULL;
	sdma_desc.src_mmu = src_mmu;
	sdma_desc.dest_mmu = dst_mmu;
	sdma_desc.desc_loc = 0;
	sdma_desc.d2d = 1;
	sdma_desc.priv = priv;
	sdma_desc.sync_tail = 1;

	elem.size = length;
	elem.desc_type = 0;
	list_add_tail(&(elem.node), &(sdma_desc.elem_head));

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				   "salve2master_sdma_asyn_pid: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "salve2master_sdma_asyn_pid() transfer completoin!\n");

	va_dma_free(priv, dma);
	return ret;
dma_free:
	va_dma_free(priv, dma);
	
	return ret;
}

static dma_addr_t sg_addr(struct vastai_pci_info *pci_info,
				struct scatterlist *sg)
{
	return iommu_is_enable(pci_info) ? sg_dma_address(sg)
					: sg_phys(sg);
}

static u32 sg_len(struct vastai_pci_info *pci_info,
				struct scatterlist *sg)
{
	return iommu_is_enable(pci_info) ? sg_dma_len(sg)
					: sg->length;

}

/* sg_gather:
	find max size whicth can be gathered.
	return dma_addr and size.
*/

static int sg_gather(struct vastai_pci_info *pci_info ,
			struct vastai_sg_gather_info *sg_info,
			u32 max_gather_size)
{
	u32 len, len_next, cur_sg_remain_len;
	dma_addr_t addr, addr_next;
	struct scatterlist *sg, *sg_next_p;
	u32 i = sg_info->cur_sg_index;

	sg = sg_info->cur_sg;
	if (sg_info->cur_sg_remain_len == 0) {
		len = sg_len(pci_info, sg); 
		sg_info->gather_addr = sg_addr(pci_info, sg); 
		sg_info->gather_len = 0;
		sg_info->cur_sg_remain_len = len;
	}

	if (sg_info->cur_sg_remain_len > max_gather_size) {
		len = max_gather_size;
		addr = sg_info->gather_addr + sg_info->gather_len;
		cur_sg_remain_len = sg_info->cur_sg_remain_len - len;
	} else {
		len = sg_info->cur_sg_remain_len;
		addr = sg_info->gather_addr + sg_info->gather_len;
		while ((i + 1) < sg_info->sg_len) {
			sg_next_p = sg_next(sg);
			addr_next = sg_addr(pci_info, sg_next_p);
			len_next = sg_len(pci_info, sg_next_p);
			if (len_next == 0)
				break;
			if (addr + len == addr_next) {
				i++;
				sg = sg_next_p;
				if (len + len_next > max_gather_size) {
					cur_sg_remain_len = len + len_next -
							max_gather_size;
					len = max_gather_size;
					break;
				} else {
					len += len_next;
					cur_sg_remain_len = 0;
				}
			} else {
				i++;
				sg = sg_next_p;
				cur_sg_remain_len = 0;
				break;
			}
		}
	}
	sg_info->cur_sg = sg;
	sg_info->cur_sg_index = i;
	sg_info->gather_addr = addr;
	sg_info->gather_len = len;
	sg_info->cur_sg_remain_len = cur_sg_remain_len;

	return 0;
}

static int vastai_dma_set_desc_memset(struct vastai_dmadesc* desc,
            dma_addr_t dma_bus_addr,
            int i,
            u64 axi_addr,
            u32 done_size,
            u32 is_host_to_dev,
            u32 is_src_not_user_mem,
            u32 is_src_dma_addr,
            u32 size)
{
	int ret = 0;
	desc[i].is_host_to_dev = is_host_to_dev;
	desc[i].is_src_not_user_mem = is_src_not_user_mem;
	desc[i].is_src_dma_addr = is_src_dma_addr;
	desc[i].dev_addr = axi_addr+ done_size;

    if(size < VASTAI_MAX_MEMSET_BUF){

        desc[i].dma_lenth = size;
    }else{

        desc[i].dma_lenth = VASTAI_MAX_MEMSET_BUF;
    }

	desc[i].host_addr.dma_addr = dma_bus_addr;
	return ret;
}
int vastai_dma_sg_set_desc(struct vastai_dmadesc* desc,
            struct vastai_sg_gather_info sg_info,
            int i,
            u64 axi_addr,
            u32 done_size,
            u32 is_host_to_dev,
            u32 is_src_not_user_mem,
            u32 is_src_dma_addr)
{
	int ret=0;
	desc[i].is_host_to_dev = is_host_to_dev;
	desc[i].is_src_not_user_mem = is_src_not_user_mem;
	desc[i].is_src_dma_addr = is_src_dma_addr;
	desc[i].dev_addr = axi_addr+ done_size;
	desc[i].dma_lenth = sg_info.gather_len;
	desc[i].host_addr.dma_addr = sg_info.gather_addr;
	return ret;
}

int vastai_sdma_transfer_memset(void *pci_info,
			int die_index,
			union core_bitmap core_id,
			dma_addr_t dma_bus_addr,
			u32 is_src_dma_addr,
			u32 is_host_to_dev,
			u32 is_src_not_user_mem,
			u64 axi_addr,
			u32 size)
{
	u32 done_size = 0;
	int ret = 0;
	int i = 0;
	struct vastai_dmadesc *desc;
	u32 entry_count = DIV_ROUND_UP(size, VASTAI_MAX_MEMSET_BUF);

	desc = kvmalloc(sizeof(struct vastai_dmadesc) * entry_count,
					GFP_KERNEL);
	if (!desc){
		VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
				"sdma_transfer_memset: desc alloc failed\n");
		return -ENOMEM;
	}
	while (done_size < size) {
		if (ret)
			break;
        ret = vastai_dma_set_desc_memset(desc,
                                         dma_bus_addr,
                                         i,
                                         axi_addr,
                                         done_size,
                                         is_host_to_dev,
                                         is_src_not_user_mem,
                                         is_src_dma_addr,
                                         size);

		if (ret)
		  break;
		++i;
        if(size < VASTAI_MAX_MEMSET_BUF){

            done_size += size;
        }else{

            done_size += VASTAI_MAX_MEMSET_BUF;
        }
	}
	ret = vastai_pci_dma_transfer_sync(
                pci_info, die_index, core_id,
				desc, i);
	kvfree(desc);
	return ret;
}

int vastai_sdma_transfer_sg(void *pci_info,
			int die_index,
			union core_bitmap core_id,
			struct vastai_dma_buf *dm,
			u32 is_src_dma_addr,
			u32 is_host_to_dev,
			u32 is_src_not_user_mem,
			u64 axi_addr,
			u32 size)
{
	u32 max_gather_size;
	u32 done_size = 0;
	int ret = 0;
	int i=0;
	struct scatterlist *sg;
	struct vastai_sg_gather_info sg_info;
	struct vastai_dmadesc *desc;
	u32 entry_count = DIV_ROUND_UP(size, VASTAI_MAX_DMA_PAYLOAD);
	/*...................Optimize desc to become an array and record its size..............*/
	memset(&sg_info, 0, sizeof(struct vastai_sg_gather_info));
	sg_info.cur_sg = dm->sg_table.sgl;
	sg_info.sg_len = dm->sg_table.nents;
	sg = sg_info.cur_sg;

	desc = kvmalloc(sizeof(struct vastai_dmadesc) * entry_count,
					GFP_KERNEL);
	if (!desc){
		VASTAI_PCI_ERR(pci_info,DUMMY_DIE_ID,
				"sdma_transfer_sg: desc alloc failed\n");
		return -ENOMEM;
	}
	while (done_size < size) {
		max_gather_size = min((u32)VASTAI_MAX_DMA_PAYLOAD,
							size - done_size);
		ret = sg_gather(pci_info, &sg_info, max_gather_size);
		if (ret)
			break;
        ret=vastai_dma_sg_set_desc(desc,
                                sg_info,
                                i,
                                axi_addr,
                                done_size,
                                is_host_to_dev,
                                is_src_not_user_mem,
                                is_src_dma_addr);

		if (ret)
		  break;
		++i;
		done_size += sg_info.gather_len;
	}
	ret = vastai_pci_dma_transfer_sync(
                pci_info, die_index, core_id,
				desc, i);
	kvfree(desc);
	return ret;
}

int vastai_sdma_transfer_by_uaddr(struct vastai_pci_info *pci_info,
			int die_index,
			union core_bitmap core_id,
			u32 is_host_to_dev,
			u64 axi_addr,
			u64 user_addr,
			u32 size)
{
	struct device *dev = &(pci_info->dev->dev);
	struct sg_table *sgt;
	int num_pinned_pages = 0;
	unsigned long offset = user_addr & ((1 << PAGE_SHIFT) - 1);
	unsigned int num_pages = (size + offset + PAGE_SIZE - 1) >> PAGE_SHIFT;
	unsigned int gup_flags = 0;
	int i;
	int ret;
	struct vastai_dma_buf *dm =
		kzalloc(sizeof(struct vastai_dma_buf), GFP_KERNEL);
	if (!dm)
		return -ENOMEM;

	dm->pages = kvmalloc_array(num_pages, sizeof(struct page *),
			GFP_KERNEL | __GFP_ZERO);
	if (!dm->pages) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"kvmalloc_array failed\n");

		ret = -ENOMEM;
		goto ERROR_FREE_DM;
	}

	dm->pci_info = pci_info;
	dm->num_pages = num_pages;
	gup_flags = is_host_to_dev ? 0 : FOLL_WRITE;
	num_pinned_pages = get_user_pages_fast(
			user_addr,
			(int)num_pages,
			gup_flags,
			dm->pages);
	if (num_pinned_pages != num_pages)
	{
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"get_user_pages_fast failed: (%d - %u)\n",
			num_pinned_pages, num_pages);

		ret = -EINVAL;
		goto ERROR_FREE_PAGE;
	}

	ret = sg_alloc_table_from_pages(&dm->sg_table, dm->pages,
				dm->num_pages, 0, size, GFP_KERNEL);
	if (ret) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"sg_alloc_table_from_pages failed:%d \n", ret);

		goto ERROR_FREE_PIN_PAGE;
	}
	sgt = &dm->sg_table;
	sgt->nents = dma_map_sg(dev, sgt->sgl, sgt->orig_nents, 0);
	if (!sgt->nents) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"dma_map_sg failed:%d \n", ret);

		goto ERROR_FREE_TABLE;
	}
#if 1
	ret = vastai_sdma_transfer_sg(pci_info,
				  die_index,
				  core_id, dm,
				  1,
				  is_host_to_dev,
				  1,
				  axi_addr,
				  size);
	if (ret) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"sdma_transfer_sg failed:%d \n", ret);
	}
#endif
//	printk("transfer ok\n");

	dma_unmap_sg(dev, sgt->sgl, sgt->orig_nents, 0);
ERROR_FREE_TABLE:
	sg_free_table(&dm->sg_table);
ERROR_FREE_PIN_PAGE:
	for (i = 0; i < num_pinned_pages; i++)
		put_page(dm->pages[i]);
ERROR_FREE_PAGE:
	kvfree(dm->pages);
ERROR_FREE_DM:
	kfree(dm);

	return ret;
}

int get_vastai_pid(struct vastai_pci_info *pci_info)
{
	pid_t usr_pid = task_tgid_nr(current);
	u32 va_pid;

	if (!pci_info->iommu_en)
	{
		va_pid = va_create_pid(pci_info, usr_pid);
		if (va_pid < 0)
		{
				VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
								"%s: map vastai pid(%d) failed\n", __func__, usr_pid);
				return -ENOMEM;
		}
	}

	return va_pid;
}

int free_vastai_pid(struct vastai_pci_info *pci_info, u32 va_pid)
{
	int ret;

	if (!pci_info->iommu_en)
	{
		ret = va_kill_pid(pci_info, va_pid);
		if (ret < 0)
		{
			VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
					"%s: unmap vastai pid(%d) failed\n", __func__, va_pid);
			return -ENOMEM;
		}
	}

	return ret;
}

int vastai_sdma_transfer_by_uaddr_withmmu(struct vastai_pci_info *pci_info,
			int die_index,
			union core_bitmap core_id,
			u32 is_host_to_dev,
			u64 dev_addr,
			u64 user_addr,
			u32 size, u32 pid)
{
	struct device *dev = &(pci_info->dev->dev);
	struct sg_table *sgt;
	int i, ret, src_mmu, dst_mmu, num_pinned_pages = 0;
	int offset = user_addr & ((1 << PAGE_SHIFT) - 1);
	unsigned int num_pages = (size + offset + PAGE_SIZE - 1) >> PAGE_SHIFT;
	unsigned int gup_flags = 0;
	u64 va, end, mapping_size;
	va_dma_t *dma;
	struct scatterlist *s;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem elem;
	struct completion completion;
	struct vastai_dma_buf *dm =
		kzalloc(sizeof(struct vastai_dma_buf), GFP_KERNEL); 

	u64 total_length = 0;
	int length = 0;

	if (!dm)
		return -ENOMEM;

	dm->pages = kvmalloc_array(num_pages, sizeof(struct page *),
			GFP_KERNEL | __GFP_ZERO);
	if (!dm->pages) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID, "kvmalloc_array failed\n");
		ret = -ENOMEM;
		goto ERROR_FREE_DM;
	}

	dm->pci_info = pci_info;
	dm->num_pages = num_pages;
	gup_flags = is_host_to_dev ? 0 : FOLL_WRITE;

	num_pinned_pages = get_user_pages_fast(
			user_addr,
			(int)num_pages,
			gup_flags,
			dm->pages);
	if (num_pinned_pages != num_pages)
	{
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
			"get_user_pages_fast failed: (%d - %u)\n",
			num_pinned_pages, num_pages);

		ret = -EINVAL;
		goto ERROR_FREE_PAGE;
	}

	ret = sg_alloc_table_from_pages(&dm->sg_table, dm->pages,
				dm->num_pages, 0, size, GFP_KERNEL);
	if (ret) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"sg_alloc_table_from_pages failed:%d \n", ret);

		goto ERROR_FREE_PIN_PAGE;
	}
	sgt = &dm->sg_table; 

#if 0
	for_each_sg(sgt->sgl, s, sgt->orig_nents, i)
	{
		VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID,
				"[Alloc_page]i:%3d, va:0x%16lx, offset:0x%x, pa:0x%10llx, length:0x%6x, dma_len:0x%6x\n", 
				i, s->page_link, s->offset, s->dma_address, s->length, s->dma_length);
	}
#endif

	sgt->nents = dma_map_sg(dev, sgt->sgl, sgt->orig_nents, 0);
	if (!sgt->nents) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"dma_map_sg failed:%d \n", sgt->nents);

		goto ERROR_FREE_TABLE;
	}

	/* start to set sdma descriptor */
	dma = va_dma_alloc(pci_info, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				"sdma_transfer_by_uaddr_withmmu: dma alloc failed\n");
		return -EAGAIN;
	}

	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));

#if 0
	for_each_sg(sgt->sgl, s, sgt->orig_nents, i)
	{
		VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID,
				"[After_mapped]i:%3d, va:0x%16lx, offset:0x%x, pa:0x%10llx, length:0x%6x, dma_len:0x%6x\n", 
				i, s->page_link, s->offset, s->dma_address, s->length, s->dma_length);
	}
#endif

	if (!pci_info->iommu_en)
	{
		mutex_lock(&pci_info->trans_mutex);		

		VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID, "sdma_transfer_by_uaddr_withmmu: pid[%d] va(0x%llx) size(0x%x), num_pages(%d), offset(0x%x)\n",
                                pid, user_addr, size, num_pages, offset);
		total_length = 0;
		va = user_addr & VASTAI_PAGE_MASK;
		end = roundup(user_addr + size, VASTAI_PAGE_SIZE);
		mapping_size = end - va;

		for_each_sg(sgt->sgl, s, sgt->orig_nents, i)
		{
			if (i == sgt->orig_nents - 1)
				length = mapping_size - total_length;
			else
				length = roundup(s->dma_length, VASTAI_PAGE_SIZE);
			va_create_page_mapping(pci_info, s->dma_address&VASTAI_PAGE_MASK , va, length, pid, 1);
			total_length += length;
			va += length;
			VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID,
					"i:%3d, va:0x%llx, phy:0x%llx, length:0x%x, mapped_total_len:0x%llx\n",
					i, va, s->dma_address, length, total_length);
		}
		mutex_unlock(&pci_info->trans_mutex);
		if (is_host_to_dev)
		{
			sdma_desc.direction = DMA_MEM_TO_DEV;
			src_mmu = 1;
			dst_mmu = 0;
		}
		else
		{
			sdma_desc.direction = DMA_DEV_TO_MEM;
			src_mmu = 0;
			dst_mmu = 1;
		}
			elem.cpu_addr = user_addr;
			elem.dev_addr = dev_addr;
	}
	else
	{
		if (is_host_to_dev)
		{
			sdma_desc.direction = DMA_MEM_TO_DEV;
		}
		else
		{
			sdma_desc.direction = DMA_DEV_TO_MEM;
		}
		elem.cpu_addr = dm->sg_table.sgl[0].dma_address + offset;
		elem.dev_addr = dev_addr;

		src_mmu = 0;
		dst_mmu = 0;
	}

	init_completion(&completion);

	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.elem_num = 1;
	sdma_desc.callback = vastai_sdma_complete;
	sdma_desc.callback_param = &completion;
	sdma_desc.pid = pid;
	sdma_desc.src_mmu = src_mmu;
	sdma_desc.dest_mmu = dst_mmu;
	sdma_desc.desc_loc = 0;
	sdma_desc.d2d = 0;
	sdma_desc.priv = pci_info;
	sdma_desc.sync_tail = 0;
	sdma_desc.priority = 0;

	elem.size = size;
	elem.desc_type = 0;
	list_add_tail(&(elem.node), &(sdma_desc.elem_head));

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				   "sdma_transfer_by_uaddr_withmmu: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

	ret = wait_for_completion_interruptible(&completion);
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				   "sdma_transfer_by_uaddr_withmmu: wait completion error %d\n", ret);
		mdelay(2000);
		goto dma_free;
	}

	VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID, "sdma_transfer_by_uaddr_withmmu() transfer completoin!\n");

	if (!pci_info->iommu_en)
	{
		mutex_lock(&pci_info->trans_mutex);
		va_free_page_maping(pci_info, user_addr & VASTAI_PAGE_MASK, total_length, pid);
		mutex_unlock(&pci_info->trans_mutex);
	}

dma_free:
	va_dma_free(pci_info, dma);

	dma_unmap_sg(dev, sgt->sgl, sgt->orig_nents, 0);
ERROR_FREE_TABLE:
	sg_free_table(&dm->sg_table);
ERROR_FREE_PIN_PAGE:
	for (i = 0; i < num_pinned_pages; i++)
		put_page(dm->pages[i]);
ERROR_FREE_PAGE:
	kvfree(dm->pages);
ERROR_FREE_DM:
	kfree(dm);

	return ret;
}

static u32 add_sg_to_elem(struct vastai_pci_info *pci_info, struct sg_table *sgt, struct va_dma_descriptor *desc, 
						struct va_dma_desc_elem *elem, u32 offset, u64 dev_addr)
{
	struct scatterlist *s;
	int i;

	for_each_sg(sgt->sgl, s, sgt->nents, i)
	{
		if (i == 0 && (s->dma_length - offset) > 0){
			elem[i].size = s->dma_length - offset;
			elem[i].cpu_addr = s->dma_address + offset;
		}else if(i == sgt->nents - 1 && (s->dma_length + offset) > 0){
			elem[i].size = s->dma_length + offset;
			elem[i].cpu_addr = s->dma_address;
		}else{
			elem[i].size = s->dma_length;
			elem[i].cpu_addr = s->dma_address;
		}
		elem[i].dev_addr = dev_addr;
		desc->total_len += elem[i].size;
		dev_addr += elem[i].size;
		VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID,
				"i:%3d, dev_addr:0x%llx, cpu_addr:0x%llx, length:0x%x, total_len:0x%x\n", 
				i, elem[i].dev_addr, elem[i].cpu_addr, elem[i].size , desc->total_len);
		list_add_tail(&(elem[i].node), &(desc->elem_head));
		desc->elem_num++;	
	}

	return desc->total_len;
}

int vastai_sdma_transfer_by_sg(struct vastai_pci_info *pci_info,
			int die_index,
			union core_bitmap core_id,
			u32 is_host_to_dev,
			u64 dev_addr,
			u64 user_addr,
			u32 size)
{
	struct device *dev = &(pci_info->dev->dev);
	struct sg_table *sgt;
	int i, ret, num_pinned_pages = 0;
	int offset = user_addr & ((1 << PAGE_SHIFT) - 1);
	unsigned int num_pages = (size + offset + PAGE_SIZE - 1) >> PAGE_SHIFT;
	unsigned int gup_flags = 0;
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem *elem;
	struct completion completion;
	struct vastai_dma_buf *dm =
		kzalloc(sizeof(struct vastai_dma_buf), GFP_KERNEL); 

	if (!dm)
		return -ENOMEM;

	dm->pages = kvmalloc_array(num_pages, sizeof(struct page *),
			GFP_KERNEL | __GFP_ZERO);
	if (!dm->pages) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID, "kvmalloc_array failed\n");
		ret = -ENOMEM;
		goto ERROR_FREE_DM;
	}

	dm->pci_info = pci_info;
	dm->num_pages = num_pages;
	gup_flags = is_host_to_dev ? 0 : FOLL_WRITE;

	num_pinned_pages = get_user_pages_fast(
			user_addr,
			(int)num_pages,
			gup_flags,
			dm->pages);
	if (num_pinned_pages != num_pages){
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
			"get_user_pages_fast failed: (%d - %u)\n",
			num_pinned_pages, num_pages);

		ret = -EINVAL;
		goto ERROR_FREE_PAGE;
	}

	ret = sg_alloc_table_from_pages(&dm->sg_table, dm->pages,
				dm->num_pages, 0, size, GFP_KERNEL);
	if (ret) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"sg_alloc_table_from_pages failed:%d \n", ret);

		goto ERROR_FREE_PIN_PAGE;
	}
	sgt = &dm->sg_table; 

	sgt->nents = dma_map_sg(dev, sgt->sgl, sgt->orig_nents, 0);
	if (!sgt->nents) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"dma_map_sg failed:%d \n", sgt->nents);

		goto ERROR_FREE_TABLE;
	}

	/* start to set sdma descriptor */
	dma = va_dma_alloc(pci_info, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				"sdma_transfer_by_sg: dma alloc failed\n");
		goto DMA_FREE;
	}

	elem = kzalloc(sgt->orig_nents * sizeof(struct va_dma_desc_elem), GFP_KERNEL); 
	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));
	init_completion(&completion);

	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.direction = is_host_to_dev ? DMA_MEM_TO_DEV:DMA_DEV_TO_MEM;
	sdma_desc.elem_num = 0;
	sdma_desc.callback = vastai_sdma_complete;
	sdma_desc.callback_param = &completion;
	sdma_desc.pid = DEAFULT_VASTAI_PID_NUM;
	sdma_desc.src_mmu = 0;
	sdma_desc.dest_mmu = 0;
	sdma_desc.desc_loc = 1;
	sdma_desc.d2d = 0;
	sdma_desc.priv = pci_info;
	sdma_desc.sync_tail = 0;
	sdma_desc.priority = 0;
#ifdef CONFIG_DDK_ALLOC_MEM
    sdma_desc.desc_addr = pci_info->AllocMemFromGfxHeap(pci_info->heapPtr, sgt->orig_nents * 32);
#endif

	ret = add_sg_to_elem(pci_info, sgt, &sdma_desc, elem, offset, dev_addr);
	if(size != ret)
	{
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				"sdma_transfer_by_sg: add sg to elem failed!, size:%#x, elem_total_size:%#x\n", size, sdma_desc.total_len);
		goto ELEM_FREE;
	}

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				   "sdma_transfer_by_sg: push_dma_desc error %d\n", ret);
		goto ELEM_FREE;
	}

	ret = wait_for_completion_interruptible(&completion);
	if (ret) {
		VASTAI_PCI_ERR(pci_info, DUMMY_DIE_ID,
				   "sdma_transfer_by_sg: wait completion error %d\n", ret);
		mdelay(2000);
		goto ELEM_FREE;
	}
#ifdef CONFIG_DDK_ALLOC_MEM
    pci_info->FreeMemToGfxHeap(pci_info->heapPtr, sdma_desc.desc_addr);
#endif

	VASTAI_PCI_DBG(pci_info, DUMMY_DIE_ID, "sdma_transfer_by_sg() transfer completoin!\n");

ELEM_FREE:
	kvfree(elem);
DMA_FREE:
	va_dma_free(pci_info, dma);
	dma_unmap_sg(dev, sgt->sgl, sgt->orig_nents, 0);
ERROR_FREE_TABLE:
	sg_free_table(&dm->sg_table);
ERROR_FREE_PIN_PAGE:
	for (i = 0; i < num_pinned_pages; i++)
		put_page(dm->pages[i]);
ERROR_FREE_PAGE:
	kvfree(dm->pages);
ERROR_FREE_DM:
	kfree(dm);

	return ret;
}

int vastai_pcie_dma_raw_transfer_by_uaddr(struct vastai_pci_info *pci_info,
			int die_index,
			union core_bitmap core_id,
			u32 is_host_to_dev,
			u64 axi_addr,
			u64 user_addr,
			u32 size)
{
	struct device *dev = &(pci_info->dev->dev);
	int num_pinned_pages = 0;
	unsigned long offset = user_addr & ((1 << PAGE_SHIFT) - 1);
	unsigned int num_pages = (size + offset + PAGE_SIZE - 1) >> PAGE_SHIFT;
	unsigned int gup_flags = 0;
	int i;
	int ret;
	struct vastai_dma_buf *dm =
		kzalloc(sizeof(struct vastai_dma_buf), GFP_KERNEL);

	struct va_dma_descriptor pcie_dma_desc;
	struct va_dma_desc_elem *elem = NULL;
	unsigned int elem_num = 0;
	struct scatterlist *sg, *sg_next_p;
	u64 len, len_next;
	dma_addr_t addr, addr_next;

	if (!dm)
		return -ENOMEM;

	dm->pages = kvmalloc_array(num_pages, sizeof(struct page *),
			GFP_KERNEL | __GFP_ZERO);
	if (!dm->pages) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"kvmalloc_array failed\n");

		ret = -ENOMEM;
		goto ERROR_FREE_DM;
	}

	dm->pci_info = pci_info;
	dm->num_pages = num_pages;
	gup_flags = is_host_to_dev ? 0 : FOLL_WRITE;
	num_pinned_pages = get_user_pages_fast(
			user_addr,
			(int)num_pages,
			gup_flags,
			dm->pages);
	if (num_pinned_pages != num_pages)
	{
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"get_user_pages_fast failed: (%d - %u)\n",
			num_pinned_pages, num_pages);

		ret = -EINVAL;
		goto ERROR_FREE_PAGE;
	}

	ret = sg_alloc_table_from_pages(&dm->sg_table, dm->pages,
				dm->num_pages, offset, size, GFP_KERNEL);
	if (ret) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"sg_alloc_table_from_pages failed:%d \n", ret);

		goto ERROR_FREE_PIN_PAGE;
	}
	pcie_dma_desc.direction = is_host_to_dev ? DMA_MEM_TO_DEV : DMA_DEV_TO_MEM;
	dm->sg_table.nents = dma_map_sg(dev, dm->sg_table.sgl, dm->sg_table.orig_nents, (enum dma_data_direction)pcie_dma_desc.direction);
	if (!dm->sg_table.nents) {
		VASTAI_PCI_ERR(
			pci_info,
			vastai_pci_get_die_id(pci_info, die_index),
			"dma_map_sg failed:%d \n", ret);

		goto ERROR_FREE_TABLE;
	}

	printk("XT:sg total nents=%u, orig_nents=%u\n", dm->sg_table.nents, dm->sg_table.orig_nents);

	for_each_sg(dm->sg_table.sgl, sg, dm->sg_table.nents, i) {
		printk("XT: sg(%d) mem_addr=%llx, len=%x, offset=%x, page_link=%lx\n",
				i, sg_dma_address(sg), sg_dma_len(sg), sg->offset, sg->page_link);
	}

	INIT_LIST_HEAD(&pcie_dma_desc.elem_head);

	for_each_sg(dm->sg_table.sgl, sg, dm->sg_table.nents, i) {
		len = sg_dma_len(sg);
		addr = sg_dma_address(sg);

		if (len == 0)
			break;
	
		while ((i + 1) < dm->sg_table.nents) {
			sg_next_p = sg_next(sg);
			len_next = sg_dma_len(sg_next_p);
			addr_next = sg_dma_address(sg_next_p);
	
			if (len_next == 0)
				break;

			if (addr + len == addr_next) {
				len += len_next;
				i++;
				sg = sg_next_p;
			} else {
				break;
			}
		}

		elem = kzalloc(sizeof(struct va_dma_desc_elem), GFP_KERNEL);
		if (!elem)
		{
			printk("elem alloc error\n");
			goto ERROR_FREE_ELEM;
		}
		elem_num++;
		elem->size = len;
		elem->cpu_addr = addr;
		elem->dev_addr = axi_addr;
		pcie_dma_desc.total_len += len;
		list_add_tail(&(elem->node), &(pcie_dma_desc.elem_head));
		axi_addr += len;
	}

	printk("pcie_dma_raw_transfer_by_uaddr %d elem_num=%d\n", __LINE__, elem_num);
	if (!elem_num)
		goto ERROR_FREE_TABLE;

	pcie_dma_desc.elem_num = elem_num;
	ret = pcie2_dma_desc_submit_raw(pci_info, &pcie_dma_desc);
	if (ret)
	{
		printk("pcie dma desc submit raw failed \n");
	}
	else
	{
		printk("transfer ok\n");
	}

ERROR_FREE_TABLE:
	dma_unmap_sg(dev, dm->sg_table.sgl, dm->sg_table.orig_nents, 0);
	sg_free_table(&dm->sg_table);
ERROR_FREE_PIN_PAGE:
	for (i = 0; i < num_pinned_pages; i++)
		put_page(dm->pages[i]);
ERROR_FREE_PAGE:
	kvfree(dm->pages);
ERROR_FREE_DM:
	kfree(dm);
ERROR_FREE_ELEM:
	kfree(elem);

	return ret;
}

int vastai_sdma_fill_data(struct vastai_pci_info *priv,
			int die_index,
			union core_bitmap core_id,
			u64 axi_addr,
			u64 data,
			u32 size)
{
	va_dma_t *dma;
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem *elem;
	struct completion completion;
	int ret = 0;
	int src_mmu = 0;
	int dst_mmu = 0;
	int pid = 0;

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "sdma_fill_data: dma alloc failed\n");
		return -EAGAIN;
	}

	elem = kvmalloc(sizeof(*elem), GFP_KERNEL);
	if (!elem) {
		ret = -ENOMEM;
		goto dma_free;
	}

	init_completion(&completion);
	memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));

	sdma_desc.direction = DMA_MEM_TO_DEV;
	INIT_LIST_HEAD(&sdma_desc.elem_head);
	sdma_desc.elem_num = 1;
	sdma_desc.callback = vastai_sdma_complete;
	sdma_desc.callback_param = &completion;
	sdma_desc.src_mmu = src_mmu;
	sdma_desc.dest_mmu = dst_mmu;
	sdma_desc.pid = pid;
	sdma_desc.desc_loc = 0;
	sdma_desc.d2d = 0;
	sdma_desc.priv = priv;
	sdma_desc.sync_tail = 0;

	elem->cpu_addr = data;
	elem->desc_type = 2;
	elem->dev_addr = axi_addr;
	elem->size = size;
	list_add_tail(&(elem->node), &(sdma_desc.elem_head));

	ret = va_dma_desc_submit(dma, &sdma_desc);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "sdma_fill_data: push_dma_desc error %d\n", ret);
		goto dma_free;
	}

	ret = wait_for_completion_interruptible(&completion);
	if (ret) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
			       "sdma_fill_data: wait completion error %d\n", ret);
		mdelay(2000);
		goto dma_free;
	}

	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "sdma_fill_data transfer completoin!\n");

dma_free:
	va_dma_free(priv, dma);
	return ret;
}

struct vastai_dma_pos_info {
	u32 current_channel;
	u32 current_line;
	u32 current_line_pos;
	u32 next_dst_offset;
	int complete;
};

/**
 * ��channel buf ������ݰ��˵�һ�� dma buf�С���ͨ��ret val�� ʹ�Լ�����һ�εĽ���м��䡣
 */
static int vastai_dma_copy_from_user(struct vastai_channel_buf *channel_buf,
					 int channel_num,
					 struct vastai_dma_buf *buf,
					 struct vastai_dma_pos_info *meta)
{
	int channel_i = meta->current_channel;
	u32 line_i = meta->current_line;
	int ret = 0;
	u32 dst_done_size = meta->next_dst_offset;
	u32 src_done_size =
		(channel_buf[meta->current_channel].src_width_offset +
		 channel_buf[meta->current_channel].width) *
			meta->current_line +
		meta->current_line_pos;

	if (meta->complete == 1)
		return 0;

	if (meta->current_line_pos) {
		u32 real_size;
		if (channel_buf[channel_i].width < meta->current_line_pos)
			return -EINVAL;
		real_size =
			channel_buf[channel_i].width - meta->current_line_pos;
		ret = vastai_pci_copy_from_user(buf->vir + dst_done_size,
					 (void*) (channel_buf[channel_i].user_buf +
						 src_done_size),
					 real_size);
		if (ret)
			return ret;

		dst_done_size +=
			(real_size + channel_buf[channel_i].dst_width_offset);
		src_done_size +=
			(real_size + channel_buf[channel_i].src_width_offset);
		line_i += 1;
	}

	for (; channel_i < channel_num; channel_i++) {
		for (; line_i < channel_buf[channel_i].high; line_i++) {
			if ((channel_buf[channel_i].width +
				 channel_buf[channel_i].dst_width_offset) >
				(buf->size - dst_done_size)) {
				u32 real_size =
					min(channel_buf[channel_i].width,
						buf->size - dst_done_size);
				u32 line_remain = channel_buf[channel_i].width -
						  real_size;

				ret = vastai_pci_copy_from_user(
					buf->vir + dst_done_size,
					(void*) (channel_buf[channel_i].user_buf +
						src_done_size),
					real_size);
				if (ret)
					return ret;

				dst_done_size += real_size;
				meta->current_channel = channel_i;
				meta->current_line_pos =
					line_remain ? real_size : 0;
				meta->next_dst_offset =
					line_remain ?
						0 :
						(channel_buf[channel_i]
							 .dst_width_offset -
						 (buf->size - dst_done_size));
				meta->current_line = meta->next_dst_offset ?
								 ++line_i :
								 line_i;
				return 0;
			}
			ret = vastai_pci_copy_from_user(buf->vir + dst_done_size,
						 (void*) (channel_buf[channel_i].user_buf +
							 src_done_size),
						 channel_buf[channel_i].width);
			if (ret)
				return ret;

			dst_done_size +=
				(channel_buf[channel_i].width +
				 channel_buf[channel_i].dst_width_offset);
			src_done_size +=
				(channel_buf[channel_i].width +
				 channel_buf[channel_i].src_width_offset);
		}
		line_i = 0;
		src_done_size = 0;
		dst_done_size += (channel_buf[channel_i].width +
				  channel_buf[channel_i].dst_width_offset) *
				 channel_buf[channel_i].dst_high_offset;
		if (dst_done_size > buf->size && channel_i < channel_num) {
			meta->current_channel = ++channel_i;
			meta->current_line = 0;
			meta->current_line_pos = 0;
			meta->next_dst_offset = dst_done_size - buf->size;
			return 0;
		}
	}
	/* all is done */
	meta->current_channel = 0;
	meta->current_line = 0;
	meta->current_line_pos = 0;
	meta->next_dst_offset = 0;
	meta->complete = 1;
	return 0;
}

int vastai_dma_video(void *priv, int die_index, union core_bitmap core_id,
						 struct vastai_channel_buf *channel_buf, int channel_num,
						 u64 dev_addr, int pid)
{
	unsigned long channel_buf_total_size = 0;
	int ret, dma_num, i = 0, src_mmu = 0, dst_mmu = 0;
	unsigned long dma_transfer_total_len = 0;
	struct vastai_dma_pos_info meta = { 0 };
	struct va_dma_descriptor sdma_desc;
	struct va_dma_desc_elem *elem = NULL;
	va_dma_t *dma;

	dma = va_dma_alloc(priv, 0, VA_DMA_ATTR_DIR_MEM_TO_DEV | VA_DMA_ATTR_DIR_DEV_TO_MEM);
	if (!dma) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"dma_video: dma alloc failed\n");
		return -EAGAIN;
	}

	memset(&meta, 0, sizeof(struct vastai_dma_pos_info));
	
	for (i = 0; i < channel_num; i++) {
		channel_buf_total_size +=
		(channel_buf[i].width +
		 channel_buf[i].dst_width_offset) *
		(channel_buf[i].high + channel_buf[i].dst_high_offset);
	}

	//Align the image data size to 128B
	channel_buf_total_size = DIV_ROUND_UP(channel_buf_total_size, VASTAI_DMA_ALIGN_SIZE) * VASTAI_DMA_ALIGN_SIZE;

	while (channel_buf_total_size - dma_transfer_total_len) 
	{
		unsigned long remain_size = channel_buf_total_size - dma_transfer_total_len;
		struct completion completion;
		init_completion(&completion);

		dma_num = DIV_ROUND_UP(remain_size, VASTAI_MAX_DMA_BUF);
		if (dma_num > 4)
			dma_num = 4;

		VASTAI_PCI_DBG(priv, DUMMY_DIE_ID,
			       "dma_video: transfer dma num:%d\n", dma_num);

		memset(&sdma_desc, 0, sizeof(struct va_dma_descriptor));
		INIT_LIST_HEAD(&sdma_desc.elem_head);
		sdma_desc.direction = DMA_MEM_TO_DEV;
		sdma_desc.elem_num = 0;
		sdma_desc.callback = vastai_sdma_complete;
		sdma_desc.callback_param = &completion;
		sdma_desc.src_mmu = src_mmu;
		sdma_desc.dest_mmu = dst_mmu;
		sdma_desc.pid = pid;
		sdma_desc.desc_loc = 0;
		sdma_desc.priv = priv;
		sdma_desc.elem_num = dma_num;
		sdma_desc.sync_tail = 0;

		elem = kvmalloc(sizeof(struct va_dma_desc_elem) * dma_num,GFP_KERNEL);
		if (!elem) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
				"dma_video elem alloc failed\n");
			ret = -ENOMEM;
			goto UP_DMA_SEM;
		}

		for (i = 0; i < dma_num; i++) {
			u32 dma_len = min(remain_size, (unsigned long)VASTAI_MAX_DMA_BUF);
			if (iommu_is_enable(priv)) {
				elem[i].buf = vastai_dma_buf_sg_get(priv,
						dma_len);
			} else {
				elem[i].buf = vastai_mempool_get(priv,dma_len);
			}
			if (!elem[i].buf) {
				dma_num = i;
				ret = -ENOMEM;
				goto CLEAN_DMA_BUF;
			}

			ret = vastai_dma_copy_from_user(channel_buf, channel_num, elem[i].buf, &meta);
			if (ret) {
				/* TODO ERROR HANDLE */
				VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					"dma_video dma_copy_from_user failed, err:%d\n", ret);
				dma_num = i;
				goto CLEAN_DMA_BUF;
			}
			elem[i].cpu_addr = elem[i].buf->dma_bus_addr;
			elem[i].size = dma_len;
			elem[i].dev_addr = dev_addr + dma_transfer_total_len;
			elem[i].desc_type = 0;

			list_add_tail(&(elem[i].node),
						&(sdma_desc.elem_head));
			remain_size -= dma_len;
			dma_transfer_total_len += dma_len;
		}

		ret = va_dma_desc_submit(dma, &sdma_desc);
		if (ret) {
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					   "dma_video: push_dma_desc error %d\n", ret);
			goto CLEAN_DMA_BUF;
		}

		ret = wait_for_completion_interruptible(&completion);
		if (ret) {
			mdelay(2000);
			VASTAI_PCI_ERR(priv, DUMMY_DIE_ID,
					   "dma_video: wait completion error %d\n", ret);
			goto CLEAN_DMA_BUF;
		}

		for (i = 0; i < dma_num; i++) {
			clean_dma_buf(priv, &(elem[i]), 0);
		}
		kvfree(elem);
		elem = NULL;
	}
	VASTAI_PCI_DBG(priv, DUMMY_DIE_ID, "dma_video, video dma stride ok!\n");

	va_dma_free(priv, dma);

	return 0;
CLEAN_DMA_BUF:
	for (i = 0; i < dma_num; i++) {
		VASTAI_PCI_ERR(priv, DUMMY_DIE_ID, "i:%d, desc_type:%d, cpu_addr:0x%llx, device_addr:0x%llx, size:0x%x\n", 
								i, elem[i].desc_type, elem[i].cpu_addr, elem[i].dev_addr, elem[i].size);
	
		clean_dma_buf(priv, &(elem[i]), 0);
	}
	dump_current_fn_hw_cfg_info(priv);
	kvfree(elem);

	va_dma_free(priv, dma);
UP_DMA_SEM:
	return ret;
}
