/*************************************************************************/ /*!
@File
@Title          System Configuration
@Copyright      Copyright (c) Imagination Technologies Ltd. All Rights Reserved
@Description    System Configuration functions
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/

#include <linux/version.h>
#include <linux/types.h>

#include "sysinfo.h"

#include "pvrsrv.h"
#include "pvrsrv_device.h"
#include "rgxdevice.h"
#include "syscommon.h"
#include "physmem.h"

#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>

#include "vastai_pcie_public.h"
#include "vastai_pcie_export.h"

/* Must be consecutive and start from 0 */
#define PHY_HEAP_CARD_GPU 0
#define PHY_HEAP_CARD_EXT 1
/* QiShao: dedicated FW tuningmem with UMA heap type */
#if defined(SUPPORT_DEDICATED_UMA_FW_TUNINGMEM)
#define PHY_HEAP_FW_TUNIG 2
#define PHY_HEAP_SYSTEM   3
#elif defined(SUPPORT_DEDICATED_UMA_FW_MAINMEM)
#define PHY_HEAP_FW_MAIN  2
#define PHY_HEAP_SYSTEM   3
#else
#define PHY_HEAP_SYSTEM   2
#endif

/* QiShao: dedicated FW tuning/main mem with UMA heap type */
#if defined(SUPPORT_DEDICATED_UMA_FW_TUNINGMEM) || defined(SUPPORT_DEDICATED_UMA_FW_MAINMEM)
#define PHY_HEAP_LMA_NUM  3
#define PHY_HEAP_HYBRID_NUM 4
#else
#define PHY_HEAP_LMA_NUM 2
#define PHY_HEAP_HYBRID_NUM 3
#endif


#define SYS_RGX_ACTIVE_POWER_LATENCY_MS (10)


static void TCLocalCpuPAddrToDevPAddr(IMG_HANDLE hPrivData,
				      IMG_UINT32 ui32NumOfAddr,
				      IMG_DEV_PHYADDR *psDevPAddr,
				      IMG_CPU_PHYADDR *psCpuPAddr);

static void TCLocalDevPAddrToCpuPAddr(IMG_HANDLE hPrivData,
				      IMG_UINT32 ui32NumOfAddr,
				      IMG_CPU_PHYADDR *psCpuPAddr,
				      IMG_DEV_PHYADDR *psDevPAddr);

static PHYS_HEAP_FUNCTIONS gsLocalPhysHeapFuncs = {
	.pfnCpuPAddrToDevPAddr = TCLocalCpuPAddrToDevPAddr,
	.pfnDevPAddrToCpuPAddr = TCLocalDevPAddrToCpuPAddr,
};

static void TCHostCpuPAddrToDevPAddr(IMG_HANDLE hPrivData,
				     IMG_UINT32 ui32NumOfAddr,
				     IMG_DEV_PHYADDR *psDevPAddr,
				     IMG_CPU_PHYADDR *psCpuPAddr);

static void TCHostDevPAddrToCpuPAddr(IMG_HANDLE hPrivData,
				     IMG_UINT32 ui32NumOfAddr,
				     IMG_CPU_PHYADDR *psCpuPAddr,
				     IMG_DEV_PHYADDR *psDevPAddr);

static PHYS_HEAP_FUNCTIONS gsHostPhysHeapFuncs = {
	.pfnCpuPAddrToDevPAddr = TCHostCpuPAddrToDevPAddr,
	.pfnDevPAddrToCpuPAddr = TCHostDevPAddrToCpuPAddr,
};

typedef struct _SYS_DATA_ SYS_DATA;

struct _SYS_DATA_ {
	struct platform_device *pdev;

	struct va_platform_data *pdata;

	struct resource *registers;
};

static IMG_CHAR *GetDeviceVersionString(SYS_DATA *psSysData)
{
	return "vastai porting img ddk to sg100";
}

static void TCLocalCpuPAddrToDevPAddr(IMG_HANDLE hPrivData,
				      IMG_UINT32 ui32NumOfAddr,
				      IMG_DEV_PHYADDR *psDevPAddr,
				      IMG_CPU_PHYADDR *psCpuPAddr)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig = (PVRSRV_DEVICE_CONFIG *)hPrivData;
	IMG_UINT32 ui32Idx;

	for (ui32Idx = 0; ui32Idx < ui32NumOfAddr; ui32Idx++) {
		psDevPAddr[ui32Idx].uiAddr = psCpuPAddr[ui32Idx].uiAddr -
			psDevConfig->pasPhysHeaps[PHY_HEAP_CARD_GPU]
				.sStartAddr.uiAddr +
			psDevConfig->pasPhysHeaps[PHY_HEAP_CARD_GPU]
				.sCardBase.uiAddr;
	}
}

static void TCLocalDevPAddrToCpuPAddr(IMG_HANDLE hPrivData,
				      IMG_UINT32 ui32NumOfAddr,
				      IMG_CPU_PHYADDR *psCpuPAddr,
				      IMG_DEV_PHYADDR *psDevPAddr)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig = (PVRSRV_DEVICE_CONFIG *)hPrivData;
	IMG_UINT32 ui32Idx;

	for (ui32Idx = 0; ui32Idx < ui32NumOfAddr; ui32Idx++) {
		psCpuPAddr[ui32Idx].uiAddr = psDevPAddr[ui32Idx].uiAddr -
			psDevConfig->pasPhysHeaps[PHY_HEAP_CARD_GPU]
				.sCardBase.uiAddr +
			psDevConfig->pasPhysHeaps[PHY_HEAP_CARD_GPU]
				.sStartAddr.uiAddr;
	}
}

#define SG100_HOST_MEM_DEV_BASE 0x6000000000LLU //384GB

static void TCHostCpuPAddrToDevPAddr(IMG_HANDLE hPrivData,
				     IMG_UINT32 uiNumOfAddr,
				     IMG_DEV_PHYADDR *psDevPAddr,
				     IMG_CPU_PHYADDR *psCpuPAddr)
{
	IMG_UINT32 ui32Idx;

	for (ui32Idx = 0; ui32Idx < uiNumOfAddr; ui32Idx++) {
		PVR_ASSERT(psCpuPAddr[ui32Idx].uiAddr < SG100_HOST_MEM_DEV_BASE);

		psDevPAddr[ui32Idx].uiAddr = psCpuPAddr[ui32Idx].uiAddr
					     + SG100_HOST_MEM_DEV_BASE;
	}
}

static void TCHostDevPAddrToCpuPAddr(IMG_HANDLE hPrivData,
				     IMG_UINT32 uiNumOfAddr,
				     IMG_CPU_PHYADDR *psCpuPAddr,
				     IMG_DEV_PHYADDR *psDevPAddr)
{
	IMG_UINT32 ui32Idx;

	for (ui32Idx = 0; ui32Idx < uiNumOfAddr; ui32Idx++) {
		PVR_ASSERT(psDevPAddr[ui32Idx].uiAddr >= SG100_HOST_MEM_DEV_BASE);

		psCpuPAddr[ui32Idx].uiAddr = psDevPAddr[ui32Idx].uiAddr
					     - SG100_HOST_MEM_DEV_BASE;
	}
}

static PVRSRV_ERROR InitLocalHeap(PHYS_HEAP_CONFIG *psPhysHeap,
				  IMG_UINT64 uiBaseAddr, IMG_UINT64 uiStartAddr,
				  IMG_UINT64 uiSize,
				  PHYS_HEAP_FUNCTIONS *psFuncs,
				  PHYS_HEAP_USAGE_FLAGS ui32Flags)
{
	psPhysHeap->sCardBase.uiAddr = uiBaseAddr;
	psPhysHeap->sStartAddr.uiAddr =
		IMG_CAST_TO_CPUPHYADDR_UINT(uiStartAddr);
	psPhysHeap->uiSize = uiSize;
	psPhysHeap->eType = PHYS_HEAP_TYPE_LMA;
	psPhysHeap->pszPDumpMemspaceName = "LMA";
	psPhysHeap->psMemFuncs = psFuncs;
	psPhysHeap->ui32UsageFlags = ui32Flags;

	return PVRSRV_OK;
}

static PVRSRV_ERROR CreateCardGPUHeaps(const SYS_DATA *psSysData,
				       PHYS_HEAP_CONFIG *pasPhysHeaps,
				       PHYS_HEAP_FUNCTIONS *psHeapFuncs)
{
	IMG_UINT64 ui64CardAddr = 0;
	IMG_UINT64 ui64StartAddr = psSysData->pdata->heap_memory_base;
	IMG_UINT64 ui64RogueHeapSize = psSysData->pdata->heap_memory_size;
	PVRSRV_ERROR eError;

	if (psSysData->pdata->mem_mode == TC_MEMORY_LOCAL) {
		ui64CardAddr = psSysData->pdata->heap_dev_base;
	}

	eError = InitLocalHeap(&pasPhysHeaps[PHY_HEAP_CARD_GPU], ui64CardAddr,
			       IMG_CAST_TO_CPUPHYADDR_UINT(ui64StartAddr),
			       ui64RogueHeapSize, psHeapFuncs,
			       PHYS_HEAP_USAGE_GPU_LOCAL);
	if (eError != PVRSRV_OK) {
		return eError;
	}

	return PVRSRV_OK;
}

static PVRSRV_ERROR CreateCardEXTHeap(const SYS_DATA *psSysData,
					PHYS_HEAP_CONFIG *pasPhysHeaps,
					PHYS_HEAP_FUNCTIONS *psHeapFuncs)
{
	IMG_UINT64 ui64CardAddr = 0;
	IMG_UINT64 ui64StartAddr = psSysData->pdata->display_heap_memory_base;
	IMG_UINT64 ui64Size = psSysData->pdata->display_heap_memory_size;
	PVRSRV_ERROR eError;

	if (psSysData->pdata->mem_mode == TC_MEMORY_LOCAL) {
		ui64CardAddr = psSysData->pdata->display_heap_dev_base;
	}

	eError = InitLocalHeap(&pasPhysHeaps[PHY_HEAP_CARD_EXT], ui64CardAddr,
			       IMG_CAST_TO_CPUPHYADDR_UINT(ui64StartAddr),
			       ui64Size, psHeapFuncs,
			       PHYS_HEAP_USAGE_EXTERNAL | PHYS_HEAP_USAGE_DISPLAY);
	if (eError != PVRSRV_OK) {
		return eError;
	}

	return PVRSRV_OK;
}

static PVRSRV_ERROR InitLocalHeaps(const SYS_DATA *psSysData,
				   PHYS_HEAP_CONFIG *pasPhysHeaps)
{
	PHYS_HEAP_FUNCTIONS *psHeapFuncs;
	PVRSRV_ERROR eError;

	if (psSysData->pdata->mem_mode == TC_MEMORY_HYBRID) {
		psHeapFuncs = &gsHostPhysHeapFuncs;
	} else {
		psHeapFuncs = &gsLocalPhysHeapFuncs;
	}

	eError = CreateCardGPUHeaps(psSysData, pasPhysHeaps, psHeapFuncs);
	if (eError != PVRSRV_OK) {
		return eError;
	}

	eError = CreateCardEXTHeap(psSysData, pasPhysHeaps, psHeapFuncs);
	if (eError != PVRSRV_OK)
	{
		return eError;
	}

	return PVRSRV_OK;
}

static PVRSRV_ERROR InitHostHeaps(const SYS_DATA *psSysData,
				  PHYS_HEAP_CONFIG *pasPhysHeaps)
{
	PVR_UNREFERENCED_PARAMETER(psSysData);

	pasPhysHeaps[PHY_HEAP_SYSTEM].eType = PHYS_HEAP_TYPE_UMA;
	pasPhysHeaps[PHY_HEAP_SYSTEM].pszPDumpMemspaceName = "SYSMEM";
	pasPhysHeaps[PHY_HEAP_SYSTEM].psMemFuncs = &gsHostPhysHeapFuncs;
	pasPhysHeaps[PHY_HEAP_SYSTEM].ui32UsageFlags =
		PHYS_HEAP_USAGE_CPU_LOCAL;

	return PVRSRV_OK;
}

/* QiShao: dedicated FW tuning mem with UMA heap type */
#if defined(SUPPORT_DEDICATED_UMA_FW_TUNINGMEM)
static PVRSRV_ERROR InitFwTuningHeapUMA(const SYS_DATA*   psSysData,
                                  		PHYS_HEAP_CONFIG* pasPhysHeaps)
{
    PVR_UNREFERENCED_PARAMETER(psSysData);

    pasPhysHeaps[PHY_HEAP_FW_TUNIG].eType                = PHYS_HEAP_TYPE_UMA;
    pasPhysHeaps[PHY_HEAP_FW_TUNIG].pszPDumpMemspaceName = "DEDICATEDFWTUNINGMEM";
    pasPhysHeaps[PHY_HEAP_FW_TUNIG].psMemFuncs           = &gsHostPhysHeapFuncs;
    pasPhysHeaps[PHY_HEAP_FW_TUNIG].ui32UsageFlags       = PHYS_HEAP_USAGE_FW_TUNING;

    return PVRSRV_OK;
}
#elif defined(SUPPORT_DEDICATED_UMA_FW_MAINMEM)
static PVRSRV_ERROR InitFwMainHeapUMA(const SYS_DATA*   psSysData,
                                        PHYS_HEAP_CONFIG* pasPhysHeaps)
{
    PVR_UNREFERENCED_PARAMETER(psSysData);

    pasPhysHeaps[PHY_HEAP_FW_MAIN].eType                = PHYS_HEAP_TYPE_UMA;
    pasPhysHeaps[PHY_HEAP_FW_MAIN].pszPDumpMemspaceName = "DEDICATEDFWMAINMEM";
    pasPhysHeaps[PHY_HEAP_FW_MAIN].psMemFuncs           = &gsHostPhysHeapFuncs;
    pasPhysHeaps[PHY_HEAP_FW_MAIN].ui32UsageFlags       = PHYS_HEAP_USAGE_FW_MAIN;

    return PVRSRV_OK;
}
#endif

static PVRSRV_ERROR PhysHeapsInit(const SYS_DATA *psSysData,
				  PHYS_HEAP_CONFIG *pasPhysHeaps,
				  void *pvPrivData)
{
	PVRSRV_ERROR eError;
	IMG_UINT32 i;
	IMG_UINT32 ui32NumHeaps =
		psSysData->pdata->mem_mode == TC_MEMORY_HYBRID ?
			PHY_HEAP_HYBRID_NUM :
			PHY_HEAP_LMA_NUM;

	eError = InitLocalHeaps(psSysData, pasPhysHeaps);
	if (eError != PVRSRV_OK) {
		return eError;
	}

	if (psSysData->pdata->mem_mode == TC_MEMORY_HYBRID) {
		eError = InitHostHeaps(psSysData, pasPhysHeaps);
		if (eError != PVRSRV_OK) {
			return eError;
		}
	}

/* QiShao: dedicated UMA tuning mem heap type */
#if defined(SUPPORT_DEDICATED_UMA_FW_TUNINGMEM)
	eError = InitFwTuningHeapUMA(psSysData, pasPhysHeaps);
	if (eError != PVRSRV_OK)
	{
		return eError;
	}
	++ui32NumHeaps;
#elif defined(SUPPORT_DEDICATED_UMA_FW_MAINMEM)
	eError = InitFwMainHeapUMA(psSysData, pasPhysHeaps);
	if (eError != PVRSRV_OK)
	{
		return eError;
	}
	++ui32NumHeaps;
#endif

	/* Initialise fields that don't change between memory modes.
	 * Fix up heap IDs. This is needed for multi-testchip systems to
	 * ensure the heap IDs are unique as this is what Services expects.
	 */
	for (i = 0; i < ui32NumHeaps; i++) {
		pasPhysHeaps[i].hPrivData = pvPrivData;
	}

	return PVRSRV_OK;
}

static PVRSRV_ERROR PhysHeapsCreate(const SYS_DATA *psSysData, void *pvPrivData,
				    PHYS_HEAP_CONFIG **ppasPhysHeapsOut,
				    IMG_UINT32 *puiPhysHeapCountOut)
{
	PHYS_HEAP_CONFIG *pasPhysHeaps;
	PVRSRV_ERROR eError;
	IMG_UINT32 ui32NumHeaps =
		psSysData->pdata->mem_mode == TC_MEMORY_HYBRID ?
			PHY_HEAP_HYBRID_NUM :
			PHY_HEAP_LMA_NUM;

	pasPhysHeaps = OSAllocMem(sizeof(*pasPhysHeaps) * ui32NumHeaps);
	if (!pasPhysHeaps) {
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}

	eError = PhysHeapsInit(psSysData, pasPhysHeaps, pvPrivData);
	if (eError != PVRSRV_OK) {
		OSFreeMem(pasPhysHeaps);
		return eError;
	}

	*ppasPhysHeapsOut = pasPhysHeaps;
	*puiPhysHeapCountOut = ui32NumHeaps;

	return PVRSRV_OK;
}

static void DeviceConfigDestroy(PVRSRV_DEVICE_CONFIG *psDevConfig)
{
	OSFreeMem(psDevConfig->pasPhysHeaps);

	OSFreeMem(psDevConfig);
}

static void odinTCDevPhysAddr2DmaAddr(PVRSRV_DEVICE_CONFIG *psDevConfig,
				      IMG_DMA_ADDR *psDmaAddr,
				      IMG_DEV_PHYADDR *psDevPAddr,
				      IMG_BOOL *pbValid, IMG_UINT32 ui32NumAddr,
				      IMG_BOOL bSparseAlloc)
{
	IMG_UINT32 ui32Idx;

	/* Fast path */
	if (!bSparseAlloc) {
		psDmaAddr->uiAddr = psDevPAddr->uiAddr;
	} else {
		for (ui32Idx = 0; ui32Idx < ui32NumAddr; ui32Idx++) {
			if (pbValid[ui32Idx]) {
				psDmaAddr[ui32Idx].uiAddr = psDevPAddr[ui32Idx].uiAddr;
			} else {
				/* Invalid DMA address marker */
				psDmaAddr[ui32Idx].uiAddr = ~((IMG_UINT64)0x0);
			}
		}
	}
}

static void va_hw_dma_chan_free(PVRSRV_DEVICE_CONFIG *psDevConfig, void *channel)
{
	(void)psDevConfig;
	(void)channel;
}

static void *va_hw_dma_chan_get(PVRSRV_DEVICE_CONFIG *psDevConfig, char *name)
{
	struct device *psDev = (struct device *)psDevConfig->pvOSDevice;
	return va_dma_get_byname(psDev->parent, name);
}

static PVRSRV_ERROR DeviceConfigCreate(SYS_DATA *psSysData,
				       PVRSRV_DEVICE_CONFIG **ppsDevConfigOut)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig;
	RGX_DATA *psRGXData;
	RGX_TIMING_INFORMATION *psRGXTimingInfo;
	PHYS_HEAP_CONFIG *pasPhysHeaps;
	IMG_UINT32 uiPhysHeapCount;
	PVRSRV_ERROR eError;

	psDevConfig = OSAllocZMem(sizeof(*psDevConfig) + sizeof(*psRGXData) +
				  sizeof(*psRGXTimingInfo));
	if (!psDevConfig) {
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}

	psRGXData =
		(RGX_DATA *)((IMG_CHAR *)psDevConfig + sizeof(*psDevConfig));
	psRGXTimingInfo = (RGX_TIMING_INFORMATION *)((IMG_CHAR *)psRGXData +
						     sizeof(*psRGXData));

	eError = PhysHeapsCreate(psSysData, psDevConfig, &pasPhysHeaps,
				 &uiPhysHeapCount);
	if (eError != PVRSRV_OK) {
		goto ErrorFreeDevConfig;
	}

	/* Setup RGX specific timing data */
	/* Volcanic FPGA 94MHz with divisor 16 = ~6MHz real clock */
	psRGXTimingInfo->ui32CoreClockSpeed = (94 * 1000 * 1000) / 16;
	psRGXTimingInfo->bEnableActivePM = IMG_FALSE;
	psRGXTimingInfo->bEnableRDPowIsland = IMG_FALSE;
	psRGXTimingInfo->ui32ActivePMLatencyms =
		SYS_RGX_ACTIVE_POWER_LATENCY_MS;

	/* Set up the RGX data */
	psRGXData->psRGXTimingInfo = psRGXTimingInfo;

	/* Setup the device config */
	psDevConfig->pvOSDevice = &psSysData->pdev->dev;
	psDevConfig->pszName = "sg100";
	psDevConfig->pszVersion = GetDeviceVersionString(psSysData);

	psDevConfig->sRegsCpuPBase.uiAddr = psSysData->registers->start;
	psDevConfig->ui32RegsSize = resource_size(psSysData->registers);

	/* gpu interrupt line */
	psDevConfig->ui32IRQ = COMMON_MSIX_GFX_VECTOR;

	psDevConfig->pasPhysHeaps = pasPhysHeaps;
	psDevConfig->ui32PhysHeapCount = uiPhysHeapCount;
	psDevConfig->eDefaultHeap = PVRSRV_PHYS_HEAP_GPU_LOCAL;

	/* Only required for LMA but having this always set shouldn't be a problem */
	// FIXME: is what?
	psDevConfig->bDevicePA0IsValid = IMG_TRUE;

	psDevConfig->hDevData = psRGXData;
	psDevConfig->hSysData = psSysData;

	psDevConfig->pfnSysDevFeatureDepInit = NULL;

#if defined(SUPPORT_LINUX_DVFS) || defined(SUPPORT_PDVFS)
	/* Fake DVFS configuration used purely for testing purposes */
	psDevConfig->sDVFS.sDVFSDeviceCfg.pasOPPTable = asOPPTable;
	psDevConfig->sDVFS.sDVFSDeviceCfg.ui32OPPTableSize = LEVEL_COUNT;
	psDevConfig->sDVFS.sDVFSDeviceCfg.pfnSetFrequency = SetFrequency;
	psDevConfig->sDVFS.sDVFSDeviceCfg.pfnSetVoltage = SetVoltage;
#endif
#if defined(SUPPORT_LINUX_DVFS)
	psDevConfig->sDVFS.sDVFSDeviceCfg.ui32PollMs = 1000;
	psDevConfig->sDVFS.sDVFSDeviceCfg.bIdleReq = IMG_TRUE;
	psDevConfig->sDVFS.sDVFSGovernorCfg.ui32UpThreshold = 90;
	psDevConfig->sDVFS.sDVFSGovernorCfg.ui32DownDifferential = 10;
#endif

	psDevConfig->bHasFBCDCVersion31 = IMG_FALSE;

	/* DMA channel config */
	psDevConfig->pfnSlaveDMAGetChan = va_hw_dma_chan_get;
	psDevConfig->pfnSlaveDMAFreeChan = va_hw_dma_chan_free;
	psDevConfig->pfnDevPhysAddr2DmaAddr = odinTCDevPhysAddr2DmaAddr;
	psDevConfig->pszDmaTxChanName = psSysData->pdata->dma_tx_chan_name;
	psDevConfig->pszDmaRxChanName = psSysData->pdata->dma_rx_chan_name;
	psDevConfig->bHasDma = true;
	/* Following two values are expressed in number of bytes */
	psDevConfig->ui32DmaTransferUnit = 1;
	psDevConfig->ui32DmaAlignment = 1;

	*ppsDevConfigOut = psDevConfig;

	return PVRSRV_OK;

ErrorFreeDevConfig:
	OSFreeMem(psDevConfig);
	return eError;
}

PVRSRV_ERROR SysDevInit(void *pvOSDevice, PVRSRV_DEVICE_CONFIG **ppsDevConfig)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig;
	SYS_DATA *psSysData;
	resource_size_t uiRegistersSize;
	PVRSRV_ERROR eError;
	//int err = 0;

	PVR_ASSERT(pvOSDevice);

	psSysData = OSAllocZMem(sizeof(*psSysData));
	if (psSysData == NULL) {
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}

	psSysData->pdev = to_platform_device((struct device *)pvOSDevice);
	psSysData->pdata = psSysData->pdev->dev.platform_data;

	// FIXME
	PVR_ASSERT(TC_MEMORY_CONFIG == psSysData->pdata->mem_mode);

	/*
	 * The device cannot address system memory, so there is no DMA
	 * limitation.
	 */
	if (psSysData->pdata->mem_mode == TC_MEMORY_LOCAL) {
		dma_set_mask(pvOSDevice, DMA_BIT_MASK(64));
	} else {
		dma_set_mask(pvOSDevice, DMA_BIT_MASK(32));
	}

	//err = tc_enable(psSysData->pdev->dev.parent);
	//if (err) {
	//	PVR_DPF((PVR_DBG_ERROR, "%s: Failed to enable PCI device (%d)",
	//		 __func__, err));
	//	eError = PVRSRV_ERROR_PCI_CALL_FAILED;
	//	goto ErrFreeSysData;
	//}

	psSysData->registers = platform_get_resource_byname(
		psSysData->pdev, IORESOURCE_MEM, "rogue-regs");
	if (!psSysData->registers) {
		PVR_DPF((PVR_DBG_ERROR,
			 "%s: Failed to get Rogue register information",
			 __func__));
		eError = PVRSRV_ERROR_PCI_REGION_UNAVAILABLE;
		goto ErrorDevDisable;
	}

	/* Check the address range is large enough. */
	uiRegistersSize = resource_size(psSysData->registers);
	if (uiRegistersSize < VA_GPU_REG_REGION_SIZE) {
		PVR_DPF((
			PVR_DBG_ERROR,
			"%s: Rogue register region isn't big enough (was %pa, required 0x%08x)",
			__func__, &uiRegistersSize, VA_GPU_REG_REGION_SIZE));

		eError = PVRSRV_ERROR_PCI_REGION_TOO_SMALL;
		goto ErrorDevDisable;
	}

#if 0
	/* Reserve the address range */
	if (!request_mem_region(psSysData->registers->start,
				resource_size(psSysData->registers),
				SYS_RGX_DEV_NAME)) {
		PVR_DPF((PVR_DBG_ERROR,
			 "%s: Rogue register memory region not available",
			 __func__));
		eError = PVRSRV_ERROR_PCI_CALL_FAILED;

		goto ErrorDevDisable;
	}
#endif

	eError = DeviceConfigCreate(psSysData, &psDevConfig);
	if (eError != PVRSRV_OK) {
		goto ErrorReleaseMemRegion;
	}

	/* Set psDevConfig->pfnSysDevErrorNotify callback */
	psDevConfig->pfnSysDevErrorNotify = SysRGXErrorNotify;

	*ppsDevConfig = psDevConfig;

	return PVRSRV_OK;

ErrorReleaseMemRegion:
	//release_mem_region(psSysData->registers->start,
	//		   resource_size(psSysData->registers));
ErrorDevDisable:
	//tc_disable(psSysData->pdev->dev.parent);
//ErrFreeSysData:
	OSFreeMem(psSysData);
	return eError;
}

void SysDevDeInit(PVRSRV_DEVICE_CONFIG *psDevConfig)
{
	SYS_DATA *psSysData = (SYS_DATA *)psDevConfig->hSysData;

	DeviceConfigDestroy(psDevConfig);

	//release_mem_region(psSysData->registers->start,
	//		   resource_size(psSysData->registers));

	//tc_disable(psSysData->pdev->dev.parent);

	OSFreeMem(psSysData);
}

PVRSRV_ERROR SysDebugInfo(PVRSRV_DEVICE_CONFIG *psDevConfig,
			  DUMPDEBUG_PRINTF_FUNC *pfnDumpDebugPrintf,
			  void *pvDumpDebugFile)
{
	PVR_UNREFERENCED_PARAMETER(psDevConfig);
	PVR_UNREFERENCED_PARAMETER(pfnDumpDebugPrintf);

	//TODO
	PVR_DPF((PVR_DBG_ERROR, "%s: TODO\n", __func__));

	return PVRSRV_OK;
}

typedef struct {
	struct device *psDev;
	int iInterruptID;
	void *pvData;
	PFN_LISR pfnLISR;
} LISR_DATA;

static void VA_GPU_InterruptHandler(void *pvData)
{
	LISR_DATA *psLISRData = pvData;
	psLISRData->pfnLISR(psLISRData->pvData);
}

PVRSRV_ERROR SysInstallDeviceLISR(IMG_HANDLE hSysData,
				  IMG_UINT32 ui32IRQ,
				  const IMG_CHAR *pszName,
				  PFN_LISR pfnLISR,
				  void *pvData,
				  IMG_HANDLE *phLISRData)
{
	SYS_DATA *psSysData = (SYS_DATA *)hSysData;
	LISR_DATA *psLISRData;
	PVRSRV_ERROR eError;
	int err;

	if (ui32IRQ != COMMON_MSIX_GFX_VECTOR)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: No device matching IRQ %d",
					__func__, ui32IRQ));
		return PVRSRV_ERROR_UNABLE_TO_INSTALL_ISR;
	}

	psLISRData = OSAllocZMem(sizeof(*psLISRData));
	if (!psLISRData)
	{
		eError = PVRSRV_ERROR_OUT_OF_MEMORY;
		goto err_out;
	}

	psLISRData->pfnLISR = pfnLISR;
	psLISRData->pvData = pvData;
	psLISRData->iInterruptID = ui32IRQ;
	psLISRData->psDev = psSysData->pdev->dev.parent;

	err = va_register_gpu_interrupt(psLISRData->psDev,
		psLISRData->iInterruptID, VA_GPU_InterruptHandler, psLISRData);
	if (err)
	{
		PVR_DPF((PVR_DBG_ERROR,
			"%s: va_register_gpu_interrupt() failed (%d)",
			__func__, err));
		eError = PVRSRV_ERROR_UNABLE_TO_INSTALL_ISR;
		goto err_free_data;
	}

	err = va_enable_gpu_interrupt(psLISRData->psDev, psLISRData->iInterruptID);
	if (err)
	{
		PVR_DPF((PVR_DBG_ERROR,
			"%s: va_enable_gpu_interrupt() failed (%d)",
			__func__, err));
		eError = PVRSRV_ERROR_UNABLE_TO_INSTALL_ISR;
		goto err_unset_interrupt_handler;
	}

	*phLISRData = psLISRData;
	eError = PVRSRV_OK;

	PVR_TRACE(("Installed device LISR " IMG_PFN_FMTSPEC " to irq %u",
			pfnLISR, ui32IRQ));

err_out:
	return eError;
err_unset_interrupt_handler:
	va_unregister_gpu_interrupt(psLISRData->psDev,
				    psLISRData->iInterruptID);
err_free_data:
	OSFreeMem(psLISRData);
	goto err_out;
}

PVRSRV_ERROR SysUninstallDeviceLISR(IMG_HANDLE hLISRData)
{
	LISR_DATA *psLISRData = (LISR_DATA *) hLISRData;
	int err;

	err = va_disable_gpu_interrupt(psLISRData->psDev,
					psLISRData->iInterruptID);
	if (err)
	{
		PVR_DPF((PVR_DBG_ERROR,
			"%s: va_disable_gpu_interrupt() failed (%d)",
			__func__, err));
	}

	err = va_unregister_gpu_interrupt(psLISRData->psDev,
					  psLISRData->iInterruptID);
	if (err)
	{
		PVR_DPF((PVR_DBG_ERROR,
			"%s: va_unregister_gpu_interrupt() failed (%d)",
			__func__, err));
	}

	PVR_TRACE(("Uninstalled device LISR " IMG_PFN_FMTSPEC " from irq %u",
		psLISRData->pfnLISR, psLISRData->iInterruptID));

	OSFreeMem(psLISRData);

	return PVRSRV_OK;
}
