#ifndef __VASTAI_PCIE_JTAG_API_H__
#define __VASTAI_PCIE_JTAG_API_H__

enum TAP_STATE {
	TAP_UNINIT      = -1,
	TAP_RESET       = 0xF,
	TAP_IDLE        = 0xC,

	TAP_DR_SCAN     = 0x7,
	TAP_CAPTURE_DR  = 0x6,
	TAP_SHIFT_DR    = 0x2,
	TAP_EXIT1_DR    = 0x1,
	TAP_PAUSE_DR    = 0x3,
	TAP_EXIT2_DR    = 0x0,
	TAP_UPDATE_DR   = 0x5,

	TAP_IR_SCAN     = 0x4,
	TAP_CAPTURE_IR  = 0xE,
	TAP_SHIFT_IR    = 0xA,
	TAP_EXIT1_IR    = 0x9,
	TAP_PAUSE_IR    = 0xB,
	TAP_EXIT2_IR    = 0x8,
	TAP_UPDATE_IR   = 0xD,
};

/* IR CODE, IR len is 5 bit */
#define VASTAI_IR_LEN                       5
#define IR_HOSTIJTAG_AXI_MASTER             0x08 //jtag2axi config TDR/TAP
#define IR_HOSTIJTAG_DEBUG_MODE             0x09 //used as bscan bypass
#define IR_HOSTIJTAG_TEST_REG               0x0A //used as Test reg TDR
#define IR_HOSTIJTAG_BYPASS_SLAVE_TAP       0x0B //used as slave tap sel
#define IR_HOSTIJTAG_TILE                   0x0C //used as all tile TDR access
#define IR_HOSTIJTAG_FRA                    0x0D //secure jtag
#define IR_HOSTIJTAG_AXI_SLAVE              0x0E //JTAG2AXI status reg
#define IR_MAIN_IDCODE                      0X0F //main tap idcode
#define IR_MAIN_BYPASS                      0x1F


/*
 * Description:
 *      Sets the TCK pin to HIGH level.
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_set_tck(uint32_t card_id);

/*
 * Description:
 *      Sets the TCK pin to LOW level.
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_clr_tck(uint32_t card_id);

/*
 * Description:
 *      This function sets the test data input to logical 1
 *  (VTref = target reference voltage).
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_set_tdi(uint32_t card_id);

/*
 * Description:
 *      This function sets the test data input to logical 0
 *  (VTref = target reference voltage).
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_clr_tdi(uint32_t card_id);

/*
 * Description:
 *      This function sets the test data input to logical 1
 *  (VTref = target reference voltage).
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_set_tms(uint32_t card_id);

/*
 * Description:
 *      This function sets the test data input to logical 0
 *  (VTref = target reference voltage).
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_clr_tms(uint32_t card_id);

/*
 * Description:
 *      This function sets the TRST pin of the J-Link target interface
 *  to HIGH (deasserts TRST).
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_set_trst(uint32_t card_id);

/*
 * Description:
 *      This function sets the TRST pin of the J-Link target interface
 *  to LOW (deasserts TRST).
 * Parameter:
 *      card_id     : card id.
 */
void vastai_jink_clr_trst(uint32_t card_id);

/*
 * Description:
 *      Creates one JTAG clock on TCK.
 * Parameter:
 *      card_id     : card id.
 * return value:
 *      Status of the TDO pin. Either 0 or 1.
 */
uint8_t vastai_jink_toggle_clk(uint32_t card_id);

/*
 * Description:
 *      This api can be used to walk to a different state 
 *  in the JTAG state machine.
 * Parameter:
 *      card_id     : card id.
 *      tms         : this buf stores the tms signal which want to input.
 *      num_bits    : the number of signals that need shift.
 */
void vastai_jtag_shift_tms(uint32_t card_id, const uint8_t * tms, uint32_t num_bits);

/*
 * Description:
 *      This api can be used to set ir reg,
 *  needn't to pay attention to the tap status before invoking this api,
 *  after invoking this api, tap will in idle state.
 * Parameter:
 *      card_id     : card id.
 *      input_buf   : this buf stores the tdi signal which want to input.
 *      output_buf  : this buf stores the tdo signal which will be output.
 *      ir_len      : the number of ir signals that need shift.
 */
void vastai_jtag_set_ir(uint32_t card_id, const uint8_t * input_buf, uint8_t * output_buf, uint32_t ir_len);

/*
 * Description:
 *      This api can be used to set dr reg,
 *  needn't to pay attention to the tap status before invoking this api,
 *  after invoking this api, tap will in idle state.
 * Parameter:
 *      card_id     : card id.
 *      input_buf   : this buf stores the tdi signal which want to input.
 *      output_buf  : this buf stores the tdo signal which will be output.
 *      dr_len      : the number of dr signals that need shift.
 */
void vastai_jtag_set_dr(uint32_t card_id, const uint8_t * input_buf, uint8_t * output_buf, uint32_t dr_len);

/*
 * Description:
 *      The functionality of this api is consistent with vastai_jtag_set_ir,
 *  needn't to pay attention to the tap status before invoking this api,
 *  after invoking this api, tap will in idle state.
 * Parameter:
 *      card_id     : card id.
 *      value       : this value could be IR CODE, it can be found in vastai_pcie_jtag_api.h.
 */
void vastai_jtag_main_tap_ir_sel(uint32_t card_id, uint8_t value);

/*
 * Description:
 *      Returns the jtag id of sg100.
 * Parameter:
 *      card_id     : card id.
 * return value:
 *      Jtag id of sg100.
 */
int vastai_jtag_get_device_id(uint32_t card_id);

/*
 * Description:
 *      This api can switch tap to the new_state.
 * Parameter:
 *      card_id     : card id.
 *      new_state   : tap will switch to new state.
 * return value:
 *      Tap old_state.
 */
int vastai_jtag_switch_tap_state(uint32_t card_id, int new_state);

/*
 * Description:
 *      Returns the current tap state.
 * Parameter:
 *      card_id     : card id.
 * return value:
 *      Current tap state.
 */
int vastai_jtag_get_tap_state(uint32_t card_id);

/*
 * Description:
 *      This api can switch jtag clock to the new freq,
 *  this value is limited to 1~500 Mhz.
 * Parameter:
 *      card_id     : card id.
 *      clock_freq  : jtag freq.
 * return value:
 *      Current jtag freq.
 */
uint32_t vastai_jtag_set_clock(uint32_t card_id, uint32_t clock_freq);

/*
 * Description:
 *      This api can be used to read noc information.
 * Parameter:
 *      card_id : card id.
 *      address : noc address.
 *      value   : return read value.
 * return value:
 *      0       : OK
 *      -1      : read failed.
 */
int vastai_jtag_axi_read(uint32_t card_id, uint64_t address, uint32_t *value);

/*
 * Description:
 *      This api can be used to write noc.
 * Parameter:
 *      card_id : card id.
 *      address : noc address.
 *      value   : the value need to write.
 * return value:
 *      0       : OK
 *      -1      : write failed.
 */
int vastai_jtag_axi_write(uint32_t card_id, uint64_t address, uint32_t value);

/*
 * Description:
 *      This api can be used to read efuse sram information.
 * Parameter:
 *      card_id : card id.
 *      address : noc address(it must be 0x10000~0x101FC, align 4).
 *      value   : return read value.
 * return value:
 *      0       : OK
 *      -1      : read failed.
 */
int vastai_jtag_fuse_sram_read(uint32_t card_id, uint32_t address, uint32_t *value);

/*
 * Description:
 *      This api can be used to write efuse sram.
 * Parameter:
 *      card_id : card id.
 *      address : noc address(it must be 0x10000~0x101FC, align 4).
 *      value   : the value need to write.
 * return value:
 *      0       : OK
 *      -1      : write failed.
 */
int vastai_jtag_fuse_sram_write(uint32_t card_id, uint32_t address, uint32_t value);

/*
 * Description:
 *      This api can be used to read efuse sram information.
 * Parameter:
 *      card_id : card id.
 *      address : noc address(it must be 0x10000~0x107FC, align 4).
 *      len     : the num of words want to read.
 *      buf     : return read data.
 * return value:
 *      0       : OK
 *      -1      : read failed.
 */
int vastai_jtag_fuse_sram_read_buf(uint32_t card_id, uint32_t address, uint32_t len, uint32_t *buf);

/*
 * Description:
 *      This api can be used to write efuse sram.
 * Parameter:
 *      card_id : card id.
 *      address : noc address(it must be 0x10000~0x107FC, align 4).
 *      len     : the num of words want to write.
 *      buf     : the data want to write.
 * return value:
 *      0       : OK
 *      -1      : write failed.
 */
int vastai_jtag_fuse_sram_write_buf(uint32_t card_id, uint32_t address, uint32_t len, uint32_t *buf);

/*
 * Description:
 *      This function will trst the vastai pcie jtag.
 * Parameter:
 *      card_id : card id.
 */
void vastai_jtag_trst(uint32_t card_id);

/*
 * Description:
 *      This function opens the connection to vastai pcie jtag and reset it.
 *  It’s always the first function you have to call to use the vastai jtag.
 * Parameter:
 *      card_id : card id.
 * return value:
 *      0       : OK
 *      -1      : vastai pcie2jtag device open failed.
 */
int vastai_jtag_open(uint32_t card_id);

/*
 * Description:
 *      This function closes the connection to vastai pcie jtag.
 * Parameter:
 *      card_id : card id.
 */
void vastai_jtag_close(uint32_t card_id);

/*
 * Description:
 *      This function can convert dbf information to card id information.
 * Parameter:
 *      dev_bdf     : bus, device, function info(such as : 81:00.0).
 * return value:
 *      others      : card id
 *      -1          : not find.
 */
int vastai_jtag_map_bdf_to_cardid(char *dev_bdf);

#endif /* __VASTAI_PCIE_JTAG_API_H__ */
