#ifndef __VASTAI_PCIE_API_H__
#define __VASTAI_PCIE_API_H__

#pragma pack(1)

#include <pthread.h>
#include "vastai_export_enum_define.h"

#define VASTAI_VER_LEN (64)
#define DIV_ROUND(x, len)	   (((x) + (len)-1) / (len))
#define ROUND_UP(x, align)	   DIV_ROUND(x, align) * (align)
#define BIT(nr)			   ((1) << (nr))
#define FOUR_K_SIZE		   (4096)
#define ERR_CODE(ret)		   (-(ret | (BIT(__COUNTER__))))

//#define VASTAI_PCI_RW_BUF_LEN_ONCE 1024 * 128 //128k  // Veloce will hang due to too large sz.
#define VASTAI_PCI_RW_BUF_LEN_ONCE 0x1000  //4k

#define VASTAI_DIE1_BASE	   (0x2000000000UL)
char *get_version(void);
unsigned long get_time(void);

typedef enum RST_TYPE_ENUM {
    RST_TYPE_HOTR_RST = 0,
    RST_TYPE_COND_REINIT_RST = 1,
    RST_TYPE_UNCOND_REINIT_RST = 2,
    RST_TYPE_FLR = 3,
}RST_TYPE_ENUM_T;

/*****************************************
 * Name-Description
 * 	vastai_do_reset - do hotR rst/cond_reinit rst / uncond_reinit rst / FLR.
 * Synopsis:
 * 	int vastai_do_reset(RST_TYPE_ENUM_T reset_type, int dev_id, int die_id);
 * Arguments:
 * 	RST_TYPE_ENUM reset_type
 * 		reset type.
 * 	unsigned int value
 * 		register value, 32bit
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_do_reset(RST_TYPE_ENUM_T reset_type, int dev_id, int die_id);

int vastai_pci_modify_reg32(unsigned long long address,unsigned int mask_bit,
								unsigned int mode,int dev_id);
/*****************************************
 * Name-Description
 * 	vastai_pcie_write_reg_32bit - write one register, width is 32bit.
 * Synopsis:
 * 	int vastai_pcie_write_reg_32bit(unsigned long long address, unsigned int value, int dev_id, int die_id);
 * Arguments:
 * 	unsigned long long address
 * 		register address, 64bit, have to 4Byte align
 * 	unsigned int value
 * 		register value, 32bit
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_write_reg_32bit(unsigned long long address, unsigned int value,
				int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_write_reg_64bit - write one register, width is 64bit.
 * Synopsis:
 * 	int vastai_pcie_write_reg_64bit(unsigned long long address, unsigned long long value, int dev_id, int die_id);
 * Arguments:
 * 	unsigned long long address
 * 		register address, 64bit, have to 4Byte align
 * 	unsigned long long value
 * 		register value, 64bit
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_write_reg_64bit(unsigned long long address,
				unsigned long long value, int dev_id,
				int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_write_reg_buf - write range buf into register
 * Synopsis:
 * 	int vastai_pcie_write_reg_buf(unsigned long long address, const void *buf, int buf_size_byte, int dev_id, int
 *die_id); Arguments: unsigned long long address register address, 64bit, have to 4Byte align const void *buf range buf
 *for write int buf_size_byte size of range buf, have to 4Byte align int dev_id select device id int die_id select die
 *id Return Value: 0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_write_reg_buf(unsigned long long address, const void *buf,
			      int buf_size_byte, int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_write_buf_unalign - write range buf into ddr
 * Synopsis:
 * 	int vastai_pcie_write_buf_unalign(unsigned long long address, const void *buf, int buf_size_byte, int dev_id, int
 *die_id); Arguments: unsigned long long address register address, 64bit, have to 4Byte align const void *buf range buf
 *for write int buf_size_byte size of range buf,  int dev_id select device id int die_id select die
 *id Return Value: 0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_write_buf_unalign(unsigned long long address, const void *buf,
				  int buf_size_byte, int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_write_reg_file - write binary file into register
 * Synopsis:
 * 	int vastai_pcie_write_reg_file(const char *file_path, int offset, unsigned long long address, int len_byte, int
 *dev_id, int die_id); Arguments: const char *file_path binary file's path, string. int offset offset with binary
 * 	unsigned long long address
 * 		register address, 64bit, have to 4Byte align
 * 	int len_byte
 * 		size of file, have to 4Byte align
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_write_reg_file(const char *file_path, int offset,
			       unsigned long long address, int len_byte,
			       int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_read_reg_32bit - read one register, width is 32bit.
 * Synopsis:
 * 	int vastai_pcie_read_reg_32bit(unsigned long long address, unsigned int *value, int dev_id, int die_id);
 * Arguments:
 * 	unsigned long long address
 * 		register address, 64bit, have to 4Byte align
 * 	unsigned int *value
 * 		return value, register value, 32bit, must __NOT__ NULL point
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_read_reg_32bit(unsigned long long address, unsigned int *value,
			       int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_read_reg_64bit - read one register, width is 64bit.
 * Synopsis:
 * 	int vastai_pcie_read_reg_64bit(unsigned long long address, unsigned long long *value, int dev_id, int die_id);
 * Arguments:
 * 	unsigned long long address
 * 		register address, 64bit, have to 4Byte align
 * 	unsigned int *value
 * 		return value, register value, 64bit, must __NOT__ NULL point
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_read_reg_64bit(unsigned long long address,
			       unsigned long long *value, int dev_id,
			       int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_read_reg_buf - read register, fill into range buf
 * Synopsis:
 * 	int vastai_pcie_read_reg_buf(unsigned long long address, void *buf, int buf_size_byte, int dev_id, int die_id);
 * Arguments:
 * 	unsigned long long address
 * 		register address, 64bit, have to 4Byte align
 * 	void *buf
 * 		range buf for read, must __NOT__ NULL point
 * 	int buf_size_byte
 * 		size of range buf, have to 4Byte align
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_read_reg_buf(unsigned long long address, void *buf,
			     int buf_size_byte, int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_read_buf_unalign - read ddr or csram, fill into range buf
 * Synopsis:
 * 	int vastai_pcie_read_buf_unalign(unsigned long long address, void *buf, int buf_size_byte, int dev_id, int die_id);
 * Arguments:
 * 	unsigned long long address
 * 		register address, 64bit,
 * 	void *buf
 * 		range buf for read, must __NOT__ NULL point
 * 	int buf_size_byte
 * 		size of range buf
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_read_buf_unalign(unsigned long long address, void *buf,
				 int buf_size_byte, int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_pcie_read_reg_file - read register, fill into binary file
 * Synopsis:
 * 	int vastai_pcie_read_reg_file(const char *file_path, int offset, unsigned long long address, int len_byte, int
 *dev_id, int die_id); Arguments: const char *file_path binary file's path, string. int offset offset with binary
 * 	unsigned long long address
 * 		register address, 64bit, have to 4Byte align
 * 	int len_byte
 * 		size of file, have to 4Byte align
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_read_reg_file(const char *file_path, int offset,
			      unsigned long long address, int len_byte,
			      int dev_id, int die_id);

#define BOOT_CMD "boot" /* for pcie boot */

/*****************************************
 * Name-Description
 * 	vastai_pcie_cmd - send cmd to pcie node
 * Synopsis:
 * 	int vastai_pcie_cmd(const char *cmd_line, int dev_id, int die_id);
 * Arguments:
 * 	const char *cmd_line
 * 		cmd string
 * 	int dev_id
 * 		select device id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_pcie_cmd(const char *cmd_line, int dev_id);

/*****************************************
 * Name-Description
 * 	compare_file - help user compare whether two binary files are consistent
 * Synopsis:
 * 	int compare_file(const char *path_a, const char *path_b, int offset, int compare_size_byte);
 * Arguments:
 * 	const char *path_a
 * 		A file path, string.
 * 	const char *path_b
 * 		B file path, string.
 * 	int offset
 * 		file offset. (effect both two files. If you need independent control, Pls contact me.)
 * 	int compare_size_byte
 * 		how many size you need compare.
 * Return Value:
 * 	0 - two files are __NOT__ consistent.
 * 	1 - two files are consistend.
 * 	other value is failed and errcode.
 *****************************************/
int compare_file(const char *path_a, const char *path_b, int offset,
		 int compare_size_byte);


/*****************************************
 * Name-Description
 * 	compare_ddr_gold_val - help user compare whether ddr values are consistent with gold value and output the difference, then exit comparing.
 * Vastai:
 * 	 int compare_ddr_gold_val(unsigned long long ddr_addr, unsigned long long (*gold_ddr_value)[2], unsigned long long byte_size, int dev_id, int die_id);
 * Arguments:
 * 	unsigned long long ddr_addr
 * 		starting ddr address to read, unsigned long long.
 * 	unsigned long long (*gold_ddr_value)[2]
 * 		gold ddr value in the form gold[1][2]={{addr, value}}, like gold_ref[1][2] ={{0x02500060, 0x00000002}}, unsigned long long[][2].
 * 	unsigned long long byte_size
 * 		byte size to read from DDR & compare, unsigned long long.
 * 	int dev_id
 * 		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	 0 - Both content are  consistent.
 * 	-2 - Both content are __NOT__ consistend.
 * 	other value is failed and errcode.
 *****************************************/
int compare_ddr_gold_val(unsigned long long ddr_addr, unsigned long long (*gold_ddr_value)[2], unsigned long long byte_size, int dev_id, int die_id);


typedef unsigned short u16;
typedef unsigned int u32;
typedef unsigned long long u64;
typedef unsigned char u8;

struct desc_node {
	u64 dev_addr;
	union {
		void *data;
		u64 addr;
	};
	u32 length;
	u32 :32;
};

struct vastai_mem_info{
	int dma_buf_fd;
	u64 dma_phy_addr;
	int size;
};

/*****************************************
 * Name-Description
 * 	vastai_pci_mmio_write_buf - help user mmio write buf
 * Synopsis:
 * 	int vastai_pci_mmio_write_buf(struct vastai_mem_info *mem_info,unsigned long long address, const void *buf,
 *			      int buf_size_byte, int dev_id, int die_id)
 * Arguments:
 * 	struct vastai_mem_info *mem_info
 * 			mem_info for copy from/to 
 *  unsigned long long address
 * 			device ddr address, 64bit, not have to 4Byte align
 *  const void *buf
 * 			user space buffer address
 *  int buf_size_byte
 * 			buffer length
 * 	int dev_id
 *		select device id
 * 	int die_id
 * 		select die id
 * 		
 * 	other value is failed and errcode.
 *****************************************/
int vastai_pci_mmio_write_buf_perf(struct vastai_mem_info *mem_info,unsigned long long address, void *buf,
			      int buf_size_byte, int dev_id, int die_id);

int vastai_pci_mmio_read_buf_perf(struct vastai_mem_info *mem_info, unsigned long long address, void *buf,
			      int buf_size_byte, int dev_id, int die_id);

/*****************************************
 * Name-Description
 * 	vastai_get_file_size - help user to get file size
 * Synopsis:
 * 	int vastai_get_file_size(const char *file_path);;
 * Arguments:
 * 	const char *path_a
 * 		a file path, string.
 * 	other value is failed and errcode.
 *****************************************/
int vastai_get_file_size(const char *file_path);

/*****************************************
 * Name-Description
 * 	vastai_sdma_read_buf_by_sg - trigger sdma by user addr
 * Synopsis:
 * 	int vastai_trigger_sdma_by_user_addr(void *user_addr, u64 dev_addr, 
 			u32 length, u32 is_dev_to_host, u32 dev_id);
 * Arguments:
 * 	void *user_addr
 * 		user address, host data buffer 
 * 	u64 dev_addr
 * 		device ddr address
 * 	u32 length
 * 		select the read data length
 *	u32 dev_id
 *		select device id
 * 	other value is failed and errcode.
 *****************************************/
int vastai_sdma_read_buf_by_sg(void *user_addr, u64 dev_addr, u32 length, u32 dev_id);

/*****************************************
 * Name-Description
 * 	vastai_sdma_write_buf_by_sg - trigger sdma by user addr
 * Synopsis:
 * 	int vastai_trigger_sdma_by_user_addr(void *user_addr, u64 dev_addr, 
 			u32 length, u32 is_dev_to_host, u32 dev_id);
 * Arguments:
 * 	void *user_addr
 * 		user address, host data buffer 
 * 	u64 dev_addr
 * 		device ddr address
 * 	u32 length
 * 		select the read data length
 *	u32 dev_id
 *		select device id
 * 	other value is failed and errcode.
 *****************************************/
int vastai_sdma_write_buf_by_sg(void *user_addr, u64 dev_addr, u32 length, u32 dev_id);

/*****************************************
 * Name-Description
 * 	vastai_sdma_read_buf_by_user_addr - trigger sdma by user addr
 * Synopsis:
 * 	int vastai_trigger_sdma_by_user_addr(void *user_addr, u64 dev_addr, 
 			u32 length, u32 is_dev_to_host, u32 dev_id);
 * Arguments:
 * 	void *user_addr
 * 		user address, host data buffer 
 * 	u64 dev_addr
 * 		device ddr address
 * 	u32 length
 * 		select the read data length
 *	u32 dev_id
 *		select device id
 * 	other value is failed and errcode.
 *****************************************/
int vastai_get_va_pid(u32 dev_id);
int vastai_free_va_pid(int va_pid, u32 dev_id);
int vastai_sdma_read_buf_by_user_addr(void *user_addr, u64 dev_addr, u32 length, u32 dev_id, u32 pid);

/*****************************************
 * Name-Description
 * 	vastai_sdma_write_buf_by_user_addr - trigger sdma by user addr
 * Synopsis:
 * 	int vastai_trigger_sdma_by_user_addr(void *user_addr, u64 dev_addr, 
 			u32 length, u32 is_dev_to_host, u32 dev_id);
 * Arguments:
 * 	void *user_addr
 * 		user address, host data buffer 
 * 	u64 dev_addr
 * 		device ddr address
 * 	u32 length
 * 		select the read data length
 *	u32 dev_id
 *		select device id
 * 	other value is failed and errcode.
 *****************************************/
int vastai_sdma_write_buf_by_user_addr(void *user_addr, u64 dev_addr, u32 length, u32 dev_id, u32 pid);

/*****************************************
 * Name-Description
 * 	vastai_sdma_read_buf - read the specified length data from device DDR into buf.
 * vastai:
 * 	int vastai_sdma_read_buf(unsigned long long address, void *buf,
 				int buf_size_byte, int dev_id, int die_index);
 * Arguments:
 * 	unsigned long long address
 * 		device ddr address, 64bit, have to 4Byte align
 * 	void *buf
 * 		store data to buf
 * 	int buf_size_byte
 * 		select the read data length
 *	int dev_id
 *		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_sdma_read_buf(unsigned long long address, void *buf,
				int buf_size_byte, int dev_id, int die_index);

int vastai_pcie_dma_read_buf(unsigned long long address, void *buf,
				int buf_size_byte, int dev_id, int die_index);

/*****************************************
 * Name-Description
 * 	vastai_sdma_write_buf - write data of the specified length in buf to device DDR.
 * vastai:
 * 	int vastai_sdma_write_buf(unsigned long long address, 
 			void *buf, int buf_size_byte, int dev_id, int die_index);
 * 	unsigned long long address
 * 		device ddr address, 64bit, have to 4Byte align
 * 	void *buf
 * 		register value, 64bit
 * 	int buf_size_byte
 * 		select the read data length
 *	int dev_id
 *		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_sdma_write_buf(unsigned long long address, void *buf,
				int buf_size_byte, int dev_id, int die_index);

int vastai_pcie_dma_write_buf(unsigned long long address, void *buf,
				int buf_size_byte, int dev_id, int die_index);

/*****************************************
 * Name-Description
 * 	vastai_pcie_dma_raw_transfer - write data of the specified length by pcie dma.
 * vastai:
 * 	int vastai_pcie_dma_raw_transfer(unsigned long long address, void *buf,
 *				int buf_size_byte, uint8_t direction, int dev_id, int die_index);
 * 	unsigned long long address
 * 		device ddr address, 64bit, have to 4Byte align
 * 	void *buf
 * 		register value, 64bit
 * 	int buf_size_byte
 * 		select the data length
 *  uint8_t direction
 * 		1 - host mem to device
 * 		0 - device to host mem
 *	int dev_id
 *		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 * 
 * Note: Make sure PCIE2_DMA_ENABLE(in va_hw_config.h) is 1 before use.
 * 
 *****************************************/
int vastai_pcie_dma_raw_transfer(unsigned long long address, void *buf,
				int buf_size_byte, int direction, int dev_id, int die_index);

/*****************************************
 * Name-Description
 * 	vastai_sdma_fill_data_2_ddr - write data of the specified length to device DDR.
 * vastai:
 * 	int vastai_sdma_fill_data_2_ddr(unsigned long long address, unsigned long long data, 
 			int size_byte, int dev_id, int die_index);
 * 	unsigned long long address
 * 		device ddr address, 64bit, have to 4Byte align
 * 	unsigned long long data
 * 		data, have to 8Byte align
 * 	int buf_size_byte
 * 		select the read data length
 *	int dev_id
 *		select device id
 * 	int die_id
 * 		select die id
 * Return Value:
 * 	0 - success. other value is failed and errcode.
 *****************************************/
int vastai_sdma_fill_data_2_ddr(unsigned long long address, unsigned long long data,
			int size_byte, int dev_id, int die_index);

/*****************************************
* Name-Description
*	vastai_sdma_write_from_file - write data of file to device DDR.
* vastai:
*int vastai_sdma_write_from_file(const char *file_path, int offset, 
			unsigned long long address, int len_byte, int dev_id, int die_id);
*const char *file_path
*	the path of file
*unsigned long long data
*	the data of offset in file
*unsigned long long address
*	device ddr address, 64bit, have to 4Byte align
*int buf_size_byte
*	select the read data length
*int dev_id
*	select device id
*int die_id
*	select die id
* Return Value:
*	0 - success. other value is failed and errcode.
*****************************************/
	int vastai_sdma_write_from_file(const char *file_path, int offset,
					   unsigned long long address, int dev_id, int die_id);

/*****************************************
* Name-Description
*	vastai_sdma_read_to_file - read data from device DDR to file.
* vastai:
*int vastai_sdma_read_to_file(const char *file_path, int offset, 
		unsigned long long address, int len_byte, int dev_id, int die_id);
*const char *file_path
*	the path of file
*unsigned long long data
*	the data of offset in file
*unsigned long long address
*	device ddr address, 64bit, have to 4Byte align
*int buf_size_byte
*	select the read data length
*int dev_id
*	select device id
*int die_id
*	select die id
* Return Value:
*	0 - success. other value is failed and er
*****************************************/
int vastai_sdma_read_to_file(const char *file_path, int offset,
					unsigned long long address, int len_byte, int dev_id, int die_id);

int vastai_sdma_d2d_transfer(u64 src_addr, u64 dst_addr, u64 length, u32 die_index, u32 dev_id);

int vastai_sdma_s2m_read(u64 src_addr, u64 dst_addr, u64 length, u32 dev_id);
int vastai_sdma_s2m_write(u64 src_addr, u64 dst_addr, u64 length, u32 dev_id);

int vastai_alloc_mem(struct vastai_mem_info *mem_info, int dev_id);
void vastai_free_mem(struct vastai_mem_info *mem_info, int dev_id);
int vastai_set_mem(struct vastai_mem_info *mem_info, int val, int dev_id);

int vastai_pcie_fread_reg_buf(unsigned long long address, void *buf,
			     int buf_size_byte, int dev_id, int die_id);
int vastai_pcie_fread_reg_32bit(unsigned long long address, unsigned int *value,
			    int dev_id, int die_id);
int vastai_pcie_fread_reg_64bit(unsigned long long address, unsigned long long *value,
			    int dev_id, int die_id);
int vastai_pcie_fwrite_reg_buf(unsigned long long address, const void *buf,
			      int buf_size_byte, int dev_id, int die_id);
int vastai_pcie_fwrite_reg_32bit(unsigned long long address, unsigned int value,
				int dev_id, int die_id);
int vastai_pcie_fwrite_reg_64bit(unsigned long long address, unsigned long long value,
				int dev_id, int die_id);

#define VASTAI_DEV_BUF_DESC_NAME_LEN 16
struct vastai_dev_buf_desc {
	char name[VASTAI_DEV_BUF_DESC_NAME_LEN];
	int filep;
	pthread_mutex_t mutex;
};

struct vastai_fw_version {
	unsigned char ver[VASTAI_VER_LEN];
	unsigned char smcu_ver[VASTAI_VER_LEN];
	unsigned char bmcu_ver[VASTAI_VER_LEN];
	unsigned char vdmcu_ver[VASTAI_VER_LEN];
	unsigned char vemcu_ver[VASTAI_VER_LEN];
	unsigned char vdsp_ver[VASTAI_VER_LEN];
	unsigned char cmcu_ver[VASTAI_VER_LEN];
	unsigned char lmcu_ver[VASTAI_VER_LEN];
	unsigned char odsp_ver[VASTAI_VER_LEN];
	unsigned char driver_ver[VASTAI_VER_LEN];
};
int vastai_get_fw_version(int dev_id, void *buf, int buf_size);

int vastai_get_cedar_asy_sts(int dev_id, void *buf, int buf_size);
int vastai_get_cedar_sym_sts(int dev_id, void *buf, int buf_size);

struct vastai_dev_buf_desc *open_vastai_dev_buf(int dev_id, char *name);
int close_vastai_dev_buf(struct vastai_dev_buf_desc *desc);

int vastai_dev_buf_read(struct vastai_dev_buf_desc *desc,
			unsigned long long address, void *buf,
			int buf_size_byte, int die_id);
int vastai_get_device_num(void);
int vastai_get_die_info(int dev_id, int die_index[], int *die_num);
int vastai_send_smi_interrupt(int die_index, int dev_id);

#define DIE_0 0
#define DIE_1 1

union die_index_data {
	u32 val;
	struct {
		u32 die_id : 8;
		u32 dev_id : 16;
		u32 seq_num : 8;
	};
};

int vastai_get_log_status(void *log_ctl_buf, u32 die_index);

struct vastai_dev_info {
	int dev_id;
	u64 unid;
	char pcie_slot[16]; /* 0000:01:00.00 */
	char pcie_dev_name[16]; /* kchar:0 */
	char device_name[16]; /* sv100 */

	union pci_id {
		u32 val;
		struct {
			u16 vender_id;
			u16 device_id;
		};
	} pci_id;

	union pci_sub_id {
		u32 val;
		struct {
			u16 sub_vender_id;
			u16 sub_device_id;
		};
	} pci_sub_id;

	union major_minor {
		u32 val;
		struct {
			u16 minor;
			u16 major;
		};
	} major_minor;

	struct ai_video_cap {
		int ai_cap;
		int video_cap;
	} cap;

	int die_num;
	struct vastai_die_info {
		union die_index_data die_index;
		char vacc_name[16]; /* vacc0 */
		char render_name[16]; /* render129 */
	} die_info[4];
};

int vastai_get_dev_info(struct vastai_dev_info dev_info[], int *dev_num);

#define CSRAM_BASE_ADDR		  0x08C00000
#define PERFORMANCE_AI_BASE	  (CSRAM_BASE_ADDR + 0x000F4000) // 8 bytes
#define PCIE_BANDWIDTH_UPSTREAM	  (CSRAM_BASE_ADDR + 0x000F40F0) // 8 bytes
#define PCIE_BANDWIDTH_DOWNSTREAM (CSRAM_BASE_ADDR + 0x000F40F8) // 8 bytes

#define DDR_ECC_1BIT 0x8CF6400
#define DDR_ECC_2BIT 0x8CF6800

struct vastai_performance_info {
	struct {
		u32 pcie_tx_vdsp_count[4];
		u32 pcie_rx_vdsp_count[4];
		u64 pcie_bandwidth_upstream;
		u64 pcie_bandwidth_downstream;
	} pcie_performance;
	u8 csram_info[0];
};

int vastai_get_performance(u32 die_index, struct vastai_performance_info *info,
			   u32 info_size);
u64 vastai_get_upstream_count(u32 die_index);
u64 vastai_get_downstream_count(u32 die_index);
int vastai_memset(u32 die_index, u64 address, char setval, u32 size);
int vastai_ioctl_memset(int fd, u32 die_index, u64 axi_addr, char val, u32 length);

struct ecc_entry {
	u32 addr_1;
	u32 addr_h;
	u32 axi_id;
	u32 axi_data_l;
	u32 axi_data_h;
	u32 errbit;
};

struct vastai_ecc_table {
	u32 entry_count;
	u32 rsvd_0;
	u32 rsvd_1;
	u32 rsvd_2;
	struct ecc_entry entry[0];
};

int vastai_get_ddr_1bit(u32 die_index, struct vastai_ecc_table *ecc_table,
			u32 table_size);
int vastai_get_ddr_2bit(u32 die_index, struct vastai_ecc_table *ecc_table,
			u32 table_size);

int vastai_dma_slave_2_master_bulk(int dev_id, int die_id, int ch,
				   struct desc_node *node, int num,
				   long int *time);
int vastai_dma_master_2_slave_bulk(int dev_id, int die_id, int ch,
				   struct desc_node *node, int num,
				   long int *time);

#define CEDAR_SVC_INFO                  (0x008E4000)
#define CEDAR_SVC_RNG_SYM_INFO          (0x008E4400) /* 512BYTE */
#define CEDAR_SVC_HASH_RES              (0x008E4600)
#define CEDAR_SVC_DONE_FLAG             (0x008E47E4)

typedef enum cedarPkaMode {
	pka224bForEcc = 0,
	pka256bForEcc = 1,
	pka384bForEcc = 2,
	pka521bForEcc = 3,
	pka2048bForRsa = 4,
	pka3072bForRsa = 5,
	pka4096bForRsa = 6,
} cedarPkaMode_t;

void big2little(u8 *data, u64 blen, u32 blockBlen);
int ce_sha2(u64 input, u64 output, u32 blen, svc_t type, int dev_id, int die_id);
int ce_sm3(u64 input, u64 output, u32 blen, svc_t type, int dev_id, int die_id);
int ce_sha2_img(u64 input, u64 output, u32 width, u32 height, svc_t type,int dev_id, int die_id);
int ce_sha3_img(u64 input, u64 output, u32 width, u32 height, svc_t type, int dev_id, int die_id);
int ce_sm3_img(u64 input, u64 output, u32 width, u32 height, svc_t type, int die_id);
int ce_aes_ecb(u64 input, u64 output, u32 blen, u64 key, u32 keyblen, u32 mode, svc_t type,int dev_id, int die_id);
int ce_aes_cbc(u64 input, u64 output, u32 blen, u64 key, u32 keyblen, u32 mode, u64 iv, svc_t type, int dev_id, int die_id);
int ce_aes_ctr(u64 input, u64 output, u32 blen, u64 key, u32 keyblen, u32 mode, u64 iv, svc_t type,int dev_id, int die_id);
int ce_aes_gcm(u64 input, u64 output, u32 blen,  u64 key, u32 keyblen,u32 mode,
		u64 iv, u32 ivblen, u64 msg, u32 msgblen,
		u64 tag, u32 tblen, svc_t type, int dev_id, int die_id);
int ce_ecc_enc(u64 input, u64 output, u32 blen, u64 random, u64 pubkey, u32 pkaMode,
		svc_t type, int dev_id,int die_id);
int ce_ecc_dec(u64 input, u64 output, u32 blen, u64 prikey, u32 pkaMode, svc_t type,int dev_id, int die_id);
int ce_ecc_sign(u64 input, u64 signature, u32 blen, u64 random, u64 prikey, u32 pkaMode, svc_t type,int dev_id, int die_id);
int ce_ecc_verify(u64 input, u64 signature, u32 blen, u64 pubkey, u32 pkaMode, svc_t type, int dev_id,int die_id);
int ce_sm4_ecb(u64 input, u64 output, u32 blen, u64 key, u32 keyblen, u32 mode, svc_t type , int dev_id, int die_id);
int ce_sm4_cbc(u64 input, u64 output, u32 blen, u64 key, u32 keyblen,u32 mode, u64 iv, svc_t type,int dev_id, int die_id);
int ce_sm4_ctr(u64 input, u64 output, u32 blen, u64 key, u32 keyblen,u32 mode, u64 iv, svc_t type, int dev_id,int die_id);
int ce_sm4_gcm(u64 input, u64 output, u32 blen,  u64 key, u32 keyblen,u32 mode,u64 iv, u32 ivblen, u64 msg, u32 msgblen,u64 tag, u32 tblen, svc_t type, int dev_id,int die_id);

#pragma pack()

#endif
