"""vastai_pcie_py.
	this mode support a python-api to access vastai sv100 pcie port

	write_4b write 4byte into sv100
	read_4b read 4byte from sv100

	dma_XXXX a famliy api for access sv100 pcie dma api
"""
import ctypes
#from typing import Iterable

class dma_node(ctypes.Structure):
	"""dma_node the abstraction of cadence's dma descript

		dev_addr : type: c u64. AXI address in sv100
		data	: type: python bytes. A host buf, for dma write/read
	"""
	_fields_ = [("dev_addr", ctypes.c_uint64),
			("data", ctypes.c_char_p),
			("length", ctypes.c_int32)]

def dev_2_dev_node(master_addr, slave_addr, length):
	master_addr_temp = ctypes.cast(slave_addr, ctypes.c_char_p)
	return dma_node(slave_addr, master_addr, length)
class vastai_pcie:

	def __init__(self, soPath, dev_id):
		""" load so lib, and open pcie_file
			pls set right so lib path. and dev_id you want.

		Args:
			soPath: so lib path.
		dev_id:
			what's dev you want access. See `ls /dev/kchar*`
		"""
		self.__vp_so__ = ctypes.CDLL(soPath)
		self.__dev_id__ = dev_id
		self.__c_dev_id__ = ctypes.c_int32(dev_id)

	def write_4b(self, addr, val, die_id):
		""" write 4 byte(val) into addr 
		
		Args:
			addr: axi address, u64
			val: write val
		Return:
			error code from c api.
		"""
		c_addr = ctypes.c_uint64(addr)
		c_val = ctypes.c_uint32(val)
		c_die_id = ctypes.c_int32(die_id)
		return self.__vp_so__.vastai_pcie_write_reg_32bit(c_addr, c_val, self.__c_dev_id__, c_die_id)

	def read_4b(self, addr, die_id):
		""" read 4 byte(val)
		
		Args:
			addr: axi address, u64
		
		Return:
			value: read from sv100.
			error: error code from c api.
		"""
		c_addr = ctypes.c_uint64(addr)
		c_val = ctypes.c_int(0)
		c_val_p = ctypes.pointer(c_val)
		c_die_id = ctypes.c_int32(die_id)
		error = self.__vp_so__.vastai_pcie_read_reg_32bit(c_addr, c_val_p, self.__c_dev_id__, c_die_id)
		return c_val.value, error

	def dma_host_2_dev_bulk(self, die_id, ch, node):
		"""dma transfer, host->dev, bulk mode

		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dma_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		#if not isinstance(node, Iterable): 
		#	raise TypeError("node must can Iterable, (pls use list)")
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_host_2_dev_bulk(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value

	def dma_dev_2_host_bulk(self, die_id, ch, node):
		"""dma transfer, dev->host, bulk mode

		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dma_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_dev_2_host_bulk(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value

	def dma_host_2_dev_scatter(self, die_id, ch, node):
		"""dma transfer, host->dev, scatter mode

		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dma_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_host_2_dev_scatter(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value

	def dma_dev_2_host_scatter(self, die_id, ch, node):
		"""dma transfer, dev->host, scatter mode

		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dma_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_dev_2_host_scatter(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value
	
	def dma_host_2_dev_gather(self, die_id, ch, node):
		"""dma transfer, host->dev, gather mode

		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dma_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_host_2_dev_gather(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value

	def dma_dev_2_host_gather(self, die_id, ch, node):
		"""dma transfer, dev->host, gather mode

		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dma_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_dev_2_host_gather(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value

	def dma_master_2_slave_bulk(self, die_id, ch, node):
		"""dma transfer, master->slave, bulk mode
	
		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dev_2_dev_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_master_2_slave_bulk(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value

	def dma_slave_2_master_bulk(self, die_id, ch, node):
		"""dma transfer, master->slave, bulk mode
	
		Args:
			die_id: which die you want access.
			ch: which dma ch you want access.
			node: type is class dev_2_dev_node
		Return:
			time: The time consumed by dma.
			error code from c api
		"""
		c_die_id = ctypes.c_int(die_id)
		c_ch = ctypes.c_int(ch)
		c_node_size = ctypes.c_int(len(node))
		c_dma_node_array = (dma_node * len(node))(*node)
		c_time = ctypes.c_int(0)
		c_time_p = ctypes.pointer(c_time)
		error = self.__vp_so__.vastai_dma_slave_2_master_bulk(self.__c_dev_id__, c_die_id, c_ch, c_dma_node_array, c_node_size, c_time_p)
		return error, c_time.value

if __name__ == "__main__":
	print("python self test")

	#Pls change so path as you need
	vp = vastai_pcie("../out/libvastai_pcie_user_api.so", 0)

	#case 1: write and read a reg
	write_value = 0x123
	test_addr = 0x840000000
	error_no = vp.write_4b(test_addr, write_value, 0)
	if error_no != 0:
		print("va pcie write failed\n")

	read_value, error = vp.read_4b(test_addr, 0)
	if error_no != 0:
		print("va pcie read failed\n")
	
	if read_value != write_value:
		print("write_value != read_value! test failed\n")
	
	#case 2: start bulk dma loop (host -> dev, dev -> host, valid) one node
	host_buf_1 = b'3'*0x200
	node_write = [dma_node(test_addr, host_buf_1, len(host_buf_1))]

	error_no, time = vp.dma_host_2_dev_bulk(0, 0, node_write)
	if error_no != 0:
		print("va pcie dma host 2 dev failed\n")

	print("case 2 host_2_dev time:%d"%time)
	host_buf_2 = b'1'*0x200
	node_read = [dma_node(test_addr, host_buf_2, len(host_buf_2))]

	error_no, time = vp.dma_dev_2_host_bulk(0, 0, node_read)
	if error_no != 0:
		print("va pcie dma dev 2 host failed\n")

	if host_buf_1 != host_buf_2:
		print("dma_buf data mismatch test failed\n")
		print("host_buf_1", host_buf_1)
		print("host_buf_2", host_buf_2)
	print("case 2 dev_2_host time:%d"%time)

	#case 3: start bulk dma loop (host -> dev, dev -> host, valid) four list
	test_addr = 0x840000000
	host_buf_write_list = [b'3'*0x200 for x in range(4)]

	node_write = [dma_node(x * 0x300 + test_addr, host_buf_write_list[x], len(host_buf_write_list[x])) for x in range(4)]

	error_no, time = vp.dma_host_2_dev_bulk(0, 0, node_write)
	if error_no != 0:
		print("va pcie dma host 2 dev failed\n")
	print("case 3 host_2_dev time:%d"%time)

	host_buf_read_list = [b'3'*0x200 for x in range(4)]
	node_read = [dma_node(x * 0x300 + test_addr, host_buf_read_list[x], len(host_buf_read_list[x])) for x in range(4)]

	error_no, time = vp.dma_dev_2_host_bulk(0, 0, node_read)
	if error_no != 0:
		print("va pcie dma dev 2 host failed\n")

	for read,write in zip(host_buf_read_list, host_buf_write_list):
		if read != write:
			print("dma_buf data mismatch test failed\n")
			print("read", read)
			print("write", write)
	print("case 3 dev_2_host time:%d"%time)
	
	#case 4: start scatter && gather dma loop (host -scatter-> dev, dev -gather-> host, valid)
	test_addr = 0x840000000
	host_buf_write = b'5'*0x200
	node_write = [dma_node(0x0, host_buf_write, len(host_buf_write)),#fire node as a pre_fetch
			dma_node(test_addr, ctypes.c_char_p(), 0x100), #second node as a post_write
			dma_node(test_addr + 0x400, ctypes.c_char_p(), 0x100)] #third node as a post_write

	error_no, time = vp.dma_host_2_dev_scatter(0, 0, node_write)
	if error_no != 0:
		print("va pcie dma host 2 dev failed\n")
	
	print("case 4 host_2_dev time:%d"%time)
	host_buf_read = b'7'*0x200
	node_read = [dma_node(test_addr, ctypes.c_char_p(), 0x100), #first node as a pre_fetch
			dma_node(test_addr + 0x400, ctypes.c_char_p(), 0x100), #second node as a pre_fetch
			dma_node(0x0, host_buf_read, len(host_buf_read))] #thrid node as a post_write

	error_no, time = vp.dma_dev_2_host_gather(0, 0, node_read)
	if error_no != 0:
		print("va pcie dma dev 2 host failed\n")
	if host_buf_read != host_buf_write:
		print("dma_buf data mismatch test failed\n")
		print("read", host_buf_write)
		print("write", host_buf_read)
	print("case 4 dev_2_host time:%d"%time)

	#case 5: start die1 bulk dma loop (host -> dev, dev -> host, valid) one node
	host_buf_1 = b'3'*0x200
	node_write = [dma_node(test_addr, host_buf_1, len(host_buf_1))]

	error_no, time = vp.dma_host_2_dev_bulk(1, 0, node_write)
	if error_no != 0:
		print("va pcie dma host 2 dev (die1) failed\n")

	print("case 5 host_2_dev time:%d"%time)
	host_buf_2 = b'1'*0x200
	node_read = [dma_node(test_addr, host_buf_2, len(host_buf_2))]

	error_no, time = vp.dma_dev_2_host_bulk(1, 0, node_read)
	if error_no != 0:
		print("va pcie dma dev 2 host failed\n")

	if host_buf_1 != host_buf_2:
		print("dma_buf data mismatch test failed\n")
		print("host_buf_1", host_buf_1)
		print("host_buf_2", host_buf_2)
	print("case 5 dev1_2_host time:%d"%time)


#case 6: start master dev 2 slave dev bulk dma loop (host -> dev, dev -> dev, dev -> host, valid) one node
	host_buf_1 = b'e'*0x200
	node_write = [dma_node(test_addr, host_buf_1, len(host_buf_1))]

	error_no, time = vp.dma_host_2_dev_bulk(0, 0, node_write)
	if error_no != 0:
		print("va pcie dma host 2 dev (die1) failed\n")

	print("case 6 host_2_dev time:%d"%time)

	dev_2_dev_desc = [dev_2_dev_node(test_addr, test_addr, len(host_buf_1))]
	error_no, time = vp.dma_master_2_slave_bulk(0, 0, dev_2_dev_desc)
	if error_no != 0:
		print("va pcie dma dev 2 dev (die1) failed\n")

	print("case 6 master_2_slave time:%d"%time)

	host_buf_2 = b'1'*0x200
	node_read = [dma_node(test_addr, host_buf_2, len(host_buf_2))]

	error_no, time = vp.dma_dev_2_host_bulk(1, 0, node_read)
	if error_no != 0:
		print("va pcie dma dev 2 host failed\n")

	print("case 6 dev_2_host time:%d"%time)

	if host_buf_1 != host_buf_2:
		print("dma_buf data mismatch test failed\n")
		print("host_buf_1", host_buf_1)
		print("host_buf_2", host_buf_2)

#case 7: start slave 2 master bulk dma loop (host -> dev, dev -> dev, dev -> host, valid) one node
	host_buf_1 = b'b'*0x200
	node_write = [dma_node(test_addr, host_buf_1, len(host_buf_1))]

	error_no, time = vp.dma_host_2_dev_bulk(1, 0, node_write)
	if error_no != 0:
		print("va pcie dma host 2 dev (die1) failed\n")

	print("case 7 host_2_dev time:%d"%time)

	dev_2_dev_desc = [dev_2_dev_node(test_addr, test_addr, len(host_buf_1))]
	error_no, time = vp.dma_slave_2_master_bulk(0, 0, dev_2_dev_desc)
	if error_no != 0:
		print("va pcie dma dev 2 dev (die1) failed\n")

	print("case 7 slave_2_master time:%d"%time)

	host_buf_2 = b'1'*0x200
	node_read = [dma_node(test_addr, host_buf_2, len(host_buf_2))]

	error_no, time = vp.dma_dev_2_host_bulk(0, 0, node_read)
	if error_no != 0:
		print("va pcie dma dev 2 host failed\n")

	print("case 7 dev_2_host time:%d"%time)

	if host_buf_1 != host_buf_2:
		print("dma_buf data mismatch test failed\n")
		print("host_buf_1", host_buf_1)
		print("host_buf_2", host_buf_2)
