
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

typedef unsigned int u32;
typedef unsigned long long int u64;

#define VASTAI_PCI_IOCTL_ALLOC	   0
#define VASTAI_PCI_IOCTL_DMA_START 1
#define VASTAI_PCI_IOCTL_DMA_TRANS 3
#define VASTAI_PCI_IOCTL_POLL_LOG  0xa

struct va_ioctk_kchar_cmd {
	union {
		struct {
			u32 size;
			int dma_buf_fd;
		} alloc_cmd;

		struct {
			u32 is_dev_to_host;
			u32 dma_buf_fd;
			u64 axi_addr;
			u32 length;
			u32 die_index;
		} dma_start_cmd;
		struct {
			u32 is_dev_to_host;
			void *vir_addr;
			u64 axi_addr;
			u32 length;
			u32 die_index;
			int pid;
		} dma_trans_cmd;
		struct {
			u32 is_dev_to_host;
			u32 die_index;
			u64 axi_addr;
			struct vastai_channel_buf *channel_buf;
			int channel_num;
			int pid;
		} dma_trans_sg_cmd;
		struct {
			u32 die_index;
			u32 timeout_water;
			u32 is_timeout;
			void *log_ring_ctl;
		} log_poll_cmd;
	};
};

union die_index_data {
	u32 val;
	struct {
		u32 die_id : 8;
		u32 dev_id : 16;
		u32 RESV : 8;
	};
};

static int open_pcie_kchar(int dev_id)
{
	char file_path[64] = { 0 };
	int filep;

	snprintf(file_path, sizeof(file_path), "/dev/va%d_ctl", dev_id);
	filep = open(file_path, O_RDWR);
	if (filep < 0) {
		/*error handle*/
		char temp_buf[128] = { 0 };
		snprintf(temp_buf, sizeof(temp_buf), "open %s failed!\n",
			 file_path);
		perror(temp_buf);
	}
	return filep;
}

static int vastai_alloc_dma_buf(int size, u32 die_index)
{
	union die_index_data temp = { .val = die_index };
	int kchar_fd = open_pcie_kchar(temp.dev_id);
	struct va_ioctk_kchar_cmd kcmd = { .alloc_cmd.size = size,
				  .alloc_cmd.dma_buf_fd = 0 };
	if (kchar_fd < 0) {
		printf("\t\topen kchar is failed %d\n", kchar_fd);
		return kchar_fd;
	}

	int ret = ioctl(kchar_fd, VASTAI_PCI_IOCTL_ALLOC, &kcmd);
	if (ret < 0) {
		printf("\t\tioctl is failed %d\n", ret);
	} else {
		ret = kcmd.alloc_cmd.dma_buf_fd;
	}
	close(kchar_fd);
	return ret;
}

/* is sync */
static int vastai_trigger_dma(int dma_buf, u64 axi_addr, u32 length,
			      u32 is_dev_to_host, u32 die_index)
{
	union die_index_data temp = { .val = die_index };
	int kchar_fd = open_pcie_kchar(temp.dev_id);
	struct va_ioctk_kchar_cmd kcmd = { .dma_start_cmd.is_dev_to_host =
					  is_dev_to_host,
				  .dma_start_cmd.dma_buf_fd = dma_buf,
				  .dma_start_cmd.axi_addr = axi_addr,
				  .dma_start_cmd.length = length,
				  .dma_start_cmd.die_index = die_index };
	if (kchar_fd < 0) {
		printf("\t\topen kchar is failed %d\n", kchar_fd);
		return kchar_fd;
	}

	int ret = ioctl(kchar_fd, VASTAI_PCI_IOCTL_DMA_START, &kcmd);
	close(kchar_fd);
	return ret;
}

/**
 * @brief this function will set device's memory by pcie dma.
 * 
 * @param addr : device's ddr address.
 * @param val : char val. memset val.
 * @param alloc_size :  buf size alloc of pc.
 * @param size : memset buf size of device ddr.
 * @param die_index
 * @return int : success is 0. other is errno.
 */
int dma_memset(u64 addr, char val, unsigned long alloc_size, unsigned long size,
	       unsigned int die_index)
{
	printf("\t\tstep 1: trigger dma host2dev\n");
	int dma_buf_fd = vastai_alloc_dma_buf(alloc_size, die_index);
	int ret;
	if (dma_buf_fd < 0) {
		printf("\t\tget dma_buf_fd error %d\n", dma_buf_fd);
		return -1;
	}

	void *mem = mmap(NULL, size, PROT_WRITE | PROT_READ, MAP_SHARED,
			 dma_buf_fd, 0);
	if (mem == MAP_FAILED) {
		printf("\t\tmap failed %lld\n", (long long)mem);
		return -1;
	}

	memset(mem, val, size);

	munmap(mem, size);
	ret = vastai_trigger_dma(dma_buf_fd, addr, size, 0 /* host to dev */,
				 die_index);
	if (ret < 0) {
		printf("\t\ttrigger dma is failed %d\n", ret);
		return -1;
	}
	close(dma_buf_fd);
	return 0;
}

/* read dma buf back and check */
int dma_memory_check(u64 addr, char val, unsigned long size,
		     unsigned int die_index)
{
	printf("\t\tstep 2: trigger dma dev2host\n");
	int dma_buf_fd = vastai_alloc_dma_buf(size, 0);
	if (dma_buf_fd < 0) {
		printf("\t\tget dma_buf_fd error %d\n", dma_buf_fd);
		return -1;
	}

	int ret = vastai_trigger_dma(dma_buf_fd, addr, size,
				     1 /* dev to host */, die_index);
	if (ret < 0) {
		printf("\t\ttrigger dma is failed %d\n", ret);
		return -1;
	}

	void *mem = mmap(NULL, size, PROT_WRITE | PROT_READ, MAP_SHARED,
			 dma_buf_fd, 0);
	if (mem == MAP_FAILED) {
		printf("\t\tmap failed %lld\n", (long long)mem);
		return -1;
	}

	/* step 3. check data */
	printf("\t\tstep 3: check dma data\n");
	for (unsigned long i = 0; i < size; i++) {
		if (((char *)mem)[i] != val) {
			printf("\t\tdata is not match\n");
			return -1;
		}
	}
	munmap(mem, size);
	close(dma_buf_fd);
	return 0;
}

int dma_buf_basic_test(void)
{
	int alloc_size = 1024 * 1024;
	printf("test case: dma loopback, size: 0x%x\n", alloc_size);

	/* step 1. host to dev */
	int ret = dma_memset(0x840000000, 0x53, alloc_size, alloc_size, 0);
	if (ret < 0) {
		printf("\ttest case: dma loopback, size: 0x%x failed: alloc failed\n",
		       alloc_size);
		return ret;
	}
	ret = dma_memory_check(0x840000000, 0x53, alloc_size, 0);
	if (ret < 0) {
		printf("\ttest case: dma loopback, size: 0x%x failed: check failed\n",
		       alloc_size);
		return ret;
	}

	printf("\ttest success\n");
	return ret;
}

int dma_buf_alloc_mmap_test(void)
{
	int dma_buf_fd[2];
	void *mem[2];
	unsigned long alloc_size = 0x440000;

	printf("test case: dma buf alloc, mmap\n");
	dma_buf_fd[0] = vastai_alloc_dma_buf(alloc_size, 0);
	if (dma_buf_fd[0] < 0) {
		printf("\tdma buf alloc test failed : alloc failed\n");
		return -1;
	}

	mem[0] = mmap(NULL, alloc_size, PROT_WRITE | PROT_READ, MAP_SHARED,
		      dma_buf_fd[0], 0);

	dma_buf_fd[1] = vastai_alloc_dma_buf(alloc_size, 0);
	if (dma_buf_fd[1] < 0) {
		printf("\tdma buf alloc test failed : alloc failed\n");
		return -1;
	}

	mem[1] = mmap(NULL, alloc_size, PROT_WRITE | PROT_READ, MAP_SHARED,
		      dma_buf_fd[1], 0);
	munmap(mem[0], alloc_size);
	if (!(mem[1])) {
		printf("\tdma buf alloc test failed : mmap failed\n");
		return -1;
	}

	mem[0] = mmap(NULL, alloc_size, PROT_WRITE | PROT_READ, MAP_SHARED,
		      dma_buf_fd[0], 0);
	if (!(mem[1])) {
		printf("\tdma buf alloc test failed : mmap failed\n");
		return -1;
	}

	munmap(mem[0], alloc_size);
	munmap(mem[1], alloc_size);
	close(dma_buf_fd[0]);
	close(dma_buf_fd[1]);

	printf("\ttest success\n");
	return 0;
}

int test_multi_write(void)
{
	int ret;

	printf("test case: #OP2245 alloc bigger size than use\n");

	ret = dma_memset(0x840000000, 0x1, 8 * 1024 * 1024, 5 * 1024 * 1024, 0);
	if (ret < 0) {
		printf("\ttest case: #OP2245 alloc bigger size than use failed: send failed\n");
		return ret;
	}
	ret = dma_memset(0x840000000 + 5 * 1024 * 1024, 0x2, 8 * 1024 * 1024,
			 5 * 1024 * 1024, 0);
	if (ret < 0) {
		printf("\ttest case: #OP2245 alloc bigger size than use failed: send failed\n");
		return ret;
	}
	ret = dma_memory_check(0x840000000, 0x1, 5 * 1024 * 1024, 0);
	if (ret < 0) {
		printf("\ttest case: #OP2245 alloc bigger size than use failed: check failed\n");
		return ret;
	}
	ret = dma_memory_check(0x840000000 + 5 * 1024 * 1024, 0x2,
			       5 * 1024 * 1024, 0);
	if (ret < 0) {
		printf("\ttest case: #OP2245 alloc bigger size than use failed: check failed\n");
		return ret;
	}
	ret = dma_memset(0x840000000, 0x1, 8 * 1024 * 1024, 5 * 1024 * 1024, 0);
	if (ret < 0) {
		printf("\ttest case: #OP2245 alloc bigger size than use failed: send failed\n");
		return ret;
	}
	ret = dma_memory_check(0x840000000, 0x1, 5 * 1024 * 1024, 0);
	if (ret < 0) {
		printf("\ttest case: #OP2245 alloc bigger size than use failed: check failed\n");
		return ret;
	}
	ret = dma_memory_check(0x840000000 + 5 * 1024 * 1024, 0x2,
			       5 * 1024 * 1024, 0);
	if (ret < 0) {
		printf("\ttest case: #OP2245 alloc bigger size than use failed: check failed\n");
		return ret;
	}
	printf("\ttest success\n");
	return ret;
}

int main(void)
{
	int ret;

	ret = dma_buf_alloc_mmap_test();
	if (ret) {
		return ret;
	}

	ret = dma_buf_basic_test();
	if (ret) {
		return ret;
	}

	ret = test_multi_write();
	if (ret) {
		return ret;
	}

	printf("dma buf test success\n");
	return 0;
}
