#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <dirent.h>
#include <getopt.h>
#include <malloc.h>
#include <signal.h>
#include "../include/vastai_pcie_dump_api.h"

char *g_dump_file = NULL;
uint64_t g_dump_addr = 0;
uint32_t g_dump_size = 0;
uint32_t g_dump_value = 0;
uint32_t g_read_en = 1;
uint32_t g_align_word = 1;
uint32_t g_line_cnt = 1;
uint32_t g_card_id = 0;

static void vastai_dump_print_usage(void)
{
	printf( "vadumper Version:%s\n", DUMP_VERSION);
	printf( "Usage: vadumper <command> [<args>]\n"
			"    [-h, --help] [-b, --byte] [-r, --read] [-w, --write]\n"
			"    [-a <dumper_address>] default: -a 0\n"
			"    [-l <dumper_length>] default: -l 0\n"
			"    [-v <dumper_value>] default: -v 0\n"
			"    [-f <file_name>]\n"
			"    [-c <card_id>] default: -c 0\n"
			"    [-C <line_counts>] default: -C 1\n\n");
	printf( "Mandatory arguments to long options are mandatory for short options too.\n");
	printf( "    -h, --help                     Help manual.\n");
	printf( "    -a, --address=dumper_address   The start address of the dump.\n");
	printf( "    -l, --length=dumper_length     The length to dump.\n");
	printf( "    -f, --file=file_name           The dump content needs to be saved to a bin file.\n");
	printf( "    -b, --byte                     The dump content is displayed as byte.\n");
	printf( "    -r, --read                     Read NOC info.\n");
	printf( "    -w, --write                    Write to NOC.\n");
	printf( "    -v, --value                    The value want write to NOC.\n");
	printf( "    -c, --card                     The card id.\n");
	printf( "    -C, --line-counts=cnt          Number of dump items displayed per line.\n");
}

static int vastai_mem_dump_arg_parse(int argc, char *argv[])
{
	int c;
	int option_index = 0;
	static struct option long_options[] = {
		{"address",     required_argument,  0,  'a'},
		{"length",      required_argument,  0,  'l'},
		{"value",       required_argument,  0,  'v'},
		{"file",        required_argument,  0,  'f'},
		{"card",        required_argument,  0,  'c'},
		{"line-counts", required_argument,  0,  'C'},
		{"byte",        no_argument,        0,  'b'},
		{"read",        no_argument,        0,  'r'},
		{"write",       no_argument,        0,  'w'},
		{"help",        no_argument,        0,  'h'},
		{0,             0,                  0,  0}
	};

	while (1) {
		c = getopt_long(argc, argv, "a:l:v:f:c:C:brwh", long_options, &option_index);
		if (c == -1) break;

		switch (c) {
		case 'a':
			if (optarg) {
				if ((!strncmp(optarg, "0x", 2)) || (!strncmp(optarg, "0X", 2)) ) {
					g_dump_addr = strtol(optarg, NULL, 16);
				}
				else {
					g_dump_addr = strtol(optarg, NULL, 10);
				}
			}
			break;
		case 'l':
			if (optarg) {
				if ((!strncmp(optarg, "0x", 2)) || (!strncmp(optarg, "0X", 2)) ) {
					g_dump_size = strtol(optarg, NULL, 16);
				}
				else {
					g_dump_size = strtol(optarg, NULL, 10);
				}
			}
			break;
		case 'v':
			if (optarg) {
				if ((!strncmp(optarg, "0x", 2)) || (!strncmp(optarg, "0X", 2)) ) {
					g_dump_value = strtol(optarg, NULL, 16);
				}
				else {
					g_dump_value = strtol(optarg, NULL, 10);
				}
			}
			break;
		case 'f':
			if (optarg) {
				g_dump_file = optarg;
			}
			break;
		case 'c':
			if (optarg) {
				g_card_id = strtol(optarg, NULL, 10);
			}
			break;
		case 'C':
			if (optarg) {
				if ((!strncmp(optarg, "0x", 2)) || (!strncmp(optarg, "0X", 2)) ) {
					g_line_cnt = strtol(optarg, NULL, 16);
				}
				else {
					g_line_cnt = strtol(optarg, NULL, 10);
				}
			}
			break;
		case 'b':
			g_align_word = 0;
			break;
		case 'r':
			g_read_en = 1;
			break;
		case 'w':
			g_read_en = 0;
			break;
		case 'h':
			vastai_dump_print_usage();
			return -1;
		default:
			break;
		}
	}

	return 0;
}

static void exit_handle(int signo) 
{
	vastai_dump_close(g_card_id);
	exit(0);
}

int main(int argc, char *argv[])
{
	int ret;
	int dump_fd;
	uint32_t tmp_count;
	uint8_t *data_buf;
	uint32_t *data_buf_w;
	uint32_t *dest;

	signal(SIGINT, exit_handle);

	if (0 > vastai_mem_dump_arg_parse(argc, argv))
		return -1;

	g_dump_size = g_dump_size & 0xFFFFFFFC;

	ret = vastai_dump_open(g_card_id);
	if (ret) {
		printf("ERROR: va_dump_open failed\n");
		return -1;
	}

	if (g_read_en) {
		/* write to file */
		if (NULL != g_dump_file) {
			printf("INFO: card_id:%d, dump_addr:0x%lx, dump_size:0x%x, dump_file:%s\n", g_card_id, g_dump_addr, g_dump_size, g_dump_file);
			ret = vastai_dump_download_file(g_card_id, g_dump_addr, g_dump_size, g_dump_file);
			if (ret) {
				printf("ERROR: Download failed\n");
				vastai_dump_close(g_card_id);
				return -1;
			}
		}
		/* write to std */
		else {
			printf("INFO: card_id:%d, dump_addr:0x%lx, dump_size:0x%x, line_cnt:%u, align_word:%s\n",
				g_card_id, g_dump_addr, g_dump_size, g_line_cnt, g_align_word ? "yes" : "no");
			data_buf = (uint8_t *)malloc(g_dump_size);
			if (NULL == data_buf) {
				printf("malloc failed\n");
				vastai_dump_close(g_card_id);
				return -1;
			}
			vastai_dump_read_buf(g_card_id, g_dump_addr, g_dump_size, data_buf);
			tmp_count = 0;
			printf("Dump begin :");
			while (g_dump_size > tmp_count) {
				if (g_align_word) {
					if (0 == (tmp_count % (g_line_cnt * 4)))
						printf("\n0x%lx :", (uint64_t)(g_dump_addr + tmp_count));
					printf(" 0x%08x", *(uint32_t *)((uint64_t)data_buf + tmp_count));
					tmp_count += 4;
				}
				else {
					if (0 == (tmp_count % g_line_cnt))
						printf("\n0x%lx :", (uint64_t)(g_dump_addr + tmp_count));
					printf(" 0x%02x", *(unsigned char *)((uint64_t)data_buf + tmp_count));
					tmp_count += 1;
				}
			}
			printf("\n");
			free(data_buf);
			printf("Dump finish!\n");
		}
	}
	else {
		/* load file to noc */
		if (NULL != g_dump_file) {
			printf("INFO: load_addr:0x%lx, load_file:%s\n", g_dump_addr, g_dump_file);
			ret = vastai_dump_load_file(g_card_id, g_dump_addr, g_dump_file);
			if (ret) {
				printf("ERROR: Load failed\n");
				vastai_dump_close(g_card_id);
				return -1;
			}
		}
		/* write to noc */
		else {
			printf("INFO: load_addr:0x%lx, load_size:0x%x, load_value:0x%x\n", g_dump_addr, g_dump_size, g_dump_value);
			data_buf_w = (uint32_t *)malloc(g_dump_size);
			if (NULL == data_buf_w) {
				printf("malloc failed\n");
				vastai_dump_close(g_card_id);
				return -1;
			}
			tmp_count = g_dump_size / 4;
			dest = (uint32_t *)data_buf_w;
			while (tmp_count--)
				*dest++ = g_dump_value;
			vastai_dump_write_buf(g_card_id, g_dump_addr, g_dump_size, (uint8_t *)data_buf_w);
			free(data_buf_w);
			printf("Load finish!\n");
		}
	}

	vastai_dump_close(g_card_id);

	return 0;
}

