#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include "vastai_pcie_dump_api.h"

#define MAX_CARD_NUM						16

#define SEM_KEY								0x7a891fbc

#define VENDOR_ID							"0x1ec6"
#define VENDOR_ID_DC						"0x1f4f"
#define DEVICE_ID_0							"0x0200"
#define DEVICE_ID_1							"0x0201"
#define DEVICE_ID_2							"0x0202"
#define DEVICE_ID_3							"0x0204"
#define SUB_VENDOR_ID						"0x1ec6"
#define SUB_VENDOR_ID_DC					"0x1f4f"

#define BRA0_SPACE_SIZE_256K				0x40000
#define BRA1_SPACE_SIZE_8K					0x2000
#define ATU_REGION_CTRL_1_OFF_INBOUND		(0x10000 + 0x100)
#define ATU_REGION_CTRL_2_OFF_INBOUND		(0x10000 + 0x104)
#define ATU_LWR_BASE_ADDR_OFF_INBOUND		(0x10000 + 0x108)
#define ATU_UPPER_BASE_ADDR_OFF_INBOUND		(0x10000 + 0x10C)
#define ATU_LIMIT_ADDR_OFF_INBOUND			(0x10000 + 0x110)
#define ATU_LWR_TARGET_ADDR_OFF_INBOUND		(0x10000 + 0x114)
#define ATU_UPPER_TARGET_ADDR_OFF_INBOUND	(0x10000 + 0x118)
#define ATU_REGION_CTRL_3_OFF_INBOUND		(0x10000 + 0x11C)
#define ATU_UPPR_LIMIT_ADDR_OFF_INBOUND		(0x10000 + 0x120)

typedef union semun
{
	int val;
}semun;

typedef struct vastai_dump_ctl_s
{
	int semid;
	int is_get_lock;
	int is_dump_open;
	int pci_bar0_fd;
	int pci_bar1_fd;
	uint64_t pci_bar0_map_addr;
	uint64_t pci_bar1_map_addr;
}vastai_dump_ctl_t;

vastai_dump_ctl_t g_dump_ctl[MAX_CARD_NUM];

static uint32_t dump_bra0_readl(uint32_t card_id, uint64_t addr)
{
	return *(volatile uint32_t *)(g_dump_ctl[card_id].pci_bar0_map_addr + addr);
}

static void dump_bra0_writel(uint32_t card_id, uint32_t data, uint64_t addr)
{
	*(volatile uint32_t *)(g_dump_ctl[card_id].pci_bar0_map_addr + addr) = data;
	return;
}

static uint32_t dump_bra1_readl(uint32_t card_id, uint64_t addr)
{
	return *(volatile uint32_t *)(g_dump_ctl[card_id].pci_bar1_map_addr + addr);
}

static void dump_bra1_writel(uint32_t card_id, uint32_t data, uint64_t addr)
{
	*(volatile uint32_t *)(g_dump_ctl[card_id].pci_bar1_map_addr + addr) = data;
	return;
}

static void dump_lock_init(uint32_t card_id)
{
	semun mu;

	g_dump_ctl[card_id].semid = semget((key_t)SEM_KEY + card_id, 0, 0666);
	if(g_dump_ctl[card_id].semid == -1)
	{
		g_dump_ctl[card_id].semid = semget((key_t)SEM_KEY + card_id, 1, 0666 | IPC_CREAT);
		mu.val=1;
		semctl(g_dump_ctl[card_id].semid, 0, SETVAL, mu);
	}
}

static void dump_lock_wait(uint32_t card_id)
{
	struct sembuf buff;

	buff.sem_num = 0;
	buff.sem_op = -1;
	buff.sem_flg = SEM_UNDO;
	semop(g_dump_ctl[card_id].semid, &buff, 1);
	g_dump_ctl[card_id].is_get_lock = 1;
}

static void dump_lock_release(uint32_t card_id)
{
	struct sembuf buff;

	buff.sem_num = 0;
	buff.sem_op=1;
	buff.sem_flg = SEM_UNDO;
	semop(g_dump_ctl[card_id].semid, &buff, 1);
	g_dump_ctl[card_id].is_get_lock = 0;
}

static char *dump_memcpy(void * dst, const void *src, uint32_t num)
{
	uint8_t *dest = dst;
	const uint8_t *source = src;
	uint32_t *aligned_dest;
	const uint32_t *aligned_src;

	/* If the size is small, or either SRC or DST is unaligned,
	then punt into the byte copy loop.  This should be rare. */
	if (!(num < 4) && !(((uint64_t)source & 3) | ((uint64_t)dest & 3))) {
		aligned_dest = (uint32_t *)dest;
		aligned_src = (uint32_t *)source;

		/* Copy 4X long words at a time if possible. */
		while (num >= 16) {
			*aligned_dest++ = *aligned_src++;
			*aligned_dest++ = *aligned_src++;
			*aligned_dest++ = *aligned_src++;
			*aligned_dest++ = *aligned_src++;
			num -= 16;
		}

		/* Copy one long word at a time if possible. */
		while (num >= 4) {
			*aligned_dest++ = *aligned_src++;
			num -= 4;
		}

		/* Pick up any residual with a byte copier. */
		dest = (uint8_t *)aligned_dest;
		source = (uint8_t *)aligned_src;
	}

	while (num--)
		*dest++ = *source++;

	return dst;
}

static int get_dev_path(uint32_t card_id, char *dev_path)
{
	DIR *dir = opendir("/sys/bus/pci/devices");
	struct dirent *pdirent = NULL;
	char file_name[128];
	int fd;
	char buf[16];
	uint32_t count = 0;
	
	while((pdirent = readdir(dir)) != NULL) {
		if (NULL == strstr(pdirent->d_name, ".0"))
			continue;
		sprintf(dev_path, "/sys/bus/pci/devices/%s", pdirent->d_name);

		sprintf(file_name, "%s/vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, VENDOR_ID)) && (NULL == strstr(buf, VENDOR_ID_DC)))
			continue;

		sprintf(file_name, "%s/device", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, DEVICE_ID_0)) && (NULL == strstr(buf, DEVICE_ID_1))
			&& (NULL == strstr(buf, DEVICE_ID_2)) && (NULL == strstr(buf, DEVICE_ID_3)))
			continue;

		sprintf(file_name, "%s/subsystem_vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, SUB_VENDOR_ID)) && (NULL == strstr(buf, SUB_VENDOR_ID_DC)))
			continue;

		if (count == card_id) {
			closedir(dir);
			return 0;
		}
		
		count++;
	}

	closedir(dir);

	return -1;
}

static int mmap_pcie_resource(uint32_t card_id, const char * const dev_path)
{
	char dev_name[128];
	int pci_config_fd;
	uint32_t val;

	sprintf(dev_name, "%s/config", dev_path);
	pci_config_fd = open(dev_name, O_RDWR | O_SYNC);
	if (pci_config_fd < 0) {
		printf("ERROR: %s open failed\n", dev_name);
		return -1;
	}
	lseek(pci_config_fd, 4, SEEK_SET); 
	read(pci_config_fd, &val, 4);
	val |= 0x2;
	lseek(pci_config_fd, 4, SEEK_SET); 
	write(pci_config_fd, &val, 4);
	close(pci_config_fd);

	sprintf(dev_name, "%s/resource0", dev_path);
	g_dump_ctl[card_id].pci_bar0_fd = open(dev_name, O_RDWR | O_SYNC);
	if (g_dump_ctl[card_id].pci_bar0_fd < 0) {
		printf("ERROR: %s open failed\n", dev_name);
		return -1;
	}
	g_dump_ctl[card_id].pci_bar0_map_addr =
			(uint64_t)mmap(NULL, BRA0_SPACE_SIZE_256K, PROT_READ | PROT_WRITE, MAP_SHARED, g_dump_ctl[card_id].pci_bar0_fd, 0);
	if (MAP_FAILED == (void *)g_dump_ctl[card_id].pci_bar0_map_addr) {
			printf("ERROR: %s, bar0 map failed\n", __func__);
			return -1;
	}
	
	sprintf(dev_name, "%s/resource1", dev_path);
	g_dump_ctl[card_id].pci_bar1_fd = open(dev_name, O_RDWR | O_SYNC);
	if (g_dump_ctl[card_id].pci_bar1_fd < 0) {
		printf("ERROR: %s open failed\n", dev_name);
		return -1;
	}
	g_dump_ctl[card_id].pci_bar1_map_addr =
			(uint64_t)mmap(NULL, BRA1_SPACE_SIZE_8K, PROT_READ | PROT_WRITE, MAP_SHARED, g_dump_ctl[card_id].pci_bar1_fd, 0);
	if (MAP_FAILED == (void *)g_dump_ctl[card_id].pci_bar1_map_addr) {
			printf("ERROR: %s, bar1 map failed\n", __func__);
			return -1;
	}

	return 0;
}

static void restore_bar1_atu0(uint32_t card_id)
{
	uint32_t temp_val;

	/* disable_atu_inbound_region */
	temp_val = dump_bra0_readl(card_id, ATU_REGION_CTRL_2_OFF_INBOUND);
	temp_val &= 0x7FFFFFFF;
	dump_bra0_writel(card_id, temp_val, ATU_REGION_CTRL_2_OFF_INBOUND);

	/* config BAR0 ATU Inbound Region, bar match mode only for BAR0 */
	dump_bra0_writel(card_id, 0x942000, ATU_LWR_TARGET_ADDR_OFF_INBOUND);
	dump_bra0_writel(card_id, 0, ATU_UPPER_TARGET_ADDR_OFF_INBOUND);

	/* define the type of the region to be MEM; and Set the function number */
	dump_bra0_writel(card_id, 0, ATU_REGION_CTRL_1_OFF_INBOUND);

	/* region enable, BAR Match Mode for Bar[0-bar0:6-rom] & enable Function Number Match */
	temp_val = 0xC0000000 | (1 << 8) | (1 << 19);
	dump_bra0_writel(card_id, temp_val, ATU_REGION_CTRL_2_OFF_INBOUND);
	usleep(10000);

	return;
}

static uint64_t switch_bar1_atu0(uint32_t card_id, uint64_t target_addr)
{
	uint64_t real_target_addr = 0;
	uint32_t temp_val = 0;

	real_target_addr = target_addr / BRA1_SPACE_SIZE_8K * BRA1_SPACE_SIZE_8K;

	/* disable_atu_inbound_region */
	temp_val = dump_bra0_readl(card_id, ATU_REGION_CTRL_2_OFF_INBOUND);
	temp_val &= 0x7FFFFFFF;
	dump_bra0_writel(card_id, temp_val, ATU_REGION_CTRL_2_OFF_INBOUND);

	/* config BARx ATU Inbound Region, bar match mode only for BARx */
	dump_bra0_writel(card_id, (uint32_t)(real_target_addr & 0xffffffff), ATU_LWR_TARGET_ADDR_OFF_INBOUND);
	dump_bra0_writel(card_id, (uint32_t)(real_target_addr >> 32), ATU_UPPER_TARGET_ADDR_OFF_INBOUND);

	/* define the type of the region to be MEM; and Set the function number */
	dump_bra0_writel(card_id, 0, ATU_REGION_CTRL_1_OFF_INBOUND);

	/* region enable, BAR Match Mode for Bar[0-bar0:6-rom] & enable Function Number Match */
	temp_val = 0xC0000000 | (1 << 8) | (1 << 19);
	dump_bra0_writel(card_id, temp_val, ATU_REGION_CTRL_2_OFF_INBOUND);
	usleep(10000);

	return real_target_addr;
}

int vastai_dump_read(uint32_t card_id, uint64_t address, uint32_t *value)
{
	uint64_t target_addr;

	if (0 == g_dump_ctl[card_id].is_dump_open) {
		printf("ERROR: dump not open.\n");
		return -1;
	}

	if (NULL == value) {
		printf("ERROR: dump_read failed, empty point\n");
		return -1;
	}

	dump_lock_wait(card_id);
	target_addr = switch_bar1_atu0(card_id, address);
	*value = dump_bra1_readl(card_id, address - target_addr);
	restore_bar1_atu0(card_id);
	dump_lock_release(card_id);

	return 0;
}

int vastai_dump_write(uint32_t card_id, uint64_t address, uint32_t value)
{
	uint64_t target_addr;

	if (0 == g_dump_ctl[card_id].is_dump_open) {
		printf("ERROR: dump not open.\n");
		return -1;
	}

	dump_lock_wait(card_id);
	target_addr = switch_bar1_atu0(card_id, address);
	dump_bra1_writel(card_id, value, address - target_addr);
	restore_bar1_atu0(card_id);
	dump_lock_release(card_id);

	return 0;
}

int vastai_dump_read_buf(uint32_t card_id, uint64_t address, uint32_t len, uint8_t *buf)
{
	uint64_t target_addr;
	uint32_t read_len;
	uint32_t record_len = 0;

	if (0 == g_dump_ctl[card_id].is_dump_open) {
		printf("ERROR: dump not open.\n");
		return -1;
	}

	if (NULL == buf) {
		printf("ERROR: dump_read_buf failed, empty point\n");
		return -1;
	}

	printf("dump_read_buf : address[0x%lx], len[0x%x]\n", address, len);
	printf("read [%3d%%]", 0);

	while (record_len < len) {
		dump_lock_wait(card_id);
		target_addr = switch_bar1_atu0(card_id, address + record_len);
		read_len = (BRA1_SPACE_SIZE_8K - (address + record_len - target_addr));
		read_len = ((len - record_len) > read_len) ? read_len : (len - record_len);
		dump_memcpy((void *)(buf + record_len),
				(void *)(g_dump_ctl[card_id].pci_bar1_map_addr + (address + record_len - target_addr)),
				read_len);
		record_len += read_len;
		printf("\b\b\b\b\b\b[%3d%%]", record_len * 100 / len);
		fflush(stdout);
		restore_bar1_atu0(card_id);
		dump_lock_release(card_id);
	}
	printf("\n");
	return 0;
}

int vastai_dump_write_buf(uint32_t card_id, uint64_t address, uint32_t len, uint8_t *buf)
{
	uint64_t target_addr;
	uint32_t write_len;
	uint32_t record_len = 0;

	if (0 == g_dump_ctl[card_id].is_dump_open) {
		printf("ERROR: dump not open.\n");
		return -1;
	}

	if (NULL == buf) {
		printf("ERROR: dump_write_buf failed, empty point\n");
		return -1;
	}

	while (record_len < len) {
		dump_lock_wait(card_id);
		target_addr = switch_bar1_atu0(card_id, address + record_len);
		write_len = (BRA1_SPACE_SIZE_8K - (address + record_len - target_addr));
		write_len = ((len - record_len) > write_len) ? write_len : (len - record_len);
		dump_memcpy((void *)(g_dump_ctl[card_id].pci_bar1_map_addr + (address + record_len - target_addr)),
				(void *)(buf + record_len),
				write_len);
		record_len += write_len;
		restore_bar1_atu0(card_id);
		dump_lock_release(card_id);
	}

	return 0;
}

int vastai_dump_download_file(uint32_t card_id, uint64_t address, uint32_t len, char *file_name)
{
	uint64_t target_addr;
	uint32_t read_len;
	uint64_t record_len = 0;
	int file_fd;

	if (0 == g_dump_ctl[card_id].is_dump_open) {
		printf("ERROR: dump not open.\n");
		return -1;
	}

	if (NULL == file_name) {
		printf("ERROR: dump_download_file failed, empty point\n");
		return -1;
	}

	file_fd = open(file_name, O_RDWR | O_CREAT);
	if (file_fd < 0) {
		printf("ERROR: %s open failed\n", file_name);
		return -1;
	}
	lseek(file_fd, 0, SEEK_SET);
	printf("Begin download, file name : %s, NOC address : 0x%lx, length : 0x%x\n", file_name, address, len);
	printf("downloading [%3d%%]", 0);
	while (record_len < len) {
		dump_lock_wait(card_id);
		target_addr = switch_bar1_atu0(card_id, address + record_len);
		read_len = (BRA1_SPACE_SIZE_8K - (address + record_len - target_addr));
		read_len = ((len - record_len) > read_len) ? read_len : (len - record_len);
		write(file_fd, (void *)(g_dump_ctl[card_id].pci_bar1_map_addr + (address + record_len - target_addr)), read_len);
		record_len += read_len;
		restore_bar1_atu0(card_id);
		dump_lock_release(card_id);
		printf("\b\b\b\b\b\b[%3u%%]", (uint32_t)(record_len * 100 / len));
		fflush(stdout);
	}
	printf("\n");
	close(file_fd);

	return 0;
}

int vastai_dump_load_file(uint32_t card_id, uint64_t address, char *file_name)
{
	uint64_t target_addr;
	uint32_t write_len;
	uint64_t record_len = 0;
	int file_fd;
	struct stat file_stat;
	uint32_t len;

	if (0 == g_dump_ctl[card_id].is_dump_open) {
		printf("ERROR: dump not open.\n");
		return -1;
	}

	if (NULL == file_name) {
		printf("ERROR: dump_load_file failed, empty point\n");
		return -1;
	}

	file_fd = open(file_name, O_RDWR);
	if (file_fd < 0) {
		printf("ERROR: %s open failed\n", file_name);
		return -1;
	}
	fstat(file_fd, &file_stat);
	len = file_stat.st_size;
	lseek(file_fd, 0, SEEK_SET);
	printf("Begin load, file name : %s, NOC address : 0x%lx, file size : 0x%x\n", file_name, address, len);
	printf("loading [%3d%%]", 0);
	while (record_len < len) {
		dump_lock_wait(card_id);
		target_addr = switch_bar1_atu0(card_id, address + record_len);
		write_len = (BRA1_SPACE_SIZE_8K - (address + record_len - target_addr));
		write_len = ((len - record_len) > write_len) ? write_len : (len - record_len);
		read(file_fd, (void *)(g_dump_ctl[card_id].pci_bar1_map_addr + (address + record_len - target_addr)), write_len);
		record_len += write_len;
		restore_bar1_atu0(card_id);
		dump_lock_release(card_id);
		printf("\b\b\b\b\b\b[%3d%%]", (uint32_t)(record_len * 100 / len));
		fflush(stdout);
	}
	printf("\n");
	close(file_fd);

	return 0;
}

int vastai_dump_open(uint32_t card_id)
{
	char dev_path[128];
	int ret;

	if (1 == g_dump_ctl[card_id].is_dump_open)
		return 0;

	/* find vastai pci device */
	ret = get_dev_path(card_id, dev_path);
	if (ret) {
		printf("ERROR: cann't find pci dev, card_id:%d\n", card_id);
		return -1;
	}

	printf("INFO: Card ID:%d, Device path:%s\n", card_id, dev_path);

	/* get mmap pcie bar1 */
	ret = mmap_pcie_resource(card_id, dev_path);
	if (ret) {
		printf("ERROR: cann't mmap_pcie_resource, card_id:%d\n", card_id);
		return -1;
	}

	dump_lock_init(card_id);

	g_dump_ctl[card_id].is_dump_open = 1;

	return 0;
}

void vastai_dump_close(uint32_t card_id)
{
	if (0 == g_dump_ctl[card_id].is_dump_open)
		return;

	if (0 == g_dump_ctl[card_id].is_get_lock)
		dump_lock_wait(card_id);

	restore_bar1_atu0(card_id);
	dump_lock_release(card_id);

	munmap((void *)g_dump_ctl[card_id].pci_bar0_map_addr, BRA0_SPACE_SIZE_256K);
	close(g_dump_ctl[card_id].pci_bar0_fd);

	munmap((void *)g_dump_ctl[card_id].pci_bar1_map_addr, BRA1_SPACE_SIZE_8K);
	close(g_dump_ctl[card_id].pci_bar1_fd);

	g_dump_ctl[card_id].is_dump_open = 0;

	return;
}

int vastai_dump_map_bdf_to_cardid(char *dev_bdf)
{
	DIR *dir = opendir("/sys/bus/pci/devices");
	struct dirent *pdirent = NULL;
	char dev_path[384];
	char file_name[512];
	int fd;
	char buf[16];
	uint32_t count = 0;
	
	while((pdirent = readdir(dir)) != NULL) {
		if (NULL == strstr(pdirent->d_name, ".0"))
			continue;
		sprintf(dev_path, "/sys/bus/pci/devices/%s", pdirent->d_name);

		sprintf(file_name, "%s/vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, VENDOR_ID)) && (NULL == strstr(buf, VENDOR_ID_DC)))
			continue;

		sprintf(file_name, "%s/device", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, DEVICE_ID_0)) && (NULL == strstr(buf, DEVICE_ID_1))
			&& (NULL == strstr(buf, DEVICE_ID_2)) && (NULL == strstr(buf, DEVICE_ID_3)))
			continue;

		sprintf(file_name, "%s/subsystem_vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, SUB_VENDOR_ID)) && (NULL == strstr(buf, SUB_VENDOR_ID_DC)))
			continue;

		if (NULL != strstr(pdirent->d_name, dev_bdf)) {
			closedir(dir);
			return count;
		}
		
		count++;
	}

	closedir(dir);

	return -1;
}


