#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <time.h>
#include <sys/mman.h>
#include "vastai_pcie_jtag_api.h"

#define MAX_CARD_NUM		16

#define VENDOR_ID			"0x1ec6"
#define VENDOR_ID_DC		"0x1f4f"
#define DEVICE_ID_0			"0x0200"
#define DEVICE_ID_1			"0x0201"
#define DEVICE_ID_2			"0x0202"
#define DEVICE_ID_3			"0x0204"
#define SUB_VENDOR_ID		"0x1ec6"
#define SUB_VENDOR_ID_DC	"0x1f4f"

typedef union jtag_bus_status
{
	uint32_t val;
	struct {
		uint32_t tdo		:1;
		uint32_t tdi		:1;
		uint32_t tms		:1;
		uint32_t trst		:1;
		uint32_t tck		:1;
		uint32_t cap		:1;
		uint32_t rsv		:26;
	} bit;
}jtag_bus_status;

typedef union jtag2axi_master_dr
{
	uint32_t val[3];
	struct {
		uint32_t opt		:1;
		uint32_t data0_30	:31;
		uint32_t data31		:1;
		uint32_t addr0_30	:31;
		uint32_t addr31_39	:9;
		uint32_t rsv		:23;
	} bit;
}jtag2axi_master_dr;

typedef struct jtag2axi_slave_dr
{
	uint32_t data;
	uint8_t rsv0			:4;
	uint8_t ready			:1;
	uint8_t rsv1			:3;
}jtag2axi_slave_dr;

typedef struct vastai_jtag_ctl_s
{
	uint32_t jtag_clock;
	jtag_bus_status jtag_bus_status;
	int pci_fd;
	uint64_t pci_map_addr;
	volatile uint32_t * volatile jtag_reg_addr;
	volatile int cur_tap_state;
}vastai_jtag_ctl_t;

vastai_jtag_ctl_t g_jtag_ctl[MAX_CARD_NUM];

static int get_dev_path(uint32_t card_id, char *dev_path)
{
	DIR *dir = opendir("/sys/bus/pci/devices");
	struct dirent *pdirent = NULL;
	char file_name[128];
	int fd;
	char buf[16];
	uint32_t count = 0;
	
	while((pdirent = readdir(dir)) != NULL) {
		if (NULL == strstr(pdirent->d_name, ".0"))
			continue;
		sprintf(dev_path, "/sys/bus/pci/devices/%s", pdirent->d_name);

		sprintf(file_name, "%s/vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, VENDOR_ID)) && (NULL == strstr(buf, VENDOR_ID_DC)))
			continue;

		sprintf(file_name, "%s/device", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, DEVICE_ID_0)) && (NULL == strstr(buf, DEVICE_ID_1))
			&& (NULL == strstr(buf, DEVICE_ID_2)) && (NULL == strstr(buf, DEVICE_ID_3)))
			continue;

		sprintf(file_name, "%s/subsystem_vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, SUB_VENDOR_ID)) && (NULL == strstr(buf, SUB_VENDOR_ID_DC)))
			continue;

		if (count == card_id) {
			closedir(dir);
			return 0;
		}
		
		count++;
	}

	closedir(dir);

	return -1;
}

static int get_jtag_reg_addr(uint32_t card_id, const char * const dev_path)
{
	char dev_name[128];
	int pci_config_fd;
	uint32_t val;

	sprintf(dev_name, "%s/config", dev_path);
	pci_config_fd = open(dev_name, O_RDWR | O_SYNC);
	if (pci_config_fd < 0) {
		printf("ERROR: %s open failed\n", dev_name);
		return -1;
	}
	lseek(pci_config_fd, 4, SEEK_SET); 
	read(pci_config_fd, &val, 4);
	val |= 0x2;
	lseek(pci_config_fd, 4, SEEK_SET); 
	write(pci_config_fd, &val, 4);
	close(pci_config_fd);

	sprintf(dev_name, "%s/resource1", dev_path);
	g_jtag_ctl[card_id].pci_fd = open(dev_name, O_RDWR | O_SYNC);
	if (g_jtag_ctl[card_id].pci_fd < 0) {
		printf("ERROR: %s open failed\n", dev_name);
		return -1;
	}

	g_jtag_ctl[card_id].pci_map_addr = (uint64_t)mmap(NULL, 0x2000, PROT_READ | PROT_WRITE, MAP_SHARED, g_jtag_ctl[card_id].pci_fd, 0);
	g_jtag_ctl[card_id].jtag_reg_addr = (volatile uint32_t * volatile)(g_jtag_ctl[card_id].pci_map_addr + 0x14);

	return 0;
}

static uint8_t config_one_jtag_cycle(uint32_t card_id)
{
	uint8_t ret;
	struct timespec ts;

	ts.tv_sec = 0;
	ts.tv_nsec = 1000 / g_jtag_ctl[card_id].jtag_clock;

	g_jtag_ctl[card_id].jtag_bus_status.bit.tck = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;
	nanosleep(&ts, NULL);

	g_jtag_ctl[card_id].jtag_bus_status.bit.tck = 1;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;
	nanosleep(&ts, NULL);

	ret = (*g_jtag_ctl[card_id].jtag_reg_addr)&0x1;

	g_jtag_ctl[card_id].jtag_bus_status.bit.tck = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;
	nanosleep(&ts, NULL);

	return ret;
}

static void run_idle_loop(uint32_t card_id, uint32_t cycle_num)
{
	uint32_t i;

	if (TAP_IDLE != g_jtag_ctl[card_id].cur_tap_state)
		return;

	for (i = 0; i < cycle_num; i++) {
		g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 0;
		config_one_jtag_cycle(card_id);
	}

	return;
}

static void set_tap_reset(uint32_t card_id)
{
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].cur_tap_state = TAP_RESET;

	return;
}

static void set_tap_idle(uint32_t card_id)
{
	set_tap_reset(card_id);
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 0; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].cur_tap_state = TAP_IDLE;

	return;
}

void jtag_shift_ir(uint32_t card_id, const uint8_t * input_buf, uint8_t * output_buf, uint32_t ir_len)
{
	uint8_t ret;
	uint32_t i = 0;
	uint8_t tms = 0b00000001;

	do {
		if(i == (ir_len - 1)) {
			g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1;
			g_jtag_ctl[card_id].jtag_bus_status.bit.tdi = *(input_buf + (i / 8)) >> (i % 8);
			ret = config_one_jtag_cycle(card_id);
		} else {
			g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 0;
			g_jtag_ctl[card_id].jtag_bus_status.bit.tdi = *(input_buf + (i / 8)) >> (i % 8);
			ret = config_one_jtag_cycle(card_id);
		}

		if (NULL != output_buf) {
			*(output_buf + (i / 8)) += ret << (i % 8);
		}
		i++;
	} while(i < ir_len);

	vastai_jtag_shift_tms(card_id, &tms, 2);
	g_jtag_ctl[card_id].cur_tap_state = TAP_IDLE;

	return;
}

void jtag_shift_dr(uint32_t card_id, const uint8_t * input_buf, uint8_t * output_buf, uint32_t dr_len)
{
	uint8_t ret;
	uint32_t i = 0;
	uint8_t tms = 0b00000001;

	do {
		if(i == (dr_len - 1)){
			g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1;
			g_jtag_ctl[card_id].jtag_bus_status.bit.tdi = *(input_buf + (i / 8)) >> (i % 8);
			ret = config_one_jtag_cycle(card_id);
		} else {
			g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 0;
			g_jtag_ctl[card_id].jtag_bus_status.bit.tdi = *(input_buf + (i / 8))>>(i % 8);
			ret = config_one_jtag_cycle(card_id);
		}

		if (NULL != output_buf) {
			*(output_buf + (i / 8)) += ret << (i % 8);
		}
		i++;
	} while(i < dr_len);

	vastai_jtag_shift_tms(card_id, &tms, 2);
	g_jtag_ctl[card_id].cur_tap_state = TAP_IDLE;

	return;
}

void vastai_jink_set_tck(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tck = 1;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

void vastai_jink_clr_tck(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tck = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

void vastai_jink_set_tdi(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tdi = 1;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

void vastai_jink_clr_tdi(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tdi = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

void vastai_jink_set_tms(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 1;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

void vastai_jink_clr_tms(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tms = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

void vastai_jink_set_trst(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.trst = 1;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

void vastai_jink_clr_trst(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.trst = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	return;
}

uint8_t vastai_jink_toggle_clk(uint32_t card_id)
{
	struct timespec ts;

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return 0;
	}

	ts.tv_sec = 0;
	ts.tv_nsec = 1000 / g_jtag_ctl[card_id].jtag_clock;

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tck = 1;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;
	nanosleep(&ts, NULL);

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;
	g_jtag_ctl[card_id].jtag_bus_status.bit.tck = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;
	nanosleep(&ts, NULL);
	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;

	return g_jtag_ctl[card_id].jtag_bus_status.bit.tdo;
}

void vastai_jtag_shift_tms(uint32_t card_id, const uint8_t * tms, uint32_t num_bits)
{
	uint32_t i;

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = *g_jtag_ctl[card_id].jtag_reg_addr;

	for (i = 0; i < num_bits; i++) {
		g_jtag_ctl[card_id].jtag_bus_status.bit.tms = *(tms + (i / 8)) >> (i % 8);
		g_jtag_ctl[card_id].jtag_bus_status.bit.tdi = 0;
		config_one_jtag_cycle(card_id);
	}

	g_jtag_ctl[card_id].cur_tap_state = TAP_UNINIT;

	return;
}

void vastai_jtag_set_ir(uint32_t card_id, const uint8_t * input_buf, uint8_t * output_buf, uint32_t ir_len)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	vastai_jtag_switch_tap_state(card_id, TAP_SHIFT_IR);
	jtag_shift_ir(card_id, input_buf, output_buf, ir_len);

	return;
}

void vastai_jtag_set_dr(uint32_t card_id, const uint8_t * input_buf, uint8_t * output_buf, uint32_t dr_len)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	vastai_jtag_switch_tap_state(card_id, TAP_SHIFT_DR);
	jtag_shift_dr(card_id, input_buf, output_buf, dr_len);

	return;
}

void vastai_jtag_main_tap_ir_sel(uint32_t card_id, uint8_t value)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	vastai_jtag_set_ir(card_id, &value, NULL, VASTAI_IR_LEN);

	return;
}

int vastai_jtag_get_device_id(uint32_t card_id)
{
	uint8_t tdi[4] = {0x0};
	uint8_t tdo[4] = {0x0};

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return 0;
	}

	vastai_jtag_main_tap_ir_sel(card_id, IR_MAIN_IDCODE);
	vastai_jtag_set_dr(card_id, &tdi[0], &tdo[0], 32);

	return *((unsigned long *)tdo);
}

int vastai_jtag_switch_tap_state(uint32_t card_id, int new_state)
{
	unsigned char tms[16] = {0x0};
	int num_bits;

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return TAP_UNINIT;
	}

	if (g_jtag_ctl[card_id].cur_tap_state == TAP_UNINIT) {
		switch (new_state) {
			case TAP_RESET:
				tms[0] = 0b00011111;
				num_bits = 5;
				break;
			case TAP_IDLE:
				tms[0] = 0b00011111;
				num_bits = 6;
				break;
			case TAP_DR_SCAN:
				tms[0] = 0b01011111;
				num_bits = 7;
				break;
			case TAP_CAPTURE_DR:
				tms[0] = 0b01011111;
				num_bits = 8;
				break;
			case TAP_SHIFT_DR:
				tms[0] = 0b01011111;
				tms[1] = 0b00000000;
				num_bits = 9;
				break;
			case TAP_EXIT1_DR:
				tms[0] = 0b01011111;
				tms[1] = 0b00000010;
				num_bits = 10;
				break;
			case TAP_PAUSE_DR:
				tms[0] = 0b01011111;
				tms[1] = 0b00000010;
				num_bits = 11;
				break;
			case TAP_EXIT2_DR:
				tms[0] = 0b01011111;
				tms[1] = 0b00001010;
				num_bits = 12;
				break;
			case TAP_UPDATE_DR:
				tms[0] = 0b01011111;
				tms[1] = 0b00000110;
				num_bits = 11;
				break;

			case TAP_IR_SCAN:
				tms[0] = 0b11011111;
				num_bits = 8;
				break;
			case TAP_CAPTURE_IR:
				tms[0] = 0b11011111;
				tms[1] = 0b00000000;
				num_bits = 9;
				break;
			case TAP_SHIFT_IR:
				tms[0] = 0b11011111;
				tms[1] = 0b00000000;
				num_bits = 10;
				break;
			case TAP_EXIT1_IR:
				tms[0] = 0b11011111;
				tms[1] = 0b00000100;
				num_bits = 11;
				break;
			case TAP_PAUSE_IR:
				tms[0] = 0b11011111;
				tms[1] = 0b00000100;
				num_bits = 12;
				break;
			case TAP_EXIT2_IR:
				tms[0] = 0b11011111;
				tms[1] = 0b00010100;
				num_bits = 13;
				break;
			case TAP_UPDATE_IR:
				tms[0] = 0b11011111;
				tms[1] = 0b00001100;
				num_bits = 12;
				break;

			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	} 
	else if ((g_jtag_ctl[card_id].cur_tap_state == TAP_IDLE) || (g_jtag_ctl[card_id].cur_tap_state == TAP_UPDATE_DR) || (g_jtag_ctl[card_id].cur_tap_state == TAP_UPDATE_IR)) {
		switch (new_state) {
			case TAP_DR_SCAN:
				tms[0] = 0b00000001;
				num_bits = 1;
				break;
			case TAP_CAPTURE_DR:
				tms[0] = 0b00000001;
				num_bits = 2;
				break;
			case TAP_SHIFT_DR:
				tms[0] = 0b00000001;
				num_bits = 3;
				break;
			case TAP_IR_SCAN:
				tms[0] = 0b00000011;
				num_bits = 2;
				break;
			case TAP_CAPTURE_IR:
				tms[0] = 0b00000011;
				num_bits = 3;
				break;
			case TAP_SHIFT_IR:
				tms[0] = 0b00000011;
				num_bits = 4;
				break;
			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	} 
	else if (g_jtag_ctl[card_id].cur_tap_state == TAP_CAPTURE_DR) {
		switch (new_state) {
			case TAP_SHIFT_DR:
				tms[0] = 0b00000000;
				num_bits = 1;
				break;
			case TAP_EXIT1_DR:
				tms[0] = 0b00000001;
				num_bits = 1;
				break;
			case TAP_UPDATE_DR:
				tms[0] = 0b00000011;
				num_bits = 2;
				break;
			case TAP_IDLE:
				tms[0] = 0b00000011;
				num_bits = 3;
				break;
			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	} 
	else if (g_jtag_ctl[card_id].cur_tap_state == TAP_SHIFT_DR) {
		switch (new_state) {
			case TAP_PAUSE_DR:
				tms[0] = 0b00000001;
				num_bits = 2;
				break;
			case TAP_UPDATE_DR:
				tms[0] = 0b00000011;
				num_bits = 2;
				break;
			case TAP_IDLE:
				tms[0] = 0b00000011;
				num_bits = 3;
				break;
			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	}
	else if (g_jtag_ctl[card_id].cur_tap_state == TAP_PAUSE_DR) {
		switch (new_state) {
			case TAP_SHIFT_DR:
				tms[0] = 0b00000001;
				num_bits = 2;
				break;
			case TAP_UPDATE_DR:
				tms[0] = 0b00000011;
				num_bits = 2;
				break;
			case TAP_IDLE:
				tms[0] = 0b00000011;
				num_bits = 3;
				break;
			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	}
	else if (g_jtag_ctl[card_id].cur_tap_state == TAP_CAPTURE_IR) {
		switch (new_state) {
			case TAP_SHIFT_IR:
				tms[0] = 0b00000000;
				num_bits = 1;
				break;
			case TAP_EXIT1_IR:
				tms[0] = 0b00000001;
				num_bits = 1;
				break;
			case TAP_UPDATE_IR:
				tms[0] = 0b00000011;
				num_bits = 2;
				break;
			case TAP_IDLE:
				tms[0] = 0b00000011;
				num_bits = 3;
				break;
			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	}
	else if (g_jtag_ctl[card_id].cur_tap_state == TAP_SHIFT_IR) {
		switch (new_state) {
			case TAP_PAUSE_IR:
				tms[0] = 0b00000001;
				num_bits = 2;
				break;
			case TAP_UPDATE_IR:
				tms[0] = 0b00000011;
				num_bits = 2;
				break;
			case TAP_IDLE:
				tms[0] = 0b00000011;
				num_bits = 3;
				break;
			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	}
	else if (g_jtag_ctl[card_id].cur_tap_state == TAP_PAUSE_IR) {
		switch (new_state) {
			case TAP_SHIFT_IR:
				tms[0] = 0b00000001;
				num_bits = 2;
				break;
			case TAP_UPDATE_IR:
				tms[0] = 0b00000011;
				num_bits = 2;
				break;
			case TAP_IDLE:
				tms[0] = 0b00000011;
				num_bits = 3;
				break;
			default:
				printf("Invalid arg, old_state:%d, new_state:%d\n", g_jtag_ctl[card_id].cur_tap_state, new_state);
				return -1;
		}
	}

	vastai_jtag_shift_tms(card_id, tms, num_bits);
	g_jtag_ctl[card_id].cur_tap_state = new_state;

	return 0;
}

int vastai_jtag_get_tap_state(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return TAP_UNINIT;
	}

	return g_jtag_ctl[card_id].cur_tap_state;
}

uint32_t vastai_jtag_set_clock(uint32_t card_id, uint32_t clock_freq)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return 0;
	}

	if ((50 < clock_freq ) || (clock_freq < 1)) {
		printf("ERROR: jtag_set_clock failed, clock %dMhz\n", clock_freq);
		return g_jtag_ctl[card_id].jtag_clock;
	}

	g_jtag_ctl[card_id].jtag_clock = clock_freq;

	return g_jtag_ctl[card_id].jtag_clock;
}

int vastai_jtag_axi_read(uint32_t card_id, uint64_t address, uint32_t *value)
{
	uint8_t tdi[5] = {0};
	int i = 0;
	jtag2axi_slave_dr slv_rcv;
	jtag2axi_master_dr mst_cfg;

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return -1;
	}

	if (NULL == value) {
		printf("ERROR: jtag_axi_read failed, empty point\n");
		return -1;
	}

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_AXI_MASTER);
	memset((void *)mst_cfg.val, 0, sizeof(mst_cfg));
	mst_cfg.bit.opt = 1;
	mst_cfg.bit.addr0_30 = address & 0x7FFFFFFF;
	mst_cfg.bit.addr31_39 = (address >> 31) & 0x1FF;
	vastai_jtag_set_dr(card_id, (uint8_t *)(mst_cfg.val), NULL, 73);

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_AXI_SLAVE);
	memset((void *)&slv_rcv, 0, sizeof(slv_rcv));
	do {
		vastai_jtag_set_dr(card_id, &tdi[0], (uint8_t *)(&slv_rcv), 37);
		if (i++ > 100) {
			printf("ERROR: jtag_axi_read failed\n");
			return -1;
		}
	} while (slv_rcv.ready == 0);
	*value = slv_rcv.data;

	return 0;
}

int vastai_jtag_axi_write(uint32_t card_id, uint64_t address, uint32_t value)
{
	uint8_t tdi[5] = {0};
	int i = 0;
	jtag2axi_slave_dr slv_rcv;
	jtag2axi_master_dr mst_cfg;

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return -1;
	}

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_AXI_MASTER);
	memset((void *)mst_cfg.val, 0, sizeof(mst_cfg));
	mst_cfg.bit.opt = 0;
	mst_cfg.bit.data0_30 = value & 0x7FFFFFFF;
	mst_cfg.bit.data31 = (value >> 31) & 0x1;
	mst_cfg.bit.addr0_30 = address & 0x7FFFFFFF;
	mst_cfg.bit.addr31_39 = (address >> 31) & 0x1FF;
	vastai_jtag_set_dr(card_id, (uint8_t *)(mst_cfg.val), NULL, 73);

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_AXI_SLAVE);
	memset((void *)&slv_rcv, 0, sizeof(slv_rcv));
	do {
		vastai_jtag_set_dr(card_id, &tdi[0], (uint8_t *)(&slv_rcv), 37);
		if (i++ > 100) {
			printf("ERROR: jtag_axi_write failed\n");
			return -1;
		}
	} while (slv_rcv.ready == 0);

	return 0;
}

int vastai_jtag_fuse_sram_read(uint32_t card_id, uint32_t address, uint32_t *value)
{
	uint8_t tdo[23]         = {0};
	uint8_t tdi_bypass[3]   = {0b00000000, 0b00000000, 0b1110000};
	uint8_t tdi_0[10]       = {0b00000000, 0b00000000, 0b00000001, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b000000};
	uint8_t tdi_1[10]       = {0b00000000, 0b00000000, 0b00000101, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000};
	uint8_t tdi_2[11]       = {0b00000000, 0b00000000, 0b10000101, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000};
	uint8_t tdi_3[23]       = {0b00000000, 0b00000000, 0b10000101, 0b00000001, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00100100, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00};
	uint8_t tdi_4[23]       = {0b00000000, 0b00000000, 0b10000101, 0b11111111, 0b00000011, 0b00000000, 0b00000000, 0b00000000, 0b10000100, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00};
	uint8_t tdi_5[23]       = {0b00000000, 0b00000000, 0b00000000, 0b11111111, 0b00000011, 0b00000000, 0b00000000, 0b00000000, 0b10000100, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00};
	uint32_t addr;

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return -1;
	}

	if ((address >= 0x10800) || (address < 0x10000) || (0 != (address % 4))) {
		printf("ERROR: jtag_fuse_sram_read failed, error addr:0x%x\n", address);
		return -1;
	}

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_BYPASS_SLAVE_TAP);
	vastai_jtag_set_dr(card_id, &tdi_bypass[0], NULL, 23);

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_TILE);
	vastai_jtag_set_dr(card_id, &tdi_0[0], NULL, 78);
	run_idle_loop(card_id, 5);
	vastai_jtag_set_dr(card_id, &tdi_1[0], NULL, 80);
	run_idle_loop(card_id, 5);
	vastai_jtag_set_dr(card_id, &tdi_2[0], NULL, 85);
	run_idle_loop(card_id, 5);
	vastai_jtag_set_dr(card_id, &tdi_3[0], NULL, 178);
	run_idle_loop(card_id, 20);
	/* fuse read config */
	addr = ((address - 0x10000) / 4) & 0x1FF;
	tdi_4[3] = ((addr & 0x7F) << 1) + (tdi_4[3] & 0b00000001);
	tdi_4[4] = ((addr >> 7) & 0x3) + (tdi_4[4] & 0b11111100);
	vastai_jtag_set_dr(card_id, &tdi_4[0], NULL, 178);
	run_idle_loop(card_id, 20);
	/* fuse read out */
	memset((void *)tdo, 0, sizeof(tdo));
	vastai_jtag_set_dr(card_id, &tdi_5[0], tdo, 178);
	run_idle_loop(card_id, 20);
	if (tdo[14] & 0x4) {
		printf("ERROR: jtag_fuse_sram_read failed, efuse busy\n");
		return -1;
	}
	*value = (tdo[10] >> 2) + (tdo[11] << 6) + (tdo[12] << 14) + (tdo[13] << 22) + ((tdo[14] & 0x3) << 30);

	vastai_jtag_main_tap_ir_sel(card_id, IR_MAIN_IDCODE);

	return 0;
}

int vastai_jtag_fuse_sram_write(uint32_t card_id, uint32_t address, uint32_t value)
{
	uint8_t tdi_bypass[3]   = {0b00000000, 0b00000000, 0b1110000};
	uint8_t tdi_0[10]       = {0b00000000, 0b00000000, 0b00000001, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b000000};
	uint8_t tdi_1[10]       = {0b00000000, 0b00000000, 0b00000101, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000};
	uint8_t tdi_2[11]       = {0b00000000, 0b00000000, 0b10000101, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000};
	uint8_t tdi_3[23]       = {0b00000000, 0b00000000, 0b10000101, 0b00000001, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00100100, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00};
	uint8_t tdi_4[23]       = {0b00000000, 0b00000000, 0b00000000, 0b11111111, 0b00111111, 0b11111110, 0b11010011, 0b11100010, 0b10100101, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00000000, 0b00};
	uint32_t addr;

	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return -1;
	}

	if ((address >= 0x10800) || (address < 0x10000) || (0 != (address % 4))) {
		printf("ERROR: jtag_fuse_sram_write failed, error addr:0x%x\n", address);
		return -1;
	}

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_BYPASS_SLAVE_TAP);
	vastai_jtag_set_dr(card_id, &tdi_bypass[0], NULL, 23);

	vastai_jtag_main_tap_ir_sel(card_id, IR_HOSTIJTAG_TILE);
	vastai_jtag_set_dr(card_id, &tdi_0[0], NULL, 78);
	run_idle_loop(card_id, 5);
	vastai_jtag_set_dr(card_id, &tdi_1[0], NULL, 80);
	run_idle_loop(card_id, 5);
	vastai_jtag_set_dr(card_id, &tdi_2[0], NULL, 85);
	run_idle_loop(card_id, 5);
	vastai_jtag_set_dr(card_id, &tdi_3[0], NULL, 178);
	run_idle_loop(card_id, 20);
	/* fuse write */
	addr = ((address - 0x10000) / 4) & 0x1FF;
	tdi_4[3] = ((addr & 0x7F) << 1) + (tdi_4[3] & 0b00000001);
	tdi_4[4] = ((addr >> 7) & 0x3) + ((value & 0b111111) << 2);
	tdi_4[5] = (value >> 6);
	tdi_4[6] = (value >> 14);
	tdi_4[7] = (value >> 22);
	tdi_4[8] = (value >> 30) + (tdi_4[8] & 0b11111100);
	vastai_jtag_set_dr(card_id, &tdi_4[0], NULL, 178);
	run_idle_loop(card_id, 20);

	vastai_jtag_main_tap_ir_sel(card_id, IR_MAIN_IDCODE);

	return 0;
}

int vastai_jtag_fuse_sram_read_buf(uint32_t card_id, uint32_t address, uint32_t len, uint32_t *buf)
{
	uint32_t i;
	int ret;

	if (NULL == buf) {
		printf("ERROR: jtag_fuse_sram_read_buf failed, empty point\n");
		return -1;
	}

	for (i = 0; i < len; i++) {
		ret = vastai_jtag_fuse_sram_read(card_id, address + i * 4, buf + i);
		if (ret) {
			printf("ERROR: jtag_fuse_sram_read_buf failed\n");
			return -1;
		}
	}

	return 0;
}

int vastai_jtag_fuse_sram_write_buf(uint32_t card_id, uint32_t address, uint32_t len, uint32_t *buf)
{
	uint32_t i;
	int ret;

	if (NULL == buf) {
		printf("ERROR: jtag_fuse_sram_write_buf failed, empty point\n");
		return -1;
	}

	for (i = 0; i < len; i++) {
		ret = vastai_jtag_fuse_sram_write(card_id, address + i * 4, *(buf + i));
		if (ret) {
			printf("ERROR: jtag_fuse_sram_write_buf failed\n");
			return -1;
		}
	}

	return 0;
}

void vastai_jtag_trst(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr) {
		printf("ERROR: jtag not open.\n");
		return;
	}

	g_jtag_ctl[card_id].jtag_bus_status.bit.trst = 0; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].jtag_bus_status.bit.trst = 1; config_one_jtag_cycle(card_id);
	g_jtag_ctl[card_id].cur_tap_state = TAP_RESET;

	return;
}

int vastai_jtag_open(uint32_t card_id)
{
	char dev_path[128];
	int ret;

	if (NULL != g_jtag_ctl[card_id].jtag_reg_addr)
		return 0;

	/* find vastai pci device */
	ret = get_dev_path(card_id, dev_path);
	if (ret) {
		printf("ERROR: cann't find pci dev\n");
		return -1;
	}

	printf("INFO: Card ID:%d, Device path:%s\n", card_id, dev_path);

	/* get jtag reg addr */
	ret = get_jtag_reg_addr(card_id, dev_path);
	if (ret) {
		printf("ERROR: cann't get jtag reg\n");
		return -1;
	}

	g_jtag_ctl[card_id].jtag_bus_status.val = 0;
	g_jtag_ctl[card_id].jtag_clock = 10;			/* 10Mhz */
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;
	vastai_jtag_trst(card_id);

	set_tap_idle(card_id);

	return 0;
}

void vastai_jtag_close(uint32_t card_id)
{
	if (NULL == g_jtag_ctl[card_id].jtag_reg_addr)
		return;

	vastai_jtag_trst(card_id);
	g_jtag_ctl[card_id].jtag_bus_status.val = 0;
	*g_jtag_ctl[card_id].jtag_reg_addr = g_jtag_ctl[card_id].jtag_bus_status.val;

	g_jtag_ctl[card_id].jtag_reg_addr = NULL;
	munmap((void *)g_jtag_ctl[card_id].pci_map_addr, 0x2000);
	close(g_jtag_ctl[card_id].pci_fd);

	return;
}

int vastai_jtag_map_bdf_to_cardid(char *dev_bdf)
{
	DIR *dir = opendir("/sys/bus/pci/devices");
	struct dirent *pdirent = NULL;
	char dev_path[384];
	char file_name[512];
	int fd;
	char buf[16];
	uint32_t count = 0;
	
	while((pdirent = readdir(dir)) != NULL) {
		if (NULL == strstr(pdirent->d_name, ".0"))
			continue;
		sprintf(dev_path, "/sys/bus/pci/devices/%s", pdirent->d_name);

		sprintf(file_name, "%s/vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, VENDOR_ID)) && (NULL == strstr(buf, VENDOR_ID_DC)))
			continue;

		sprintf(file_name, "%s/device", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, DEVICE_ID_0)) && (NULL == strstr(buf, DEVICE_ID_1))
			&& (NULL == strstr(buf, DEVICE_ID_2)) && (NULL == strstr(buf, DEVICE_ID_3)))
			continue;

		sprintf(file_name, "%s/subsystem_vendor", dev_path);
		fd = open(file_name, O_RDONLY);
		if (fd < 0)
			continue;
		lseek(fd, 0, SEEK_SET);
		read(fd, buf, 16);
		close(fd);
		if ((NULL == strstr(buf, SUB_VENDOR_ID)) && (NULL == strstr(buf, SUB_VENDOR_ID_DC)))
			continue;

		if (NULL != strstr(pdirent->d_name, dev_bdf)) {
			closedir(dir);
			return count;
		}
		
		count++;
	}

	closedir(dir);

	return -1;
}

