﻿#include "vatools.h"
#include "vatools_log.h"

// runtime 实现了动态开关DLC功能. 在DLC 关闭时, 不能通过perfmon读取数据.
// 会因为访问未知地址 hang. 暂时注释下面代码, 不读取DLC和ODMA
#define VATOOLS_PROFILER_DLC_ODMA (0)

// 使用全局变量标识是否开启profiler, fifo地址
unsigned int profiler_timeline_enable = 0;
void *gp_profiler_dma_fifo = NULL;
void *gp_profiler_msg_fifo = NULL;
struct mutex pcie_timeline_dma_mutex;
struct mutex pcie_timeline_msg_mutex;

/// 读取pcie dma 性能数据, 函数执行频繁, 要求执行时间尽可能短, 不可以阻塞
void vatools_profiler_dma_checkpoint(
	struct s_profiler_dma_checkpoint *dma_section)
{
	V_UNREFERENCE(dma_section);

#ifdef ENABLE_VATOOLS_PROFILER
	struct vastai_fifo *fifo = (struct vastai_fifo *)gp_profiler_dma_fifo;
	u8 *fifo_p = NULL;
	int locked;

	//debug
	VATOOLS_DBG(
		NULL, DUMMY_DIE_ID,
		"dma_checkpoint[%d]: %u %u %u | %llu %u %u | %llu %llu %u %u %llu\n",
		profiler_timeline_enable, dma_section->dev_id,
		dma_section->die_id, dma_section->die_index,
		dma_section->dma_axi_addr, dma_section->dma_data_size,
		dma_section->dma_dir, dma_section->time_of_driver_recv_dma_req,
		dma_section->time_of_host_trigger_dma,
		dma_section->time_of_dma_run, dma_section->time_of_dma_finish,
		dma_section->time_of_driver_notify_host_dma_done);

	//TODO: 如果不开启timeline 数据采集, 直接返回
	if (!profiler_timeline_enable) {
		VATOOLS_DBG(NULL, DUMMY_DIE_ID,
			    "profiler pcie timeline is disable\n");
		return;
	}

	if (fifo == NULL) {
		VATOOLS_INFO(NULL, DUMMY_DIE_ID, "dma fifo is NULL\n");
		return;
	}

	//if (vastai_fifo_is_full(fifo) || !vastai_is_fifo_valid(fifo)) {
	if (!vastai_is_fifo_valid(fifo)) {
		VATOOLS_INFO(NULL, DUMMY_DIE_ID, "dma fifo is not valid\n");
		return;
	}

	if (vastai_fifo_is_full(fifo)) {
		VATOOLS_INFO(NULL, DUMMY_DIE_ID, "dma fifo is full\n");
		locked = mutex_trylock(&pcie_timeline_dma_mutex);
		if (!locked) {
			VATOOLS_DBG(NULL, DUMMY_DIE_ID, "dma trylock fail=%d\n",
				    locked);
			return;
		}
	}

	fifo_p = (u8 *)fifo + vastai_fifo_wr_next(fifo);
	memcpy(fifo_p, dma_section, fifo->elem_size);
	mutex_unlock(&pcie_timeline_dma_mutex);
#endif
	return;
}

/// 清空 dma buffer
static void vatools_profiler_dma_checkpoint_clear(void)
{
#ifdef ENABLE_VATOOLS_PROFILER
	struct vastai_fifo *fifo = (struct vastai_fifo *)gp_profiler_dma_fifo;
	mutex_trylock(
		&pcie_timeline_dma_mutex); //TODO: change mutex_trylock to mutex_lock
	fifo->rd = 0;
	fifo->wr = 0;
	memset(fifo->buf, 0, fifo->elem_count * fifo->elem_size);
	mutex_unlock(&pcie_timeline_dma_mutex);
#endif
}

/// 读取pcie msg 性能数据, 函数执行频繁, 要求执行时间尽可能短, 不可以阻塞
void vatools_profiler_msg_checkpoint(
	struct s_profiler_msg_checkpoint *msg_section)
{
	V_UNREFERENCE(msg_section);
#ifdef ENABLE_VATOOLS_PROFILER
	struct vastai_fifo *fifo = (struct vastai_fifo *)gp_profiler_msg_fifo;
	u8 *fifo_p = NULL;
	int locked;

	VATOOLS_DBG(NULL, DUMMY_DIE_ID,
		    "msg_checkpoint[%d]: %u %u %u | %llu %llu\n",
		    profiler_timeline_enable, msg_section->dev_id,
		    msg_section->die_id, msg_section->die_index,
		    msg_section->time_of_host_send_msg,
		    msg_section->time_of_fw_recv_msg);

	//TODO: 如果不开启timeline 数据采集, 直接返回
	if (!profiler_timeline_enable) {
		VATOOLS_DBG(NULL, DUMMY_DIE_ID,
			    "profiler pcie timeline is disable\n");
		return;
	}

	if (fifo == NULL) {
		VATOOLS_INFO(NULL, DUMMY_DIE_ID, "msg fifo is NULL\n");
		return;
	}

	//if (vastai_fifo_is_full(fifo) || !vastai_is_fifo_valid(fifo)) {
	if (!vastai_is_fifo_valid(fifo)) {
		VATOOLS_INFO(NULL, DUMMY_DIE_ID, "msg fifo is not valid\n");
		return;
	}
	if (vastai_fifo_is_full(fifo)) {
		locked = mutex_trylock(&pcie_timeline_msg_mutex);
		if (!locked) {
			VATOOLS_DBG(NULL, DUMMY_DIE_ID, "msg trylock fail=%d\n",
				    locked);
			return;
		}
	}

	fifo_p = (u8 *)fifo + vastai_fifo_wr_next(fifo);
	memcpy(fifo_p, msg_section, fifo->elem_size);
	mutex_unlock(&pcie_timeline_msg_mutex);
#endif
	return;
}

/// 清空 msg buffer
static void vatools_profiler_msg_checkpoint_clear(void)
{
#ifdef ENABLE_VATOOLS_PROFILER
	struct vastai_fifo *fifo = (struct vastai_fifo *)gp_profiler_msg_fifo;
	mutex_trylock(
		&pcie_timeline_msg_mutex); //TODO: change mutex_trylock to mutex_lock
	fifo->rd = 0;
	fifo->wr = 0;
	memset(fifo->buf, 0, fifo->elem_count * fifo->elem_size);
	mutex_unlock(&pcie_timeline_msg_mutex);
#endif
}

loff_t profiler_llseek(struct file *filp, loff_t offset, int orig)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(offset);
	V_UNREFERENCE(orig);
	return 0;
}

ssize_t profiler_read(struct file *filp, char __user *buf, size_t size,
		      loff_t *pos)
{
	// struct vatools_node*    node        = vatools_file_get_node( filp );
	struct vatools_reader *reader = vatools_file_get_reader(filp);
	T_SMI_BLOCK *p_smi_block = NULL;
	ssize_t ret = 0;
	struct vastai_pci_info *priv = NULL;
	int die_index = 0;
	int device_id = 0;

	size_t length = 0;
	struct T_PERFMON_DATA data;

	// read data from fw
	T_PERFMON_FW perf_buffer;
	int index = 0;
#if (VATOOLS_PROFILER_DLC_ODMA == 1)
	u64 odma_addr[MAX_ODMA_NUM] = { csr_odma0_perfmon_en,
					csr_odma1_perfmon_en };

	u64 oak_addr[MAX_DLC_NUM] = {
		csr_oak0_perfmon_en, csr_oak1_perfmon_en, csr_oak2_perfmon_en,
		csr_oak3_perfmon_en, csr_oak4_perfmon_en, csr_oak5_perfmon_en,
		csr_oak6_perfmon_en, csr_oak7_perfmon_en
	};
#endif

	u64 mc_addr[MAX_MC_NUM] = { csr_mc0_axi_w_r_ostd, csr_mc1_axi_w_r_ostd,
				    csr_mc2_axi_w_r_ostd,
				    csr_mc3_axi_w_r_ostd };
	T_PERFMON_MC_FW mc_buffer;

	V_UNREFERENCE(pos);
	VATOOLS_FUNC_ENTERY;

	memset(&data, 0x00, sizeof(struct T_PERFMON_DATA));
	memset(&perf_buffer, 0x00, sizeof(T_PERFMON_FW));
	memset(&mc_buffer, 0x00, sizeof(T_PERFMON_MC_FW));

	VATOOLS_DBG(priv, die_index,
		    "app_category=%d  block_id=%d  device_id=%d  die_id=%d"
		    "die_index=0x%x\n",
		    reader->trans_category.app_category,
		    reader->trans_category.block_id,
		    reader->trans_category.device.dev_id,
		    reader->trans_category.device.die_id,
		    reader->trans_category.device.die_index);

	p_smi_block = vatools_fw_smi_get_block_via_app_category(
		reader->trans_category.app_category);

	ret = 0;
	priv = vatools_get_vastai_pci_device_info(
		(char)reader->trans_category.device.dev_id);
	if (priv == NULL) {
		VATOOLS_INFO(priv, die_index, "priv is NULL, dev_id=0x%x\n",
			     reader->trans_category.device.dev_id);
		ret = -ENODEV;
		goto profiler_read_out;
	}
	if (vatools_check_pci_state(priv, VASTAI_NORMAL_STATE,
				    VASTAI_DEBUG_STATE, -1)) {
		VATOOLS_DBG(priv, die_index, "priv->pci_state=0x%x\n",
			    atomic_read(&priv->pci_state));
		ret = -ENODEV;
		goto profiler_read_out;
	}

	die_index = reader->trans_category.device.die_index;
	device_id = reader->trans_category.device.dev_id;

	length = min(size, sizeof(struct T_PERFMON_DATA));

	VATOOLS_DBG(priv, die_index,
		    "priv=0x%p, device_id=%d, die_index=0x%x\n", priv,
		    device_id, die_index);

	VATOOLS_DBG(priv, die_index,
		    "sizeof(T_PERFMON_DATA)=%ld, size=%ld, min length=%ld\n",
		    sizeof(struct T_PERFMON_DATA), size, length);
#if (VATOOLS_PROFILER_DLC_ODMA == 1)
	// runtime 实现了动态开关DLC功能. 在DLC 关闭时, 不能通过perfmon读取数据.
	// 会因为访问未知地址 hang. 暂时注释下面代码, 不读取DLC和ODMA

	// ODMA0
	for (index = 0; index < sizeof(odma_addr) / sizeof(u64); index++) {
		memset(&perf_buffer, 0x00, sizeof(T_PERFMON_FW));
		ret = vatools_fw_smcu_read(p_smi_block->block_id, priv,
					   die_index, odma_addr[index],
					   &perf_buffer, sizeof(T_PERFMON_FW));
		if (ret < 0) {
			VATOOLS_INFO(priv, die_index,
				     "[odma index %d] ret=%ld\n", index, ret);
			goto profiler_read_out;
		}
		memcpy(&(data.tODMA[index]), &perf_buffer,
		       sizeof(T_PERFMON_FW));
	}

	// OAK
	for (index = 0; index < MAX_DLC_NUM; index++) {
		memset(&perf_buffer, 0x00, sizeof(T_PERFMON_FW));
		ret = vatools_fw_smcu_read(p_smi_block->block_id, priv,
					   die_index, oak_addr[index],
					   &perf_buffer, sizeof(T_PERFMON_FW));
		if (ret < 0) {
			VATOOLS_INFO(priv, die_index,
				     "[oak index %d] ret=%ld\n", index, ret);
			goto profiler_read_out;
		}
		memcpy(&(data.tDLC[index]), &perf_buffer, sizeof(T_PERFMON_FW));
	}
#endif
	// MC
	for (index = 0; index < MAX_MC_NUM; index++) {
		memset(&mc_buffer, 0x00, sizeof(T_PERFMON_MC_FW));
		ret = vatools_fw_smcu_read(p_smi_block->block_id, priv,
					   die_index, mc_addr[index],
					   &mc_buffer, sizeof(T_PERFMON_MC_FW));
		if (ret < 0) {
			VATOOLS_INFO(priv, die_index, "[mc index %d] ret=%ld\n",
				     index, ret);
			goto profiler_read_out;
		}
		memcpy(&(data.tMC[index]), &mc_buffer, sizeof(T_PERFMON_MC_FW));
	}

	VATOOLS_DUMP_BRIEF("profiler_read buffer: ", &data, length);

	if (copy_to_user_ex((void __user *)buf, &data, length)) {
		ret = -EFAULT;
		VATOOLS_ERR(priv, die_index, "ret=%ld\n", ret);
		goto profiler_read_out;
	}
	ret = length;

profiler_read_out:
	VATOOLS_DBG(priv, die_index, "ret=%ld\n", ret);
	VATOOLS_FUNC_EXIT;
	return ret;
}

ssize_t profiler_write(struct file *filp, const char __user *buf, size_t size,
		       loff_t *pos)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(buf);
	V_UNREFERENCE(size);
	V_UNREFERENCE(pos);

	return 0;
}

ssize_t profiler_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
	V_UNREFERENCE(iocb);
	V_UNREFERENCE(from);

	return 0;
}

unsigned int profiler_poll(struct file *filp, struct poll_table_struct *wait)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(wait);

	return 0;
}

long profiler_set_performance_csr(struct file *filp, unsigned int cmd,
				  IOCTL_ARG_T arg)
{
	//struct vatools_node* node = vatools_file_get_node( filp );
	// struct vatools_reader*  reader   = vatools_file_get_reader( filp );
	long ret = -EINVAL;
	struct vastai_pci_info *priv = NULL;
	void __user *argp = (void __user *)arg;
	int i = 0;
	u32 *kp_value = NULL;
	void *pk_buf = NULL;
	struct T_ADDR_VAL val = { 0 };
	u32 die_index = 0;

	T_SMI_IOCTL_TRANS_DATA kdata = { 0 };

	V_UNREFERENCE(filp);
	V_UNREFERENCE(cmd);
	VATOOLS_FUNC_ENTERY;
	VATOOLS_DBG(priv, die_index, "cmd=0x%x\n", cmd);

	memset(&kdata, 0x00, sizeof(T_SMI_IOCTL_TRANS_DATA));
	ret = copy_from_user_ex(&kdata, (void __user *)argp, sizeof(kdata));
	if (ret) {
		VATOOLS_DBG(priv, die_index, "copy_from_user_ex ret=%ld\n",
			    ret);
		ret = -EFAULT;
		goto out;
	}

	VATOOLS_DBG(priv, die_index,
		    "block_id=%d  device_id=%d  die_id=%d  flag=0x%x  addr= "
		    "0x%llx\n",
		    kdata.block_id, kdata.device.dev_id, kdata.device.die_id,
		    kdata.flag, kdata.address);
	VATOOLS_DBG(
		priv, die_index,
		"read_len=%d  read_buf=0x%llx  write_len=%d  write_buf=0x%llx\n",
		kdata.output_buf.buf_size, kdata.output_buf.buf_addr,
		kdata.input_buf.buf_size, kdata.input_buf.buf_addr);

	priv = vatools_get_vastai_pci_device_info((char)kdata.device.dev_id);
	die_index = kdata.device.die_index;
	if (priv == NULL) {
		VATOOLS_INFO(priv, die_index, "priv is NULL, dev_id=0x%x\n",
			     kdata.device.dev_id);
		ret = -ENODEV;
		goto out;
	}
	if (vatools_check_pci_state(priv, VASTAI_NORMAL_STATE,
				    VASTAI_DEBUG_STATE, -1)) {
		VATOOLS_DBG(priv, die_index, "priv->pci_state=0x%x\n",
			    atomic_read(&priv->pci_state));
		ret = -ENODEV;
		goto out;
	}

	ret = copy_from_user_ex(&val, (void __user *)kdata.input_buf.buf_addr,
				sizeof(struct T_ADDR_VAL));
	if (ret) {
		VATOOLS_ERR(priv, die_index, "copy_from_user_ex ret=%ld\n",
			    ret);
		ret = -EFAULT;
		goto out;
	}

	VATOOLS_DBG(priv, die_index,
		    "ret=%ld, addr=0x%llx, len=%d, val.pcBuf=0x%p\n", ret,
		    val.nAddr, val.nLen, val.pcBuf);

	pk_buf = kzalloc(val.nLen, GFP_KERNEL);
	if (pk_buf == NULL) {
		ret = -ENOMEM;
		VATOOLS_ERR(priv, die_index, "kzalloc ret=%ld\n", ret);
		goto out;
	}

	ret = copy_from_user_ex(pk_buf, (void __user *)val.pcBuf, val.nLen);
	if (ret) {
		VATOOLS_ERR(priv, die_index, "copy_from_user_ex ret=%ld\n",
			    ret);
		ret = -EFAULT;
		goto out_kfree;
	}

	kp_value = (u32 *)pk_buf;
	for (i = 0; i < val.nLen / sizeof(u32); i++) {
		u32 v = *(kp_value + i);
		u64 addr = (u64)(val.nAddr + i * 4);
		// VATOOLS_DBG(priv, die_index, "[%d] addr=0x%08llx, val=0x%08x\n",
		// 	    i, addr, v);
#if (VATOOLS_PROFILER_DLC_ODMA == 0)
		//TODO:
		// runtime 实现了动态开关DLC功能. 在DLC 关闭时, 不能通过perfmon读取数据.
		// 会因为访问未知地址 hang. 暂时过滤地址, 不访问DLC和ODMA
		if (addr >= csr_mc0_axi_w_r_ostd &&
		    addr <= csr_mc3_axi_enable_clr_mode)
#endif
		{
			VATOOLS_DBG(priv, die_index,
				    "[%d] addr=0x%08llx, val=0x%08x\n", i, addr,
				    v);
			ret = vatools_fw_smcu_write(kdata.block_id, priv,
						    kdata.device.die_index,
						    addr, &v, sizeof(u32));
		}
	}

out_kfree:
	kvfree(pk_buf);
out:
	VATOOLS_DBG(priv, die_index, "out: ret=%ld\n", ret);
	VATOOLS_FUNC_EXIT;

	return ret;
}

int profiler_release_checkpoint(struct file *filp)
{
	int ret = 0;
	struct vatools_node *node = NULL;
	struct vatools_reader *reader = NULL;
	struct vatools_node *p_current = NULL, *p_next = NULL;
	struct vastai_pci_info *priv = NULL;
	ring_buf_elem_t ring_buf_elem;
	int core_index = 0;
	int can_release_checkpoint = 1;
	u32 die_index = 0;

	VATOOLS_FUNC_ENTERY;

	node = vatools_file_get_node(filp);
	reader = vatools_file_get_reader(filp);
	priv = vatools_get_vastai_pci_device_info(
		(char)reader->trans_category.device.dev_id);

	VATOOLS_DBG(
		priv, die_index,
		"filp=0x%p  node=0x%p  reader=0x%p  reader->profiler_checkpoint_corebitmap=0x%x\n",
		filp, node, reader, reader->profiler_checkpoint_corebitmap);

	// disable profiler checkpoint
	// TODO: 每个打开的fd, 如果调用 PROFILER_CHECKPOINT_MSG_START , 则计算器增加. 在释放时, 如果计算器为0, 则释放.
	if (reader->profiler_checkpoint_corebitmap != 0) {
		// TODO: 循环检查所有的die, 如果die_index相同并且 profiler_checkpoint_corebitmap 设置了值, 不为0,
		// 则表示有其他程序在使用profiler, 不能释放

		list_for_each_entry_safe (p_current, p_next,
					  vatools_get_vastai_head(),
					  list_nodes) {
			struct vatools_reader *p_current_reader = NULL,
					      *p_next_reader = NULL;

			for (core_index = 0; core_index < CORE_NUMBER;
			     core_index++) {
				can_release_checkpoint = 1;

				list_for_each_entry_safe (
					p_current_reader, p_next_reader,
					&p_current->list_node_readers,
					list_readers) {
					VATOOLS_DBG(
						priv, die_index,
						"p_current_reader=0x%p  p_current_node=0x%p\n",
						p_current_reader, p_current);
					VATOOLS_DBG(
						priv, die_index,
						"profiler_checkpoint_corebitmap=0x%x  category=%d dev=%d die=%d die_index=0x%x\n",
						p_current_reader
							->profiler_checkpoint_corebitmap,
						p_current_reader->trans_category
							.app_category,
						p_current_reader->trans_category
							.device.dev_id,
						p_current_reader->trans_category
							.device.die_id,
						p_current_reader->trans_category
							.device.die_index);

					if ((reader != p_current_reader) &&
					    (reader->trans_category
						     .app_category ==
					     p_current_reader->trans_category
						     .app_category) &&
					    (reader->trans_category.device
						     .dev_id ==
					     p_current_reader->trans_category
						     .device.dev_id) &&
					    (reader->trans_category.device
						     .die_id ==
					     p_current_reader->trans_category
						     .device.die_id) &&
					    (reader->trans_category.device
						     .die_index ==
					     p_current_reader->trans_category
						     .device.die_index)) {
						// corebitmap loop check
						if ((((p_current_reader
							       ->profiler_checkpoint_corebitmap >>
						       core_index) &
						      0x1))) {
							can_release_checkpoint =
								0;
							break;
						}
					}
				}
				VATOOLS_DBG(priv, die_index,
					    "can_release_checkpoint=0x%x\n",
					    can_release_checkpoint);
				if (can_release_checkpoint &&
				    (((reader->profiler_checkpoint_corebitmap >>
				       core_index) &
				      0x1))) {
					// release checkpoint
					//发送消息到 core, 停止profiler checkpoints功能
					memset(&ring_buf_elem, 0,
					       sizeof(ring_buf_elem));

					ring_buf_elem.whole = 0;
					ring_buf_elem.cmd =
						VATOOLS_INTERRUPT_TOOLS_CATEGARY_CMD;
					ring_buf_elem.use_ringbuf =
						VATOOLS_INTERRUPT_USE_RINGBUF;
					ring_buf_elem.seq_no = 0;
					ring_buf_elem.sub_cmd =
						VATOOLS_INTERRUPT_PROFILER_CMD;
					ring_buf_elem.val =
						PROFILER_CHECKPOINT_MSG_STOP;

					VATOOLS_DBG(
						priv, die_index,
						"die_index=0x%x, core_index=0x%x ring_buf_elem.whole=0x%llx\n",
						reader->trans_category.device
							.die_index,
						core_index,
						ring_buf_elem.whole);

					// 发送中断到 cmcu
#ifdef ENABLE_VATOOLS_PROFILER
					if (priv == NULL) {
						VATOOLS_INFO(
							priv, die_index,
							"priv is NULL, dev_id=0x%x\n",
							reader->trans_category
								.device.dev_id);
						ret = -ENODEV;
						goto out;
					}
					if (vatools_check_pci_state(
						    priv, VASTAI_NORMAL_STATE,
						    -1)) {
						VATOOLS_INFO(
							priv, die_index,
							"priv->pci_state=0x%x\n",
							atomic_read(
								&priv->pci_state));
						ret = -ENODEV;
						goto out;
					}
					ret = vatools_send_ctrl_cmd(
						priv,
						reader->trans_category.device
							.die_index,
						0x1 << core_index,
						ring_buf_elem.whole);
#else
					ret = -EINVAL;
#endif

					VATOOLS_DBG(
						priv, die_index,
						"send ctrl cmd return 0x%x\n",
						ret);
					if (ret < 0) {
						VATOOLS_INFO(
							priv, die_index,
							"send ctrl cmd failed. ret=%d\n",
							ret);
						goto out;
					}

					reader->profiler_checkpoint_corebitmap &=
						~(0x1 << core_index);
					VATOOLS_DBG(
						priv, die_index,
						"reader->profiler_checkpoint_corebitmap=0x%x\n",
						reader->profiler_checkpoint_corebitmap);
				}
			}
		}
	}
out:
	reader->profiler_checkpoint_corebitmap = 0;
	VATOOLS_DBG(priv, die_index,
		    "reader->profiler_checkpoint_corebitmap=0x%x\n",
		    reader->profiler_checkpoint_corebitmap);
	VATOOLS_FUNC_EXIT;

	return 0;
}

int profiler_ioctl_set(struct file *filp, unsigned int cmd, IOCTL_ARG_T arg)
{
	int ret = 0;
	struct vatools_node *node = NULL;
	struct vatools_reader *reader = NULL;
	void __user *argp = (void __user *)arg;
	T_SMI_IOCTL_TRANS_DATA kdata = { 0 };
	struct vastai_pci_info *priv = NULL;
	int die_index = 0;
	int ret2 = 0;
	ring_buf_elem_t ring_buf_elem;
	u32 corebitmap = 0;
	u32 coreindex = 0;

	V_UNREFERENCE(cmd);

	node = vatools_file_get_node(filp);
	reader = vatools_file_get_reader(filp);

	VATOOLS_DBG(
		priv, die_index,
		"ioctl cmd=0x%x  filp=0x%p  node=0x%p  reader=0x%p corebitmap=0x%x\n",
		cmd, filp, node, reader,
		reader->profiler_checkpoint_corebitmap);

	if (!node) {
		VATOOLS_INFO(priv, die_index, "node is NULL\n");
		ret = -EINVAL;
		goto out;
	}

	memset(&kdata, 0x00, sizeof(T_SMI_IOCTL_TRANS_DATA));
	ret = copy_from_user_ex(&kdata, (void __user *)argp, sizeof(kdata));
	if (ret) {
		VATOOLS_ERR(priv, die_index, "copy_from_user_ex ret=%d\n", ret);
		ret = -EFAULT;
		goto out;
	}

	coreindex = (u32)(kdata.address & 0xFFFFFFFF);

	VATOOLS_DBG(
		priv, die_index,
		"block_id=0x%x  device_id=0x%x  die_id=0x%x  die_id=0x%x  flag=0x%x  addr= "
		"0x%llx\n",
		kdata.block_id, kdata.device.dev_id, kdata.device.die_id,
		kdata.device.die_index, kdata.flag, kdata.address);
	VATOOLS_DBG(priv, die_index,
		    "read_len=%d  read_buf=0x%llx  write_len=%d"
		    "write_buf=0x%llx\n",
		    kdata.output_buf.buf_size, kdata.output_buf.buf_addr,
		    kdata.input_buf.buf_size, kdata.input_buf.buf_addr);

	priv = vatools_get_vastai_pci_device_info((char)kdata.device.dev_id);
	die_index = kdata.device.die_index;
	if (priv == NULL) {
		VATOOLS_INFO(priv, die_index, "priv is NULL, dev_id=0x%x\n",
			     kdata.device.dev_id);
		ret = -ENODEV;
		goto out;
	}
	if (vatools_check_pci_state(priv, VASTAI_NORMAL_STATE, -1)) {
		VATOOLS_DBG(priv, die_index, "priv->pci_state=0x%x\n",
			    atomic_read(&priv->pci_state));
		ret = -ENODEV;
		goto out;
	}

	die_index = vatools_get_vastai_pci_die_index((char)kdata.device.dev_id,
						     kdata.device.die_id);
	if (die_index < 0) {
		ret = -ENODEV;
		goto out;
	}

	switch (kdata.flag) {
	case PROFILER_CHECKPOINT_MSG_START: {
		VATOOLS_DBG(priv, die_index,
			    "PROFILER_CHECKPOINT_MSG_START: coreindex=%d\n",
			    coreindex);
		// TODO: 标识这个fd 开启了profiler checkpoints 功能, fd release释放时, 需要检查, 并发送stop 命令到fw,
		// 取消这个功能. 避免影响 fw 性能.

		//TODO: 判断coreindex如果是 CORE_HOST_PCIE , 开启pcie timeline功能,
		if (CORE_HOST_PCIE == coreindex) {
			// clear pcie buffer
			vatools_profiler_dma_checkpoint_clear();
			vatools_profiler_msg_checkpoint_clear();
			profiler_timeline_enable = 1;
		} else {
			//发送消息到 core
			memset(&ring_buf_elem, 0, sizeof(ring_buf_elem));

			ring_buf_elem.whole = 0;
			ring_buf_elem.cmd =
				VATOOLS_INTERRUPT_TOOLS_CATEGARY_CMD;
			ring_buf_elem.use_ringbuf =
				VATOOLS_INTERRUPT_USE_RINGBUF;
			ring_buf_elem.seq_no = 0;
			ring_buf_elem.sub_cmd = VATOOLS_INTERRUPT_PROFILER_CMD;
			ring_buf_elem.val = kdata.flag;
			corebitmap = (0x1 << (coreindex)) &
				     ((0x1 << CORE_NUMBER) - 1);

			VATOOLS_DBG(
				priv, die_index,
				"START: die_index=0x%x, corebitmap=0x%x kdata.flag=0x%x ring_buf_elem.whole=0x%llx\n",
				kdata.device.die_index, corebitmap, kdata.flag,
				ring_buf_elem.whole);

#ifdef ENABLE_VATOOLS_PROFILER
			// 发送中断到 cmcu // 1 << CORE_CMCU,
			ret = vatools_send_ctrl_cmd(priv,
						    kdata.device.die_index,
						    corebitmap,
						    ring_buf_elem.whole);
#else
			ret = -EINVAL;
#endif
			VATOOLS_DBG(priv, die_index,
				    "send ctrl cmd return 0x%x\n", ret);
			if (ret < 0) {
				VATOOLS_INFO(priv, die_index,
					     "send ctrl cmd failed. ret=%d\n",
					     ret);
				goto out;
			}
			// 记录开启profiler checkpoint 的core , 使用 corebitmap记录, 每个core 一个bit
			reader->profiler_checkpoint_corebitmap |= corebitmap;
		}
	} break;

	case PROFILER_CHECKPOINT_MSG_STOP: {
		VATOOLS_DBG(priv, die_index,
			    "PROFILER_CHECKPOINT_MSG_STOP: coreindex=%d\n",
			    coreindex);
		//TODO: 判断coreindex如果是 CORE_HOST_PCIE , 开启pcie timeline功能,
		if (CORE_HOST_PCIE == coreindex) {
			profiler_timeline_enable = 0;
			// clear pcie buffer
			vatools_profiler_dma_checkpoint_clear();
			vatools_profiler_msg_checkpoint_clear();
		} else {
			//只释放一个core
			memset(&ring_buf_elem, 0, sizeof(ring_buf_elem));

			ring_buf_elem.whole = 0;
			ring_buf_elem.cmd =
				VATOOLS_INTERRUPT_TOOLS_CATEGARY_CMD;
			ring_buf_elem.use_ringbuf =
				VATOOLS_INTERRUPT_USE_RINGBUF;
			ring_buf_elem.seq_no = 0;
			ring_buf_elem.sub_cmd = VATOOLS_INTERRUPT_PROFILER_CMD;
			ring_buf_elem.val = kdata.flag;
			corebitmap = (0x1 << (coreindex)) &
				     ((0x1 << CORE_NUMBER) - 1);

			VATOOLS_DBG(
				priv, die_index,
				"STOP: die_index=0x%x, corebitmap=0x%x kdata.flag=0x%x ring_buf_elem.whole=0x%llx\n",
				kdata.device.die_index, corebitmap, kdata.flag,
				ring_buf_elem.whole);

			// 发送中断到 core
#ifdef ENABLE_VATOOLS_PROFILER
			ret = vatools_send_ctrl_cmd(priv,
						    kdata.device.die_index,
						    corebitmap,
						    ring_buf_elem.whole);
#else
			ret = -EINVAL;
#endif
			VATOOLS_DBG(priv, die_index,
				    "send ctrl cmd return 0x%x\n", ret);
			if (ret < 0) {
				VATOOLS_INFO(priv, die_index,
					     "send ctrl cmd failed. ret=%d\n",
					     ret);
				goto out;
			}
			// 记录开启profiler checkpoint 的core , 使用 corebitmap记录, 每个core 一个bit
			reader->profiler_checkpoint_corebitmap &= ~corebitmap;
		}

	} break;

	case PROFILER_CHECKPOINT_MSG_READ_RPWP:
	case PROFILER_CHECKPOINT_MSG_READ_DATA: {
		void *temp_buf =
			(void *)kvmalloc(kdata.output_buf.buf_size, GFP_KERNEL);
		if (!temp_buf) {
			VATOOLS_ERR(priv, die_index,
				    " kvmalloc error. size=%d\n",
				    kdata.output_buf.buf_size);
			ret = -ENOMEM;
			goto out;
		}

		ret = vatools_pci_mem_read(priv, die_index, kdata.address,
					   temp_buf, kdata.output_buf.buf_size);
		if (ret < 0) {
			ret = -EFAULT;
			VATOOLS_INFO(priv, die_index,
				     " vatools_pci_mem_read error. ret=%d\n",
				     ret);
			goto PROFILER_CHECKPOINT_MSG_READ_RPWP_vfree;
		}
		ret = copy_to_user_ex((void *)kdata.output_buf.buf_addr,
				      temp_buf, kdata.output_buf.buf_size);
		if (ret) {
			VATOOLS_ERR(priv, die_index,
				    "copy_to_user_ex ret=%d byte=%d\n", ret,
				    kdata.output_buf.buf_size);
			ret = -EFAULT;
			goto PROFILER_CHECKPOINT_MSG_READ_RPWP_vfree;
		}

	PROFILER_CHECKPOINT_MSG_READ_RPWP_vfree:
		kvfree(temp_buf);

		goto out;
	} break;

	case PROFILER_CHECKPOINT_MSG_WRITE_RPWP: {
		u8 *temp_buf =
			(u8 *)kvmalloc(kdata.input_buf.buf_size, GFP_KERNEL);
		if (!temp_buf) {
			VATOOLS_ERR(priv, die_index,
				    " kvmalloc error. size=%d\n",
				    kdata.input_buf.buf_size);
			ret = -ENOMEM;
			goto out;
		}

		if (copy_from_user_ex(temp_buf,
				      (void *)kdata.input_buf.buf_addr,
				      kdata.input_buf.buf_size)) {
			ret = -EFAULT;
			VATOOLS_ERR(priv, die_index,
				    "copy_from_user_ex ret=%d\n", ret);
			goto PROFILER_CHECKPOINT_MSG_WRITE_RPWP_vfree;
		}

		ret = vatools_pci_mem_write(priv, die_index, kdata.address,
					    temp_buf, kdata.input_buf.buf_size);
		if (ret < 0) {
			ret = -EFAULT;
			goto PROFILER_CHECKPOINT_MSG_WRITE_RPWP_vfree;
		}

	PROFILER_CHECKPOINT_MSG_WRITE_RPWP_vfree:
		kvfree(temp_buf);

		goto out;
	} break;

		// pcie dma profiler timeline
	case PROFILER_CHECKPOINT_MSG_READ_PCIE_DMA: {
#ifdef ENABLE_VATOOLS_PROFILER
		// 使用 TLV 传递数据, 注意套圈数据拷贝
		T_SMI_TLV tlv;
		void *ringbuf_data_p = NULL;
		struct vastai_fifo *fifo = NULL;
		u32 count = 0;
		u32 wr_to_end_len = 0;
		u32 rd = 0;
		u32 wr = 0;

		fifo = (struct vastai_fifo *)node->profiler_dma_ringbuf;
		mutex_lock(&pcie_timeline_dma_mutex);
		rd = fifo->rd;
		wr = fifo->wr;

		if (vastai_fifo_is_empty(fifo) || !vastai_is_fifo_valid(fifo)) {
			ret = -ENOMEM;
			// copy header
			memset(&tlv, 0, sizeof(T_SMI_TLV));
			ret = copy_to_user_ex((void *)kdata.output_buf.buf_addr,
					      &tlv, sizeof(T_SMI_TLV));
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%ld\n",
					    ret, sizeof(T_SMI_TLV));
				ret = -EFAULT;
			}
			goto out_dma_mutex_unlock;
		}

		if (fifo->elem_count == 0) {
			ret = -EDOM;
			// copy header
			memset(&tlv, 0, sizeof(T_SMI_TLV));
			ret = copy_to_user_ex((void *)kdata.output_buf.buf_addr,
					      &tlv, sizeof(T_SMI_TLV));
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%ld\n",
					    ret, sizeof(T_SMI_TLV));
				ret = -EFAULT;
			}
			goto out_dma_mutex_unlock;
		}
		count = (wr + (fifo->elem_count - rd)) % fifo->elem_count;
		ringbuf_data_p =
			(void *)((u8 *)fifo + vastai_fifo_rd_pre_fetch(fifo));

		tlv.address = 0;
		tlv.length = count * fifo->elem_size;
		tlv.value[0] = 0;

		VATOOLS_DBG(
			priv, die_index,
			"read pcie dma: count=%d size=%d tlv.length=%d ringbuf_data_p=%p rd=0x%x wr=0x%x\n",
			count, fifo->elem_size, tlv.length, ringbuf_data_p, rd,
			wr);

		// copy header
		ret = copy_to_user_ex((void *)kdata.output_buf.buf_addr, &tlv,
				      sizeof(T_SMI_TLV));
		if (ret) {
			VATOOLS_ERR(priv, die_index,
				    "copy_to_user_ex ret=%d byte=%ld\n", ret,
				    sizeof(T_SMI_TLV));
			ret = -EFAULT;
			goto out_dma_mutex_unlock;
		}

		if (rd < wr) {
			// copy value
			ret = copy_to_user_ex(
				(void *)((u8 *)kdata.output_buf.buf_addr +
					 sizeof(T_SMI_TLV)),
				ringbuf_data_p, count * fifo->elem_size);
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%d\n",
					    ret, count * fifo->elem_size);
				ret = -EFAULT;
				goto out_dma_mutex_unlock;
			}
		} else {
			// rd -> end
			wr_to_end_len =
				(fifo->elem_count - rd) * fifo->elem_size;
			ret = copy_to_user_ex(
				(void *)((u8 *)kdata.output_buf.buf_addr +
					 sizeof(T_SMI_TLV)),
				ringbuf_data_p, wr_to_end_len);
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%d\n",
					    ret, wr_to_end_len);
				ret = -EFAULT;
				goto out_dma_mutex_unlock;
			}
			// 0 -> wr
			ret = copy_to_user_ex(
				(void *)((u8 *)kdata.output_buf.buf_addr +
					 sizeof(T_SMI_TLV) + wr_to_end_len),
				fifo->buf, (wr * fifo->elem_size));
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%d\n",
					    ret, wr_to_end_len);
				ret = -EFAULT;
				goto out_dma_mutex_unlock;
			}
		}
		VATOOLS_DBG(priv, die_index,
			    "read pcie DMA done. rd=%d, wr=%d\n", rd, wr);
		fifo->rd = wr;
	out_dma_mutex_unlock:
		mutex_unlock(&pcie_timeline_dma_mutex);
		goto out;
#endif
	} break;

		// pcie msg profiler timeline
	case PROFILER_CHECKPOINT_MSG_READ_PCIE_MSG: {
#ifdef ENABLE_VATOOLS_PROFILER
		// 使用 TLV 传递数据, 注意套圈数据拷贝
		T_SMI_TLV tlv;
		void *ringbuf_data_p = NULL;
		struct vastai_fifo *fifo = NULL;
		u32 count = 0;
		u32 wr_to_end_len = 0;
		u32 rd = 0;
		u32 wr = 0;

		fifo = (struct vastai_fifo *)node->profiler_msg_ringbuf;
		mutex_lock(&pcie_timeline_msg_mutex);
		rd = fifo->rd;
		wr = fifo->wr;

		if (vastai_fifo_is_empty(fifo) || !vastai_is_fifo_valid(fifo)) {
			ret = -ENOMEM;
			// copy header
			memset(&tlv, 0, sizeof(T_SMI_TLV));
			ret = copy_to_user_ex((void *)kdata.output_buf.buf_addr,
					      &tlv, sizeof(T_SMI_TLV));
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%ld\n",
					    ret, sizeof(T_SMI_TLV));
				ret = -EFAULT;
			}
			goto out_msg_mutex_unlock;
		}

		if (fifo->elem_count == 0) {
			ret = -EDOM;
			// copy header
			memset(&tlv, 0, sizeof(T_SMI_TLV));
			ret = copy_to_user_ex((void *)kdata.output_buf.buf_addr,
					      &tlv, sizeof(T_SMI_TLV));
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%ld\n",
					    ret, sizeof(T_SMI_TLV));
				ret = -EFAULT;
			}
			goto out_msg_mutex_unlock;
		}
		count = (wr + (fifo->elem_count - rd)) % fifo->elem_count;
		ringbuf_data_p =
			(void *)((u8 *)fifo + vastai_fifo_rd_pre_fetch(fifo));

		tlv.address = 0;
		tlv.length = count * fifo->elem_size;
		tlv.value[0] = 0;

		VATOOLS_DBG(
			priv, die_index,
			"read pcie msg: count=%d size=%d tlv.length=%d ringbuf_data_p=%p rd=0x%x wr=0x%x\n",
			count, fifo->elem_size, tlv.length, ringbuf_data_p, rd,
			wr);

		// copy header
		ret = copy_to_user_ex((void *)kdata.output_buf.buf_addr, &tlv,
				      sizeof(T_SMI_TLV));
		if (ret) {
			VATOOLS_ERR(priv, die_index,
				    "copy_to_user_ex ret=%d byte=%ld\n", ret,
				    sizeof(T_SMI_TLV));
			ret = -EFAULT;
			goto out_msg_mutex_unlock;
		}

		if (rd < wr) {
			// copy value
			ret = copy_to_user_ex(
				(void *)((u8 *)kdata.output_buf.buf_addr +
					 sizeof(T_SMI_TLV)),
				ringbuf_data_p, count * fifo->elem_size);
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%d\n",
					    ret, count * fifo->elem_size);
				ret = -EFAULT;
				goto out_msg_mutex_unlock;
			}
		} else {
			// rd -> end
			wr_to_end_len =
				(fifo->elem_count - rd) * fifo->elem_size;
			ret = copy_to_user_ex(
				(void *)((u8 *)kdata.output_buf.buf_addr +
					 sizeof(T_SMI_TLV)),
				ringbuf_data_p, wr_to_end_len);
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%d\n",
					    ret, wr_to_end_len);
				ret = -EFAULT;
				goto out_msg_mutex_unlock;
			}
			// 0 -> wr
			ret = copy_to_user_ex(
				(void *)((u8 *)kdata.output_buf.buf_addr +
					 sizeof(T_SMI_TLV) + wr_to_end_len),
				fifo->buf, (wr * fifo->elem_size));
			if (ret) {
				VATOOLS_ERR(priv, die_index,
					    "copy_to_user_ex ret=%d byte=%d\n",
					    ret, wr_to_end_len);
				ret = -EFAULT;
				goto out_msg_mutex_unlock;
			}
		}
		VATOOLS_DBG(priv, die_index,
			    "read pcie msg done. rd=%d, wr=%d\n", rd, wr);
		fifo->rd = wr;
	out_msg_mutex_unlock:
		mutex_unlock(&pcie_timeline_msg_mutex);
		goto out;
#endif
	} break;
	}

out:
	kdata.errcode = ret;
	ret2 = copy_to_user_ex((void __user *)argp, &kdata, sizeof(kdata));
	if (ret2) {
		VATOOLS_ERR(priv, die_index, "copy_to_user_ex ret=%d\n", ret);
		ret2 = -EFAULT;
	}
	VATOOLS_DBG(priv, die_index, "ioctl return ret=%d  ret2=%d\n", ret,
		    ret2);

	return ret2;
}

int profiler_ioctl_get(struct file *filp, unsigned int cmd, IOCTL_ARG_T arg)
{
	int ret = 0;

	VATOOLS_FUNC_ENTERY;
	ret = profiler_ioctl_set(filp, cmd, arg);
	VATOOLS_FUNC_EXIT;

	return ret;
}

long profiler_ioctl(struct file *filp, unsigned int cmd, IOCTL_ARG_T arg)
{
	long ret = -EINVAL;

	switch (cmd) {
	case VATOOLS_PROFILER_SET_PERFORMANCE_CSR:
		ret = profiler_set_performance_csr(filp, cmd, arg);
		break;
	case VATOOLS_PROFILER_SET:
		ret = profiler_ioctl_set(filp, cmd, arg);
		break;
	case VATOOLS_PROFILER_GET:
		ret = profiler_ioctl_get(filp, cmd, arg);
		break;
	}

	return ret;
}

int profiler_mmap(struct file *filp, struct vm_area_struct *pvm)
{
	V_UNREFERENCE(filp);
	V_UNREFERENCE(pvm);

	return 0;
}
int profiler_open(struct inode *inode, struct file *filp)
{
	VATOOLS_FUNC_ENTERY;
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "filp=0x%p\n", filp);

	V_UNREFERENCE(inode);
	V_UNREFERENCE(filp);

	VATOOLS_FUNC_EXIT;
	return 0;
}

int profiler_release(struct inode *ignored, struct file *filp)
{
	int ret = 0;
	struct vatools_node *node;
	struct vatools_reader *reader;
	VATOOLS_DBG(NULL, DUMMY_DIE_ID, "filp=0x%p\n", filp);

	V_UNREFERENCE(ignored);
	VATOOLS_FUNC_ENTERY;

	node = vatools_file_get_node(filp);
	reader = vatools_file_get_reader(filp);
	VATOOLS_DBG(
		NULL, DUMMY_DIE_ID,
		"filp=0x%p  node=0x%p  reader=0x%p  reader->profiler_checkpoint_corebitmap=0x%x\n",
		filp, node, reader, reader->profiler_checkpoint_corebitmap);

	ret = profiler_release_checkpoint(filp);
	return ret;
}

int profiler_fasync(int fd, struct file *filp, int mode)
{
	V_UNREFERENCE(fd);
	V_UNREFERENCE(filp);
	V_UNREFERENCE(mode);

	return 0;
}
