/*
 * vastai_pci - driver a Vastai PCI device
 * Copyright (C) 2021 Vastai Technologies All Rights Reserved.
 * This work is licensed under the terms of the GNU GPL, version 2.
 *
 * Create on: 2020/09/25
 * Author: gang.wang
 */

#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/uaccess.h>
#include <linux/random.h>
#include <linux/dma-buf.h>
#include <linux/poll.h> 
#include <linux/proc_fs.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/version.h>

#include "vastai_pci.h"
#include "vastai_vcmd.h"

#include "vastai_dec.h"
#include "vastai_enc.h"

#include "vastai_render.h"
#include "vastai_udma_test.h"
#include "vastai_dmabuf.h"

#define VIDEO_NODE_NAME	"va_video"

#define DEVICE_ADDR_SG100_32G_GFX_MIN   (0x10ba000000)
#define DEVICE_ADDR_SG100_32G_GFX_MAX   (0x1230000000)
#define DEVICE_ADDR_SG100_64G_GFX_MIN   (0x10ca000000)
#define DEVICE_ADDR_SG100_64G_GFX_MAX   (0x1430000000)

extern struct list_head dec_device_handle_head;
extern struct list_head enc_device_handle_head;

static int video_driver_inited = 0;

//vastai video debug node related code start, node name: /proc/vastai/video/debug
static struct proc_dir_entry* dir;
static char video_debug_buffer[1024];

static int do_video_debug(char* buffer, size_t count) {
	const char* delim1 = " ", *delim2 = "=";
	char* token, *param, *value;
	int die_index;

	token = strsep(&buffer, delim1);
	while (token) {
		param = strsep(&token, delim2);
		value = token;
		//printk("param=%s, value=%s\n", param, value);

		if (strncmp(param, VASTAI_DEBUG_PARAM_DIEINDEX, sizeof(VASTAI_DEBUG_PARAM_DIEINDEX)) == 0) {

			die_index = simple_strtol(value, NULL, 16);

		} else if (strncmp(param, VASTAI_DEBUG_PARAM_ENC_LOGLEVEL, sizeof(VASTAI_DEBUG_PARAM_ENC_LOGLEVEL)) == 0) {

			int enc_loglevel = simple_strtol(value, NULL, 10);
			vastai_enc_set_log_level(enc_loglevel);

		} else if (strncmp(param, VASTAI_DEBUG_PARAM_DEC_LOGLEVEL, sizeof(VASTAI_DEBUG_PARAM_DEC_LOGLEVEL)) == 0) {

			//int dec_loglevel = simple_strtol(value, NULL, 10);
			//TODO

		} else if (strncmp(param, VASTAI_DEBUG_PARAM_ENCMODE, sizeof(VASTAI_DEBUG_PARAM_ENCMODE)) == 0) {

			//encoder_workmode_t work_mode = (encoder_workmode_t)simple_strtol(value, NULL, 10);
			//vastai_enc_set_work_mode(die_index, work_mode, false);

		} else {
			pr_err("error unsupported param: %s\n", param);
			return -1;
		}

		token = strsep(&buffer, delim1);
	}

	return 0;
}

static int video_debug_open(struct inode *inode, struct file *filp)
{
	return 0;
}

static int video_debug_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static ssize_t video_debug_read(struct file *filp, char __user *buffer,
			    size_t count, loff_t *offset)
{
	int ret;
	//printk("count=%ld, offset=%lld, video_debug_buffer=%s\n", count, *offset, video_debug_buffer);
	ret = video_copy_to_user(buffer, video_debug_buffer, MIN(count, strlen(video_debug_buffer)));
	if (ret) {
		pr_err("error %d\n", ret);
		return -EFAULT;
	}

	return 0;
}

static ssize_t video_debug_write(struct file *filp, const char __user *buffer,
			    size_t count, loff_t *offset)
{
	int ret;
	static char video_debug_buffer[1024];
	//printk("count=%ld, offset=%lld, video_debug_buffer=%s\n", count, *offset, video_debug_buffer);
	ret = video_copy_from_user(video_debug_buffer, buffer, MIN(count, sizeof(video_debug_buffer)));
	if (ret) {
		pr_err("error %d\n", ret);
		return -EFAULT;
	}

	do_video_debug(video_debug_buffer, count);
	return MIN(count, sizeof(video_debug_buffer));
}

static long video_debug_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	return 0;
}
static int video_debug_init = 0;
static int video_debug_deinit = 0;

#if KERNEL_VERSION(5,6,0) <= LINUX_VERSION_CODE
static	struct proc_ops vastai_video_debug_ops = {
	.proc_open = video_debug_open,
	.proc_release = video_debug_release,
	.proc_read = video_debug_read,
	.proc_write = video_debug_write,
	.proc_ioctl = video_debug_ioctl,
};
#else
static  struct file_operations vastai_video_debug_ops = {
	.owner = THIS_MODULE,
	.open = video_debug_open,
	.release = video_debug_release,
	.read = video_debug_read,
	.write = video_debug_write,
	.unlocked_ioctl = video_debug_ioctl,
};
#endif

static int vastai_video_debug_init(void) {
	struct proc_dir_entry* file;
	if (video_debug_init)
		return 0;
	dir = proc_mkdir(VASTAI_VIDEO_DEBUG_DIR, NULL);
	if (dir == NULL) {
		pr_err("create debug dir %s failed\n", VASTAI_VIDEO_DEBUG_DIR);
		return -1;
	}

	file = proc_create(VASTAI_VIDEO_DEBUG_NAME, 0777, dir, &vastai_video_debug_ops);
	if (file == NULL) {
		pr_err("create debug dir %s failed\n", VASTAI_VIDEO_DEBUG_NAME);
		return -1;
	}
	video_debug_init = 1;
	return 0;
}

static int vastai_video_debug_deinit(void) {
	if (video_debug_deinit)
		return 0;
	remove_proc_entry(VASTAI_VIDEO_DEBUG_NAME, dir);
	remove_proc_entry(VASTAI_VIDEO_DEBUG_DIR, NULL);

	video_debug_deinit = 1;

	return 0;
}
//vastai video debug node related code end



static long vastai_common_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	int ret = 0;
	struct video_common_cmd kcmd;
	struct vastai_video_info *video_info;
	struct vastai_pci_info *pci_info;

	video_info = (struct vastai_video_info *)filp->private_data;
	if (!video_info || !video_info->render_file) {
		pr_err("%d video_info or render_file is null\n", __LINE__);
		return -EIO;
	}

	pci_info = video_info->render_file->pci_info;
	if (!pci_info) {
		pr_err("%d pci_info is null\n", __LINE__);
		return -EIO;
	}

	switch(cmd) {

		case VASTAI_COMMON_IOCH_DMA_ALLOC: {
			struct dma_buf *dmabuf;
			ret = video_copy_from_user(&kcmd, (void *)arg, sizeof(kcmd));
			if (ret) {
				pr_err("error video_copy_from_user, ret = %d\n", ret);
				return -EIO;
			}

			dmabuf = vastai_alloc_dma_buf(pci_info, kcmd.alloc_cmd.size);
			if (dmabuf) {
				int fd = dma_buf_fd(dmabuf, O_CLOEXEC);
				kcmd.alloc_cmd.dma_buf_fd = fd;
			} else {
				kcmd.alloc_cmd.dma_buf_fd = -ENOMEM;
			}

			ret = video_copy_to_user((void *)arg, &kcmd, sizeof(kcmd));
			if (ret) {
				pr_err("error video_copy_to_user, ret = %d\n", ret);
				return -EIO;
			}

			break;
		}

		case  VASTAI_COMMON_IOCH_DMA_START: {

			struct dma_buf *dmabuf = dma_buf_get(kcmd.dma_start_cmd.dma_buf_fd);
			struct vastai_dma_buf *dm = get_vastai_dma_buf(dmabuf, 0);
			
			ret = video_copy_from_user(&kcmd, (void *)arg, sizeof(kcmd));
			if (ret) {
				pr_err("error video_copy_from_user, ret = %d\n", ret);
				return -EIO;
			}

			vastai_video_dma_transfer_sync(pci_info, kcmd.dma_start_cmd.is_dev_to_host, 0, 0,  kcmd.dma_start_cmd.axi_addr,  kcmd.dma_start_cmd.length, dm);
			dma_buf_put(dmabuf);

			ret = video_copy_to_user((void *)arg, &kcmd, sizeof(kcmd));
			if (ret) {
				pr_err("error video_copy_to_user, ret = %d\n", ret);
				return -EIO;
			}

			break;
		}

		case  VASTAI_COMMON_IOCH_GET_HW_TYPE: {
			struct video_type_mode	type_mode;

			type_mode.hw_type = video_info->render_file->hw_type;
			type_mode.enc_work_mode = video_info->render_file->enc_work_mode;
			type_mode.dec_work_mode = video_info->render_file->dec_work_mode;

			ret = video_copy_to_user((void *)arg, &type_mode, sizeof(type_mode));
			if (ret) {
				pr_err("%d error video_copy_to_user, ret = %d\n", __LINE__, ret);
				return -EIO;
			}
			break;
		}

		case  VASTAI_COMMON_IOCH_GET_DDR_SZ: {

			ret = video_copy_to_user((void *)arg, &video_info->render_file->gfx_ddr, sizeof(struct vastai_gfx_limit));
			if (ret) {
				pr_err("%d error video_copy_to_user, ret = %d\n", __LINE__, ret);
				return -EIO;
			}
			break;
		}

		case VASTAI_COMMON_IOCH_CHECK_DEV_ADDR: {

			struct dev_check_info dev_info;
			uint64_t gfx_ddr_min = video_info->render_file->gfx_ddr.gfx_ddr_min;
			uint64_t gfx_ddr_max = video_info->render_file->gfx_ddr.gfx_ddr_max;

			ret = video_copy_from_user(&dev_info, (void *)arg, sizeof(dev_info));
			if (ret) {
				pr_err("%d error video_copy_from_user, ret = %d\n", __LINE__, ret);
				return -1;
			}

			if (dev_info.dev_addr < gfx_ddr_min || (dev_info.dev_addr + dev_info.size) > gfx_ddr_max) {
				pr_err("%d invalid dev addr from gfx: addr=0x%llx size=%lld\n", __LINE__, dev_info.dev_addr, dev_info.size);
				return -1;
			}

			break;
		}

		default: {
			pr_err("%d error invalid cmd 0x%x\n", __LINE__, cmd);
			return -1;
		}
	}

	return 0;
}


static int render_open(struct inode *inode, struct file *filp)
{
	struct vastai_file_info *file_info;
	struct vastai_render_file *render_file;
	struct vastai_video_info *video_info =
		kmalloc(sizeof(struct vastai_video_info), GFP_KERNEL);
	if (!video_info)
		return -ENOMEM;

	file_info = to_vastai_file_info(inode->i_cdev);
	render_file =
		container_of(file_info, struct vastai_render_file, file_info);

	video_info->render_file = render_file;
	video_info->dec_device_handle = NULL; //will set on decoder ioctl separately
	video_info->enc_device_handle = NULL; //will set on encoder ioctl separately
	filp->private_data = video_info;

	vastai_enc_open(inode, filp);

	return 0;
}

static int render_release(struct inode *inode, struct file *filp)
{
	struct vastai_video_info *video_info =
		(struct vastai_video_info *)filp->private_data;

	if (video_info->enc_device_handle == NULL && video_info->dec_device_handle != NULL) { 

		dec_device_handle_t *device_handle  = video_info->dec_device_handle;
		vastai_dec_release(filp, &device_handle->die);

	} else if (video_info->dec_device_handle == NULL && video_info->enc_device_handle != NULL) {

		vastai_enc_release(inode, filp);
		mutex_lock(&video_info->enc_device_handle->die.job_mutex);
		video_info->enc_device_handle->die.enc_job_count--;
		mutex_unlock(&video_info->enc_device_handle->die.job_mutex);

	}

	kfree(filp->private_data);

	return 0;
}

static ssize_t render_read(struct file *filp, char __user *buffer,
			    size_t count, loff_t *offset)
{
	return 0;
}
static long render_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct vastai_video_info *video_info;

    //printk("render_ioctl: type %c, id %d\n", (cmd >> 8) &  0xFF, cmd & 0xFF);

	video_info = (struct vastai_video_info *)filp->private_data;
	if (cmd == VASTAI_VCMD_IOCH_SET_DECODER) {
		video_info->dec_device_handle = video_info->render_file->dec_private_data;
		video_info->enc_device_handle = NULL;
		//printk("video driver setup decoder\n");
		return 0;
	} else {
		if (!video_info->enc_device_handle && !video_info->dec_device_handle) {
			video_info->enc_device_handle = video_info->render_file->enc_private_data;
			video_info->dec_device_handle = NULL;
			mutex_lock(&video_info->enc_device_handle->die.job_mutex);
			video_info->enc_device_handle->die.enc_job_count++;
			mutex_unlock(&video_info->enc_device_handle->die.job_mutex);
			//printk("video driver setup encoder\n");
		}

		if (cmd == VASTAI_ENC_IOCH_SET_ENCODER) {
			return 0;
		}
	}

	if (_IOC_TYPE(cmd) == VASTAI_VCMD_IOC_MAGIC)
		return vastai_dec_ioctl(filp, cmd, arg);

	if(_IOC_TYPE(cmd) == VASTAI_ENC_IOC_MAGIC)
		return vastai_enc_ioctl(filp, cmd, arg);

	if(_IOC_TYPE(cmd) == VASTAI_COMMON_IOC_MAGIC)
		return vastai_common_ioctl(filp, cmd, arg);
	return 0;
}

static unsigned int render_poll(struct file *filp, poll_table *wait)
{
	return vastai_enc_poll(filp, wait);
}

static vastai_video_workmode_t get_videomode_from_fnmode(u8 fn_mode) {

	vastai_video_workmode_t video_mode = PCIE_PF_STANDALONE_MODE;

	switch (fn_mode) {
		case ONLY_1_PF:
			video_mode = PCIE_PASSTHROUGH_MODE;
			break;
		case ONLY_2_PF:
			video_mode = PCIE_PF_2_MASTER_MODE;
			break;
		case ONLY_4_PF:
			video_mode = PCIE_PF_STANDALONE_MODE;
			break;
		case PF_4_WITH_SRIOV:
			video_mode = PCIE_VF_MODE;
			break;
		default:
			video_mode = PCIE_PF_STANDALONE_MODE;
			break;
	}

	return video_mode;
}

static const struct file_operations fop = {
	.owner = THIS_MODULE,
	.open = render_open,
	.release = render_release,
	.read = render_read,
	.unlocked_ioctl = render_ioctl,
	.poll = render_poll,
#ifdef CONFIG_COMPAT
	.compat_ioctl = render_ioctl,
#endif
};

extern struct va_gfx_cfg gfx_cfg;

static int vastai_render_file_init(struct vastai_pci_info *pci_info, u32 die_index,
				void *dec_private_data, void *enc_private_data)
{
	struct vastai_render_file *render_file =
		vmalloc(sizeof(struct vastai_render_file));
	int ret;

	if (render_file == NULL) {
		return -ENOMEM;
	}

	render_file->die_index = die_index;
	ret = vastai_file_create(&render_file->file_info, pci_info, &fop,
				 "%s%d",
				 VIDEO_NODE_NAME, pci_info->dev_id);
	render_file->dec_private_data = dec_private_data;
	render_file->enc_private_data = enc_private_data;
	render_file->pci_info = pci_info;
	render_file->hw_type  = VASTAI_HW_SG100;
	render_file->enc_work_mode = get_videomode_from_fnmode(pci_info->fn_mode);
	render_file->dec_work_mode = get_videomode_from_fnmode(pci_info->fn_mode);
	render_file->gfx_ddr.gfx_ddr_min = gfx_cfg.gfx_heap.dev_paddr;
	render_file->gfx_ddr.gfx_ddr_max = gfx_cfg.gfx_heap.dev_paddr + gfx_cfg.gfx_heap.size + gfx_cfg.disp_heap.size;

	printk("video encoder: gfx_min=0x%llx, gfx_max=0x%llx\n", render_file->gfx_ddr.gfx_ddr_min, render_file->gfx_ddr.gfx_ddr_max);

	list_add_tail(&render_file->node, &pci_info->render_head);

	vastai_video_debug_init();

	return ret;
}

int vastai_video_init(void *pcie_dev, unsigned int die_id)
{
	int ret = -1;
	u8 is_pf = 0, is_vf = 0;
	u8 devfn = 0, fn_mode = 0, pkg_id = 0, pfn = 0, vfn = 0, decoder_sriov_en = 0, total_vf_per_pf = 0;

	dec_device_handle_t *dec_device_handle = NULL;
	enc_device_handle_t *enc_device_handle = NULL;
	struct vastai_pci_info * pci_info = NULL;

	if(!video_driver_inited) {
		INIT_LIST_HEAD(&dec_device_handle_head);
		INIT_LIST_HEAD(&enc_device_handle_head);
		video_driver_inited = 1;
	}

	pci_info = (struct vastai_pci_info *)pcie_dev;

	is_pf   = pci_info->is_pf;
	is_vf   = pci_info->is_vf;
	fn_mode = pci_info->fn_mode;
	pkg_id  = pci_info->pkg_id;
	devfn   = pci_info->priv_hw_cfg->sys_cfg.devfn;
	pfn     = pci_info->priv_hw_cfg->sys_cfg.pfn;
	vfn     = pci_info->priv_hw_cfg->sys_cfg.vfn;
	decoder_sriov_en = pci_info->priv_hw_cfg->sys_cfg.decoder_sriov_en;
	total_vf_per_pf  = pci_info->priv_hw_cfg->sys_cfg.total_vf_per_pf;

	if (fn_mode == PF_4_WITH_SRIOV && is_vf == 0) {
		printk("video PF_4_WITH_SRIOV but isvf = 0, return!\n");
		return 0;
	}

	// 1. decoder driver init
	dec_device_handle = kzalloc(sizeof(dec_device_handle_t), GFP_KERNEL);
	if (!dec_device_handle) {
		VASTAI_PCI_ERR(pcie_dev, DUMMY_DIE_ID,
					"%s:alloc decoder device handle mem failed\n",
					pci_name(pcie_dev));
		goto err;
	}

	dec_device_handle->die.die_id   = die_id;
	dec_device_handle->die.vcmd_pri = pcie_dev;
	dec_device_handle->die.is_pf    = is_pf;
	dec_device_handle->die.is_vf    = is_vf;
	dec_device_handle->die.devfn    = devfn;
	dec_device_handle->die.fn_mode  = fn_mode;
	dec_device_handle->die.pkg_id   = pkg_id;
	dec_device_handle->die.pfn      = pfn;
	dec_device_handle->die.vfn      = vfn;
	dec_device_handle->die.decoder_sriov_en = decoder_sriov_en;
	dec_device_handle->die.total_vf_per_pf  = total_vf_per_pf;

	dec_device_handle->die.dec_work_mode = get_videomode_from_fnmode(pci_info->fn_mode);;
	list_add_tail(&(dec_device_handle->node), &dec_device_handle_head);

	ret = vastai_dec_init(&dec_device_handle->die);
	if (ret) {
		VASTAI_PCI_ERR(pcie_dev, DUMMY_DIE_ID,
					"%s:init decoder die info failed, ret=%d\n",
					pci_name(pcie_dev), ret);
		goto err;
	}

	// 2. encoder driver init
	enc_device_handle = (enc_device_handle_t*)kzalloc(sizeof(enc_device_handle_t), GFP_KERNEL);
	if (!enc_device_handle)
	{
		VASTAI_PCI_ERR(pcie_dev, DUMMY_DIE_ID,
					"%s:alloc encoder device handle mem failed\n",
					pci_name(pcie_dev));
		goto err;
	}

	enc_device_handle->die.die_index = die_id;
	enc_device_handle->die.priv_dev = (struct vastai_pci_info *)pcie_dev;

	enc_device_handle->die.is_pf   = is_pf;
	enc_device_handle->die.is_vf   = is_vf;
	enc_device_handle->die.devfn   = devfn;
	enc_device_handle->die.fn_mode = fn_mode;
	enc_device_handle->die.pkg_id  = pkg_id;
	enc_device_handle->die.pfn     = pfn;
	enc_device_handle->die.vfn     = vfn;

	enc_device_handle->die.enc_work_mode = get_videomode_from_fnmode(pci_info->fn_mode);;
	list_add_tail(&(enc_device_handle->node), &enc_device_handle_head);

	ret = vastai_enc_init(&enc_device_handle->die);
	if (ret) {
		VASTAI_PCI_ERR(pcie_dev, DUMMY_DIE_ID,
					"%s:init encoder die info failed, ret=%d\n",
					pci_name(pcie_dev), ret);
		goto err;
	}

	ret = vastai_render_file_init(pcie_dev, die_id, dec_device_handle, enc_device_handle);
	if (ret) {
		VASTAI_PCI_ERR(pcie_dev, DUMMY_DIE_ID,
					"%s:render file init failed ret=%d\n",
					pci_name(pcie_dev), ret);
		goto err;
	}


	printk("device[%d] die_index=%d \n", ((struct vastai_pci_info *)pcie_dev)->dev_id, die_id);

	return 0;

err:
	if (dec_device_handle)
		kfree(dec_device_handle);

	if (enc_device_handle)
		kfree(enc_device_handle);

	return ret;
}

void vastai_video_deinit(struct vastai_pci_info *pci_info)
{
	struct vastai_render_file *render_file, *temp;

	if (pci_info->fn_mode == PF_4_WITH_SRIOV && pci_info->is_vf == 0){
		printk("video PF_4_WITH_SRIOV but is_vf=0, return!\n");
		return;
	}

	vastai_enc_deinit(pci_info);
	vastai_dec_deinit(pci_info);

	list_for_each_entry_safe (
		render_file,
		temp, &pci_info->render_head, node) {
		vastai_file_destroy(&render_file->file_info);
		list_del(&render_file->node);
		vfree(render_file);
	}

	vastai_video_debug_deinit();
}
