#include <linux/ioctl.h>
#include <linux/types.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/uaccess.h>

#include "vastai_video_utils.h"
#include "vastai_pcie_export.h"
#include "vastai_enc.h"


//#define VASTAI_CMDQUEUE_DEBUG

#undef PDEBUG
#ifdef VASTAI_CMDQUEUE_DEBUG
#   define PDEBUG(fmt, args...) printk( KERN_INFO "va_cmdid_queue: " fmt, ## args)
#else
#   define PDEBUG(fmt, args...)  /* not debugging: nothing */
#endif

void vastai_cmdid_init(VastCmdbufQueue *IdQueue, codec_type_t type)
{
    memset(IdQueue, 0, sizeof(VastCmdbufQueue));
    IdQueue->max_size = MAX_ELEM_NUM;
    IdQueue->type = type;
    spin_lock_init(&IdQueue->spin_lock);
}

int vastai_cmdid_queue_empty(VastCmdbufQueue *IdQueue)
{
	if (NULL == IdQueue){
		return -EINVAL;
	}

	if (IdQueue->rd == IdQueue->wr){
		return 1;
	}

	return 0;
}

int vastai_cmdid_queue_full(VastCmdbufQueue *IdQueue){
	u16 rd = 0, wr = 0;
	if (NULL == IdQueue) {
		return -EINVAL;
	}

	rd = IdQueue->rd;
	wr = IdQueue->wr;
	PDEBUG("type=%d, rd: %d, wr: %d\n", IdQueue->type, rd, wr);
	if (( wr + 1) % IdQueue->max_size == rd){
		return 1;
	}

	return 0;
}

int vastai_cmdid_enqueue(VastCmdbufQueue *IdQueue,  u32 CmdId, u32 irq_status)
{
    if(vastai_cmdid_queue_full(IdQueue)) {
        pr_err("type=%d, cmdbuf id queue is full!\n", IdQueue->type);
        return -ENOMEM;
    }
    PDEBUG("before type=%d, rd=%d, wr=%d\n", IdQueue->type, IdQueue->rd, IdQueue->wr);
    IdQueue->cmdbuf_id_fifo[IdQueue->wr] = CmdId;
    IdQueue->cmdbuf_irq_status[IdQueue->wr] = irq_status;
    IdQueue->wr = (IdQueue->wr + 1) % IdQueue->max_size;
    PDEBUG("after type=%d, rd=%d, wr=%d\n", IdQueue->type, IdQueue->rd, IdQueue->wr);
    return 0;
}

int vastai_cmdid_dequeue(VastCmdbufQueue *IdQueue, u32* CmdId, u32* irq_status)
 {
     if(vastai_cmdid_queue_empty(IdQueue)) {
         //printk("%s: cmdbuf id queue is empty!\n");
         return -EAGAIN;
     }
     PDEBUG("before type=%d, rd=%d, wr=%d\n", IdQueue->type, IdQueue->rd, IdQueue->wr);
     *CmdId = IdQueue->cmdbuf_id_fifo[IdQueue->rd];
     *irq_status = IdQueue->cmdbuf_irq_status[IdQueue->rd];
     IdQueue->rd = (IdQueue->rd + 1) % IdQueue->max_size;
     PDEBUG("after type=%d, rd=%d, wr=%d\n", IdQueue->type, IdQueue->rd, IdQueue->wr);
     return 0;
 }




 void init_bi_list(bi_list* list)
{
  list->head = NULL;
  list->tail = NULL; 
}

bi_list_node* bi_list_create_node(void)
{
  bi_list_node* node=NULL;
  node=(bi_list_node*)vmalloc(sizeof(bi_list_node));
  if(node==NULL)
  {
    PDEBUG ("%s\n","vmalloc for node fail!");
    return node;
  }
  memset(node,0,sizeof(bi_list_node)); 
  return node;
}
void bi_list_free_node(bi_list_node* node)
{
  //free current node
  vfree(node);
  return;
}

void bi_list_insert_node_tail(bi_list* list,bi_list_node* current_node)
{
  if(current_node==NULL)
  {
    PDEBUG ("%s\n","insert node tail  NULL");
    return;
  }
  if(list->tail)
  {
   current_node->previous=list->tail;
   list->tail->next=current_node;
   list->tail=current_node;
   list->tail->next=NULL;
  }
  else
  {
   list->head=current_node;
   list->tail=current_node;
   current_node->next=NULL;
   current_node->previous=NULL;
  }
  return;
}

void bi_list_insert_node_before(bi_list* list,bi_list_node* base_node,bi_list_node* new_node)
{
  bi_list_node* temp_node_previous=NULL;
  if(new_node==NULL)
  {
    PDEBUG ("%s\n","insert node before new node NULL");
    return;
  }
  if(base_node)
  {
   if(base_node->previous)
   {
    //at middle position
    temp_node_previous = base_node->previous;
    temp_node_previous->next=new_node;
    new_node->next = base_node;
    base_node->previous = new_node;
    new_node->previous=temp_node_previous;
   }
   else
   {
    //at head
    base_node->previous = new_node;
    new_node->next = base_node;
    list->head=new_node;
    new_node->previous = NULL;
   }
  }
  else
  {
   //at tail
   bi_list_insert_node_tail(list,new_node);
  }
  return;
}

void bi_list_remove_node(bi_list* list,bi_list_node* current_node)
{
  bi_list_node* temp_node_previous=NULL;
  bi_list_node* temp_node_next=NULL;
  if(current_node==NULL)
  {
    PDEBUG ("%s\n","remove node NULL");
    return;
  }
  temp_node_next=current_node->next;
  temp_node_previous=current_node->previous;
  
  if(temp_node_next==NULL && temp_node_previous==NULL )
  {
    //there is only one node.
    list->head=NULL;
    list->tail=NULL;
  }
  else if(temp_node_next==NULL)
  {
    //at tail
    list->tail=temp_node_previous;
    temp_node_previous->next=NULL;
  }
  else if( temp_node_previous==NULL)
  {
    //at head
    list->head=temp_node_next;
    temp_node_next->previous=NULL;
  }
  else
  {
   //at middle position
   temp_node_previous->next=temp_node_next;
   temp_node_next->previous=temp_node_previous;
  }
  return;
}


/**
 * @brief : push a elem into device fifo.
 *
 * @param pci_info : pointer to the pcie device.
 * @param die_index : union die_index_data.
 * @param fifo_addr : device address of fifo.
 * @param elem : object which need push.
 * @param fifo : return val, for the fifo cache. If it be give, this function
 * 		 will be more quickly.
 *
 * @return int : success is zero. other is errno.
 */
int vastai_video_fifo_push_elem(void *pci_info, int die_index, u64 fifo_addr,
			  void *elem, struct vastai_video_fifo *fifo)
{
	int ret = 0;
	u64 entry_addr = 0;
	u32 try_count = 2;
	struct vastai_video_fifo _fifo = { 0 };
	struct vastai_video_fifo _fifo_chk = { 0 };
	TVidEncCfgParam tvidencparam = {0};	
    
	/* user not need cache */
	if (!fifo)
		fifo = &_fifo;

	/* flash cache */
	if ((fifo == &_fifo)
	    || !vastai_video_is_fifo_valid(fifo)
	    || vastai_video_fifo_is_full(fifo)) {
		ret = vastai_pci_mem_read(pci_info, die_index, fifo_addr, fifo,
					  sizeof(struct vastai_video_fifo));
		if (ret)
			return -ret;
	}

	if (va_pci_log_level & VASTAI_PCI_DEBUG_LEVEL) {
		char print_string[128];
		sprintf(print_string, "(push) read fifo[0x%llx]:", fifo_addr);
		vastai_video_fifo_print(pci_info, die_index, fifo, print_string);
	}
	if (vastai_video_fifo_is_full(fifo))
		return -ENOMEM;
	if (!vastai_video_is_fifo_valid(fifo))
		return -EFAULT;

	entry_addr = fifo_addr + vastai_video_fifo_wr_next(fifo);
	/*ret = vastai_pci_mem_write(pci_info, die_index, entry_addr, elem,
				   fifo->elem_size);
	if (ret)
		return -ret;

	ret = vastai_pci_mem_write(pci_info, die_index,
				   fifo_addr + offsetof(struct vastai_video_fifo, wr),
				   &(fifo->wr), sizeof(fifo->wr));
	if (ret)
		return -ret;
	*/
		
	//sync windows drv code,add check fifo
	do
	{			
		ret = vastai_pci_mem_write((struct vastai_pci_info *)pci_info, die_index, entry_addr, elem,
						fifo->elem_size);
		if (ret)
			return ret;
		//ret = vastai_pci_mem_read((struct vastai_pci_info *)pci_info, die_index, entry_addr, pviv->fifo_chk_mem,
		//			fifo->elem_size);
		ret = vastai_pci_mem_read((struct vastai_pci_info *)pci_info, die_index, entry_addr, &tvidencparam,
					fifo->elem_size);
		if (ret)
			return ret;
		//if(!memcmp(pviv->fifo_chk_mem, (const void *)elem, fifo->elem_size))
		if(!memcmp(&tvidencparam, (const void *)elem, fifo->elem_size))
		{
			break;
		}
		PDEBUG(VIDEO_LOGLEVEL_ERROR,"entry_addr=0x%llx, data check err \r\n", entry_addr);
		try_count--;
	} while (try_count);
	
	ret = vastai_pci_mem_write((struct vastai_pci_info *)pci_info, die_index,
				   fifo_addr + offsetof(struct vastai_video_fifo, wr),
				  &(fifo->wr), sizeof(fifo->wr));
	if (ret)
		return ret;
	
	try_count = 5;
	do
	{
		ret = vastai_pci_mem_read((struct vastai_pci_info *)pci_info, die_index, fifo_addr, &_fifo_chk,
					  sizeof(struct vastai_video_fifo));		
		if (ret)
			return ret;
	
		/*if(_fifo_chk.wr != fifo->wr || _fifo_chk.elem_count != fifo->elem_count || _fifo_chk.elem_size != fifo->elem_size
			|| !(_fifo_chk.rd  == (fifo->rd+1)%fifo->elem_count || _fifo_chk.rd == fifo->rd))
		{
			VLOG_ERR("fifo_addr=0x%llx, header check err  chk_wr=0x%x wr=0x%x\r\n", fifo_addr, _fifo_chk.wr, fifo->wr);
			try_count--;
			if(try_count == 0) break;
			continue;
		}*/
		//write back wr for vemcu
		ret = vastai_pci_mem_write((struct vastai_pci_info *)pci_info, die_index,
				   fifo_addr + offsetof(struct vastai_video_fifo, wr_bak),
				   &(fifo->wr_bak), sizeof(fifo->wr_bak));
		
		//printk("vastai_pci_mem_write wr_bak\n");
		if (ret)
			return ret;
	} while (0);


	if (va_pci_log_level & VASTAI_PCI_DEBUG_LEVEL) {
		char print_string[128];
		sprintf(print_string, "(push) write fifo[0x%llx]:", fifo_addr);
		vastai_video_fifo_print(pci_info, die_index, fifo, print_string);
	}
	return 0;
}

/**
 * @brief pop a elem from device fifo.
 *
 * @param pci_info : pointer to the pcie device
 * @param die_index : union die_index_data
 * @param fifo_addr : device address of fifo.
 * @param elem : return value, object which be poped.
 * @param fifo : return val, for the fifo cache. If it be give, this function
 * 		 will be more quickly.
 * @return int : success is zero. other is errno.
 */
int vastai_video_fifo_pop_elem(void *pci_info, int die_index, u64 fifo_addr,
			 void *elem, struct vastai_video_fifo *fifo)
{
	int ret = 0;
	u64 entry_addr = 0;
	struct vastai_video_fifo _fifo = { 0 };
	struct vastai_video_fifo _fifo_chk = { 0 };

	/* user not need cache */
	if (!fifo)
		fifo = &_fifo;

	/* flash cache */
	if ((fifo == &_fifo)
	    || !vastai_video_is_fifo_valid(fifo)
	    || vastai_video_fifo_is_empty(fifo)) {

		ret = vastai_pci_mem_read(pci_info, die_index, fifo_addr, fifo,
					  sizeof(struct vastai_video_fifo));
		if (ret)
			return -ret;
	}

	if (va_pci_log_level & VASTAI_PCI_DEBUG_LEVEL) {
		char print_string[128];
		sprintf(print_string, "(pop) read fifo[0x%llx]:", fifo_addr);
		vastai_video_fifo_print(pci_info, die_index, fifo, print_string);
	}
	if (vastai_video_fifo_is_empty(fifo))
		return -ENOMEM;
	if (!vastai_video_is_fifo_valid(fifo))
		return -EFAULT;

	entry_addr = fifo_addr + vastai_video_fifo_rd_next(fifo);

	if (elem) {
		ret = vastai_pci_mem_read(pci_info, die_index, entry_addr, elem,
					  fifo->elem_size);
		if (ret)
			return -ret;
	}

	ret = vastai_pci_mem_write(pci_info, die_index,
				   fifo_addr + offsetof(struct vastai_video_fifo, rd),
				   &(fifo->rd), sizeof(fifo->rd));
	if (ret)
		return -ret;
	
	ret = vastai_pci_mem_read((struct vastai_pci_info *)pci_info, die_index, fifo_addr, &_fifo_chk,
					  sizeof(struct vastai_video_fifo));
	if (ret)
		return ret;

	/*if(_fifo_chk.rd != fifo->rd || _fifo_chk.elem_count != fifo->elem_count || _fifo_chk.elem_size != fifo->elem_size
	 || !(_fifo_chk.wr  == (fifo->wr+1)%fifo->elem_count || _fifo_chk.wr == fifo->wr)){
		VLOG_ERR("fifo_addr=0x%llx, header check err  chk_wr=0x%x wr=0x%x\r\n", fifo_addr, _fifo_chk.wr, fifo->wr);	
	}*/

	ret = vastai_pci_mem_write((struct vastai_pci_info *)pci_info, die_index,
				   fifo_addr + offsetof(struct vastai_video_fifo, rd_bak),
				    &(fifo->rd_bak), sizeof(fifo->rd_bak));
	if (ret)
		return -ret;

	if (va_pci_log_level & VASTAI_PCI_DEBUG_LEVEL) {
		char print_string[128];
		sprintf(print_string, "(pop) write fifo[0x%llx]:", fifo_addr);
		vastai_video_fifo_print(pci_info, die_index, fifo, print_string);
	}
	return 0;
}


int vastai_video_enqueue_data(struct vastai_video_fifo *ring_buf, uint8_t *data, int32_t len)
{
	uint32_t wr;
	uint8_t *ptr;
	int count = 0;

	if (NULL == ring_buf || len != ring_buf->elem_size) {
		PDEBUG(VIDEO_LOGLEVEL_ERROR, "enqueue data, ring buf address is null or invalid len\n");
		return -EINVAL;
	}

	while (ring_buf_full(ring_buf)) {
		msleep(1);
		count++;
		if (count == MAX_WAIT_COUNT) {
			PDEBUG(VIDEO_LOGLEVEL_ERROR, "buffer full, addr: 0x%x, len: %d, rd=%d, wr=%d\n", ring_buf->buf, len, ring_buf ->rd, ring_buf ->wr);
			return -ENOMEM;
		}
	}

	wr = ring_buf->wr;
	ptr = (uint8_t *)ring_buf->buf + wr * ring_buf->elem_size;

	memcpy(ptr, data, len);

	wr = (wr + 1) % ring_buf->elem_count;
	ring_buf->wr = wr;

	return 0;
}

int vastai_video_dequeue_data(struct vastai_video_fifo *ring_buf, uint8_t *data)
{
	int32_t rd;
	uint8_t *ptr;

	if (NULL == ring_buf) {
		PDEBUG(VIDEO_LOGLEVEL_ERROR, "dequeue data, ring buf address is null\n");
		return -EINVAL;
	}

	if (ring_buf_empty(ring_buf))
		return 0;

	rd = ring_buf ->rd;
	ptr = (uint8_t *)ring_buf->buf + rd * ring_buf->elem_size;

	memcpy(data, ptr, ring_buf->elem_size);

	rd = (rd + 1) % ring_buf->elem_count;
	ring_buf->rd = rd;

	return ring_buf->elem_size;
}


void vastai_video_dma_transfer_sync(struct vastai_pci_info *priv, u8 dir, u8 src_mmu, u8 dst_mmu, u64 dev_addr, u32 trans_len, struct vastai_dma_buf* dmb)
{
	struct vastai_dmadesc desc;
	union core_bitmap core_id = {.val = 0};

	memset(&desc, 0, sizeof(struct vastai_dmadesc));
	desc.dev_addr = dev_addr;
	desc.is_src_dma_addr = 0;
	desc.is_src_not_user_mem = 1;
	desc.host_addr.vir_addr = dmb->vir;
	desc.host_addr.dma_addr = (dma_addr_t)dmb->phy;
	desc.is_host_to_dev = dir;
	desc.dma_lenth = trans_len;

	//core_id.encode = 0x0;
	vastai_pci_dma_transfer_sync(priv, priv->dies[0].die_index, core_id, &desc, 1);
}

//#undef PDEBUG
//#define PDEBUG(level, fmt, args...) {printk( KERN_INFO "pid[%d] video_utils: " fmt, current->pid, ## args);}


/**
 * 0: 64bit
 * 1: 32bit
*/
int video_is_32_bit(void)
{
	int is_32_bit = 0;
#if (defined(__aarch64__) || defined(__arm__))
	is_32_bit = test_thread_flag(TIF_32BIT);
#else
	is_32_bit = test_thread_flag(TIF_ADDR32);
#endif
	PDEBUG(VIDEO_LOGLEVEL_INFO, "is_32_bit_thread = %d\n", is_32_bit);
	return is_32_bit;
}

unsigned long video_copy_to_user(void __user *to, const void *from,
			      unsigned long n)
{
	unsigned long ret = 0;
	PDEBUG(VIDEO_LOGLEVEL_DEBUG, "to=%p , from=%p ,n=%ld", to, from, n);
#ifdef CONFIG_COMPAT
	if (video_is_32_bit()) {
		ret = copy_to_user(
			(void __user *)compat_ptr(
				(compat_uptr_t)(((u64)to) & 0xffffffff)),
			from, n);
		if (ret) {
		    PDEBUG(VIDEO_LOGLEVEL_ERROR, "fail: 32bit: to=%p from=%p n=%ld ret=%ld",
				    to, from, n, ret);
		}
		return ret;
	} else 
#endif
    {
		ret = copy_to_user(to, from, n);
		if (ret) {
			PDEBUG(VIDEO_LOGLEVEL_ERROR,
				    "fail: 64bit: to=%p from=%p n=%ld ret=%ld",
				    to, from, n, ret);
		}
		return ret;
	}
	return 0;
}

unsigned long video_copy_from_user(void *to, const void __user *from,
				unsigned long n)
{
	unsigned long ret = 0;
	PDEBUG(VIDEO_LOGLEVEL_DEBUG, "to=%p , from=%p ,n=%ld ", to, from, n);

#ifdef CONFIG_COMPAT
	if (video_is_32_bit()) {
		ret = copy_from_user(
			to,
			(void __user *)compat_ptr(
				(compat_uptr_t)(((u64)from) & 0xffffffff)),
			n);
		if (ret) {
			PDEBUG(VIDEO_LOGLEVEL_ERROR,
				    "fail: 32bit: to=%p from=%p n=%ld ret=%ld",
				    to, from, n, ret);
		}
		return ret;
	} else 
#endif
    {
		ret = copy_from_user(to, from, n);
		if (ret) {
			PDEBUG(VIDEO_LOGLEVEL_ERROR,
				    "fail: 64bit: to=%p from=%p n=%ld ret=%ld",
				    to, from, n, ret);
		}
		return ret;
	}
	return 0;
}
