#ifndef __VASTAI_VIDEO_UTILS_H__
#define __VASTAI_VIDEO_UTILS_H__

#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/compat.h>
#include <linux/thread_info.h>

#include "vastai_pci.h"

#define MAX_ELEM_NUM 256
#define MAX_WAIT_COUNT 100

struct video_common_cmd {
	union {
		struct {
			u32 size;
			int dma_buf_fd;
		} alloc_cmd;

		struct {
			u32 is_dev_to_host;
			u32 dma_buf_fd;
			u64 axi_addr;
			u32 length;
			u32 die_index;
		} dma_start_cmd;

		struct {
			u32 is_dev_to_host;
			void * vir_addr;
			u64 axi_addr;
			u32 length;
			u32 die_index;
			int pid;
		} dma_trans_cmd;

	};
};

typedef enum{
    DECODER,
    ENCODER,
} codec_type_t;

typedef struct bi_list_node{
  void* data;
  struct bi_list_node* next; 
  struct bi_list_node* previous;
}bi_list_node;

typedef struct bi_list{
  bi_list_node* head; 
  bi_list_node* tail; 
}bi_list;

typedef struct {
    volatile u16 rd;
    volatile u16 wr;
    volatile u16 max_size;
    u32 cmdbuf_id_fifo[MAX_ELEM_NUM];
	u32 cmdbuf_irq_status[MAX_ELEM_NUM];
    spinlock_t spin_lock;
    codec_type_t type;
}VastCmdbufQueue;


struct vastai_video_fifo {
	u32 rd;
	u32 wr;
	u32 elem_count;
	u32 elem_size;
	u32 rd_bak;
	u32 wr_bak;
	u8 buf[0]; /* this will not be used at host driver */
};


/********* local ringbuffer start ************/
static inline int ring_buf_empty(struct vastai_video_fifo *ring_buf)
{
	if (NULL == ring_buf)
		return -EINVAL;

	if (ring_buf->rd == ring_buf->wr)
		return 1;

	return 0;
}

static inline int ring_buf_full(struct vastai_video_fifo *ring_buf)
{	int rd, wr;

	if (NULL == ring_buf) {
		pr_err("ringbuffer buffer is null\n");
		return -EINVAL;
	}

	rd = ring_buf->rd;
	wr = ring_buf->wr;
	if ((wr + 1) % ring_buf->elem_count == rd)
		return 1;

	return 0;
}

int vastai_video_enqueue_data(struct vastai_video_fifo *ring_buf, uint8_t *data, int32_t len);
int vastai_video_dequeue_data(struct vastai_video_fifo *ring_buf, uint8_t *data);
/********* local ringbuffer end ************/

/********* host<=>vemcu ringbuffer start ************/
static inline u8 vastai_video_fifo_is_full(struct vastai_video_fifo *fifo)
{
	return ((fifo->wr + 1) % (fifo->elem_count) == (fifo->rd));
}

static inline u8 vastai_video_fifo_is_empty(struct vastai_video_fifo *fifo)
{
	return ((fifo->wr) == (fifo->rd));
}

static inline u32 vastai_video_fifo_wr_pre_fetch(struct vastai_video_fifo *fifo)
{
	return ((u32)offsetof(struct vastai_video_fifo, buf) + fifo->elem_size * fifo->wr);
}

static inline u32 vastai_video_fifo_wr_next(struct vastai_video_fifo *fifo)
{
	u32 ret = (u32)offsetof(struct vastai_video_fifo, buf) + fifo->elem_size * fifo->wr;
	fifo->wr = (fifo->wr + 1) % (fifo->elem_count);
	fifo->wr_bak = fifo->wr;
	return ret;
}

static inline u32 vastai_video_fifo_rd_pre_fetch(struct vastai_video_fifo *fifo)
{
	return ((u32)offsetof(struct vastai_video_fifo, buf) + fifo->elem_size * fifo->rd);
}

static inline u32 vastai_video_fifo_rd_next(struct vastai_video_fifo *fifo)
{
	u32 ret = (u32)offsetof(struct vastai_video_fifo, buf) + fifo->elem_size * fifo->rd;
	fifo->rd = (fifo->rd + 1) % (fifo->elem_count);
	fifo->rd_bak = fifo->rd;
	return ret;
}

static inline int vastai_video_is_fifo_valid(struct vastai_video_fifo *fifo)
{
	return ((fifo->rd < fifo->elem_count) && (fifo->wr < fifo->elem_count));
}

static inline int vastai_video_read_pcie(void* pci_info, unsigned int die_id, void* val, void* addr, unsigned int len){
	int ret;
	ret = vastai_pci_mem_read((struct vastai_pci_info *)pci_info, die_id, (u64)(addr), val, len);
	return ret;
}

static inline int vastai_video_write_pcie(void* pci_info, unsigned int die_id, void* val, void* addr, unsigned int len){
	return vastai_pci_mem_write((struct vastai_pci_info *)pci_info, die_id, (u64)(addr), val, len);
}

static inline void vastai_video_fifo_print(struct vastai_pci_info *priv, u32 die_id,
				struct vastai_video_fifo *fifo, char *head)
{
	VASTAI_PCI_DBG(priv, die_id,
			"%s rd[0x%x] wr[0x%x] elem_count[0x%x] elem_size[0x%x]\n",
			head, fifo->rd, fifo->wr, fifo->elem_count, fifo->elem_size);
}

static inline int vastai_video_fifo_init(struct vastai_pci_info *priv, int die_index,
        u64 fifo_addr, u32 elem_count, u32 elem_size) {

	struct vastai_video_fifo fifo = {.rd = 0, .wr = 0, .elem_count=elem_count, .elem_size=elem_size};
	return vastai_pci_mem_write(priv, die_index, fifo_addr, &fifo, sizeof(struct vastai_video_fifo));
}

int vastai_video_fifo_push_elem(void *pci_info, int die_index, u64 fifo_addr,
			  void *elem, struct vastai_video_fifo *fifo);

int vastai_video_fifo_pop_elem(void *pci_info, int die_index, u64 fifo_addr,
    void *elem, struct vastai_video_fifo *fifo);
/********* host<=>vemcu ringbuffer end ************/

/*********  queue start ************/
void vastai_cmdid_init(VastCmdbufQueue *IdQueue, codec_type_t type);

int vastai_cmdid_queue_empty(VastCmdbufQueue *IdQueue);

int vastai_cmdid_queue_full(VastCmdbufQueue *IdQueue);

int vastai_cmdid_enqueue(VastCmdbufQueue *IdQueue,  u32 CmdId, u32 irq_status);

int vastai_cmdid_dequeue(VastCmdbufQueue *IdQueue, u32* CmdId, u32* irq_status);
/*********  list end ************/

/*********  list start ************/
void init_bi_list(bi_list* list);

bi_list_node* bi_list_create_node(void);

void bi_list_free_node(bi_list_node* node);

void bi_list_insert_node_tail(bi_list* list,bi_list_node* current_node);

void bi_list_insert_node_before(bi_list* list,bi_list_node* base_node,bi_list_node* new_node);

void bi_list_remove_node(bi_list* list,bi_list_node* current_node);
/*********  list end ************/

/*********  dma trans start ************/
void vastai_video_dma_transfer_sync(struct vastai_pci_info *priv, u8 dir, u8 src_mmu, u8 dst_mmu, u64 dev_addr, u32 trans_len, struct vastai_dma_buf* dmb);
/*********  dma trans end ************/


int video_is_32_bit(void);
unsigned long video_copy_from_user(void *to, const void __user *from,
                               unsigned long n);
unsigned long video_copy_to_user(void __user *to, const void *from,
                               unsigned long n);

#endif